/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 4.0.2
 *
 * This file is not intended to be easily readable and contains a number of
 * coding conventions designed to improve portability and efficiency. Do not make
 * changes to this file unless you know what you are doing--modify the SWIG
 * interface file instead.
 * ----------------------------------------------------------------------------- */


#ifndef SWIGPYTHON
#define SWIGPYTHON
#endif

#define SWIG_PYTHON_DIRECTOR_NO_VTABLE
#define SWIGPYTHON_BUILTIN


#ifdef __cplusplus
/* SwigValueWrapper is described in swig.swg */
template<typename T> class SwigValueWrapper {
  struct SwigMovePointer {
    T *ptr;
    SwigMovePointer(T *p) : ptr(p) { }
    ~SwigMovePointer() { delete ptr; }
    SwigMovePointer& operator=(SwigMovePointer& rhs) { T* oldptr = ptr; ptr = 0; delete oldptr; ptr = rhs.ptr; rhs.ptr = 0; return *this; }
  } pointer;
  SwigValueWrapper& operator=(const SwigValueWrapper<T>& rhs);
  SwigValueWrapper(const SwigValueWrapper<T>& rhs);
public:
  SwigValueWrapper() : pointer(0) { }
  SwigValueWrapper& operator=(const T& t) { SwigMovePointer tmp(new T(t)); pointer = tmp; return *this; }
  operator T&() const { return *pointer.ptr; }
  T *operator&() { return pointer.ptr; }
};

template <typename T> T SwigValueInit() {
  return T();
}
#endif

/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x560)
#  define SWIGTEMPLATEDISAMBIGUATOR template
# elif defined(__HP_aCC)
/* Needed even with `aCC -AA' when `aCC -V' reports HP ANSI C++ B3910B A.03.55 */
/* If we find a maximum version that requires this, the test would be __HP_aCC <= 35500 for A.03.55 */
#  define SWIGTEMPLATEDISAMBIGUATOR template
# else
#  define SWIGTEMPLATEDISAMBIGUATOR
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__))
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__))
# else
#   define SWIGUNUSED
# endif
#endif

#ifndef SWIG_MSC_UNSUPPRESS_4505
# if defined(_MSC_VER)
#   pragma warning(disable : 4505) /* unreferenced local function has been removed */
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if defined(__GNUC__)
#  if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#    ifndef GCC_HASCLASSVISIBILITY
#      define GCC_HASCLASSVISIBILITY
#    endif
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif

/* Deal with Microsoft's attempt at deprecating methods in the standard C++ library */
#if !defined(SWIG_NO_SCL_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_SCL_SECURE_NO_DEPRECATE)
# define _SCL_SECURE_NO_DEPRECATE
#endif

/* Deal with Apple's deprecated 'AssertMacros.h' from Carbon-framework */
#if defined(__APPLE__) && !defined(__ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES)
# define __ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES 0
#endif

/* Intel's compiler complains if a variable which was never initialised is
 * cast to void, which is a common idiom which we use to indicate that we
 * are aware a variable isn't used.  So we just silence that warning.
 * See: https://github.com/swig/swig/issues/192 for more discussion.
 */
#ifdef __INTEL_COMPILER
# pragma warning disable 592
#endif


#if defined(__GNUC__) && defined(_WIN32) && !defined(SWIG_PYTHON_NO_HYPOT_WORKAROUND)
/* Workaround for '::hypot' has not been declared', see https://bugs.python.org/issue11566 */
# include <math.h>
#endif

#if defined(_DEBUG) && defined(SWIG_PYTHON_INTERPRETER_NO_DEBUG)
/* Use debug wrappers with the Python release dll */
# undef _DEBUG
# include <Python.h>
# define _DEBUG 1
#else
# include <Python.h>
#endif

/* -----------------------------------------------------------------------------
 * swigrun.swg
 *
 * This file contains generic C API SWIG runtime support for pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* This should only be incremented when either the layout of swig_type_info changes,
   or for whatever reason, the runtime changes incompatibly */
#define SWIG_RUNTIME_VERSION "4"

/* define SWIG_TYPE_TABLE_NAME as "SWIG_TYPE_TABLE" */
#ifdef SWIG_TYPE_TABLE
# define SWIG_QUOTE_STRING(x) #x
# define SWIG_EXPAND_AND_QUOTE_STRING(x) SWIG_QUOTE_STRING(x)
# define SWIG_TYPE_TABLE_NAME SWIG_EXPAND_AND_QUOTE_STRING(SWIG_TYPE_TABLE)
#else
# define SWIG_TYPE_TABLE_NAME
#endif

/*
  You can use the SWIGRUNTIME and SWIGRUNTIMEINLINE macros for
  creating a static or dynamic library from the SWIG runtime code.
  In 99.9% of the cases, SWIG just needs to declare them as 'static'.

  But only do this if strictly necessary, ie, if you have problems
  with your compiler or suchlike.
*/

#ifndef SWIGRUNTIME
# define SWIGRUNTIME SWIGINTERN
#endif

#ifndef SWIGRUNTIMEINLINE
# define SWIGRUNTIMEINLINE SWIGRUNTIME SWIGINLINE
#endif

/*  Generic buffer size */
#ifndef SWIG_BUFFER_SIZE
# define SWIG_BUFFER_SIZE 1024
#endif

/* Flags for pointer conversions */
#define SWIG_POINTER_DISOWN        0x1
#define SWIG_CAST_NEW_MEMORY       0x2
#define SWIG_POINTER_NO_NULL       0x4

/* Flags for new pointer objects */
#define SWIG_POINTER_OWN           0x1


/*
   Flags/methods for returning states.

   The SWIG conversion methods, as ConvertPtr, return an integer
   that tells if the conversion was successful or not. And if not,
   an error code can be returned (see swigerrors.swg for the codes).

   Use the following macros/flags to set or process the returning
   states.

   In old versions of SWIG, code such as the following was usually written:

     if (SWIG_ConvertPtr(obj,vptr,ty.flags) != -1) {
       // success code
     } else {
       //fail code
     }

   Now you can be more explicit:

    int res = SWIG_ConvertPtr(obj,vptr,ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
    } else {
      // fail code
    }

   which is the same really, but now you can also do

    Type *ptr;
    int res = SWIG_ConvertPtr(obj,(void **)(&ptr),ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
      if (SWIG_IsNewObj(res) {
        ...
	delete *ptr;
      } else {
        ...
      }
    } else {
      // fail code
    }

   I.e., now SWIG_ConvertPtr can return new objects and you can
   identify the case and take care of the deallocation. Of course that
   also requires SWIG_ConvertPtr to return new result values, such as

      int SWIG_ConvertPtr(obj, ptr,...) {
        if (<obj is ok>) {
          if (<need new object>) {
            *ptr = <ptr to new allocated object>;
            return SWIG_NEWOBJ;
          } else {
            *ptr = <ptr to old object>;
            return SWIG_OLDOBJ;
          }
        } else {
          return SWIG_BADOBJ;
        }
      }

   Of course, returning the plain '0(success)/-1(fail)' still works, but you can be
   more explicit by returning SWIG_BADOBJ, SWIG_ERROR or any of the
   SWIG errors code.

   Finally, if the SWIG_CASTRANK_MODE is enabled, the result code
   allows to return the 'cast rank', for example, if you have this

       int food(double)
       int fooi(int);

   and you call

      food(1)   // cast rank '1'  (1 -> 1.0)
      fooi(1)   // cast rank '0'

   just use the SWIG_AddCast()/SWIG_CheckState()
*/

#define SWIG_OK                    (0)
#define SWIG_ERROR                 (-1)
#define SWIG_IsOK(r)               (r >= 0)
#define SWIG_ArgError(r)           ((r != SWIG_ERROR) ? r : SWIG_TypeError)

/* The CastRankLimit says how many bits are used for the cast rank */
#define SWIG_CASTRANKLIMIT         (1 << 8)
/* The NewMask denotes the object was created (using new/malloc) */
#define SWIG_NEWOBJMASK            (SWIG_CASTRANKLIMIT  << 1)
/* The TmpMask is for in/out typemaps that use temporal objects */
#define SWIG_TMPOBJMASK            (SWIG_NEWOBJMASK << 1)
/* Simple returning values */
#define SWIG_BADOBJ                (SWIG_ERROR)
#define SWIG_OLDOBJ                (SWIG_OK)
#define SWIG_NEWOBJ                (SWIG_OK | SWIG_NEWOBJMASK)
#define SWIG_TMPOBJ                (SWIG_OK | SWIG_TMPOBJMASK)
/* Check, add and del mask methods */
#define SWIG_AddNewMask(r)         (SWIG_IsOK(r) ? (r | SWIG_NEWOBJMASK) : r)
#define SWIG_DelNewMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_NEWOBJMASK) : r)
#define SWIG_IsNewObj(r)           (SWIG_IsOK(r) && (r & SWIG_NEWOBJMASK))
#define SWIG_AddTmpMask(r)         (SWIG_IsOK(r) ? (r | SWIG_TMPOBJMASK) : r)
#define SWIG_DelTmpMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_TMPOBJMASK) : r)
#define SWIG_IsTmpObj(r)           (SWIG_IsOK(r) && (r & SWIG_TMPOBJMASK))

/* Cast-Rank Mode */
#if defined(SWIG_CASTRANK_MODE)
#  ifndef SWIG_TypeRank
#    define SWIG_TypeRank             unsigned long
#  endif
#  ifndef SWIG_MAXCASTRANK            /* Default cast allowed */
#    define SWIG_MAXCASTRANK          (2)
#  endif
#  define SWIG_CASTRANKMASK          ((SWIG_CASTRANKLIMIT) -1)
#  define SWIG_CastRank(r)           (r & SWIG_CASTRANKMASK)
SWIGINTERNINLINE int SWIG_AddCast(int r) {
  return SWIG_IsOK(r) ? ((SWIG_CastRank(r) < SWIG_MAXCASTRANK) ? (r + 1) : SWIG_ERROR) : r;
}
SWIGINTERNINLINE int SWIG_CheckState(int r) {
  return SWIG_IsOK(r) ? SWIG_CastRank(r) + 1 : 0;
}
#else /* no cast-rank mode */
#  define SWIG_AddCast(r) (r)
#  define SWIG_CheckState(r) (SWIG_IsOK(r) ? 1 : 0)
#endif


#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef void *(*swig_converter_func)(void *, int *);
typedef struct swig_type_info *(*swig_dycast_func)(void **);

/* Structure to store information on one type */
typedef struct swig_type_info {
  const char             *name;			/* mangled name of this type */
  const char             *str;			/* human readable name of this type */
  swig_dycast_func        dcast;		/* dynamic cast function down a hierarchy */
  struct swig_cast_info  *cast;			/* linked list of types that can cast into this type */
  void                   *clientdata;		/* language specific type data */
  int                    owndata;		/* flag if the structure owns the clientdata */
} swig_type_info;

/* Structure to store a type and conversion function used for casting */
typedef struct swig_cast_info {
  swig_type_info         *type;			/* pointer to type that is equivalent to this type */
  swig_converter_func     converter;		/* function to cast the void pointers */
  struct swig_cast_info  *next;			/* pointer to next cast in linked list */
  struct swig_cast_info  *prev;			/* pointer to the previous cast */
} swig_cast_info;

/* Structure used to store module information
 * Each module generates one structure like this, and the runtime collects
 * all of these structures and stores them in a circularly linked list.*/
typedef struct swig_module_info {
  swig_type_info         **types;		/* Array of pointers to swig_type_info structures that are in this module */
  size_t                 size;		        /* Number of types in this module */
  struct swig_module_info *next;		/* Pointer to next element in circularly linked list */
  swig_type_info         **type_initial;	/* Array of initially generated type structures */
  swig_cast_info         **cast_initial;	/* Array of initially generated casting structures */
  void                    *clientdata;		/* Language specific module data */
} swig_module_info;

/*
  Compare two type names skipping the space characters, therefore
  "char*" == "char *" and "Class<int>" == "Class<int >", etc.

  Return 0 when the two name types are equivalent, as in
  strncmp, but skipping ' '.
*/
SWIGRUNTIME int
SWIG_TypeNameComp(const char *f1, const char *l1,
		  const char *f2, const char *l2) {
  for (;(f1 != l1) && (f2 != l2); ++f1, ++f2) {
    while ((*f1 == ' ') && (f1 != l1)) ++f1;
    while ((*f2 == ' ') && (f2 != l2)) ++f2;
    if (*f1 != *f2) return (*f1 > *f2) ? 1 : -1;
  }
  return (int)((l1 - f1) - (l2 - f2));
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if equal, -1 if nb < tb, 1 if nb > tb
*/
SWIGRUNTIME int
SWIG_TypeCmp(const char *nb, const char *tb) {
  int equiv = 1;
  const char* te = tb + strlen(tb);
  const char* ne = nb;
  while (equiv != 0 && *ne) {
    for (nb = ne; *ne; ++ne) {
      if (*ne == '|') break;
    }
    equiv = SWIG_TypeNameComp(nb, ne, tb, te);
    if (*ne) ++ne;
  }
  return equiv;
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if not equal, 1 if equal
*/
SWIGRUNTIME int
SWIG_TypeEquiv(const char *nb, const char *tb) {
  return SWIG_TypeCmp(nb, tb) == 0 ? 1 : 0;
}

/*
  Check the typename
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheck(const char *c, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *iter = ty->cast;
    while (iter) {
      if (strcmp(iter->type->name, c) == 0) {
        if (iter == ty->cast)
          return iter;
        /* Move iter to the top of the linked list */
        iter->prev->next = iter->next;
        if (iter->next)
          iter->next->prev = iter->prev;
        iter->next = ty->cast;
        iter->prev = 0;
        if (ty->cast) ty->cast->prev = iter;
        ty->cast = iter;
        return iter;
      }
      iter = iter->next;
    }
  }
  return 0;
}

/*
  Identical to SWIG_TypeCheck, except strcmp is replaced with a pointer comparison
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheckStruct(swig_type_info *from, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *iter = ty->cast;
    while (iter) {
      if (iter->type == from) {
        if (iter == ty->cast)
          return iter;
        /* Move iter to the top of the linked list */
        iter->prev->next = iter->next;
        if (iter->next)
          iter->next->prev = iter->prev;
        iter->next = ty->cast;
        iter->prev = 0;
        if (ty->cast) ty->cast->prev = iter;
        ty->cast = iter;
        return iter;
      }
      iter = iter->next;
    }
  }
  return 0;
}

/*
  Cast a pointer up an inheritance hierarchy
*/
SWIGRUNTIMEINLINE void *
SWIG_TypeCast(swig_cast_info *ty, void *ptr, int *newmemory) {
  return ((!ty) || (!ty->converter)) ? ptr : (*ty->converter)(ptr, newmemory);
}

/*
   Dynamic pointer casting. Down an inheritance hierarchy
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeDynamicCast(swig_type_info *ty, void **ptr) {
  swig_type_info *lastty = ty;
  if (!ty || !ty->dcast) return ty;
  while (ty && (ty->dcast)) {
    ty = (*ty->dcast)(ptr);
    if (ty) lastty = ty;
  }
  return lastty;
}

/*
  Return the name associated with this type
*/
SWIGRUNTIMEINLINE const char *
SWIG_TypeName(const swig_type_info *ty) {
  return ty->name;
}

/*
  Return the pretty name associated with this type,
  that is an unmangled type name in a form presentable to the user.
*/
SWIGRUNTIME const char *
SWIG_TypePrettyName(const swig_type_info *type) {
  /* The "str" field contains the equivalent pretty names of the
     type, separated by vertical-bar characters.  We choose
     to print the last name, as it is often (?) the most
     specific. */
  if (!type) return NULL;
  if (type->str != NULL) {
    const char *last_name = type->str;
    const char *s;
    for (s = type->str; *s; s++)
      if (*s == '|') last_name = s+1;
    return last_name;
  }
  else
    return type->name;
}

/*
   Set the clientdata field for a type
*/
SWIGRUNTIME void
SWIG_TypeClientData(swig_type_info *ti, void *clientdata) {
  swig_cast_info *cast = ti->cast;
  /* if (ti->clientdata == clientdata) return; */
  ti->clientdata = clientdata;

  while (cast) {
    if (!cast->converter) {
      swig_type_info *tc = cast->type;
      if (!tc->clientdata) {
	SWIG_TypeClientData(tc, clientdata);
      }
    }
    cast = cast->next;
  }
}
SWIGRUNTIME void
SWIG_TypeNewClientData(swig_type_info *ti, void *clientdata) {
  SWIG_TypeClientData(ti, clientdata);
  ti->owndata = 1;
}

/*
  Search for a swig_type_info structure only by mangled name
  Search is a O(log #types)

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_MangledTypeQueryModule(swig_module_info *start,
                            swig_module_info *end,
		            const char *name) {
  swig_module_info *iter = start;
  do {
    if (iter->size) {
      size_t l = 0;
      size_t r = iter->size - 1;
      do {
	/* since l+r >= 0, we can (>> 1) instead (/ 2) */
	size_t i = (l + r) >> 1;
	const char *iname = iter->types[i]->name;
	if (iname) {
	  int compare = strcmp(name, iname);
	  if (compare == 0) {
	    return iter->types[i];
	  } else if (compare < 0) {
	    if (i) {
	      r = i - 1;
	    } else {
	      break;
	    }
	  } else if (compare > 0) {
	    l = i + 1;
	  }
	} else {
	  break; /* should never happen */
	}
      } while (l <= r);
    }
    iter = iter->next;
  } while (iter != end);
  return 0;
}

/*
  Search for a swig_type_info structure for either a mangled name or a human readable name.
  It first searches the mangled names of the types, which is a O(log #types)
  If a type is not found it then searches the human readable names, which is O(#types).

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeQueryModule(swig_module_info *start,
                     swig_module_info *end,
		     const char *name) {
  /* STEP 1: Search the name field using binary search */
  swig_type_info *ret = SWIG_MangledTypeQueryModule(start, end, name);
  if (ret) {
    return ret;
  } else {
    /* STEP 2: If the type hasn't been found, do a complete search
       of the str field (the human readable name) */
    swig_module_info *iter = start;
    do {
      size_t i = 0;
      for (; i < iter->size; ++i) {
	if (iter->types[i]->str && (SWIG_TypeEquiv(iter->types[i]->str, name)))
	  return iter->types[i];
      }
      iter = iter->next;
    } while (iter != end);
  }

  /* neither found a match */
  return 0;
}

/*
   Pack binary data into a string
*/
SWIGRUNTIME char *
SWIG_PackData(char *c, void *ptr, size_t sz) {
  static const char hex[17] = "0123456789abcdef";
  const unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu =  u + sz;
  for (; u != eu; ++u) {
    unsigned char uu = *u;
    *(c++) = hex[(uu & 0xf0) >> 4];
    *(c++) = hex[uu & 0xf];
  }
  return c;
}

/*
   Unpack binary data from a string
*/
SWIGRUNTIME const char *
SWIG_UnpackData(const char *c, void *ptr, size_t sz) {
  unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu = u + sz;
  for (; u != eu; ++u) {
    char d = *(c++);
    unsigned char uu;
    if ((d >= '0') && (d <= '9'))
      uu = (unsigned char)((d - '0') << 4);
    else if ((d >= 'a') && (d <= 'f'))
      uu = (unsigned char)((d - ('a'-10)) << 4);
    else
      return (char *) 0;
    d = *(c++);
    if ((d >= '0') && (d <= '9'))
      uu |= (unsigned char)(d - '0');
    else if ((d >= 'a') && (d <= 'f'))
      uu |= (unsigned char)(d - ('a'-10));
    else
      return (char *) 0;
    *u = uu;
  }
  return c;
}

/*
   Pack 'void *' into a string buffer.
*/
SWIGRUNTIME char *
SWIG_PackVoidPtr(char *buff, void *ptr, const char *name, size_t bsz) {
  char *r = buff;
  if ((2*sizeof(void *) + 2) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,&ptr,sizeof(void *));
  if (strlen(name) + 1 > (bsz - (r - buff))) return 0;
  strcpy(r,name);
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackVoidPtr(const char *c, void **ptr, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      *ptr = (void *) 0;
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sizeof(void *));
}

SWIGRUNTIME char *
SWIG_PackDataName(char *buff, void *ptr, size_t sz, const char *name, size_t bsz) {
  char *r = buff;
  size_t lname = (name ? strlen(name) : 0);
  if ((2*sz + 2 + lname) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,ptr,sz);
  if (lname) {
    strncpy(r,name,lname+1);
  } else {
    *r = 0;
  }
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackDataName(const char *c, void *ptr, size_t sz, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      memset(ptr,0,sz);
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sz);
}

#ifdef __cplusplus
}
#endif

/*  Errors in SWIG */
#define  SWIG_UnknownError    	   -1
#define  SWIG_IOError        	   -2
#define  SWIG_RuntimeError   	   -3
#define  SWIG_IndexError     	   -4
#define  SWIG_TypeError      	   -5
#define  SWIG_DivisionByZero 	   -6
#define  SWIG_OverflowError  	   -7
#define  SWIG_SyntaxError    	   -8
#define  SWIG_ValueError     	   -9
#define  SWIG_SystemError    	   -10
#define  SWIG_AttributeError 	   -11
#define  SWIG_MemoryError    	   -12
#define  SWIG_NullReferenceError   -13



/* Compatibility macros for Python 3 */
#if PY_VERSION_HEX >= 0x03000000

#define PyClass_Check(obj) PyObject_IsInstance(obj, (PyObject *)&PyType_Type)
#define PyInt_Check(x) PyLong_Check(x)
#define PyInt_AsLong(x) PyLong_AsLong(x)
#define PyInt_FromLong(x) PyLong_FromLong(x)
#define PyInt_FromSize_t(x) PyLong_FromSize_t(x)
#define PyString_Check(name) PyBytes_Check(name)
#define PyString_FromString(x) PyUnicode_FromString(x)
#define PyString_Format(fmt, args)  PyUnicode_Format(fmt, args)
#define PyString_AsString(str) PyBytes_AsString(str)
#define PyString_Size(str) PyBytes_Size(str)	
#define PyString_InternFromString(key) PyUnicode_InternFromString(key)
#define Py_TPFLAGS_HAVE_CLASS Py_TPFLAGS_BASETYPE
#define PyString_AS_STRING(x) PyUnicode_AS_STRING(x)
#define _PyLong_FromSsize_t(x) PyLong_FromSsize_t(x)

#endif

#ifndef Py_TYPE
#  define Py_TYPE(op) ((op)->ob_type)
#endif

/* SWIG APIs for compatibility of both Python 2 & 3 */

#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_Python_str_FromFormat PyUnicode_FromFormat
#else
#  define SWIG_Python_str_FromFormat PyString_FromFormat
#endif


/* Warning: This function will allocate a new string in Python 3,
 * so please call SWIG_Python_str_DelForPy3(x) to free the space.
 */
SWIGINTERN char*
SWIG_Python_str_AsChar(PyObject *str)
{
#if PY_VERSION_HEX >= 0x03030000
  return (char *)PyUnicode_AsUTF8(str);
#elif PY_VERSION_HEX >= 0x03000000
  char *newstr = 0;
  str = PyUnicode_AsUTF8String(str);
  if (str) {
    char *cstr;
    Py_ssize_t len;
    if (PyBytes_AsStringAndSize(str, &cstr, &len) != -1) {
      newstr = (char *) malloc(len+1);
      if (newstr)
        memcpy(newstr, cstr, len+1);
    }
    Py_XDECREF(str);
  }
  return newstr;
#else
  return PyString_AsString(str);
#endif
}

#if PY_VERSION_HEX >= 0x03030000 || PY_VERSION_HEX < 0x03000000
#  define SWIG_Python_str_DelForPy3(x)
#else
#  define SWIG_Python_str_DelForPy3(x) free( (void*) (x) )
#endif


SWIGINTERN PyObject*
SWIG_Python_str_FromChar(const char *c)
{
#if PY_VERSION_HEX >= 0x03000000
  return PyUnicode_FromString(c); 
#else
  return PyString_FromString(c);
#endif
}

#ifndef PyObject_DEL
# define PyObject_DEL PyObject_Del
#endif

// SWIGPY_USE_CAPSULE is no longer used within SWIG itself, but some user
// interface files check for it.
# define SWIGPY_USE_CAPSULE
# define SWIGPY_CAPSULE_NAME ("swig_runtime_data" SWIG_RUNTIME_VERSION ".type_pointer_capsule" SWIG_TYPE_TABLE_NAME)

#if PY_VERSION_HEX < 0x03020000
#define PyDescr_TYPE(x) (((PyDescrObject *)(x))->d_type)
#define PyDescr_NAME(x) (((PyDescrObject *)(x))->d_name)
#define Py_hash_t long
#endif

/* -----------------------------------------------------------------------------
 * error manipulation
 * ----------------------------------------------------------------------------- */

SWIGRUNTIME PyObject*
SWIG_Python_ErrorType(int code) {
  PyObject* type = 0;
  switch(code) {
  case SWIG_MemoryError:
    type = PyExc_MemoryError;
    break;
  case SWIG_IOError:
    type = PyExc_IOError;
    break;
  case SWIG_RuntimeError:
    type = PyExc_RuntimeError;
    break;
  case SWIG_IndexError:
    type = PyExc_IndexError;
    break;
  case SWIG_TypeError:
    type = PyExc_TypeError;
    break;
  case SWIG_DivisionByZero:
    type = PyExc_ZeroDivisionError;
    break;
  case SWIG_OverflowError:
    type = PyExc_OverflowError;
    break;
  case SWIG_SyntaxError:
    type = PyExc_SyntaxError;
    break;
  case SWIG_ValueError:
    type = PyExc_ValueError;
    break;
  case SWIG_SystemError:
    type = PyExc_SystemError;
    break;
  case SWIG_AttributeError:
    type = PyExc_AttributeError;
    break;
  default:
    type = PyExc_RuntimeError;
  }
  return type;
}


SWIGRUNTIME void
SWIG_Python_AddErrorMsg(const char* mesg)
{
  PyObject *type = 0;
  PyObject *value = 0;
  PyObject *traceback = 0;

  if (PyErr_Occurred())
    PyErr_Fetch(&type, &value, &traceback);
  if (value) {
    PyObject *old_str = PyObject_Str(value);
    const char *tmp = SWIG_Python_str_AsChar(old_str);
    PyErr_Clear();
    Py_XINCREF(type);
    if (tmp)
      PyErr_Format(type, "%s %s", tmp, mesg);
    else
      PyErr_Format(type, "%s", mesg);
    SWIG_Python_str_DelForPy3(tmp);
    Py_DECREF(old_str);
    Py_DECREF(value);
  } else {
    PyErr_SetString(PyExc_RuntimeError, mesg);
  }
}

SWIGRUNTIME int
SWIG_Python_TypeErrorOccurred(PyObject *obj)
{
  PyObject *error;
  if (obj)
    return 0;
  error = PyErr_Occurred();
  return error && PyErr_GivenExceptionMatches(error, PyExc_TypeError);
}

SWIGRUNTIME void
SWIG_Python_RaiseOrModifyTypeError(const char *message)
{
  if (SWIG_Python_TypeErrorOccurred(NULL)) {
    /* Use existing TypeError to preserve stacktrace and enhance with given message */
    PyObject *newvalue;
    PyObject *type = NULL, *value = NULL, *traceback = NULL;
    PyErr_Fetch(&type, &value, &traceback);
#if PY_VERSION_HEX >= 0x03000000
    newvalue = PyUnicode_FromFormat("%S\nAdditional information:\n%s", value, message);
#else
    newvalue = PyString_FromFormat("%s\nAdditional information:\n%s", PyString_AsString(value), message);
#endif
    Py_XDECREF(value);
    PyErr_Restore(type, newvalue, traceback);
  } else {
    /* Raise TypeError using given message */
    PyErr_SetString(PyExc_TypeError, message);
  }
}

#if defined(SWIG_PYTHON_NO_THREADS)
#  if defined(SWIG_PYTHON_THREADS)
#    undef SWIG_PYTHON_THREADS
#  endif
#endif
#if defined(SWIG_PYTHON_THREADS) /* Threading support is enabled */
#  if !defined(SWIG_PYTHON_USE_GIL) && !defined(SWIG_PYTHON_NO_USE_GIL)
#    define SWIG_PYTHON_USE_GIL
#  endif
#  if defined(SWIG_PYTHON_USE_GIL) /* Use PyGILState threads calls */
#    ifndef SWIG_PYTHON_INITIALIZE_THREADS
#     define SWIG_PYTHON_INITIALIZE_THREADS  PyEval_InitThreads() 
#    endif
#    ifdef __cplusplus /* C++ code */
       class SWIG_Python_Thread_Block {
         bool status;
         PyGILState_STATE state;
       public:
         void end() { if (status) { PyGILState_Release(state); status = false;} }
         SWIG_Python_Thread_Block() : status(true), state(PyGILState_Ensure()) {}
         ~SWIG_Python_Thread_Block() { end(); }
       };
       class SWIG_Python_Thread_Allow {
         bool status;
         PyThreadState *save;
       public:
         void end() { if (status) { PyEval_RestoreThread(save); status = false; }}
         SWIG_Python_Thread_Allow() : status(true), save(PyEval_SaveThread()) {}
         ~SWIG_Python_Thread_Allow() { end(); }
       };
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   SWIG_Python_Thread_Block _swig_thread_block
#      define SWIG_PYTHON_THREAD_END_BLOCK     _swig_thread_block.end()
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   SWIG_Python_Thread_Allow _swig_thread_allow
#      define SWIG_PYTHON_THREAD_END_ALLOW     _swig_thread_allow.end()
#    else /* C code */
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   PyGILState_STATE _swig_thread_block = PyGILState_Ensure()
#      define SWIG_PYTHON_THREAD_END_BLOCK     PyGILState_Release(_swig_thread_block)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   PyThreadState *_swig_thread_allow = PyEval_SaveThread()
#      define SWIG_PYTHON_THREAD_END_ALLOW     PyEval_RestoreThread(_swig_thread_allow)
#    endif
#  else /* Old thread way, not implemented, user must provide it */
#    if !defined(SWIG_PYTHON_INITIALIZE_THREADS)
#      define SWIG_PYTHON_INITIALIZE_THREADS
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_BLOCK)
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_BLOCK)
#      define SWIG_PYTHON_THREAD_END_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_ALLOW)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_ALLOW)
#      define SWIG_PYTHON_THREAD_END_ALLOW
#    endif
#  endif
#else /* No thread support */
#  define SWIG_PYTHON_INITIALIZE_THREADS
#  define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#  define SWIG_PYTHON_THREAD_END_BLOCK
#  define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#  define SWIG_PYTHON_THREAD_END_ALLOW
#endif

/* -----------------------------------------------------------------------------
 * Python API portion that goes into the runtime
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

/* -----------------------------------------------------------------------------
 * Constant declarations
 * ----------------------------------------------------------------------------- */

/* Constant Types */
#define SWIG_PY_POINTER 4
#define SWIG_PY_BINARY  5

/* Constant information structure */
typedef struct swig_const_info {
  int type;
  const char *name;
  long lvalue;
  double dvalue;
  void   *pvalue;
  swig_type_info **ptype;
} swig_const_info;

#ifdef __cplusplus
}
#endif


/* -----------------------------------------------------------------------------
 * pyrun.swg
 *
 * This file contains the runtime support for Python modules
 * and includes code for managing global variables and pointer
 * type checking.
 *
 * ----------------------------------------------------------------------------- */

#if PY_VERSION_HEX < 0x02070000 /* 2.7.0 */
# error "This version of SWIG only supports Python >= 2.7"
#endif

#if PY_VERSION_HEX >= 0x03000000 && PY_VERSION_HEX < 0x03020000
# error "This version of SWIG only supports Python 3 >= 3.2"
#endif

/* Common SWIG API */

/* for raw pointers */
#define SWIG_Python_ConvertPtr(obj, pptr, type, flags)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, 0)
#define SWIG_ConvertPtr(obj, pptr, type, flags)         SWIG_Python_ConvertPtr(obj, pptr, type, flags)
#define SWIG_ConvertPtrAndOwn(obj,pptr,type,flags,own)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, own)

#ifdef SWIGPYTHON_BUILTIN
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(self, ptr, type, flags)
#else
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(NULL, ptr, type, flags)
#endif

#define SWIG_InternalNewPointerObj(ptr, type, flags)	SWIG_Python_NewPointerObj(NULL, ptr, type, flags)

#define SWIG_CheckImplicit(ty)                          SWIG_Python_CheckImplicit(ty) 
#define SWIG_AcquirePtr(ptr, src)                       SWIG_Python_AcquirePtr(ptr, src)
#define swig_owntype                                    int

/* for raw packed data */
#define SWIG_ConvertPacked(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewPackedObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)

/* for class or struct pointers */
#define SWIG_ConvertInstance(obj, pptr, type, flags)    SWIG_ConvertPtr(obj, pptr, type, flags)
#define SWIG_NewInstanceObj(ptr, type, flags)           SWIG_NewPointerObj(ptr, type, flags)

/* for C or C++ function pointers */
#define SWIG_ConvertFunctionPtr(obj, pptr, type)        SWIG_Python_ConvertFunctionPtr(obj, pptr, type)
#define SWIG_NewFunctionPtrObj(ptr, type)               SWIG_Python_NewPointerObj(NULL, ptr, type, 0)

/* for C++ member pointers, ie, member methods */
#define SWIG_ConvertMember(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewMemberObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)


/* Runtime API */

#define SWIG_GetModule(clientdata)                      SWIG_Python_GetModule(clientdata)
#define SWIG_SetModule(clientdata, pointer)             SWIG_Python_SetModule(pointer)
#define SWIG_NewClientData(obj)                         SwigPyClientData_New(obj)

#define SWIG_SetErrorObj                                SWIG_Python_SetErrorObj                            
#define SWIG_SetErrorMsg                        	SWIG_Python_SetErrorMsg				   
#define SWIG_ErrorType(code)                    	SWIG_Python_ErrorType(code)                        
#define SWIG_Error(code, msg)            		SWIG_Python_SetErrorMsg(SWIG_ErrorType(code), msg) 
#define SWIG_fail                        		goto fail					   


/* Runtime API implementation */

/* Error manipulation */

SWIGINTERN void 
SWIG_Python_SetErrorObj(PyObject *errtype, PyObject *obj) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK; 
  PyErr_SetObject(errtype, obj);
  Py_DECREF(obj);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

SWIGINTERN void 
SWIG_Python_SetErrorMsg(PyObject *errtype, const char *msg) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK;
  PyErr_SetString(errtype, msg);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

#define SWIG_Python_Raise(obj, type, desc)  SWIG_Python_SetErrorObj(SWIG_Python_ExceptionType(desc), obj)

/* Set a constant value */

#if defined(SWIGPYTHON_BUILTIN)

SWIGINTERN void
SwigPyBuiltin_AddPublicSymbol(PyObject *seq, const char *key) {
  PyObject *s = PyString_InternFromString(key);
  PyList_Append(seq, s);
  Py_DECREF(s);
}

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, PyObject *public_interface, const char *name, PyObject *obj) {   
  PyDict_SetItemString(d, name, obj);
  Py_DECREF(obj);
  if (public_interface)
    SwigPyBuiltin_AddPublicSymbol(public_interface, name);
}

#else

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, const char *name, PyObject *obj) {   
  PyDict_SetItemString(d, name, obj);
  Py_DECREF(obj);                            
}

#endif

/* Append a value to the result obj */

SWIGINTERN PyObject*
SWIG_Python_AppendOutput(PyObject* result, PyObject* obj) {
  if (!result) {
    result = obj;
  } else if (result == Py_None) {
    Py_DECREF(result);
    result = obj;
  } else {
    if (!PyList_Check(result)) {
      PyObject *o2 = result;
      result = PyList_New(1);
      PyList_SetItem(result, 0, o2);
    }
    PyList_Append(result,obj);
    Py_DECREF(obj);
  }
  return result;
}

/* Unpack the argument tuple */

SWIGINTERN Py_ssize_t
SWIG_Python_UnpackTuple(PyObject *args, const char *name, Py_ssize_t min, Py_ssize_t max, PyObject **objs)
{
  if (!args) {
    if (!min && !max) {
      return 1;
    } else {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got none", 
		   name, (min == max ? "" : "at least "), (int)min);
      return 0;
    }
  }  
  if (!PyTuple_Check(args)) {
    if (min <= 1 && max >= 1) {
      Py_ssize_t i;
      objs[0] = args;
      for (i = 1; i < max; ++i) {
	objs[i] = 0;
      }
      return 2;
    }
    PyErr_SetString(PyExc_SystemError, "UnpackTuple() argument list is not a tuple");
    return 0;
  } else {
    Py_ssize_t l = PyTuple_GET_SIZE(args);
    if (l < min) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at least "), (int)min, (int)l);
      return 0;
    } else if (l > max) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at most "), (int)max, (int)l);
      return 0;
    } else {
      Py_ssize_t i;
      for (i = 0; i < l; ++i) {
	objs[i] = PyTuple_GET_ITEM(args, i);
      }
      for (; l < max; ++l) {
	objs[l] = 0;
      }
      return i + 1;
    }    
  }
}

SWIGINTERN int
SWIG_Python_CheckNoKeywords(PyObject *kwargs, const char *name) {
  int no_kwargs = 1;
  if (kwargs) {
    assert(PyDict_Check(kwargs));
    if (PyDict_Size(kwargs) > 0) {
      PyErr_Format(PyExc_TypeError, "%s() does not take keyword arguments", name);
      no_kwargs = 0;
    }
  }
  return no_kwargs;
}

/* A functor is a function object with one single object argument */
#define SWIG_Python_CallFunctor(functor, obj)	        PyObject_CallFunctionObjArgs(functor, obj, NULL);

/*
  Helper for static pointer initialization for both C and C++ code, for example
  static PyObject *SWIG_STATIC_POINTER(MyVar) = NewSomething(...);
*/
#ifdef __cplusplus
#define SWIG_STATIC_POINTER(var)  var
#else
#define SWIG_STATIC_POINTER(var)  var = 0; if (!var) var
#endif

/* -----------------------------------------------------------------------------
 * Pointer declarations
 * ----------------------------------------------------------------------------- */

/* Flags for new pointer objects */
#define SWIG_POINTER_NOSHADOW       (SWIG_POINTER_OWN      << 1)
#define SWIG_POINTER_NEW            (SWIG_POINTER_NOSHADOW | SWIG_POINTER_OWN)

#define SWIG_POINTER_IMPLICIT_CONV  (SWIG_POINTER_DISOWN   << 1)

#define SWIG_BUILTIN_TP_INIT	    (SWIG_POINTER_OWN << 2)
#define SWIG_BUILTIN_INIT	    (SWIG_BUILTIN_TP_INIT | SWIG_POINTER_OWN)

#ifdef __cplusplus
extern "C" {
#endif

/* The python void return value */

SWIGRUNTIMEINLINE PyObject * 
SWIG_Py_Void(void)
{
  PyObject *none = Py_None;
  Py_INCREF(none);
  return none;
}

/* SwigPyClientData */

typedef struct {
  PyObject *klass;
  PyObject *newraw;
  PyObject *newargs;
  PyObject *destroy;
  int delargs;
  int implicitconv;
  PyTypeObject *pytype;
} SwigPyClientData;

SWIGRUNTIMEINLINE int 
SWIG_Python_CheckImplicit(swig_type_info *ty)
{
  SwigPyClientData *data = (SwigPyClientData *)ty->clientdata;
  int fail = data ? data->implicitconv : 0;
  if (fail)
    PyErr_SetString(PyExc_TypeError, "Implicit conversion is prohibited for explicit constructors.");
  return fail;
}

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_ExceptionType(swig_type_info *desc) {
  SwigPyClientData *data = desc ? (SwigPyClientData *) desc->clientdata : 0;
  PyObject *klass = data ? data->klass : 0;
  return (klass ? klass : PyExc_RuntimeError);
}


SWIGRUNTIME SwigPyClientData * 
SwigPyClientData_New(PyObject* obj)
{
  if (!obj) {
    return 0;
  } else {
    SwigPyClientData *data = (SwigPyClientData *)malloc(sizeof(SwigPyClientData));
    /* the klass element */
    data->klass = obj;
    Py_INCREF(data->klass);
    /* the newraw method and newargs arguments used to create a new raw instance */
    if (PyClass_Check(obj)) {
      data->newraw = 0;
      data->newargs = obj;
      Py_INCREF(obj);
    } else {
      data->newraw = PyObject_GetAttrString(data->klass, "__new__");
      if (data->newraw) {
	Py_INCREF(data->newraw);
	data->newargs = PyTuple_New(1);
	PyTuple_SetItem(data->newargs, 0, obj);
      } else {
	data->newargs = obj;
      }
      Py_INCREF(data->newargs);
    }
    /* the destroy method, aka as the C++ delete method */
    data->destroy = PyObject_GetAttrString(data->klass, "__swig_destroy__");
    if (PyErr_Occurred()) {
      PyErr_Clear();
      data->destroy = 0;
    }
    if (data->destroy) {
      int flags;
      Py_INCREF(data->destroy);
      flags = PyCFunction_GET_FLAGS(data->destroy);
      data->delargs = !(flags & (METH_O));
    } else {
      data->delargs = 0;
    }
    data->implicitconv = 0;
    data->pytype = 0;
    return data;
  }
}

SWIGRUNTIME void 
SwigPyClientData_Del(SwigPyClientData *data) {
  Py_XDECREF(data->newraw);
  Py_XDECREF(data->newargs);
  Py_XDECREF(data->destroy);
}

/* =============== SwigPyObject =====================*/

typedef struct {
  PyObject_HEAD
  void *ptr;
  swig_type_info *ty;
  int own;
  PyObject *next;
#ifdef SWIGPYTHON_BUILTIN
  PyObject *dict;
#endif
} SwigPyObject;


#ifdef SWIGPYTHON_BUILTIN

SWIGRUNTIME PyObject *
SwigPyObject_get___dict__(PyObject *v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;

  if (!sobj->dict)
    sobj->dict = PyDict_New();

  Py_INCREF(sobj->dict);
  return sobj->dict;
}

#endif

SWIGRUNTIME PyObject *
SwigPyObject_long(SwigPyObject *v)
{
  return PyLong_FromVoidPtr(v->ptr);
}

SWIGRUNTIME PyObject *
SwigPyObject_format(const char* fmt, SwigPyObject *v)
{
  PyObject *res = NULL;
  PyObject *args = PyTuple_New(1);
  if (args) {
    if (PyTuple_SetItem(args, 0, SwigPyObject_long(v)) == 0) {
      PyObject *ofmt = SWIG_Python_str_FromChar(fmt);
      if (ofmt) {
#if PY_VERSION_HEX >= 0x03000000
	res = PyUnicode_Format(ofmt,args);
#else
	res = PyString_Format(ofmt,args);
#endif
	Py_DECREF(ofmt);
      }
      Py_DECREF(args);
    }
  }
  return res;
}

SWIGRUNTIME PyObject *
SwigPyObject_oct(SwigPyObject *v)
{
  return SwigPyObject_format("%o",v);
}

SWIGRUNTIME PyObject *
SwigPyObject_hex(SwigPyObject *v)
{
  return SwigPyObject_format("%x",v);
}

SWIGRUNTIME PyObject *
SwigPyObject_repr(SwigPyObject *v)
{
  const char *name = SWIG_TypePrettyName(v->ty);
  PyObject *repr = SWIG_Python_str_FromFormat("<Swig Object of type '%s' at %p>", (name ? name : "unknown"), (void *)v);
  if (v->next) {
    PyObject *nrep = SwigPyObject_repr((SwigPyObject *)v->next);
# if PY_VERSION_HEX >= 0x03000000
    PyObject *joined = PyUnicode_Concat(repr, nrep);
    Py_DecRef(repr);
    Py_DecRef(nrep);
    repr = joined;
# else
    PyString_ConcatAndDel(&repr,nrep);
# endif
  }
  return repr;  
}

/* We need a version taking two PyObject* parameters so it's a valid
 * PyCFunction to use in swigobject_methods[]. */
SWIGRUNTIME PyObject *
SwigPyObject_repr2(PyObject *v, PyObject *SWIGUNUSEDPARM(args))
{
  return SwigPyObject_repr((SwigPyObject*)v);
}

SWIGRUNTIME int
SwigPyObject_compare(SwigPyObject *v, SwigPyObject *w)
{
  void *i = v->ptr;
  void *j = w->ptr;
  return (i < j) ? -1 : ((i > j) ? 1 : 0);
}

/* Added for Python 3.x, would it also be useful for Python 2.x? */
SWIGRUNTIME PyObject*
SwigPyObject_richcompare(SwigPyObject *v, SwigPyObject *w, int op)
{
  PyObject* res;
  if( op != Py_EQ && op != Py_NE ) {
    Py_INCREF(Py_NotImplemented);
    return Py_NotImplemented;
  }
  res = PyBool_FromLong( (SwigPyObject_compare(v, w)==0) == (op == Py_EQ) ? 1 : 0);
  return res;  
}


SWIGRUNTIME PyTypeObject* SwigPyObject_TypeOnce(void);

#ifdef SWIGPYTHON_BUILTIN
static swig_type_info *SwigPyObject_stype = 0;
SWIGRUNTIME PyTypeObject*
SwigPyObject_type(void) {
    SwigPyClientData *cd;
    assert(SwigPyObject_stype);
    cd = (SwigPyClientData*) SwigPyObject_stype->clientdata;
    assert(cd);
    assert(cd->pytype);
    return cd->pytype;
}
#else
SWIGRUNTIME PyTypeObject*
SwigPyObject_type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigPyObject_TypeOnce();
  return type;
}
#endif

SWIGRUNTIMEINLINE int
SwigPyObject_Check(PyObject *op) {
#ifdef SWIGPYTHON_BUILTIN
  PyTypeObject *target_tp = SwigPyObject_type();
  if (PyType_IsSubtype(op->ob_type, target_tp))
    return 1;
  return (strcmp(op->ob_type->tp_name, "SwigPyObject") == 0);
#else
  return (Py_TYPE(op) == SwigPyObject_type())
    || (strcmp(Py_TYPE(op)->tp_name,"SwigPyObject") == 0);
#endif
}

SWIGRUNTIME PyObject *
SwigPyObject_New(void *ptr, swig_type_info *ty, int own);

SWIGRUNTIME void
SwigPyObject_dealloc(PyObject *v)
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  PyObject *next = sobj->next;
  if (sobj->own == SWIG_POINTER_OWN) {
    swig_type_info *ty = sobj->ty;
    SwigPyClientData *data = ty ? (SwigPyClientData *) ty->clientdata : 0;
    PyObject *destroy = data ? data->destroy : 0;
    if (destroy) {
      /* destroy is always a VARARGS method */
      PyObject *res;

      /* PyObject_CallFunction() has the potential to silently drop
         the active exception.  In cases of unnamed temporary
         variable or where we just finished iterating over a generator
         StopIteration will be active right now, and this needs to
         remain true upon return from SwigPyObject_dealloc.  So save
         and restore. */
      
      PyObject *type = NULL, *value = NULL, *traceback = NULL;
      PyErr_Fetch(&type, &value, &traceback);

      if (data->delargs) {
        /* we need to create a temporary object to carry the destroy operation */
        PyObject *tmp = SwigPyObject_New(sobj->ptr, ty, 0);
        res = SWIG_Python_CallFunctor(destroy, tmp);
        Py_DECREF(tmp);
      } else {
        PyCFunction meth = PyCFunction_GET_FUNCTION(destroy);
        PyObject *mself = PyCFunction_GET_SELF(destroy);
        res = ((*meth)(mself, v));
      }
      if (!res)
        PyErr_WriteUnraisable(destroy);

      PyErr_Restore(type, value, traceback);

      Py_XDECREF(res);
    } 
#if !defined(SWIG_PYTHON_SILENT_MEMLEAK)
    else {
      const char *name = SWIG_TypePrettyName(ty);
      printf("swig/python detected a memory leak of type '%s', no destructor found.\n", (name ? name : "unknown"));
    }
#endif
  } 
  Py_XDECREF(next);
  PyObject_DEL(v);
}

SWIGRUNTIME PyObject* 
SwigPyObject_append(PyObject* v, PyObject* next)
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  if (!SwigPyObject_Check(next)) {
    PyErr_SetString(PyExc_TypeError, "Attempt to append a non SwigPyObject");
    return NULL;
  }
  sobj->next = next;
  Py_INCREF(next);
  return SWIG_Py_Void();
}

SWIGRUNTIME PyObject* 
SwigPyObject_next(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  if (sobj->next) {    
    Py_INCREF(sobj->next);
    return sobj->next;
  } else {
    return SWIG_Py_Void();
  }
}

SWIGINTERN PyObject*
SwigPyObject_disown(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;
  sobj->own = 0;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
SwigPyObject_acquire(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;
  sobj->own = SWIG_POINTER_OWN;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
SwigPyObject_own(PyObject *v, PyObject *args)
{
  PyObject *val = 0;
  if (!PyArg_UnpackTuple(args, "own", 0, 1, &val)) {
    return NULL;
  } else {
    SwigPyObject *sobj = (SwigPyObject *)v;
    PyObject *obj = PyBool_FromLong(sobj->own);
    if (val) {
      if (PyObject_IsTrue(val)) {
        SwigPyObject_acquire(v,args);
      } else {
        SwigPyObject_disown(v,args);
      }
    } 
    return obj;
  }
}

static PyMethodDef
swigobject_methods[] = {
  {"disown",  SwigPyObject_disown,  METH_NOARGS,  "releases ownership of the pointer"},
  {"acquire", SwigPyObject_acquire, METH_NOARGS,  "acquires ownership of the pointer"},
  {"own",     SwigPyObject_own,     METH_VARARGS, "returns/sets ownership of the pointer"},
  {"append",  SwigPyObject_append,  METH_O,       "appends another 'this' object"},
  {"next",    SwigPyObject_next,    METH_NOARGS,  "returns the next 'this' object"},
  {"__repr__",SwigPyObject_repr2,   METH_NOARGS,  "returns object representation"},
  {0, 0, 0, 0}  
};

SWIGRUNTIME PyTypeObject*
SwigPyObject_TypeOnce(void) {
  static char swigobject_doc[] = "Swig object carries a C/C++ instance pointer";

  static PyNumberMethods SwigPyObject_as_number = {
    (binaryfunc)0, /*nb_add*/
    (binaryfunc)0, /*nb_subtract*/
    (binaryfunc)0, /*nb_multiply*/
    /* nb_divide removed in Python 3 */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc)0, /*nb_divide*/
#endif
    (binaryfunc)0, /*nb_remainder*/
    (binaryfunc)0, /*nb_divmod*/
    (ternaryfunc)0,/*nb_power*/
    (unaryfunc)0,  /*nb_negative*/
    (unaryfunc)0,  /*nb_positive*/
    (unaryfunc)0,  /*nb_absolute*/
    (inquiry)0,    /*nb_nonzero*/
    0,		   /*nb_invert*/
    0,		   /*nb_lshift*/
    0,		   /*nb_rshift*/
    0,		   /*nb_and*/
    0,		   /*nb_xor*/
    0,		   /*nb_or*/
#if PY_VERSION_HEX < 0x03000000
    0,   /*nb_coerce*/
#endif
    (unaryfunc)SwigPyObject_long, /*nb_int*/
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc)SwigPyObject_long, /*nb_long*/
#else
    0, /*nb_reserved*/
#endif
    (unaryfunc)0,                 /*nb_float*/
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc)SwigPyObject_oct,  /*nb_oct*/
    (unaryfunc)SwigPyObject_hex,  /*nb_hex*/
#endif
#if PY_VERSION_HEX >= 0x03050000 /* 3.5 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_inplace_matrix_multiply */
#elif PY_VERSION_HEX >= 0x03000000 /* 3.0 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_index, nb_inplace_divide removed */
#else
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_index */
#endif
  };

  static PyTypeObject swigpyobject_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                    /* ob_size */
#endif
      "SwigPyObject",                       /* tp_name */
      sizeof(SwigPyObject),                 /* tp_basicsize */
      0,                                    /* tp_itemsize */
      (destructor)SwigPyObject_dealloc,     /* tp_dealloc */
      0,                                    /* tp_print */
      (getattrfunc)0,                       /* tp_getattr */
      (setattrfunc)0,                       /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
      0, /* tp_reserved in 3.0.1, tp_compare in 3.0.0 but not used */
#else
      (cmpfunc)SwigPyObject_compare,        /* tp_compare */
#endif
      (reprfunc)SwigPyObject_repr,          /* tp_repr */
      &SwigPyObject_as_number,              /* tp_as_number */
      0,                                    /* tp_as_sequence */
      0,                                    /* tp_as_mapping */
      (hashfunc)0,                          /* tp_hash */
      (ternaryfunc)0,                       /* tp_call */
      0,                                    /* tp_str */
      PyObject_GenericGetAttr,              /* tp_getattro */
      0,                                    /* tp_setattro */
      0,                                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,                   /* tp_flags */
      swigobject_doc,                       /* tp_doc */
      0,                                    /* tp_traverse */
      0,                                    /* tp_clear */
      (richcmpfunc)SwigPyObject_richcompare,/* tp_richcompare */
      0,                                    /* tp_weaklistoffset */
      0,                                    /* tp_iter */
      0,                                    /* tp_iternext */
      swigobject_methods,                   /* tp_methods */
      0,                                    /* tp_members */
      0,                                    /* tp_getset */
      0,                                    /* tp_base */
      0,                                    /* tp_dict */
      0,                                    /* tp_descr_get */
      0,                                    /* tp_descr_set */
      0,                                    /* tp_dictoffset */
      0,                                    /* tp_init */
      0,                                    /* tp_alloc */
      0,                                    /* tp_new */
      0,                                    /* tp_free */
      0,                                    /* tp_is_gc */
      0,                                    /* tp_bases */
      0,                                    /* tp_mro */
      0,                                    /* tp_cache */
      0,                                    /* tp_subclasses */
      0,                                    /* tp_weaklist */
      0,                                    /* tp_del */
      0,                                    /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                    /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
      0,                                    /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
      0,                                    /* tp_print */
#endif
#ifdef COUNT_ALLOCS
      0,                                    /* tp_allocs */
      0,                                    /* tp_frees */
      0,                                    /* tp_maxalloc */
      0,                                    /* tp_prev */
      0                                     /* tp_next */
#endif
    };
    swigpyobject_type = tmp;
    type_init = 1;
    if (PyType_Ready(&swigpyobject_type) < 0)
      return NULL;
  }
  return &swigpyobject_type;
}

SWIGRUNTIME PyObject *
SwigPyObject_New(void *ptr, swig_type_info *ty, int own)
{
  SwigPyObject *sobj = PyObject_NEW(SwigPyObject, SwigPyObject_type());
  if (sobj) {
    sobj->ptr  = ptr;
    sobj->ty   = ty;
    sobj->own  = own;
    sobj->next = 0;
  }
  return (PyObject *)sobj;
}

/* -----------------------------------------------------------------------------
 * Implements a simple Swig Packed type, and use it instead of string
 * ----------------------------------------------------------------------------- */

typedef struct {
  PyObject_HEAD
  void *pack;
  swig_type_info *ty;
  size_t size;
} SwigPyPacked;

SWIGRUNTIME PyObject *
SwigPyPacked_repr(SwigPyPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))) {
    return SWIG_Python_str_FromFormat("<Swig Packed at %s%s>", result, v->ty->name);
  } else {
    return SWIG_Python_str_FromFormat("<Swig Packed %s>", v->ty->name);
  }  
}

SWIGRUNTIME PyObject *
SwigPyPacked_str(SwigPyPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))){
    return SWIG_Python_str_FromFormat("%s%s", result, v->ty->name);
  } else {
    return SWIG_Python_str_FromChar(v->ty->name);
  }  
}

SWIGRUNTIME int
SwigPyPacked_compare(SwigPyPacked *v, SwigPyPacked *w)
{
  size_t i = v->size;
  size_t j = w->size;
  int s = (i < j) ? -1 : ((i > j) ? 1 : 0);
  return s ? s : strncmp((const char *)v->pack, (const char *)w->pack, 2*v->size);
}

SWIGRUNTIME PyTypeObject* SwigPyPacked_TypeOnce(void);

SWIGRUNTIME PyTypeObject*
SwigPyPacked_type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigPyPacked_TypeOnce();
  return type;
}

SWIGRUNTIMEINLINE int
SwigPyPacked_Check(PyObject *op) {
  return ((op)->ob_type == SwigPyPacked_TypeOnce()) 
    || (strcmp((op)->ob_type->tp_name,"SwigPyPacked") == 0);
}

SWIGRUNTIME void
SwigPyPacked_dealloc(PyObject *v)
{
  if (SwigPyPacked_Check(v)) {
    SwigPyPacked *sobj = (SwigPyPacked *) v;
    free(sobj->pack);
  }
  PyObject_DEL(v);
}

SWIGRUNTIME PyTypeObject*
SwigPyPacked_TypeOnce(void) {
  static char swigpacked_doc[] = "Swig object carries a C/C++ instance pointer";
  static PyTypeObject swigpypacked_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX>=0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                    /* ob_size */
#endif
      "SwigPyPacked",                       /* tp_name */
      sizeof(SwigPyPacked),                 /* tp_basicsize */
      0,                                    /* tp_itemsize */
      (destructor)SwigPyPacked_dealloc,     /* tp_dealloc */
      0,                                    /* tp_print */
      (getattrfunc)0,                       /* tp_getattr */
      (setattrfunc)0,                       /* tp_setattr */
#if PY_VERSION_HEX>=0x03000000
      0, /* tp_reserved in 3.0.1 */
#else
      (cmpfunc)SwigPyPacked_compare,        /* tp_compare */
#endif
      (reprfunc)SwigPyPacked_repr,          /* tp_repr */
      0,                                    /* tp_as_number */
      0,                                    /* tp_as_sequence */
      0,                                    /* tp_as_mapping */
      (hashfunc)0,                          /* tp_hash */
      (ternaryfunc)0,                       /* tp_call */
      (reprfunc)SwigPyPacked_str,           /* tp_str */
      PyObject_GenericGetAttr,              /* tp_getattro */
      0,                                    /* tp_setattro */
      0,                                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,                   /* tp_flags */
      swigpacked_doc,                       /* tp_doc */
      0,                                    /* tp_traverse */
      0,                                    /* tp_clear */
      0,                                    /* tp_richcompare */
      0,                                    /* tp_weaklistoffset */
      0,                                    /* tp_iter */
      0,                                    /* tp_iternext */
      0,                                    /* tp_methods */
      0,                                    /* tp_members */
      0,                                    /* tp_getset */
      0,                                    /* tp_base */
      0,                                    /* tp_dict */
      0,                                    /* tp_descr_get */
      0,                                    /* tp_descr_set */
      0,                                    /* tp_dictoffset */
      0,                                    /* tp_init */
      0,                                    /* tp_alloc */
      0,                                    /* tp_new */
      0,                                    /* tp_free */
      0,                                    /* tp_is_gc */
      0,                                    /* tp_bases */
      0,                                    /* tp_mro */
      0,                                    /* tp_cache */
      0,                                    /* tp_subclasses */
      0,                                    /* tp_weaklist */
      0,                                    /* tp_del */
      0,                                    /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                    /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
      0,                                    /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
      0,                                    /* tp_print */
#endif
#ifdef COUNT_ALLOCS
      0,                                    /* tp_allocs */
      0,                                    /* tp_frees */
      0,                                    /* tp_maxalloc */
      0,                                    /* tp_prev */
      0                                     /* tp_next */
#endif
    };
    swigpypacked_type = tmp;
    type_init = 1;
    if (PyType_Ready(&swigpypacked_type) < 0)
      return NULL;
  }
  return &swigpypacked_type;
}

SWIGRUNTIME PyObject *
SwigPyPacked_New(void *ptr, size_t size, swig_type_info *ty)
{
  SwigPyPacked *sobj = PyObject_NEW(SwigPyPacked, SwigPyPacked_type());
  if (sobj) {
    void *pack = malloc(size);
    if (pack) {
      memcpy(pack, ptr, size);
      sobj->pack = pack;
      sobj->ty   = ty;
      sobj->size = size;
    } else {
      PyObject_DEL((PyObject *) sobj);
      sobj = 0;
    }
  }
  return (PyObject *) sobj;
}

SWIGRUNTIME swig_type_info *
SwigPyPacked_UnpackData(PyObject *obj, void *ptr, size_t size)
{
  if (SwigPyPacked_Check(obj)) {
    SwigPyPacked *sobj = (SwigPyPacked *)obj;
    if (sobj->size != size) return 0;
    memcpy(ptr, sobj->pack, size);
    return sobj->ty;
  } else {
    return 0;
  }
}

/* -----------------------------------------------------------------------------
 * pointers/data manipulation
 * ----------------------------------------------------------------------------- */

static PyObject *Swig_This_global = NULL;

SWIGRUNTIME PyObject *
SWIG_This(void)
{
  if (Swig_This_global == NULL)
    Swig_This_global = SWIG_Python_str_FromChar("this");
  return Swig_This_global;
}

/* #define SWIG_PYTHON_SLOW_GETSET_THIS */

/* TODO: I don't know how to implement the fast getset in Python 3 right now */
#if PY_VERSION_HEX>=0x03000000
#define SWIG_PYTHON_SLOW_GETSET_THIS 
#endif

SWIGRUNTIME SwigPyObject *
SWIG_Python_GetSwigThis(PyObject *pyobj) 
{
  PyObject *obj;

  if (SwigPyObject_Check(pyobj))
    return (SwigPyObject *) pyobj;

#ifdef SWIGPYTHON_BUILTIN
  (void)obj;
# ifdef PyWeakref_CheckProxy
  if (PyWeakref_CheckProxy(pyobj)) {
    pyobj = PyWeakref_GET_OBJECT(pyobj);
    if (pyobj && SwigPyObject_Check(pyobj))
      return (SwigPyObject*) pyobj;
  }
# endif
  return NULL;
#else

  obj = 0;

#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
  if (PyInstance_Check(pyobj)) {
    obj = _PyInstance_Lookup(pyobj, SWIG_This());      
  } else {
    PyObject **dictptr = _PyObject_GetDictPtr(pyobj);
    if (dictptr != NULL) {
      PyObject *dict = *dictptr;
      obj = dict ? PyDict_GetItem(dict, SWIG_This()) : 0;
    } else {
#ifdef PyWeakref_CheckProxy
      if (PyWeakref_CheckProxy(pyobj)) {
	PyObject *wobj = PyWeakref_GET_OBJECT(pyobj);
	return wobj ? SWIG_Python_GetSwigThis(wobj) : 0;
      }
#endif
      obj = PyObject_GetAttr(pyobj,SWIG_This());
      if (obj) {
	Py_DECREF(obj);
      } else {
	if (PyErr_Occurred()) PyErr_Clear();
	return 0;
      }
    }
  }
#else
  obj = PyObject_GetAttr(pyobj,SWIG_This());
  if (obj) {
    Py_DECREF(obj);
  } else {
    if (PyErr_Occurred()) PyErr_Clear();
    return 0;
  }
#endif
  if (obj && !SwigPyObject_Check(obj)) {
    /* a PyObject is called 'this', try to get the 'real this'
       SwigPyObject from it */ 
    return SWIG_Python_GetSwigThis(obj);
  }
  return (SwigPyObject *)obj;
#endif
}

/* Acquire a pointer value */

SWIGRUNTIME int
SWIG_Python_AcquirePtr(PyObject *obj, int own) {
  if (own == SWIG_POINTER_OWN) {
    SwigPyObject *sobj = SWIG_Python_GetSwigThis(obj);
    if (sobj) {
      int oldown = sobj->own;
      sobj->own = own;
      return oldown;
    }
  }
  return 0;
}

/* Convert a pointer value */

SWIGRUNTIME int
SWIG_Python_ConvertPtrAndOwn(PyObject *obj, void **ptr, swig_type_info *ty, int flags, int *own) {
  int res;
  SwigPyObject *sobj;
  int implicit_conv = (flags & SWIG_POINTER_IMPLICIT_CONV) != 0;

  if (!obj)
    return SWIG_ERROR;
  if (obj == Py_None && !implicit_conv) {
    if (ptr)
      *ptr = 0;
    return (flags & SWIG_POINTER_NO_NULL) ? SWIG_NullReferenceError : SWIG_OK;
  }

  res = SWIG_ERROR;

  sobj = SWIG_Python_GetSwigThis(obj);
  if (own)
    *own = 0;
  while (sobj) {
    void *vptr = sobj->ptr;
    if (ty) {
      swig_type_info *to = sobj->ty;
      if (to == ty) {
        /* no type cast needed */
        if (ptr) *ptr = vptr;
        break;
      } else {
        swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
        if (!tc) {
          sobj = (SwigPyObject *)sobj->next;
        } else {
          if (ptr) {
            int newmemory = 0;
            *ptr = SWIG_TypeCast(tc,vptr,&newmemory);
            if (newmemory == SWIG_CAST_NEW_MEMORY) {
              assert(own); /* badly formed typemap which will lead to a memory leak - it must set and use own to delete *ptr */
              if (own)
                *own = *own | SWIG_CAST_NEW_MEMORY;
            }
          }
          break;
        }
      }
    } else {
      if (ptr) *ptr = vptr;
      break;
    }
  }
  if (sobj) {
    if (own)
      *own = *own | sobj->own;
    if (flags & SWIG_POINTER_DISOWN) {
      sobj->own = 0;
    }
    res = SWIG_OK;
  } else {
    if (implicit_conv) {
      SwigPyClientData *data = ty ? (SwigPyClientData *) ty->clientdata : 0;
      if (data && !data->implicitconv) {
        PyObject *klass = data->klass;
        if (klass) {
          PyObject *impconv;
          data->implicitconv = 1; /* avoid recursion and call 'explicit' constructors*/
          impconv = SWIG_Python_CallFunctor(klass, obj);
          data->implicitconv = 0;
          if (PyErr_Occurred()) {
            PyErr_Clear();
            impconv = 0;
          }
          if (impconv) {
            SwigPyObject *iobj = SWIG_Python_GetSwigThis(impconv);
            if (iobj) {
              void *vptr;
              res = SWIG_Python_ConvertPtrAndOwn((PyObject*)iobj, &vptr, ty, 0, 0);
              if (SWIG_IsOK(res)) {
                if (ptr) {
                  *ptr = vptr;
                  /* transfer the ownership to 'ptr' */
                  iobj->own = 0;
                  res = SWIG_AddCast(res);
                  res = SWIG_AddNewMask(res);
                } else {
                  res = SWIG_AddCast(res);		    
                }
              }
            }
            Py_DECREF(impconv);
          }
        }
      }
      if (!SWIG_IsOK(res) && obj == Py_None) {
        if (ptr)
          *ptr = 0;
        if (PyErr_Occurred())
          PyErr_Clear();
        res = SWIG_OK;
      }
    }
  }
  return res;
}

/* Convert a function ptr value */

SWIGRUNTIME int
SWIG_Python_ConvertFunctionPtr(PyObject *obj, void **ptr, swig_type_info *ty) {
  if (!PyCFunction_Check(obj)) {
    return SWIG_ConvertPtr(obj, ptr, ty, 0);
  } else {
    void *vptr = 0;
    swig_cast_info *tc;

    /* here we get the method pointer for callbacks */
    const char *doc = (((PyCFunctionObject *)obj) -> m_ml -> ml_doc);
    const char *desc = doc ? strstr(doc, "swig_ptr: ") : 0;
    if (desc)
      desc = ty ? SWIG_UnpackVoidPtr(desc + 10, &vptr, ty->name) : 0;
    if (!desc)
      return SWIG_ERROR;
    tc = SWIG_TypeCheck(desc,ty);
    if (tc) {
      int newmemory = 0;
      *ptr = SWIG_TypeCast(tc,vptr,&newmemory);
      assert(!newmemory); /* newmemory handling not yet implemented */
    } else {
      return SWIG_ERROR;
    }
    return SWIG_OK;
  }
}

/* Convert a packed pointer value */

SWIGRUNTIME int
SWIG_Python_ConvertPacked(PyObject *obj, void *ptr, size_t sz, swig_type_info *ty) {
  swig_type_info *to = SwigPyPacked_UnpackData(obj, ptr, sz);
  if (!to) return SWIG_ERROR;
  if (ty) {
    if (to != ty) {
      /* check type cast? */
      swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
      if (!tc) return SWIG_ERROR;
    }
  }
  return SWIG_OK;
}  

/* -----------------------------------------------------------------------------
 * Create a new pointer object
 * ----------------------------------------------------------------------------- */

/*
  Create a new instance object, without calling __init__, and set the
  'this' attribute.
*/

SWIGRUNTIME PyObject* 
SWIG_Python_NewShadowInstance(SwigPyClientData *data, PyObject *swig_this)
{
  PyObject *inst = 0;
  PyObject *newraw = data->newraw;
  if (newraw) {
    inst = PyObject_Call(newraw, data->newargs, NULL);
    if (inst) {
#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
      PyObject **dictptr = _PyObject_GetDictPtr(inst);
      if (dictptr != NULL) {
	PyObject *dict = *dictptr;
	if (dict == NULL) {
	  dict = PyDict_New();
	  *dictptr = dict;
	  PyDict_SetItem(dict, SWIG_This(), swig_this);
	}
      }
#else
      if (PyObject_SetAttr(inst, SWIG_This(), swig_this) == -1) {
        Py_DECREF(inst);
        inst = 0;
      }
#endif
    }
  } else {
#if PY_VERSION_HEX >= 0x03000000
    PyObject *empty_args = PyTuple_New(0);
    if (empty_args) {
      PyObject *empty_kwargs = PyDict_New();
      if (empty_kwargs) {
        inst = ((PyTypeObject *)data->newargs)->tp_new((PyTypeObject *)data->newargs, empty_args, empty_kwargs);
        Py_DECREF(empty_kwargs);
        if (inst) {
          if (PyObject_SetAttr(inst, SWIG_This(), swig_this) == -1) {
            Py_DECREF(inst);
            inst = 0;
          } else {
            Py_TYPE(inst)->tp_flags &= ~Py_TPFLAGS_VALID_VERSION_TAG;
          }
        }
      }
      Py_DECREF(empty_args);
    }
#else
    PyObject *dict = PyDict_New();
    if (dict) {
      PyDict_SetItem(dict, SWIG_This(), swig_this);
      inst = PyInstance_NewRaw(data->newargs, dict);
      Py_DECREF(dict);
    }
#endif
  }
  return inst;
}

SWIGRUNTIME int
SWIG_Python_SetSwigThis(PyObject *inst, PyObject *swig_this)
{
#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
  PyObject **dictptr = _PyObject_GetDictPtr(inst);
  if (dictptr != NULL) {
    PyObject *dict = *dictptr;
    if (dict == NULL) {
      dict = PyDict_New();
      *dictptr = dict;
    }
    return PyDict_SetItem(dict, SWIG_This(), swig_this);
  }
#endif
  return PyObject_SetAttr(inst, SWIG_This(), swig_this);
} 


SWIGINTERN PyObject *
SWIG_Python_InitShadowInstance(PyObject *args) {
  PyObject *obj[2];
  if (!SWIG_Python_UnpackTuple(args, "swiginit", 2, 2, obj)) {
    return NULL;
  } else {
    SwigPyObject *sthis = SWIG_Python_GetSwigThis(obj[0]);
    if (sthis) {
      SwigPyObject_append((PyObject*) sthis, obj[1]);
    } else {
      if (SWIG_Python_SetSwigThis(obj[0], obj[1]) != 0)
        return NULL;
    }
    return SWIG_Py_Void();
  }
}

/* Create a new pointer object */

SWIGRUNTIME PyObject *
SWIG_Python_NewPointerObj(PyObject *self, void *ptr, swig_type_info *type, int flags) {
  SwigPyClientData *clientdata;
  PyObject * robj;
  int own;

  if (!ptr)
    return SWIG_Py_Void();

  clientdata = type ? (SwigPyClientData *)(type->clientdata) : 0;
  own = (flags & SWIG_POINTER_OWN) ? SWIG_POINTER_OWN : 0;
  if (clientdata && clientdata->pytype) {
    SwigPyObject *newobj;
    if (flags & SWIG_BUILTIN_TP_INIT) {
      newobj = (SwigPyObject*) self;
      if (newobj->ptr) {
        PyObject *next_self = clientdata->pytype->tp_alloc(clientdata->pytype, 0);
        while (newobj->next)
	  newobj = (SwigPyObject *) newobj->next;
        newobj->next = next_self;
        newobj = (SwigPyObject *)next_self;
#ifdef SWIGPYTHON_BUILTIN
        newobj->dict = 0;
#endif
      }
    } else {
      newobj = PyObject_New(SwigPyObject, clientdata->pytype);
#ifdef SWIGPYTHON_BUILTIN
      newobj->dict = 0;
#endif
    }
    if (newobj) {
      newobj->ptr = ptr;
      newobj->ty = type;
      newobj->own = own;
      newobj->next = 0;
      return (PyObject*) newobj;
    }
    return SWIG_Py_Void();
  }

  assert(!(flags & SWIG_BUILTIN_TP_INIT));

  robj = SwigPyObject_New(ptr, type, own);
  if (robj && clientdata && !(flags & SWIG_POINTER_NOSHADOW)) {
    PyObject *inst = SWIG_Python_NewShadowInstance(clientdata, robj);
    Py_DECREF(robj);
    robj = inst;
  }
  return robj;
}

/* Create a new packed object */

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_NewPackedObj(void *ptr, size_t sz, swig_type_info *type) {
  return ptr ? SwigPyPacked_New((void *) ptr, sz, type) : SWIG_Py_Void();
}

/* -----------------------------------------------------------------------------*
 *  Get type list 
 * -----------------------------------------------------------------------------*/

#ifdef SWIG_LINK_RUNTIME
void *SWIG_ReturnGlobalTypeList(void *);
#endif

SWIGRUNTIME swig_module_info *
SWIG_Python_GetModule(void *SWIGUNUSEDPARM(clientdata)) {
  static void *type_pointer = (void *)0;
  /* first check if module already created */
  if (!type_pointer) {
#ifdef SWIG_LINK_RUNTIME
    type_pointer = SWIG_ReturnGlobalTypeList((void *)0);
#else
    type_pointer = PyCapsule_Import(SWIGPY_CAPSULE_NAME, 0);
    if (PyErr_Occurred()) {
      PyErr_Clear();
      type_pointer = (void *)0;
    }
#endif
  }
  return (swig_module_info *) type_pointer;
}

SWIGRUNTIME void
SWIG_Python_DestroyModule(PyObject *obj)
{
  swig_module_info *swig_module = (swig_module_info *) PyCapsule_GetPointer(obj, SWIGPY_CAPSULE_NAME);
  swig_type_info **types = swig_module->types;
  size_t i;
  for (i =0; i < swig_module->size; ++i) {
    swig_type_info *ty = types[i];
    if (ty->owndata) {
      SwigPyClientData *data = (SwigPyClientData *) ty->clientdata;
      if (data) SwigPyClientData_Del(data);
    }
  }
  Py_DECREF(SWIG_This());
  Swig_This_global = NULL;
}

SWIGRUNTIME void
SWIG_Python_SetModule(swig_module_info *swig_module) {
#if PY_VERSION_HEX >= 0x03000000
 /* Add a dummy module object into sys.modules */
  PyObject *module = PyImport_AddModule("swig_runtime_data" SWIG_RUNTIME_VERSION);
#else
  static PyMethodDef swig_empty_runtime_method_table[] = { {NULL, NULL, 0, NULL} }; /* Sentinel */
  PyObject *module = Py_InitModule("swig_runtime_data" SWIG_RUNTIME_VERSION, swig_empty_runtime_method_table);
#endif
  PyObject *pointer = PyCapsule_New((void *) swig_module, SWIGPY_CAPSULE_NAME, SWIG_Python_DestroyModule);
  if (pointer && module) {
    PyModule_AddObject(module, "type_pointer_capsule" SWIG_TYPE_TABLE_NAME, pointer);
  } else {
    Py_XDECREF(pointer);
  }
}

/* The python cached type query */
SWIGRUNTIME PyObject *
SWIG_Python_TypeCache(void) {
  static PyObject *SWIG_STATIC_POINTER(cache) = PyDict_New();
  return cache;
}

SWIGRUNTIME swig_type_info *
SWIG_Python_TypeQuery(const char *type)
{
  PyObject *cache = SWIG_Python_TypeCache();
  PyObject *key = SWIG_Python_str_FromChar(type); 
  PyObject *obj = PyDict_GetItem(cache, key);
  swig_type_info *descriptor;
  if (obj) {
    descriptor = (swig_type_info *) PyCapsule_GetPointer(obj, NULL);
  } else {
    swig_module_info *swig_module = SWIG_GetModule(0);
    descriptor = SWIG_TypeQueryModule(swig_module, swig_module, type);
    if (descriptor) {
      obj = PyCapsule_New((void*) descriptor, NULL, NULL);
      PyDict_SetItem(cache, key, obj);
      Py_DECREF(obj);
    }
  }
  Py_DECREF(key);
  return descriptor;
}

/* 
   For backward compatibility only
*/
#define SWIG_POINTER_EXCEPTION  0
#define SWIG_arg_fail(arg)      SWIG_Python_ArgFail(arg)
#define SWIG_MustGetPtr(p, type, argnum, flags)  SWIG_Python_MustGetPtr(p, type, argnum, flags)

SWIGRUNTIME int
SWIG_Python_AddErrMesg(const char* mesg, int infront)
{  
  if (PyErr_Occurred()) {
    PyObject *type = 0;
    PyObject *value = 0;
    PyObject *traceback = 0;
    PyErr_Fetch(&type, &value, &traceback);
    if (value) {
      PyObject *old_str = PyObject_Str(value);
      const char *tmp = SWIG_Python_str_AsChar(old_str);
      const char *errmesg = tmp ? tmp : "Invalid error message";
      Py_XINCREF(type);
      PyErr_Clear();
      if (infront) {
	PyErr_Format(type, "%s %s", mesg, errmesg);
      } else {
	PyErr_Format(type, "%s %s", errmesg, mesg);
      }
      SWIG_Python_str_DelForPy3(tmp);
      Py_DECREF(old_str);
    }
    return 1;
  } else {
    return 0;
  }
}
  
SWIGRUNTIME int
SWIG_Python_ArgFail(int argnum)
{
  if (PyErr_Occurred()) {
    /* add information about failing argument */
    char mesg[256];
    PyOS_snprintf(mesg, sizeof(mesg), "argument number %d:", argnum);
    return SWIG_Python_AddErrMesg(mesg, 1);
  } else {
    return 0;
  }
}

SWIGRUNTIMEINLINE const char *
SwigPyObject_GetDesc(PyObject *self)
{
  SwigPyObject *v = (SwigPyObject *)self;
  swig_type_info *ty = v ? v->ty : 0;
  return ty ? ty->str : "";
}

SWIGRUNTIME void
SWIG_Python_TypeError(const char *type, PyObject *obj)
{
  if (type) {
#if defined(SWIG_COBJECT_TYPES)
    if (obj && SwigPyObject_Check(obj)) {
      const char *otype = (const char *) SwigPyObject_GetDesc(obj);
      if (otype) {
	PyErr_Format(PyExc_TypeError, "a '%s' is expected, 'SwigPyObject(%s)' is received",
		     type, otype);
	return;
      }
    } else 
#endif      
    {
      const char *otype = (obj ? obj->ob_type->tp_name : 0); 
      if (otype) {
	PyObject *str = PyObject_Str(obj);
	const char *cstr = str ? SWIG_Python_str_AsChar(str) : 0;
	if (cstr) {
	  PyErr_Format(PyExc_TypeError, "a '%s' is expected, '%s(%s)' is received",
		       type, otype, cstr);
          SWIG_Python_str_DelForPy3(cstr);
	} else {
	  PyErr_Format(PyExc_TypeError, "a '%s' is expected, '%s' is received",
		       type, otype);
	}
	Py_XDECREF(str);
	return;
      }
    }   
    PyErr_Format(PyExc_TypeError, "a '%s' is expected", type);
  } else {
    PyErr_Format(PyExc_TypeError, "unexpected type is received");
  }
}


/* Convert a pointer value, signal an exception on a type mismatch */
SWIGRUNTIME void *
SWIG_Python_MustGetPtr(PyObject *obj, swig_type_info *ty, int SWIGUNUSEDPARM(argnum), int flags) {
  void *result;
  if (SWIG_Python_ConvertPtr(obj, &result, ty, flags) == -1) {
    PyErr_Clear();
#if SWIG_POINTER_EXCEPTION
    if (flags) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(ty), obj);
      SWIG_Python_ArgFail(argnum);
    }
#endif
  }
  return result;
}

#ifdef SWIGPYTHON_BUILTIN
SWIGRUNTIME int
SWIG_Python_NonDynamicSetAttr(PyObject *obj, PyObject *name, PyObject *value) {
  PyTypeObject *tp = obj->ob_type;
  PyObject *descr;
  PyObject *encoded_name;
  descrsetfunc f;
  int res = -1;

# ifdef Py_USING_UNICODE
  if (PyString_Check(name)) {
    name = PyUnicode_Decode(PyString_AsString(name), PyString_Size(name), NULL, NULL);
    if (!name)
      return -1;
  } else if (!PyUnicode_Check(name))
# else
  if (!PyString_Check(name))
# endif
  {
    PyErr_Format(PyExc_TypeError, "attribute name must be string, not '%.200s'", name->ob_type->tp_name);
    return -1;
  } else {
    Py_INCREF(name);
  }

  if (!tp->tp_dict) {
    if (PyType_Ready(tp) < 0)
      goto done;
  }

  descr = _PyType_Lookup(tp, name);
  f = NULL;
  if (descr != NULL)
    f = descr->ob_type->tp_descr_set;
  if (!f) {
    if (PyString_Check(name)) {
      encoded_name = name;
      Py_INCREF(name);
    } else {
      encoded_name = PyUnicode_AsUTF8String(name);
      if (!encoded_name)
        return -1;
    }
    PyErr_Format(PyExc_AttributeError, "'%.100s' object has no attribute '%.200s'", tp->tp_name, PyString_AsString(encoded_name));
    Py_DECREF(encoded_name);
  } else {
    res = f(descr, obj, value);
  }
  
  done:
  Py_DECREF(name);
  return res;
}
#endif


#ifdef __cplusplus
}
#endif

#ifdef __cplusplus
extern "C" {
#endif

SWIGINTERN Py_hash_t
SwigPyObject_hash(PyObject *obj) {
  SwigPyObject *sobj = (SwigPyObject *)obj;
  void *ptr = sobj->ptr;
  return (Py_hash_t)ptr;
}

SWIGINTERN Py_hash_t
SWIG_PyNumber_AsPyHash(PyObject *obj) {
  Py_hash_t result = -1;
#if PY_VERSION_HEX < 0x03020000
  if (PyInt_Check(obj))
    result = PyInt_AsLong(obj);
  else if (PyLong_Check(obj))
    result = PyLong_AsLong(obj);
#else
  if (PyNumber_Check(obj))
    result = PyNumber_AsSsize_t(obj, NULL);
#endif
  else
    PyErr_Format(PyExc_TypeError, "Wrong type for hash function");
  return PyErr_Occurred() ? -1 : result;
}

SWIGINTERN int
SwigPyBuiltin_BadInit(PyObject *self, PyObject *SWIGUNUSEDPARM(args), PyObject *SWIGUNUSEDPARM(kwds)) {
  PyErr_Format(PyExc_TypeError, "Cannot create new instances of type '%.300s'", self->ob_type->tp_name);
  return -1;
}

SWIGINTERN void
SwigPyBuiltin_BadDealloc(PyObject *obj) {
  SwigPyObject *sobj = (SwigPyObject *)obj;
  if (sobj->own) {
    PyErr_Format(PyExc_TypeError, "Swig detected a memory leak in type '%.300s': no callable destructor found.", obj->ob_type->tp_name);
  }
}

typedef struct {
  PyCFunction get;
  PyCFunction set;
} SwigPyGetSet;

SWIGINTERN PyObject *
SwigPyBuiltin_GetterClosure (PyObject *obj, void *closure) {
  SwigPyGetSet *getset;
  PyObject *tuple, *result;
  if (!closure)
    return SWIG_Py_Void();
  getset = (SwigPyGetSet *)closure;
  if (!getset->get)
    return SWIG_Py_Void();
  tuple = PyTuple_New(0);
  assert(tuple);
  result = (*getset->get)(obj, tuple);
  Py_DECREF(tuple);
  return result;
}

SWIGINTERN PyObject *
SwigPyBuiltin_FunpackGetterClosure (PyObject *obj, void *closure) {
  SwigPyGetSet *getset;
  PyObject *result;
  if (!closure)
    return SWIG_Py_Void();
  getset = (SwigPyGetSet *)closure;
  if (!getset->get)
    return SWIG_Py_Void();
  result = (*getset->get)(obj, NULL);
  return result;
}

SWIGINTERN int
SwigPyBuiltin_SetterClosure (PyObject *obj, PyObject *val, void *closure) {
  SwigPyGetSet *getset;
  PyObject *tuple, *result;
  if (!closure) {
    PyErr_Format(PyExc_TypeError, "Missing getset closure");
    return -1;
  }
  getset = (SwigPyGetSet *)closure;
  if (!getset->set) {
    PyErr_Format(PyExc_TypeError, "Illegal member variable assignment in type '%.300s'", obj->ob_type->tp_name);
    return -1;
  }
  tuple = PyTuple_New(1);
  assert(tuple);
  Py_INCREF(val);
  PyTuple_SET_ITEM(tuple, 0, val);
  result = (*getset->set)(obj, tuple);
  Py_DECREF(tuple);
  Py_XDECREF(result);
  return result ? 0 : -1;
}

SWIGINTERN int
SwigPyBuiltin_FunpackSetterClosure (PyObject *obj, PyObject *val, void *closure) {
  SwigPyGetSet *getset;
  PyObject *result;
  if (!closure) {
    PyErr_Format(PyExc_TypeError, "Missing getset closure");
    return -1;
  }
  getset = (SwigPyGetSet *)closure;
  if (!getset->set) {
    PyErr_Format(PyExc_TypeError, "Illegal member variable assignment in type '%.300s'", obj->ob_type->tp_name);
    return -1;
  }
  result = (*getset->set)(obj, val);
  Py_XDECREF(result);
  return result ? 0 : -1;
}

SWIGINTERN void
SwigPyStaticVar_dealloc(PyDescrObject *descr) {
  PyObject_GC_UnTrack(descr);
  Py_XDECREF(PyDescr_TYPE(descr));
  Py_XDECREF(PyDescr_NAME(descr));
  PyObject_GC_Del(descr);
}

SWIGINTERN PyObject *
SwigPyStaticVar_repr(PyGetSetDescrObject *descr) {
#if PY_VERSION_HEX >= 0x03000000

  return PyUnicode_FromFormat("<class attribute '%S' of type '%s'>", PyDescr_NAME(descr), PyDescr_TYPE(descr)->tp_name);
#else
  return PyString_FromFormat("<class attribute '%s' of type '%s'>", PyString_AsString(PyDescr_NAME(descr)), PyDescr_TYPE(descr)->tp_name);
#endif
}

SWIGINTERN int
SwigPyStaticVar_traverse(PyObject *self, visitproc visit, void *arg) {
  PyDescrObject *descr;
  descr = (PyDescrObject *)self;
  Py_VISIT((PyObject*) PyDescr_TYPE(descr));
  return 0;
}

SWIGINTERN PyObject *
SwigPyStaticVar_get(PyGetSetDescrObject *descr, PyObject *obj, PyObject *SWIGUNUSEDPARM(type)) {
  if (descr->d_getset->get != NULL)
    return descr->d_getset->get(obj, descr->d_getset->closure);
#if PY_VERSION_HEX >= 0x03000000
  PyErr_Format(PyExc_AttributeError, "attribute '%.300S' of '%.100s' objects is not readable", PyDescr_NAME(descr), PyDescr_TYPE(descr)->tp_name);
#else
  PyErr_Format(PyExc_AttributeError, "attribute '%.300s' of '%.100s' objects is not readable", PyString_AsString(PyDescr_NAME(descr)), PyDescr_TYPE(descr)->tp_name);
#endif
  return NULL;
}

SWIGINTERN int
SwigPyStaticVar_set(PyGetSetDescrObject *descr, PyObject *obj, PyObject *value) {
  if (descr->d_getset->set != NULL)
    return descr->d_getset->set(obj, value, descr->d_getset->closure);
#if PY_VERSION_HEX >= 0x03000000
  PyErr_Format(PyExc_AttributeError, "attribute '%.300S' of '%.100s' objects is not writable", PyDescr_NAME(descr), PyDescr_TYPE(descr)->tp_name);
#else
  PyErr_Format(PyExc_AttributeError, "attribute '%.300s' of '%.100s' objects is not writable", PyString_AsString(PyDescr_NAME(descr)), PyDescr_TYPE(descr)->tp_name);
#endif
  return -1;
}

SWIGINTERN int
SwigPyObjectType_setattro(PyObject *typeobject, PyObject *name, PyObject *value) {
  PyObject *attribute;
  PyTypeObject *type;
  descrsetfunc local_set;

  assert(PyType_Check(typeobject));
  type = (PyTypeObject *)typeobject;
  attribute = _PyType_Lookup(type, name);
  if (attribute != NULL) {
    /* Implement descriptor functionality, if any */
    local_set = attribute->ob_type->tp_descr_set;
    if (local_set != NULL)
      return local_set(attribute, (PyObject *)type, value);
#if PY_VERSION_HEX >= 0x03000000
    PyErr_Format(PyExc_AttributeError, "cannot modify read-only attribute '%.50s.%.400S'", type->tp_name, name);
#else 
    PyErr_Format(PyExc_AttributeError, "cannot modify read-only attribute '%.50s.%.400s'", type->tp_name, PyString_AS_STRING(name));
#endif
  } else {
#if PY_VERSION_HEX >= 0x03000000
    PyErr_Format(PyExc_AttributeError, "type '%.50s' has no attribute '%.400S'", type->tp_name, name);
#else
    PyErr_Format(PyExc_AttributeError, "type '%.50s' has no attribute '%.400s'", type->tp_name, PyString_AS_STRING(name));
#endif
  }

  return -1;
}

SWIGINTERN PyTypeObject*
SwigPyStaticVar_Type(void) {
  static PyTypeObject staticvar_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
      PyVarObject_HEAD_INIT(&PyType_Type, 0)
#else
      PyObject_HEAD_INIT(&PyType_Type)
      0,                                        /* ob_size */
#endif
      "swig_static_var_getset_descriptor",      /* tp_name */
      sizeof(PyGetSetDescrObject),              /* tp_basicsize */
      0,                                        /* tp_itemsize */
      (destructor)SwigPyStaticVar_dealloc,      /* tp_dealloc */
      0,                                        /* tp_print */
      0,                                        /* tp_getattr */
      0,                                        /* tp_setattr */
      0,                                        /* tp_compare */
      (reprfunc)SwigPyStaticVar_repr,           /* tp_repr */
      0,                                        /* tp_as_number */
      0,                                        /* tp_as_sequence */
      0,                                        /* tp_as_mapping */
      0,                                        /* tp_hash */
      0,                                        /* tp_call */
      0,                                        /* tp_str */
      PyObject_GenericGetAttr,                  /* tp_getattro */
      0,                                        /* tp_setattro */
      0,                                        /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_CLASS, /* tp_flags */
      0,                                        /* tp_doc */
      SwigPyStaticVar_traverse,                 /* tp_traverse */
      0,                                        /* tp_clear */
      0,                                        /* tp_richcompare */
      0,                                        /* tp_weaklistoffset */
      0,                                        /* tp_iter */
      0,                                        /* tp_iternext */
      0,                                        /* tp_methods */
      0,                                        /* tp_members */
      0,                                        /* tp_getset */
      0,                                        /* tp_base */
      0,                                        /* tp_dict */
      (descrgetfunc)SwigPyStaticVar_get,        /* tp_descr_get */
      (descrsetfunc)SwigPyStaticVar_set,        /* tp_descr_set */
      0,                                        /* tp_dictoffset */
      0,                                        /* tp_init */
      0,                                        /* tp_alloc */
      0,                                        /* tp_new */
      0,                                        /* tp_free */
      0,                                        /* tp_is_gc */
      0,                                        /* tp_bases */
      0,                                        /* tp_mro */
      0,                                        /* tp_cache */
      0,                                        /* tp_subclasses */
      0,                                        /* tp_weaklist */
      0,                                        /* tp_del */
      0,                                        /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                        /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
      0,                                        /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
      0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
      0,                                        /* tp_allocs */
      0,                                        /* tp_frees */
      0,                                        /* tp_maxalloc */
      0,                                        /* tp_prev */
      0                                         /* tp_next */
#endif
    };
    staticvar_type = tmp;
    type_init = 1;
    if (PyType_Ready(&staticvar_type) < 0)
      return NULL;
  }
  return &staticvar_type;
}

SWIGINTERN PyTypeObject*
SwigPyObjectType(void) {
  static char swigpyobjecttype_doc[] = "Metaclass for SWIG wrapped types";
  static PyTypeObject swigpyobjecttype_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
      PyVarObject_HEAD_INIT(&PyType_Type, 0)
#else
      PyObject_HEAD_INIT(&PyType_Type)
      0,                                        /* ob_size */
#endif
      "SwigPyObjectType",                       /* tp_name */
      PyType_Type.tp_basicsize,                 /* tp_basicsize */
      0,                                        /* tp_itemsize */
      0,                                        /* tp_dealloc */
      0,                                        /* tp_print */
      0,                                        /* tp_getattr */
      0,                                        /* tp_setattr */
      0,                                        /* tp_compare */
      0,                                        /* tp_repr */
      0,                                        /* tp_as_number */
      0,                                        /* tp_as_sequence */
      0,                                        /* tp_as_mapping */
      0,                                        /* tp_hash */
      0,                                        /* tp_call */
      0,                                        /* tp_str */
      0,                                        /* tp_getattro */
      SwigPyObjectType_setattro,                /* tp_setattro */
      0,                                        /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_CLASS, /* tp_flags */
      swigpyobjecttype_doc,                     /* tp_doc */
      0,                                        /* tp_traverse */
      0,                                        /* tp_clear */
      0,                                        /* tp_richcompare */
      0,                                        /* tp_weaklistoffset */
      0,                                        /* tp_iter */
      0,                                        /* tp_iternext */
      0,                                        /* tp_methods */
      0,                                        /* tp_members */
      0,                                        /* tp_getset */
      0,                                        /* tp_base */
      0,                                        /* tp_dict */
      0,                                        /* tp_descr_get */
      0,                                        /* tp_descr_set */
      0,                                        /* tp_dictoffset */
      0,                                        /* tp_init */
      0,                                        /* tp_alloc */
      0,                                        /* tp_new */
      0,                                        /* tp_free */
      0,                                        /* tp_is_gc */
      0,                                        /* tp_bases */
      0,                                        /* tp_mro */
      0,                                        /* tp_cache */
      0,                                        /* tp_subclasses */
      0,                                        /* tp_weaklist */
      0,                                        /* tp_del */
      0,                                        /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                        /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
      0,                                        /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
      0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
      0,                                        /* tp_allocs */
      0,                                        /* tp_frees */
      0,                                        /* tp_maxalloc */
      0,                                        /* tp_prev */
      0                                         /* tp_next */
#endif
    };
    swigpyobjecttype_type = tmp;
    type_init = 1;
    swigpyobjecttype_type.tp_base = &PyType_Type;
    if (PyType_Ready(&swigpyobjecttype_type) < 0)
      return NULL;
  }
  return &swigpyobjecttype_type;
}

SWIGINTERN PyGetSetDescrObject *
SwigPyStaticVar_new_getset(PyTypeObject *type, PyGetSetDef *getset) {

  PyGetSetDescrObject *descr;
  descr = (PyGetSetDescrObject *)PyType_GenericAlloc(SwigPyStaticVar_Type(), 0);
  assert(descr);
  Py_XINCREF(type);
  PyDescr_TYPE(descr) = type;
  PyDescr_NAME(descr) = PyString_InternFromString(getset->name);
  descr->d_getset = getset;
  if (PyDescr_NAME(descr) == NULL) {
    Py_DECREF(descr);
    descr = NULL;
  }
  return descr;
}

SWIGINTERN void
SwigPyBuiltin_InitBases (PyTypeObject *type, PyTypeObject **bases) {
  Py_ssize_t base_count = 0;
  PyTypeObject **b;
  PyObject *tuple;
  Py_ssize_t i;

  if (!bases[0]) {
    bases[0] = SwigPyObject_type();
    bases[1] = NULL;
  }
  type->tp_base = bases[0];
  Py_INCREF((PyObject *)bases[0]);
  for (b = bases; *b != NULL; ++b)
    ++base_count;
  tuple = PyTuple_New(base_count);
  for (i = 0; i < base_count; ++i) {
    Py_INCREF((PyObject *)bases[i]);
    PyTuple_SET_ITEM(tuple, i, (PyObject *)bases[i]);
  }
  type->tp_bases = tuple;
}

SWIGINTERN PyObject *
SwigPyBuiltin_ThisClosure (PyObject *self, void *SWIGUNUSEDPARM(closure)) {
  PyObject *result;
  result = (PyObject *)SWIG_Python_GetSwigThis(self);
  Py_XINCREF(result);
  return result;
}

SWIGINTERN void
SwigPyBuiltin_SetMetaType (PyTypeObject *type, PyTypeObject *metatype)
{
#if PY_VERSION_HEX >= 0x03000000
    type->ob_base.ob_base.ob_type = metatype;
#else
    type->ob_type = metatype;
#endif
}


/* Start of callback function macros for use in PyTypeObject */

typedef PyObject *(*SwigPyWrapperFunction)(PyObject *, PyObject *);

#define SWIGPY_UNARYFUNC_CLOSURE(wrapper)		\
SWIGINTERN PyObject *					\
wrapper##_unaryfunc_closure(PyObject *a) {		\
  return SwigPyBuiltin_unaryfunc_closure(wrapper, a);	\
}
SWIGINTERN PyObject *
SwigPyBuiltin_unaryfunc_closure(SwigPyWrapperFunction wrapper, PyObject *a) {
  return wrapper(a, NULL);
}

#define SWIGPY_DESTRUCTOR_CLOSURE(wrapper)			\
SWIGINTERN void							\
wrapper##_destructor_closure(PyObject *a) {			\
    SwigPyBuiltin_destructor_closure(wrapper, #wrapper, a);	\
}
SWIGINTERN void
SwigPyBuiltin_destructor_closure(SwigPyWrapperFunction wrapper, const char *wrappername, PyObject *a) {
  SwigPyObject *sobj;
  sobj = (SwigPyObject *)a;
  Py_XDECREF(sobj->dict);
  if (sobj->own) {
    PyObject *o;
    PyObject *type = 0, *value = 0, *traceback = 0;
    PyErr_Fetch(&type, &value, &traceback);
    o = wrapper(a, NULL);
    if (!o) {
      PyObject *deallocname = PyString_FromString(wrappername);
      PyErr_WriteUnraisable(deallocname);
      Py_DECREF(deallocname);
    }
    PyErr_Restore(type, value, traceback);
    Py_XDECREF(o);
  }
  if (PyType_IS_GC(a->ob_type)) {
    PyObject_GC_Del(a);
  } else {
    PyObject_Del(a);
  }
}

#define SWIGPY_INQUIRY_CLOSURE(wrapper)			\
SWIGINTERN int						\
wrapper##_inquiry_closure(PyObject *a) {		\
  return SwigPyBuiltin_inquiry_closure(wrapper, a);	\
}
SWIGINTERN int
SwigPyBuiltin_inquiry_closure(SwigPyWrapperFunction wrapper, PyObject *a) {
  PyObject *pyresult;
  int result;
  pyresult = wrapper(a, NULL);
  result = pyresult && PyObject_IsTrue(pyresult) ? 1 : 0;
  Py_XDECREF(pyresult);
  return result;
}

#define SWIGPY_GETITERFUNC_CLOSURE(wrapper)		\
SWIGINTERN PyObject *					\
wrapper##_getiterfunc_closure(PyObject *a) {		\
  return SwigPyBuiltin_getiterfunc_closure(wrapper, a);	\
}
SWIGINTERN PyObject *
SwigPyBuiltin_getiterfunc_closure(SwigPyWrapperFunction wrapper, PyObject *a) {
  return wrapper(a, NULL);
}

#define SWIGPY_BINARYFUNC_CLOSURE(wrapper)			\
SWIGINTERN PyObject *						\
wrapper##_binaryfunc_closure(PyObject *a, PyObject *b) {	\
  return SwigPyBuiltin_binaryfunc_closure(wrapper, a, b);	\
}
SWIGINTERN PyObject *
SwigPyBuiltin_binaryfunc_closure(SwigPyWrapperFunction wrapper, PyObject *a, PyObject *b) {
  PyObject *tuple, *result;
  tuple = PyTuple_New(1);
  assert(tuple);
  Py_INCREF(b);
  PyTuple_SET_ITEM(tuple, 0, b);
  result = wrapper(a, tuple);
  Py_DECREF(tuple);
  return result;
}

typedef ternaryfunc ternarycallfunc;

#define SWIGPY_TERNARYFUNC_CLOSURE(wrapper)				\
SWIGINTERN PyObject *							\
wrapper##_ternaryfunc_closure(PyObject *a, PyObject *b, PyObject *c) {	\
  return SwigPyBuiltin_ternaryfunc_closure(wrapper, a, b, c);		\
}
SWIGINTERN PyObject *
SwigPyBuiltin_ternaryfunc_closure(SwigPyWrapperFunction wrapper, PyObject *a, PyObject *b, PyObject *c) {
  PyObject *tuple, *result;
  tuple = PyTuple_New(2);
  assert(tuple);
  Py_INCREF(b);
  PyTuple_SET_ITEM(tuple, 0, b);
  Py_INCREF(c);
  PyTuple_SET_ITEM(tuple, 1, c);
  result = wrapper(a, tuple);
  Py_DECREF(tuple);
  return result;
}

#define SWIGPY_TERNARYCALLFUNC_CLOSURE(wrapper)					\
SWIGINTERN PyObject *								\
wrapper##_ternarycallfunc_closure(PyObject *a, PyObject *b, PyObject *c) {	\
  return SwigPyBuiltin_ternarycallfunc_closure(wrapper, a, b, c);		\
}
SWIGINTERN PyObject *
SwigPyBuiltin_ternarycallfunc_closure(SwigPyWrapperFunction wrapper, PyObject *a, PyObject *b, PyObject *c) {
  (void) c;
  return wrapper(a, b);
}

#define SWIGPY_LENFUNC_CLOSURE(wrapper)			\
SWIGINTERN Py_ssize_t					\
wrapper##_lenfunc_closure(PyObject *a) {		\
  return SwigPyBuiltin_lenfunc_closure(wrapper, a);	\
}
SWIGINTERN Py_ssize_t
SwigPyBuiltin_lenfunc_closure(SwigPyWrapperFunction wrapper, PyObject *a) {
  PyObject *resultobj;
  Py_ssize_t result;
  resultobj = wrapper(a, NULL);
  result = PyNumber_AsSsize_t(resultobj, NULL);
  Py_DECREF(resultobj);
  return result;
}

#define SWIGPY_SSIZESSIZEARGFUNC_CLOSURE(wrapper)				\
SWIGINTERN PyObject *								\
wrapper##_ssizessizeargfunc_closure(PyObject *a, Py_ssize_t b, Py_ssize_t c) {	\
  return SwigPyBuiltin_ssizessizeargfunc_closure(wrapper, a, b, c);		\
}
SWIGINTERN PyObject *
SwigPyBuiltin_ssizessizeargfunc_closure(SwigPyWrapperFunction wrapper, PyObject *a, Py_ssize_t b, Py_ssize_t c) {
  PyObject *tuple, *result;
  tuple = PyTuple_New(2);
  assert(tuple);
  PyTuple_SET_ITEM(tuple, 0, _PyLong_FromSsize_t(b));
  PyTuple_SET_ITEM(tuple, 1, _PyLong_FromSsize_t(c));
  result = wrapper(a, tuple);
  Py_DECREF(tuple);
  return result;
}

#define SWIGPY_SSIZESSIZEOBJARGPROC_CLOSURE(wrapper)						\
SWIGINTERN int											\
wrapper##_ssizessizeobjargproc_closure(PyObject *a, Py_ssize_t b, Py_ssize_t c, PyObject *d) {	\
  return SwigPyBuiltin_ssizessizeobjargproc_closure(wrapper, a, b, c, d);			\
}
SWIGINTERN int
SwigPyBuiltin_ssizessizeobjargproc_closure(SwigPyWrapperFunction wrapper, PyObject *a, Py_ssize_t b, Py_ssize_t c, PyObject *d) {
  PyObject *tuple, *resultobj;
  int result;
  tuple = PyTuple_New(d ? 3 : 2);
  assert(tuple);
  PyTuple_SET_ITEM(tuple, 0, _PyLong_FromSsize_t(b));
  PyTuple_SET_ITEM(tuple, 1, _PyLong_FromSsize_t(c));
  if (d) {
    Py_INCREF(d);
    PyTuple_SET_ITEM(tuple, 2, d);
  }
  resultobj = wrapper(a, tuple);
  result = resultobj ? 0 : -1;
  Py_DECREF(tuple);
  Py_XDECREF(resultobj);
  return result;
}

#define SWIGPY_SSIZEARGFUNC_CLOSURE(wrapper)				\
SWIGINTERN PyObject *							\
wrapper##_ssizeargfunc_closure(PyObject *a, Py_ssize_t b) {		\
  return SwigPyBuiltin_funpack_ssizeargfunc_closure(wrapper, a, b);	\
}
SWIGINTERN PyObject *
SwigPyBuiltin_funpack_ssizeargfunc_closure(SwigPyWrapperFunction wrapper, PyObject *a, Py_ssize_t b) {
  PyObject *tuple, *result;
  tuple = PyTuple_New(1);
  assert(tuple);
  PyTuple_SET_ITEM(tuple, 0, _PyLong_FromSsize_t(b));
  result = wrapper(a, tuple);
  Py_DECREF(tuple);
  return result;
}

#define SWIGPY_FUNPACK_SSIZEARGFUNC_CLOSURE(wrapper)			\
SWIGINTERN PyObject *							\
wrapper##_ssizeargfunc_closure(PyObject *a, Py_ssize_t b) {		\
  return SwigPyBuiltin_ssizeargfunc_closure(wrapper, a, b);		\
}
SWIGINTERN PyObject *
SwigPyBuiltin_ssizeargfunc_closure(SwigPyWrapperFunction wrapper, PyObject *a, Py_ssize_t b) {
  PyObject *arg, *result;
  arg = _PyLong_FromSsize_t(b);
  result = wrapper(a, arg);
  Py_DECREF(arg);
  return result;
}

#define SWIGPY_SSIZEOBJARGPROC_CLOSURE(wrapper)					\
SWIGINTERN int									\
wrapper##_ssizeobjargproc_closure(PyObject *a, Py_ssize_t b, PyObject *c) {	\
  return SwigPyBuiltin_ssizeobjargproc_closure(wrapper, a, b, c);		\
}
SWIGINTERN int
SwigPyBuiltin_ssizeobjargproc_closure(SwigPyWrapperFunction wrapper, PyObject *a, Py_ssize_t b, PyObject *c) {
  PyObject *tuple, *resultobj;
  int result;
  tuple = PyTuple_New(2);
  assert(tuple);
  PyTuple_SET_ITEM(tuple, 0, _PyLong_FromSsize_t(b));
  Py_INCREF(c);
  PyTuple_SET_ITEM(tuple, 1, c);
  resultobj = wrapper(a, tuple);
  result = resultobj ? 0 : -1;
  Py_XDECREF(resultobj);
  Py_DECREF(tuple);
  return result;
}

#define SWIGPY_OBJOBJPROC_CLOSURE(wrapper)			\
SWIGINTERN int							\
wrapper##_objobjproc_closure(PyObject *a, PyObject *b) {	\
  return SwigPyBuiltin_objobjproc_closure(wrapper, a, b);	\
}
SWIGINTERN int
SwigPyBuiltin_objobjproc_closure(SwigPyWrapperFunction wrapper, PyObject *a, PyObject *b) {
  int result;
  PyObject *pyresult;
  PyObject *tuple;
  tuple = PyTuple_New(1);
  assert(tuple);
  Py_INCREF(b);
  PyTuple_SET_ITEM(tuple, 0, b);
  pyresult = wrapper(a, tuple);
  result = pyresult ? (PyObject_IsTrue(pyresult) ? 1 : 0) : -1;
  Py_XDECREF(pyresult);
  Py_DECREF(tuple);
  return result;
}

#define SWIGPY_FUNPACK_OBJOBJPROC_CLOSURE(wrapper)			\
SWIGINTERN int								\
wrapper##_objobjproc_closure(PyObject *a, PyObject *b) {		\
  return SwigPyBuiltin_funpack_objobjproc_closure(wrapper, a, b);	\
}
SWIGINTERN int
SwigPyBuiltin_funpack_objobjproc_closure(SwigPyWrapperFunction wrapper, PyObject *a, PyObject *b) {
  int result;
  PyObject *pyresult;
  pyresult = wrapper(a, b);
  result = pyresult ? (PyObject_IsTrue(pyresult) ? 1 : 0) : -1;
  Py_XDECREF(pyresult);
  return result;
}

#define SWIGPY_OBJOBJARGPROC_CLOSURE(wrapper)					\
SWIGINTERN int									\
wrapper##_objobjargproc_closure(PyObject *a, PyObject *b, PyObject *c) {	\
  return SwigPyBuiltin_objobjargproc_closure(wrapper, a, b, c);			\
}
SWIGINTERN int
SwigPyBuiltin_objobjargproc_closure(SwigPyWrapperFunction wrapper, PyObject *a, PyObject *b, PyObject *c) {
  PyObject *tuple, *resultobj;
  int result;
  tuple = PyTuple_New(c ? 2 : 1);
  assert(tuple);
  Py_INCREF(b);
  PyTuple_SET_ITEM(tuple, 0, b);
  if (c) {
    Py_INCREF(c);
    PyTuple_SET_ITEM(tuple, 1, c);
  }
  resultobj = wrapper(a, tuple);
  result = resultobj ? 0 : -1;
  Py_XDECREF(resultobj);
  Py_DECREF(tuple);
  return result;
}

#define SWIGPY_REPRFUNC_CLOSURE(wrapper)		\
SWIGINTERN PyObject *					\
wrapper##_reprfunc_closure(PyObject *a) {		\
  return SwigPyBuiltin_reprfunc_closure(wrapper, a);	\
}
SWIGINTERN PyObject *
SwigPyBuiltin_reprfunc_closure(SwigPyWrapperFunction wrapper, PyObject *a) {
  return wrapper(a, NULL);
}

#define SWIGPY_HASHFUNC_CLOSURE(wrapper)		\
SWIGINTERN Py_hash_t					\
wrapper##_hashfunc_closure(PyObject *a) {		\
  return SwigPyBuiltin_hashfunc_closure(wrapper, a);	\
}
SWIGINTERN Py_hash_t
SwigPyBuiltin_hashfunc_closure(SwigPyWrapperFunction wrapper, PyObject *a) {
  PyObject *pyresult;
  Py_hash_t result;
  pyresult = wrapper(a, NULL);
  if (!pyresult)
    return -1;
  result = SWIG_PyNumber_AsPyHash(pyresult);
  Py_DECREF(pyresult);
  return result;
}

#define SWIGPY_ITERNEXTFUNC_CLOSURE(wrapper)		\
SWIGINTERN PyObject *					\
wrapper##_iternextfunc_closure(PyObject *a) {		\
  return SwigPyBuiltin_iternextfunc_closure(wrapper, a);\
}
SWIGINTERN PyObject *
SwigPyBuiltin_iternextfunc_closure(SwigPyWrapperFunction wrapper, PyObject *a) {
  return wrapper(a, NULL);
}

/* End of callback function macros for use in PyTypeObject */

#ifdef __cplusplus
}
#endif




#define SWIG_exception_fail(code, msg) do { SWIG_Error(code, msg); SWIG_fail; } while(0) 

#define SWIG_contract_assert(expr, msg) if (!(expr)) { SWIG_Error(SWIG_RuntimeError, msg); SWIG_fail; } else 



#ifdef __cplusplus
extern "C" {
#endif

/* Method creation and docstring support functions */

SWIGINTERN PyMethodDef *SWIG_PythonGetProxyDoc(const char *name);
SWIGINTERN PyObject *SWIG_PyInstanceMethod_New(PyObject *SWIGUNUSEDPARM(self), PyObject *func);
SWIGINTERN PyObject *SWIG_PyStaticMethod_New(PyObject *SWIGUNUSEDPARM(self), PyObject *func);

#ifdef __cplusplus
}
#endif


  #define SWIG_exception(code, msg) do { SWIG_Error(code, msg); SWIG_fail;; } while(0) 


/* -------- TYPES TABLE (BEGIN) -------- */

#define SWIGTYPE_p_SwigPyObject swig_types[0]
#define SWIGTYPE_p_char swig_types[1]
#define SWIGTYPE_p_egglib__AbstractBaseAlphabet swig_types[2]
#define SWIGTYPE_p_egglib__AbstractTemplateAlphabetT_char_p_t swig_types[3]
#define SWIGTYPE_p_egglib__AbstractTemplateAlphabetT_char_t swig_types[4]
#define SWIGTYPE_p_egglib__AbstractTemplateAlphabetT_int_t swig_types[5]
#define SWIGTYPE_p_egglib__AlleleStatus swig_types[6]
#define SWIGTYPE_p_egglib__BaseFormatter swig_types[7]
#define SWIGTYPE_p_egglib__BedParser swig_types[8]
#define SWIGTYPE_p_egglib__CaseInsensitiveCharAlphabet swig_types[9]
#define SWIGTYPE_p_egglib__CaseInsensitiveStringAlphabet swig_types[10]
#define SWIGTYPE_p_egglib__Coalesce swig_types[11]
#define SWIGTYPE_p_egglib__CodonAlphabet swig_types[12]
#define SWIGTYPE_p_egglib__ComputeV swig_types[13]
#define SWIGTYPE_p_egglib__CustomStringAlphabet swig_types[14]
#define SWIGTYPE_p_egglib__DNAAlphabet swig_types[15]
#define SWIGTYPE_p_egglib__DataHolder swig_types[16]
#define SWIGTYPE_p_egglib__Diversity1 swig_types[17]
#define SWIGTYPE_p_egglib__Diversity2 swig_types[18]
#define SWIGTYPE_p_egglib__EHH swig_types[19]
#define SWIGTYPE_p_egglib__Event swig_types[20]
#define SWIGTYPE_p_egglib__Export swig_types[21]
#define SWIGTYPE_p_egglib__FastaFormatter swig_types[22]
#define SWIGTYPE_p_egglib__FastaParser swig_types[23]
#define SWIGTYPE_p_egglib__Feature swig_types[24]
#define SWIGTYPE_p_egglib__FiniteAlphabetT_char_p_t swig_types[25]
#define SWIGTYPE_p_egglib__FiniteAlphabetT_char_t swig_types[26]
#define SWIGTYPE_p_egglib__FiniteAlphabetT_int_t swig_types[27]
#define SWIGTYPE_p_egglib__FreqHolder swig_types[28]
#define SWIGTYPE_p_egglib__FreqSet swig_types[29]
#define SWIGTYPE_p_egglib__GFF3 swig_types[30]
#define SWIGTYPE_p_egglib__GeneticCode swig_types[31]
#define SWIGTYPE_p_egglib__Genotypes swig_types[32]
#define SWIGTYPE_p_egglib__Haplotypes swig_types[33]
#define SWIGTYPE_p_egglib__IntersperseAlign swig_types[34]
#define SWIGTYPE_p_egglib__MatrixLD swig_types[35]
#define SWIGTYPE_p_egglib__Migration swig_types[36]
#define SWIGTYPE_p_egglib__Node swig_types[37]
#define SWIGTYPE_p_egglib__PairwiseLD swig_types[38]
#define SWIGTYPE_p_egglib__ParalogPi swig_types[39]
#define SWIGTYPE_p_egglib__Params swig_types[40]
#define SWIGTYPE_p_egglib__RangeAlphabet swig_types[41]
#define SWIGTYPE_p_egglib__Rd swig_types[42]
#define SWIGTYPE_p_egglib__SiteDiversity swig_types[43]
#define SWIGTYPE_p_egglib__SiteHolder swig_types[44]
#define SWIGTYPE_p_egglib__StringAlphabet swig_types[45]
#define SWIGTYPE_p_egglib__StructureCluster swig_types[46]
#define SWIGTYPE_p_egglib__StructureHolder swig_types[47]
#define SWIGTYPE_p_egglib__StructureIndiv swig_types[48]
#define SWIGTYPE_p_egglib__StructurePopulation swig_types[49]
#define SWIGTYPE_p_egglib__Tree swig_types[50]
#define SWIGTYPE_p_egglib__VcfIndex swig_types[51]
#define SWIGTYPE_p_egglib__VcfParser swig_types[52]
#define SWIGTYPE_p_egglib__VcfWindowBED swig_types[53]
#define SWIGTYPE_p_egglib__VcfWindowBase swig_types[54]
#define SWIGTYPE_p_egglib__VcfWindowSlider swig_types[55]
#define SWIGTYPE_p_egglib__VcfWindowSliderBase swig_types[56]
#define SWIGTYPE_p_egglib__VcfWindowSliderPerSite swig_types[57]
#define SWIGTYPE_p_egglib__VectorInt swig_types[58]
#define SWIGTYPE_p_egglib__WSite swig_types[59]
#define SWIGTYPE_p_egglib__vcf__Alt swig_types[60]
#define SWIGTYPE_p_egglib__vcf__Filter swig_types[61]
#define SWIGTYPE_p_egglib__vcf__FlagInfo swig_types[62]
#define SWIGTYPE_p_egglib__vcf__Format swig_types[63]
#define SWIGTYPE_p_egglib__vcf__Info swig_types[64]
#define SWIGTYPE_p_egglib__vcf__Meta swig_types[65]
#define SWIGTYPE_p_egglib__vcf__SampleInfo swig_types[66]
#define SWIGTYPE_p_egglib__vcf__StringInfo swig_types[67]
#define SWIGTYPE_p_egglib__vcf__TypeInfoT_char_p_t swig_types[68]
#define SWIGTYPE_p_egglib__vcf__TypeInfoT_char_t swig_types[69]
#define SWIGTYPE_p_egglib__vcf__TypeInfoT_double_t swig_types[70]
#define SWIGTYPE_p_egglib__vcf__TypeInfoT_int_t swig_types[71]
#define SWIGTYPE_p_int swig_types[72]
#define SWIGTYPE_p_std__istream swig_types[73]
#define SWIGTYPE_p_std__string swig_types[74]
#define SWIGTYPE_p_unsigned_int swig_types[75]
static swig_type_info *swig_types[77];
static swig_module_info swig_module = {swig_types, 76, 0, 0, 0, 0};
#define SWIG_TypeQuery(name) SWIG_TypeQueryModule(&swig_module, &swig_module, name)
#define SWIG_MangledTypeQuery(name) SWIG_MangledTypeQueryModule(&swig_module, &swig_module, name)

/* -------- TYPES TABLE (END) -------- */

#ifdef SWIG_TypeQuery
# undef SWIG_TypeQuery
#endif
#define SWIG_TypeQuery SWIG_Python_TypeQuery

/*-----------------------------------------------
              @(target):= _eggwrapper.so
  ------------------------------------------------*/
#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_init    PyInit__eggwrapper

#else
#  define SWIG_init    init_eggwrapper

#endif
#define SWIG_name    "_eggwrapper"

#define SWIGVERSION 0x040002 
#define SWIG_VERSION SWIGVERSION


#define SWIG_as_voidptr(a) const_cast< void * >(static_cast< const void * >(a)) 
#define SWIG_as_voidptrptr(a) ((void)SWIG_as_voidptr(*a),reinterpret_cast< void** >(a)) 


#include <stdexcept>


namespace swig {
  class SwigPtr_PyObject {
  protected:
    PyObject *_obj;

  public:
    SwigPtr_PyObject() :_obj(0)
    {
    }

    SwigPtr_PyObject(const SwigPtr_PyObject& item) : _obj(item._obj)
    {
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      Py_XINCREF(_obj);      
      SWIG_PYTHON_THREAD_END_BLOCK;
    }
    
    SwigPtr_PyObject(PyObject *obj, bool initial_ref = true) :_obj(obj)
    {
      if (initial_ref) {
        SWIG_PYTHON_THREAD_BEGIN_BLOCK;
        Py_XINCREF(_obj);
        SWIG_PYTHON_THREAD_END_BLOCK;
      }
    }
    
    SwigPtr_PyObject & operator=(const SwigPtr_PyObject& item) 
    {
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      Py_XINCREF(item._obj);
      Py_XDECREF(_obj);
      _obj = item._obj;
      SWIG_PYTHON_THREAD_END_BLOCK;
      return *this;      
    }
    
    ~SwigPtr_PyObject() 
    {
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      Py_XDECREF(_obj);
      SWIG_PYTHON_THREAD_END_BLOCK;
    }
    
    operator PyObject *() const
    {
      return _obj;
    }

    PyObject *operator->() const
    {
      return _obj;
    }
  };
}


namespace swig {
  struct SwigVar_PyObject : SwigPtr_PyObject {
    SwigVar_PyObject(PyObject* obj = 0) : SwigPtr_PyObject(obj, false) { }
    
    SwigVar_PyObject & operator = (PyObject* obj)
    {
      Py_XDECREF(_obj);
      _obj = obj;
      return *this;      
    }
  };
}


#include <stddef.h>


#include <string>


// this is to suppress warnings from eggwrapper.cpp
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wmaybe-uninitialized"


#include "egglib.hpp"
#include "AlleleStatus.hpp"
#include "Alphabet.hpp"
#include "Coalesce.hpp"
#include "DataHolder.hpp"
#include "Rd.hpp"
#include "Diversity1.hpp"
#include "Diversity2.hpp"
#include "Fs.hpp"
#include "ParalogPi.hpp"
#include "EHH.hpp"
#include "Fasta.hpp"
#include "Export.hpp"
#include "GeneticCode.hpp"
#include "GFF3.hpp"
#include "Haplotypes.hpp"
#include "LD.hpp"
#include "Lineage.hpp"
#include "Params.hpp"
#include "SiteDiversity.hpp"
#include "SiteHolder.hpp"
#include "FreqHolder.hpp"
#include "Structure.hpp"
#include "Tree.hpp"
#include "VCF.hpp"
#include "ComputeV.hpp"
#include "VcfWindow.hpp"
#include "BED.hpp"

// this is to fix an include fail of swig
#include <cstddef>


#include <limits.h>
#if !defined(SWIG_NO_LLONG_MAX)
# if !defined(LLONG_MAX) && defined(__GNUC__) && defined (__LONG_LONG_MAX__)
#   define LLONG_MAX __LONG_LONG_MAX__
#   define LLONG_MIN (-LLONG_MAX - 1LL)
#   define ULLONG_MAX (LLONG_MAX * 2ULL + 1ULL)
# endif
#endif


SWIGINTERN int
SWIG_AsVal_double (PyObject *obj, double *val)
{
  int res = SWIG_TypeError;
  if (PyFloat_Check(obj)) {
    if (val) *val = PyFloat_AsDouble(obj);
    return SWIG_OK;
#if PY_VERSION_HEX < 0x03000000
  } else if (PyInt_Check(obj)) {
    if (val) *val = (double) PyInt_AsLong(obj);
    return SWIG_OK;
#endif
  } else if (PyLong_Check(obj)) {
    double v = PyLong_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    double d = PyFloat_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = d;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      long v = PyLong_AsLong(obj);
      if (!PyErr_Occurred()) {
	if (val) *val = v;
	return SWIG_AddCast(SWIG_AddCast(SWIG_OK));
      } else {
	PyErr_Clear();
      }
    }
  }
#endif
  return res;
}


#include <float.h>


#include <math.h>


SWIGINTERNINLINE int
SWIG_CanCastAsInteger(double *d, double min, double max) {
  double x = *d;
  if ((min <= x && x <= max)) {
   double fx = floor(x);
   double cx = ceil(x);
   double rd =  ((x - fx) < 0.5) ? fx : cx; /* simple rint */
   if ((errno == EDOM) || (errno == ERANGE)) {
     errno = 0;
   } else {
     double summ, reps, diff;
     if (rd < x) {
       diff = x - rd;
     } else if (rd > x) {
       diff = rd - x;
     } else {
       return 1;
     }
     summ = rd + x;
     reps = diff/summ;
     if (reps < 8*DBL_EPSILON) {
       *d = rd;
       return 1;
     }
   }
  }
  return 0;
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_long (PyObject *obj, unsigned long *val) 
{
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(obj)) {
    long v = PyInt_AsLong(obj);
    if (v >= 0) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      return SWIG_OverflowError;
    }
  } else
#endif
  if (PyLong_Check(obj)) {
    unsigned long v = PyLong_AsUnsignedLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      return SWIG_OverflowError;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    unsigned long v = PyLong_AsUnsignedLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, 0, ULONG_MAX)) {
	if (val) *val = (unsigned long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_int (PyObject * obj, unsigned int *val)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v > UINT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = static_cast< unsigned int >(v);
    }
  }  
  return res;
}


SWIGINTERNINLINE PyObject*
  SWIG_From_unsigned_SS_int  (unsigned int value)
{
  return PyInt_FromSize_t((size_t) value);
}


SWIGINTERN int
SWIG_AsVal_long (PyObject *obj, long* val)
{
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(obj)) {
    if (val) *val = PyInt_AsLong(obj);
    return SWIG_OK;
  } else
#endif
  if (PyLong_Check(obj)) {
    long v = PyLong_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      return SWIG_OverflowError;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    long v = PyInt_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, LONG_MIN, LONG_MAX)) {
	if (val) *val = (long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_int (PyObject * obj, int *val)
{
  long v;
  int res = SWIG_AsVal_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < INT_MIN || v > INT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = static_cast< int >(v);
    }
  }  
  return res;
}


SWIGINTERNINLINE PyObject*
  SWIG_From_int  (int value)
{
  return PyInt_FromLong((long) value);
}


SWIGINTERNINLINE PyObject*
  SWIG_From_bool  (bool value)
{
  return PyBool_FromLong(value ? 1 : 0);
}


SWIGINTERN swig_type_info*
SWIG_pchar_descriptor(void)
{
  static int init = 0;
  static swig_type_info* info = 0;
  if (!init) {
    info = SWIG_TypeQuery("_p_char");
    init = 1;
  }
  return info;
}


SWIGINTERN int
SWIG_AsCharPtrAndSize(PyObject *obj, char** cptr, size_t* psize, int *alloc)
{
#if PY_VERSION_HEX>=0x03000000
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
  if (PyBytes_Check(obj))
#else
  if (PyUnicode_Check(obj))
#endif
#else  
  if (PyString_Check(obj))
#endif
  {
    char *cstr; Py_ssize_t len;
    int ret = SWIG_OK;
#if PY_VERSION_HEX>=0x03000000
#if !defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
    if (!alloc && cptr) {
        /* We can't allow converting without allocation, since the internal
           representation of string in Python 3 is UCS-2/UCS-4 but we require
           a UTF-8 representation.
           TODO(bhy) More detailed explanation */
        return SWIG_RuntimeError;
    }
    obj = PyUnicode_AsUTF8String(obj);
    if (!obj)
      return SWIG_TypeError;
    if (alloc)
      *alloc = SWIG_NEWOBJ;
#endif
    if (PyBytes_AsStringAndSize(obj, &cstr, &len) == -1)
      return SWIG_TypeError;
#else
    if (PyString_AsStringAndSize(obj, &cstr, &len) == -1)
      return SWIG_TypeError;
#endif
    if (cptr) {
      if (alloc) {
	if (*alloc == SWIG_NEWOBJ) {
	  *cptr = reinterpret_cast< char* >(memcpy(new char[len + 1], cstr, sizeof(char)*(len + 1)));
	  *alloc = SWIG_NEWOBJ;
	} else {
	  *cptr = cstr;
	  *alloc = SWIG_OLDOBJ;
	}
      } else {
#if PY_VERSION_HEX>=0x03000000
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
	*cptr = PyBytes_AsString(obj);
#else
	assert(0); /* Should never reach here with Unicode strings in Python 3 */
#endif
#else
	*cptr = SWIG_Python_str_AsChar(obj);
        if (!*cptr)
          ret = SWIG_TypeError;
#endif
      }
    }
    if (psize) *psize = len + 1;
#if PY_VERSION_HEX>=0x03000000 && !defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
    Py_XDECREF(obj);
#endif
    return ret;
  } else {
#if defined(SWIG_PYTHON_2_UNICODE)
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
#error "Cannot use both SWIG_PYTHON_2_UNICODE and SWIG_PYTHON_STRICT_BYTE_CHAR at once"
#endif
#if PY_VERSION_HEX<0x03000000
    if (PyUnicode_Check(obj)) {
      char *cstr; Py_ssize_t len;
      if (!alloc && cptr) {
        return SWIG_RuntimeError;
      }
      obj = PyUnicode_AsUTF8String(obj);
      if (!obj)
        return SWIG_TypeError;
      if (PyString_AsStringAndSize(obj, &cstr, &len) != -1) {
        if (cptr) {
          if (alloc) *alloc = SWIG_NEWOBJ;
          *cptr = reinterpret_cast< char* >(memcpy(new char[len + 1], cstr, sizeof(char)*(len + 1)));
        }
        if (psize) *psize = len + 1;

        Py_XDECREF(obj);
        return SWIG_OK;
      } else {
        Py_XDECREF(obj);
      }
    }
#endif
#endif

    swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
    if (pchar_descriptor) {
      void* vptr = 0;
      if (SWIG_ConvertPtr(obj, &vptr, pchar_descriptor, 0) == SWIG_OK) {
	if (cptr) *cptr = (char *) vptr;
	if (psize) *psize = vptr ? (strlen((char *)vptr) + 1) : 0;
	if (alloc) *alloc = SWIG_OLDOBJ;
	return SWIG_OK;
      }
    }
  }
  return SWIG_TypeError;
}





SWIGINTERNINLINE PyObject *
SWIG_FromCharPtrAndSize(const char* carray, size_t size)
{
  if (carray) {
    if (size > INT_MAX) {
      swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
      return pchar_descriptor ? 
	SWIG_InternalNewPointerObj(const_cast< char * >(carray), pchar_descriptor, 0) : SWIG_Py_Void();
    } else {
#if PY_VERSION_HEX >= 0x03000000
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
      return PyBytes_FromStringAndSize(carray, static_cast< Py_ssize_t >(size));
#else
      return PyUnicode_DecodeUTF8(carray, static_cast< Py_ssize_t >(size), "surrogateescape");
#endif
#else
      return PyString_FromStringAndSize(carray, static_cast< Py_ssize_t >(size));
#endif
    }
  } else {
    return SWIG_Py_Void();
  }
}


SWIGINTERNINLINE PyObject * 
SWIG_FromCharPtr(const char *cptr)
{ 
  return SWIG_FromCharPtrAndSize(cptr, (cptr ? strlen(cptr) : 0));
}


SWIGINTERNINLINE PyObject *
SWIG_From_char  (char c) 
{ 
  return SWIG_FromCharPtrAndSize(&c,1);
}


SWIGINTERN int
SWIG_AsCharArray(PyObject * obj, char *val, size_t size)
{ 
  char* cptr = 0; size_t csize = 0; int alloc = SWIG_OLDOBJ;
  int res = SWIG_AsCharPtrAndSize(obj, &cptr, &csize, &alloc);
  if (SWIG_IsOK(res)) {
    /* special case of single char conversion when we don't need space for NUL */
    if (size == 1 && csize == 2 && cptr && !cptr[1]) --csize;
    if (csize <= size) {
      if (val) {
	if (csize) memcpy(val, cptr, csize*sizeof(char));
	if (csize < size) memset(val + csize, 0, (size - csize)*sizeof(char));
      }
      if (alloc == SWIG_NEWOBJ) {
	delete[] cptr;
	res = SWIG_DelNewMask(res);
      }      
      return res;
    }
    if (alloc == SWIG_NEWOBJ) delete[] cptr;
  }
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_char (PyObject * obj, char *val)
{    
  int res = SWIG_AsCharArray(obj, val, 1);
  if (!SWIG_IsOK(res)) {
    long v;
    res = SWIG_AddCast(SWIG_AsVal_long (obj, &v));
    if (SWIG_IsOK(res)) {
      if ((CHAR_MIN <= v) && (v <= CHAR_MAX)) {
	if (val) *val = static_cast< char >(v);
      } else {
	res = SWIG_OverflowError;
      }
    }
  }
  return res;
}


SWIGINTERN int
SWIG_AsVal_bool (PyObject *obj, bool *val)
{
  int r;
  if (!PyBool_Check(obj))
    return SWIG_ERROR;
  r = PyObject_IsTrue(obj);
  if (r == -1)
    return SWIG_ERROR;
  if (val) *val = r ? true : false;
  return SWIG_OK;
}


  #define SWIG_From_double   PyFloat_FromDouble 


SWIGINTERNINLINE PyObject *
SWIG_From_std_string  (const std::string& s)
{
  return SWIG_FromCharPtrAndSize(s.data(), s.size());
}


  #define SWIG_From_long   PyInt_FromLong 


SWIGINTERNINLINE PyObject* 
SWIG_From_unsigned_SS_long  (unsigned long value)
{
  return (value > LONG_MAX) ?
    PyLong_FromUnsignedLong(value) : PyInt_FromLong(static_cast< long >(value));
}


SWIGINTERN int
SWIG_AsPtr_std_string (PyObject * obj, std::string **val) 
{
  char* buf = 0 ; size_t size = 0; int alloc = SWIG_OLDOBJ;
  if (SWIG_IsOK((SWIG_AsCharPtrAndSize(obj, &buf, &size, &alloc)))) {
    if (buf) {
      if (val) *val = new std::string(buf, size - 1);
      if (alloc == SWIG_NEWOBJ) delete[] buf;
      return SWIG_NEWOBJ;
    } else {
      if (val) *val = 0;
      return SWIG_OLDOBJ;
    }
  } else {
    static int init = 0;
    static swig_type_info* descriptor = 0;
    if (!init) {
      descriptor = SWIG_TypeQuery("std::string" " *");
      init = 1;
    }
    if (descriptor) {
      std::string *vptr;
      int res = SWIG_ConvertPtr(obj, (void**)&vptr, descriptor, 0);
      if (SWIG_IsOK(res) && val) *val = vptr;
      return res;
    }
  }
  return SWIG_ERROR;
}

#ifdef __cplusplus
extern "C" {
#endif
SWIGINTERN int _wrap_new_VectorInt__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::VectorInt *result = 0 ;
  
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  {
    try {
      result = (egglib::VectorInt *)new egglib::VectorInt();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__VectorInt, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_VectorInt(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VectorInt *arg1 = (egglib::VectorInt *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_VectorInt", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VectorInt, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_VectorInt" "', argument " "1"" of type '" "egglib::VectorInt *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VectorInt * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN int _wrap_new_VectorInt__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::VectorInt *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  egglib::VectorInt *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_egglib__VectorInt,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_VectorInt" "', argument " "1"" of type '" "egglib::VectorInt const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_VectorInt" "', argument " "1"" of type '" "egglib::VectorInt const &""'"); 
  }
  arg1 = reinterpret_cast< egglib::VectorInt * >(argp1);
  {
    try {
      result = (egglib::VectorInt *)new egglib::VectorInt((egglib::VectorInt const &)*arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__VectorInt, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_VectorInt(PyObject *self, PyObject *args, PyObject *kwargs) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_VectorInt")) SWIG_fail;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_VectorInt", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_VectorInt__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_egglib__VectorInt, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_VectorInt__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_VectorInt'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::VectorInt::VectorInt()\n"
    "    egglib::VectorInt::VectorInt(egglib::VectorInt const &)\n");
  return -1;
}


SWIGINTERN PyObject *_wrap_VectorInt_set_num_values(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VectorInt *arg1 = (egglib::VectorInt *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VectorInt, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VectorInt_set_num_values" "', argument " "1"" of type '" "egglib::VectorInt *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VectorInt * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VectorInt_set_num_values" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_num_values(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VectorInt_get_num_values(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VectorInt *arg1 = (egglib::VectorInt *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VectorInt_get_num_values", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VectorInt, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VectorInt_get_num_values" "', argument " "1"" of type '" "egglib::VectorInt const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VectorInt * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VectorInt const *)arg1)->get_num_values();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VectorInt_set_item(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VectorInt *arg1 = (egglib::VectorInt *) 0 ;
  unsigned int arg2 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VectorInt_set_item", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VectorInt, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VectorInt_set_item" "', argument " "1"" of type '" "egglib::VectorInt *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VectorInt * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VectorInt_set_item" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VectorInt_set_item" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  {
    try {
      (arg1)->set_item(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VectorInt_get_item(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VectorInt *arg1 = (egglib::VectorInt *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VectorInt, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VectorInt_get_item" "', argument " "1"" of type '" "egglib::VectorInt const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VectorInt * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VectorInt_get_item" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (int)((egglib::VectorInt const *)arg1)->get_item(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VectorInt_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VectorInt *arg1 = (egglib::VectorInt *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VectorInt_clear", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VectorInt, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VectorInt_clear" "', argument " "1"" of type '" "egglib::VectorInt *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VectorInt * >(argp1);
  {
    try {
      (arg1)->clear();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_VectorInt) /* defines _wrap_delete_VectorInt_destructor_closure */

SWIGINTERN PyObject *_wrap_AbstractBaseAlphabet_lock(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AbstractBaseAlphabet *arg1 = (egglib::AbstractBaseAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "AbstractBaseAlphabet_lock", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AbstractBaseAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AbstractBaseAlphabet_lock" "', argument " "1"" of type '" "egglib::AbstractBaseAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AbstractBaseAlphabet * >(argp1);
  {
    try {
      (arg1)->lock();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AbstractBaseAlphabet_is_locked(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AbstractBaseAlphabet *arg1 = (egglib::AbstractBaseAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "AbstractBaseAlphabet_is_locked", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AbstractBaseAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AbstractBaseAlphabet_is_locked" "', argument " "1"" of type '" "egglib::AbstractBaseAlphabet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AbstractBaseAlphabet * >(argp1);
  {
    try {
      result = (bool)((egglib::AbstractBaseAlphabet const *)arg1)->is_locked();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AbstractBaseAlphabet_set_name(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AbstractBaseAlphabet *arg1 = (egglib::AbstractBaseAlphabet *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AbstractBaseAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AbstractBaseAlphabet_set_name" "', argument " "1"" of type '" "egglib::AbstractBaseAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AbstractBaseAlphabet * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "AbstractBaseAlphabet_set_name" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->set_name((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_AbstractBaseAlphabet_get_name(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AbstractBaseAlphabet *arg1 = (egglib::AbstractBaseAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "AbstractBaseAlphabet_get_name", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AbstractBaseAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AbstractBaseAlphabet_get_name" "', argument " "1"" of type '" "egglib::AbstractBaseAlphabet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AbstractBaseAlphabet * >(argp1);
  {
    try {
      result = (char *)((egglib::AbstractBaseAlphabet const *)arg1)->get_name();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AbstractBaseAlphabet_num_exploitable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AbstractBaseAlphabet *arg1 = (egglib::AbstractBaseAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "AbstractBaseAlphabet_num_exploitable", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AbstractBaseAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AbstractBaseAlphabet_num_exploitable" "', argument " "1"" of type '" "egglib::AbstractBaseAlphabet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AbstractBaseAlphabet * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::AbstractBaseAlphabet const *)arg1)->num_exploitable();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AbstractBaseAlphabet_num_missing(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AbstractBaseAlphabet *arg1 = (egglib::AbstractBaseAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "AbstractBaseAlphabet_num_missing", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AbstractBaseAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AbstractBaseAlphabet_num_missing" "', argument " "1"" of type '" "egglib::AbstractBaseAlphabet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AbstractBaseAlphabet * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::AbstractBaseAlphabet const *)arg1)->num_missing();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AbstractBaseAlphabet_get_type(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AbstractBaseAlphabet *arg1 = (egglib::AbstractBaseAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "AbstractBaseAlphabet_get_type", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AbstractBaseAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AbstractBaseAlphabet_get_type" "', argument " "1"" of type '" "egglib::AbstractBaseAlphabet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AbstractBaseAlphabet * >(argp1);
  {
    try {
      result = (char *)((egglib::AbstractBaseAlphabet const *)arg1)->get_type();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AbstractBaseAlphabet_set_type(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AbstractBaseAlphabet *arg1 = (egglib::AbstractBaseAlphabet *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AbstractBaseAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AbstractBaseAlphabet_set_type" "', argument " "1"" of type '" "egglib::AbstractBaseAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AbstractBaseAlphabet * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "AbstractBaseAlphabet_set_type" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->set_type((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_AbstractBaseAlphabet_case_insensitive(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AbstractBaseAlphabet *arg1 = (egglib::AbstractBaseAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "AbstractBaseAlphabet_case_insensitive", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AbstractBaseAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AbstractBaseAlphabet_case_insensitive" "', argument " "1"" of type '" "egglib::AbstractBaseAlphabet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AbstractBaseAlphabet * >(argp1);
  {
    try {
      result = (bool)((egglib::AbstractBaseAlphabet const *)arg1)->case_insensitive();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_AbstractBaseAlphabet(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AbstractBaseAlphabet *arg1 = (egglib::AbstractBaseAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_AbstractBaseAlphabet", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AbstractBaseAlphabet, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_AbstractBaseAlphabet" "', argument " "1"" of type '" "egglib::AbstractBaseAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AbstractBaseAlphabet * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_AbstractBaseAlphabet) /* defines _wrap_delete_AbstractBaseAlphabet_destructor_closure */

SWIGINTERN PyObject *_wrap_IntBaseAlphabet_get_value(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AbstractTemplateAlphabet< int > *arg1 = (egglib::AbstractTemplateAlphabet< int > *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AbstractTemplateAlphabetT_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "IntBaseAlphabet_get_value" "', argument " "1"" of type '" "egglib::AbstractTemplateAlphabet< int > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AbstractTemplateAlphabet< int > * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "IntBaseAlphabet_get_value" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      result = (int)(arg1)->get_value(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_IntBaseAlphabet_get_code(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AbstractTemplateAlphabet< int > *arg1 = (egglib::AbstractTemplateAlphabet< int > *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AbstractTemplateAlphabetT_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "IntBaseAlphabet_get_code" "', argument " "1"" of type '" "egglib::AbstractTemplateAlphabet< int > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AbstractTemplateAlphabet< int > * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "IntBaseAlphabet_get_code" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      result = (int)(arg1)->get_code(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_IntBaseAlphabet(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AbstractTemplateAlphabet< int > *arg1 = (egglib::AbstractTemplateAlphabet< int > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_IntBaseAlphabet", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AbstractTemplateAlphabetT_int_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_IntBaseAlphabet" "', argument " "1"" of type '" "egglib::AbstractTemplateAlphabet< int > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AbstractTemplateAlphabet< int > * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_IntBaseAlphabet) /* defines _wrap_delete_IntBaseAlphabet_destructor_closure */

SWIGINTERN PyObject *_wrap_CharBaseAlphabet_get_value(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AbstractTemplateAlphabet< char > *arg1 = (egglib::AbstractTemplateAlphabet< char > *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AbstractTemplateAlphabetT_char_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CharBaseAlphabet_get_value" "', argument " "1"" of type '" "egglib::AbstractTemplateAlphabet< char > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AbstractTemplateAlphabet< char > * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CharBaseAlphabet_get_value" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      result = (char)(arg1)->get_value(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_char(static_cast< char >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CharBaseAlphabet_get_code(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AbstractTemplateAlphabet< char > *arg1 = (egglib::AbstractTemplateAlphabet< char > *) 0 ;
  char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AbstractTemplateAlphabetT_char_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CharBaseAlphabet_get_code" "', argument " "1"" of type '" "egglib::AbstractTemplateAlphabet< char > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AbstractTemplateAlphabet< char > * >(argp1);
  ecode2 = SWIG_AsVal_char(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CharBaseAlphabet_get_code" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = static_cast< char >(val2);
  {
    try {
      result = (int)(arg1)->get_code(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CharBaseAlphabet(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AbstractTemplateAlphabet< char > *arg1 = (egglib::AbstractTemplateAlphabet< char > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_CharBaseAlphabet", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AbstractTemplateAlphabetT_char_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CharBaseAlphabet" "', argument " "1"" of type '" "egglib::AbstractTemplateAlphabet< char > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AbstractTemplateAlphabet< char > * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_CharBaseAlphabet) /* defines _wrap_delete_CharBaseAlphabet_destructor_closure */

SWIGINTERN PyObject *_wrap_CStringBaseAlphabet_get_value(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AbstractTemplateAlphabet< char * > *arg1 = (egglib::AbstractTemplateAlphabet< char * > *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AbstractTemplateAlphabetT_char_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CStringBaseAlphabet_get_value" "', argument " "1"" of type '" "egglib::AbstractTemplateAlphabet< char * > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AbstractTemplateAlphabet< char * > * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CStringBaseAlphabet_get_value" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      result = (char *)(arg1)->get_value(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CStringBaseAlphabet_get_code(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AbstractTemplateAlphabet< char * > *arg1 = (egglib::AbstractTemplateAlphabet< char * > *) 0 ;
  char *arg2 = (char *) (char *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AbstractTemplateAlphabetT_char_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CStringBaseAlphabet_get_code" "', argument " "1"" of type '" "egglib::AbstractTemplateAlphabet< char * > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AbstractTemplateAlphabet< char * > * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CStringBaseAlphabet_get_code" "', argument " "2"" of type '" "char *const""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      result = (int)(arg1)->get_code(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CStringBaseAlphabet(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AbstractTemplateAlphabet< char * > *arg1 = (egglib::AbstractTemplateAlphabet< char * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_CStringBaseAlphabet", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AbstractTemplateAlphabetT_char_p_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CStringBaseAlphabet" "', argument " "1"" of type '" "egglib::AbstractTemplateAlphabet< char * > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AbstractTemplateAlphabet< char * > * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_CStringBaseAlphabet) /* defines _wrap_delete_CStringBaseAlphabet_destructor_closure */

SWIGINTERN int _wrap_new_IntAlphabet(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< int > *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_IntAlphabet")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_IntAlphabet", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::FiniteAlphabet< int > *)new egglib::FiniteAlphabet< int >();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__FiniteAlphabetT_int_t, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_IntAlphabet_get_value(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< int > *arg1 = (egglib::FiniteAlphabet< int > *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FiniteAlphabetT_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "IntAlphabet_get_value" "', argument " "1"" of type '" "egglib::FiniteAlphabet< int > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FiniteAlphabet< int > * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "IntAlphabet_get_value" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      result = (int)(arg1)->get_value(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_IntAlphabet_get_code(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< int > *arg1 = (egglib::FiniteAlphabet< int > *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FiniteAlphabetT_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "IntAlphabet_get_code" "', argument " "1"" of type '" "egglib::FiniteAlphabet< int > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FiniteAlphabet< int > * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "IntAlphabet_get_code" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      result = (int)(arg1)->get_code(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_IntAlphabet_add_exploitable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< int > *arg1 = (egglib::FiniteAlphabet< int > *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FiniteAlphabetT_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "IntAlphabet_add_exploitable" "', argument " "1"" of type '" "egglib::FiniteAlphabet< int > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FiniteAlphabet< int > * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "IntAlphabet_add_exploitable" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      (arg1)->add_exploitable(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_IntAlphabet_add_missing(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< int > *arg1 = (egglib::FiniteAlphabet< int > *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FiniteAlphabetT_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "IntAlphabet_add_missing" "', argument " "1"" of type '" "egglib::FiniteAlphabet< int > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FiniteAlphabet< int > * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "IntAlphabet_add_missing" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      (arg1)->add_missing(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_IntAlphabet_num_exploitable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< int > *arg1 = (egglib::FiniteAlphabet< int > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "IntAlphabet_num_exploitable", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FiniteAlphabetT_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "IntAlphabet_num_exploitable" "', argument " "1"" of type '" "egglib::FiniteAlphabet< int > const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FiniteAlphabet< int > * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::FiniteAlphabet< int > const *)arg1)->num_exploitable();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_IntAlphabet_num_missing(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< int > *arg1 = (egglib::FiniteAlphabet< int > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "IntAlphabet_num_missing", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FiniteAlphabetT_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "IntAlphabet_num_missing" "', argument " "1"" of type '" "egglib::FiniteAlphabet< int > const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FiniteAlphabet< int > * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::FiniteAlphabet< int > const *)arg1)->num_missing();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_IntAlphabet(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< int > *arg1 = (egglib::FiniteAlphabet< int > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_IntAlphabet", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FiniteAlphabetT_int_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_IntAlphabet" "', argument " "1"" of type '" "egglib::FiniteAlphabet< int > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FiniteAlphabet< int > * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_IntAlphabet) /* defines _wrap_delete_IntAlphabet_destructor_closure */

SWIGINTERN int _wrap_new_CharAlphabet(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< char > *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_CharAlphabet")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_CharAlphabet", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::FiniteAlphabet< char > *)new egglib::FiniteAlphabet< char >();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__FiniteAlphabetT_char_t, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_CharAlphabet_get_value(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< char > *arg1 = (egglib::FiniteAlphabet< char > *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FiniteAlphabetT_char_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CharAlphabet_get_value" "', argument " "1"" of type '" "egglib::FiniteAlphabet< char > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FiniteAlphabet< char > * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CharAlphabet_get_value" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      result = (char)(arg1)->get_value(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_char(static_cast< char >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CharAlphabet_get_code(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< char > *arg1 = (egglib::FiniteAlphabet< char > *) 0 ;
  char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FiniteAlphabetT_char_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CharAlphabet_get_code" "', argument " "1"" of type '" "egglib::FiniteAlphabet< char > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FiniteAlphabet< char > * >(argp1);
  ecode2 = SWIG_AsVal_char(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CharAlphabet_get_code" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = static_cast< char >(val2);
  {
    try {
      result = (int)(arg1)->get_code(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CharAlphabet_add_exploitable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< char > *arg1 = (egglib::FiniteAlphabet< char > *) 0 ;
  char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FiniteAlphabetT_char_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CharAlphabet_add_exploitable" "', argument " "1"" of type '" "egglib::FiniteAlphabet< char > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FiniteAlphabet< char > * >(argp1);
  ecode2 = SWIG_AsVal_char(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CharAlphabet_add_exploitable" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = static_cast< char >(val2);
  {
    try {
      (arg1)->add_exploitable(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CharAlphabet_add_missing(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< char > *arg1 = (egglib::FiniteAlphabet< char > *) 0 ;
  char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FiniteAlphabetT_char_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CharAlphabet_add_missing" "', argument " "1"" of type '" "egglib::FiniteAlphabet< char > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FiniteAlphabet< char > * >(argp1);
  ecode2 = SWIG_AsVal_char(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CharAlphabet_add_missing" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = static_cast< char >(val2);
  {
    try {
      (arg1)->add_missing(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CharAlphabet_num_exploitable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< char > *arg1 = (egglib::FiniteAlphabet< char > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "CharAlphabet_num_exploitable", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FiniteAlphabetT_char_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CharAlphabet_num_exploitable" "', argument " "1"" of type '" "egglib::FiniteAlphabet< char > const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FiniteAlphabet< char > * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::FiniteAlphabet< char > const *)arg1)->num_exploitable();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CharAlphabet_num_missing(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< char > *arg1 = (egglib::FiniteAlphabet< char > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "CharAlphabet_num_missing", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FiniteAlphabetT_char_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CharAlphabet_num_missing" "', argument " "1"" of type '" "egglib::FiniteAlphabet< char > const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FiniteAlphabet< char > * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::FiniteAlphabet< char > const *)arg1)->num_missing();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CharAlphabet(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< char > *arg1 = (egglib::FiniteAlphabet< char > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_CharAlphabet", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FiniteAlphabetT_char_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CharAlphabet" "', argument " "1"" of type '" "egglib::FiniteAlphabet< char > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FiniteAlphabet< char > * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_CharAlphabet) /* defines _wrap_delete_CharAlphabet_destructor_closure */

SWIGINTERN int _wrap_new_CStringAlphabet(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< char * > *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_CStringAlphabet")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_CStringAlphabet", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::FiniteAlphabet< char * > *)new egglib::FiniteAlphabet< char * >();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__FiniteAlphabetT_char_p_t, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_CStringAlphabet_get_value(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< char * > *arg1 = (egglib::FiniteAlphabet< char * > *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FiniteAlphabetT_char_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CStringAlphabet_get_value" "', argument " "1"" of type '" "egglib::FiniteAlphabet< char * > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FiniteAlphabet< char * > * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CStringAlphabet_get_value" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      result = (char *)(arg1)->get_value(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CStringAlphabet_get_code(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< char * > *arg1 = (egglib::FiniteAlphabet< char * > *) 0 ;
  char *arg2 = (char *) (char *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FiniteAlphabetT_char_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CStringAlphabet_get_code" "', argument " "1"" of type '" "egglib::FiniteAlphabet< char * > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FiniteAlphabet< char * > * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CStringAlphabet_get_code" "', argument " "2"" of type '" "char *const""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      result = (int)(arg1)->get_code(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CStringAlphabet_add_exploitable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< char * > *arg1 = (egglib::FiniteAlphabet< char * > *) 0 ;
  char *arg2 = (char *) (char *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FiniteAlphabetT_char_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CStringAlphabet_add_exploitable" "', argument " "1"" of type '" "egglib::FiniteAlphabet< char * > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FiniteAlphabet< char * > * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CStringAlphabet_add_exploitable" "', argument " "2"" of type '" "char *const""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->add_exploitable(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CStringAlphabet_add_missing(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< char * > *arg1 = (egglib::FiniteAlphabet< char * > *) 0 ;
  char *arg2 = (char *) (char *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FiniteAlphabetT_char_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CStringAlphabet_add_missing" "', argument " "1"" of type '" "egglib::FiniteAlphabet< char * > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FiniteAlphabet< char * > * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CStringAlphabet_add_missing" "', argument " "2"" of type '" "char *const""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->add_missing(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CStringAlphabet_num_exploitable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< char * > *arg1 = (egglib::FiniteAlphabet< char * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "CStringAlphabet_num_exploitable", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FiniteAlphabetT_char_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CStringAlphabet_num_exploitable" "', argument " "1"" of type '" "egglib::FiniteAlphabet< char * > const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FiniteAlphabet< char * > * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::FiniteAlphabet< char * > const *)arg1)->num_exploitable();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CStringAlphabet_num_missing(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< char * > *arg1 = (egglib::FiniteAlphabet< char * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "CStringAlphabet_num_missing", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FiniteAlphabetT_char_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CStringAlphabet_num_missing" "', argument " "1"" of type '" "egglib::FiniteAlphabet< char * > const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FiniteAlphabet< char * > * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::FiniteAlphabet< char * > const *)arg1)->num_missing();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CStringAlphabet(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FiniteAlphabet< char * > *arg1 = (egglib::FiniteAlphabet< char * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_CStringAlphabet", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FiniteAlphabetT_char_p_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CStringAlphabet" "', argument " "1"" of type '" "egglib::FiniteAlphabet< char * > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FiniteAlphabet< char * > * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_CStringAlphabet) /* defines _wrap_delete_CStringAlphabet_destructor_closure */

SWIGINTERN int _wrap_new_CaseInsensitiveCharAlphabet(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::CaseInsensitiveCharAlphabet *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_CaseInsensitiveCharAlphabet")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_CaseInsensitiveCharAlphabet", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::CaseInsensitiveCharAlphabet *)new egglib::CaseInsensitiveCharAlphabet();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__CaseInsensitiveCharAlphabet, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_CaseInsensitiveCharAlphabet_add_exploitable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::CaseInsensitiveCharAlphabet *arg1 = (egglib::CaseInsensitiveCharAlphabet *) 0 ;
  char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__CaseInsensitiveCharAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CaseInsensitiveCharAlphabet_add_exploitable" "', argument " "1"" of type '" "egglib::CaseInsensitiveCharAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::CaseInsensitiveCharAlphabet * >(argp1);
  ecode2 = SWIG_AsVal_char(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CaseInsensitiveCharAlphabet_add_exploitable" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = static_cast< char >(val2);
  {
    try {
      (arg1)->add_exploitable(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CaseInsensitiveCharAlphabet_add_missing(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::CaseInsensitiveCharAlphabet *arg1 = (egglib::CaseInsensitiveCharAlphabet *) 0 ;
  char arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__CaseInsensitiveCharAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CaseInsensitiveCharAlphabet_add_missing" "', argument " "1"" of type '" "egglib::CaseInsensitiveCharAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::CaseInsensitiveCharAlphabet * >(argp1);
  ecode2 = SWIG_AsVal_char(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CaseInsensitiveCharAlphabet_add_missing" "', argument " "2"" of type '" "char""'");
  } 
  arg2 = static_cast< char >(val2);
  {
    try {
      (arg1)->add_missing(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CaseInsensitiveCharAlphabet_case_insensitive(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::CaseInsensitiveCharAlphabet *arg1 = (egglib::CaseInsensitiveCharAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "CaseInsensitiveCharAlphabet_case_insensitive", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__CaseInsensitiveCharAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CaseInsensitiveCharAlphabet_case_insensitive" "', argument " "1"" of type '" "egglib::CaseInsensitiveCharAlphabet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::CaseInsensitiveCharAlphabet * >(argp1);
  {
    try {
      result = (bool)((egglib::CaseInsensitiveCharAlphabet const *)arg1)->case_insensitive();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CaseInsensitiveCharAlphabet(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::CaseInsensitiveCharAlphabet *arg1 = (egglib::CaseInsensitiveCharAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_CaseInsensitiveCharAlphabet", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__CaseInsensitiveCharAlphabet, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CaseInsensitiveCharAlphabet" "', argument " "1"" of type '" "egglib::CaseInsensitiveCharAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::CaseInsensitiveCharAlphabet * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_CaseInsensitiveCharAlphabet) /* defines _wrap_delete_CaseInsensitiveCharAlphabet_destructor_closure */

SWIGINTERN int _wrap_new_DNAAlphabet(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::DNAAlphabet *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_DNAAlphabet")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_DNAAlphabet", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::DNAAlphabet *)new egglib::DNAAlphabet();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__DNAAlphabet, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_DNAAlphabet(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DNAAlphabet *arg1 = (egglib::DNAAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_DNAAlphabet", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DNAAlphabet, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_DNAAlphabet" "', argument " "1"" of type '" "egglib::DNAAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DNAAlphabet * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_DNAAlphabet) /* defines _wrap_delete_DNAAlphabet_destructor_closure */

SWIGINTERN int _wrap_new_StringAlphabet(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::StringAlphabet *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_StringAlphabet")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_StringAlphabet", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::StringAlphabet *)new egglib::StringAlphabet();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__StringAlphabet, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_StringAlphabet_get_code(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StringAlphabet *arg1 = (egglib::StringAlphabet *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StringAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StringAlphabet_get_code" "', argument " "1"" of type '" "egglib::StringAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StringAlphabet * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "StringAlphabet_get_code" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      result = (int)(arg1)->get_code((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_StringAlphabet_add_exploitable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StringAlphabet *arg1 = (egglib::StringAlphabet *) 0 ;
  char *arg2 = (char *) (char *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StringAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StringAlphabet_add_exploitable" "', argument " "1"" of type '" "egglib::StringAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StringAlphabet * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "StringAlphabet_add_exploitable" "', argument " "2"" of type '" "char *const""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->add_exploitable(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_StringAlphabet_add_missing(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StringAlphabet *arg1 = (egglib::StringAlphabet *) 0 ;
  char *arg2 = (char *) (char *)0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StringAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StringAlphabet_add_missing" "', argument " "1"" of type '" "egglib::StringAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StringAlphabet * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "StringAlphabet_add_missing" "', argument " "2"" of type '" "char *const""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->add_missing(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_StringAlphabet_longest_length(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StringAlphabet *arg1 = (egglib::StringAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "StringAlphabet_longest_length", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StringAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StringAlphabet_longest_length" "', argument " "1"" of type '" "egglib::StringAlphabet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StringAlphabet * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::StringAlphabet const *)arg1)->longest_length();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_StringAlphabet(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StringAlphabet *arg1 = (egglib::StringAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_StringAlphabet", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StringAlphabet, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_StringAlphabet" "', argument " "1"" of type '" "egglib::StringAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StringAlphabet * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_StringAlphabet) /* defines _wrap_delete_StringAlphabet_destructor_closure */

SWIGINTERN int _wrap_new_CodonAlphabet(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::CodonAlphabet *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_CodonAlphabet")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_CodonAlphabet", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::CodonAlphabet *)new egglib::CodonAlphabet();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__CodonAlphabet, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_CodonAlphabet_get_code(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::CodonAlphabet *arg1 = (egglib::CodonAlphabet *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__CodonAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CodonAlphabet_get_code" "', argument " "1"" of type '" "egglib::CodonAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::CodonAlphabet * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CodonAlphabet_get_code" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      result = (int)(arg1)->get_code((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CodonAlphabet_get_code_from_bases(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::CodonAlphabet *arg1 = (egglib::CodonAlphabet *) 0 ;
  int arg2 ;
  int arg3 ;
  int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  int result;
  
  if (!SWIG_Python_UnpackTuple(args, "CodonAlphabet_get_code_from_bases", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__CodonAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CodonAlphabet_get_code_from_bases" "', argument " "1"" of type '" "egglib::CodonAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::CodonAlphabet * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CodonAlphabet_get_code_from_bases" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "CodonAlphabet_get_code_from_bases" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "CodonAlphabet_get_code_from_bases" "', argument " "4"" of type '" "int""'");
  } 
  arg4 = static_cast< int >(val4);
  {
    try {
      result = (int)(arg1)->get_code_from_bases(arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CodonAlphabet_num_exploitable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::CodonAlphabet *arg1 = (egglib::CodonAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "CodonAlphabet_num_exploitable", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__CodonAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CodonAlphabet_num_exploitable" "', argument " "1"" of type '" "egglib::CodonAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::CodonAlphabet * >(argp1);
  {
    try {
      result = (unsigned int)(arg1)->num_exploitable();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CodonAlphabet_num_missing(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::CodonAlphabet *arg1 = (egglib::CodonAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "CodonAlphabet_num_missing", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__CodonAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CodonAlphabet_num_missing" "', argument " "1"" of type '" "egglib::CodonAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::CodonAlphabet * >(argp1);
  {
    try {
      result = (unsigned int)(arg1)->num_missing();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CodonAlphabet_add_exploitable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::CodonAlphabet *arg1 = (egglib::CodonAlphabet *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__CodonAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CodonAlphabet_add_exploitable" "', argument " "1"" of type '" "egglib::CodonAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::CodonAlphabet * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CodonAlphabet_add_exploitable" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->add_exploitable((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_CodonAlphabet_add_missing(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::CodonAlphabet *arg1 = (egglib::CodonAlphabet *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__CodonAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CodonAlphabet_add_missing" "', argument " "1"" of type '" "egglib::CodonAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::CodonAlphabet * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "CodonAlphabet_add_missing" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->add_missing((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CodonAlphabet(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::CodonAlphabet *arg1 = (egglib::CodonAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_CodonAlphabet", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__CodonAlphabet, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CodonAlphabet" "', argument " "1"" of type '" "egglib::CodonAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::CodonAlphabet * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_CodonAlphabet) /* defines _wrap_delete_CodonAlphabet_destructor_closure */

SWIGINTERN int _wrap_new_CaseInsensitiveStringAlphabet(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::CaseInsensitiveStringAlphabet *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_CaseInsensitiveStringAlphabet")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_CaseInsensitiveStringAlphabet", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::CaseInsensitiveStringAlphabet *)new egglib::CaseInsensitiveStringAlphabet();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__CaseInsensitiveStringAlphabet, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_CaseInsensitiveStringAlphabet_case_insensitive(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::CaseInsensitiveStringAlphabet *arg1 = (egglib::CaseInsensitiveStringAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "CaseInsensitiveStringAlphabet_case_insensitive", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__CaseInsensitiveStringAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CaseInsensitiveStringAlphabet_case_insensitive" "', argument " "1"" of type '" "egglib::CaseInsensitiveStringAlphabet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::CaseInsensitiveStringAlphabet * >(argp1);
  {
    try {
      result = (bool)((egglib::CaseInsensitiveStringAlphabet const *)arg1)->case_insensitive();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CaseInsensitiveStringAlphabet(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::CaseInsensitiveStringAlphabet *arg1 = (egglib::CaseInsensitiveStringAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_CaseInsensitiveStringAlphabet", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__CaseInsensitiveStringAlphabet, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CaseInsensitiveStringAlphabet" "', argument " "1"" of type '" "egglib::CaseInsensitiveStringAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::CaseInsensitiveStringAlphabet * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_CaseInsensitiveStringAlphabet) /* defines _wrap_delete_CaseInsensitiveStringAlphabet_destructor_closure */

SWIGINTERN int _wrap_new_CustomStringAlphabet(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::CustomStringAlphabet *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_CustomStringAlphabet")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_CustomStringAlphabet", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::CustomStringAlphabet *)new egglib::CustomStringAlphabet();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__CustomStringAlphabet, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_CustomStringAlphabet(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::CustomStringAlphabet *arg1 = (egglib::CustomStringAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_CustomStringAlphabet", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__CustomStringAlphabet, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CustomStringAlphabet" "', argument " "1"" of type '" "egglib::CustomStringAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::CustomStringAlphabet * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_CustomStringAlphabet) /* defines _wrap_delete_CustomStringAlphabet_destructor_closure */

SWIGINTERN int _wrap_new_RangeAlphabet(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::RangeAlphabet *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_RangeAlphabet")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_RangeAlphabet", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::RangeAlphabet *)new egglib::RangeAlphabet();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__RangeAlphabet, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_RangeAlphabet_get_code(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::RangeAlphabet *arg1 = (egglib::RangeAlphabet *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__RangeAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "RangeAlphabet_get_code" "', argument " "1"" of type '" "egglib::RangeAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::RangeAlphabet * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "RangeAlphabet_get_code" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      result = (int)(arg1)->get_code(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_RangeAlphabet_get_value(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::RangeAlphabet *arg1 = (egglib::RangeAlphabet *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__RangeAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "RangeAlphabet_get_value" "', argument " "1"" of type '" "egglib::RangeAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::RangeAlphabet * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "RangeAlphabet_get_value" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      result = (int)(arg1)->get_value(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_RangeAlphabet_first_exploitable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::RangeAlphabet *arg1 = (egglib::RangeAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!SWIG_Python_UnpackTuple(args, "RangeAlphabet_first_exploitable", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__RangeAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "RangeAlphabet_first_exploitable" "', argument " "1"" of type '" "egglib::RangeAlphabet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::RangeAlphabet * >(argp1);
  {
    try {
      result = (int)((egglib::RangeAlphabet const *)arg1)->first_exploitable();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_RangeAlphabet_end_exploitable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::RangeAlphabet *arg1 = (egglib::RangeAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!SWIG_Python_UnpackTuple(args, "RangeAlphabet_end_exploitable", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__RangeAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "RangeAlphabet_end_exploitable" "', argument " "1"" of type '" "egglib::RangeAlphabet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::RangeAlphabet * >(argp1);
  {
    try {
      result = (int)((egglib::RangeAlphabet const *)arg1)->end_exploitable();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_RangeAlphabet_first_missing(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::RangeAlphabet *arg1 = (egglib::RangeAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!SWIG_Python_UnpackTuple(args, "RangeAlphabet_first_missing", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__RangeAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "RangeAlphabet_first_missing" "', argument " "1"" of type '" "egglib::RangeAlphabet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::RangeAlphabet * >(argp1);
  {
    try {
      result = (int)((egglib::RangeAlphabet const *)arg1)->first_missing();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_RangeAlphabet_end_missing(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::RangeAlphabet *arg1 = (egglib::RangeAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!SWIG_Python_UnpackTuple(args, "RangeAlphabet_end_missing", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__RangeAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "RangeAlphabet_end_missing" "', argument " "1"" of type '" "egglib::RangeAlphabet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::RangeAlphabet * >(argp1);
  {
    try {
      result = (int)((egglib::RangeAlphabet const *)arg1)->end_missing();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_RangeAlphabet_set_exploitable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::RangeAlphabet *arg1 = (egglib::RangeAlphabet *) 0 ;
  int arg2 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "RangeAlphabet_set_exploitable", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__RangeAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "RangeAlphabet_set_exploitable" "', argument " "1"" of type '" "egglib::RangeAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::RangeAlphabet * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "RangeAlphabet_set_exploitable" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "RangeAlphabet_set_exploitable" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  {
    try {
      (arg1)->set_exploitable(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_RangeAlphabet_set_missing(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::RangeAlphabet *arg1 = (egglib::RangeAlphabet *) 0 ;
  int arg2 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "RangeAlphabet_set_missing", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__RangeAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "RangeAlphabet_set_missing" "', argument " "1"" of type '" "egglib::RangeAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::RangeAlphabet * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "RangeAlphabet_set_missing" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "RangeAlphabet_set_missing" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  {
    try {
      (arg1)->set_missing(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_RangeAlphabet_min_value(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::RangeAlphabet *arg1 = (egglib::RangeAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!SWIG_Python_UnpackTuple(args, "RangeAlphabet_min_value", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__RangeAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "RangeAlphabet_min_value" "', argument " "1"" of type '" "egglib::RangeAlphabet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::RangeAlphabet * >(argp1);
  {
    try {
      result = (int)((egglib::RangeAlphabet const *)arg1)->min_value();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_RangeAlphabet_max_value(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::RangeAlphabet *arg1 = (egglib::RangeAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!SWIG_Python_UnpackTuple(args, "RangeAlphabet_max_value", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__RangeAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "RangeAlphabet_max_value" "', argument " "1"" of type '" "egglib::RangeAlphabet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::RangeAlphabet * >(argp1);
  {
    try {
      result = (int)((egglib::RangeAlphabet const *)arg1)->max_value();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_RangeAlphabet_num_exploitable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::RangeAlphabet *arg1 = (egglib::RangeAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "RangeAlphabet_num_exploitable", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__RangeAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "RangeAlphabet_num_exploitable" "', argument " "1"" of type '" "egglib::RangeAlphabet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::RangeAlphabet * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::RangeAlphabet const *)arg1)->num_exploitable();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_RangeAlphabet_num_missing(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::RangeAlphabet *arg1 = (egglib::RangeAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "RangeAlphabet_num_missing", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__RangeAlphabet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "RangeAlphabet_num_missing" "', argument " "1"" of type '" "egglib::RangeAlphabet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::RangeAlphabet * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::RangeAlphabet const *)arg1)->num_missing();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_RangeAlphabet(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::RangeAlphabet *arg1 = (egglib::RangeAlphabet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_RangeAlphabet", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__RangeAlphabet, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_RangeAlphabet" "', argument " "1"" of type '" "egglib::RangeAlphabet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::RangeAlphabet * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_RangeAlphabet) /* defines _wrap_delete_RangeAlphabet_destructor_closure */

SWIGINTERN int _wrap_new_DataHolder__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  bool arg1 ;
  bool val1 ;
  int ecode1 = 0 ;
  egglib::DataHolder *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_bool(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_DataHolder" "', argument " "1"" of type '" "bool""'");
  } 
  arg1 = static_cast< bool >(val1);
  {
    try {
      result = (egglib::DataHolder *)new egglib::DataHolder(arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__DataHolder, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_DataHolder__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::DataHolder *result = 0 ;
  
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  {
    try {
      result = (egglib::DataHolder *)new egglib::DataHolder();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__DataHolder, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_DataHolder__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  egglib::DataHolder *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_egglib__DataHolder,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_DataHolder" "', argument " "1"" of type '" "egglib::DataHolder const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_DataHolder" "', argument " "1"" of type '" "egglib::DataHolder const &""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  {
    try {
      result = (egglib::DataHolder *)new egglib::DataHolder((egglib::DataHolder const &)*arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__DataHolder, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_DataHolder(PyObject *self, PyObject *args, PyObject *kwargs) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_DataHolder")) SWIG_fail;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_DataHolder", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_DataHolder__SWIG_1(self, argc, argv);
  }
  if (argc == 1) {
    int _v;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_egglib__DataHolder, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_DataHolder__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v;
    {
      int res = SWIG_AsVal_bool(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_DataHolder__SWIG_0(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_DataHolder'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::DataHolder::DataHolder(bool)\n"
    "    egglib::DataHolder::DataHolder()\n"
    "    egglib::DataHolder::DataHolder(egglib::DataHolder const &)\n");
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_DataHolder(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_DataHolder", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_DataHolder" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_valid_phyml_names(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_valid_phyml_names", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_valid_phyml_names" "', argument " "1"" of type '" "egglib::DataHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  {
    try {
      result = (bool)((egglib::DataHolder const *)arg1)->valid_phyml_names();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_set_is_matrix(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_set_is_matrix" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_set_is_matrix" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->set_is_matrix(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_get_is_matrix(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_get_is_matrix", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_get_is_matrix" "', argument " "1"" of type '" "egglib::DataHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  {
    try {
      result = (bool)((egglib::DataHolder const *)arg1)->get_is_matrix();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_reserve(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_reserve", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_reserve" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_reserve" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "DataHolder_reserve" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "DataHolder_reserve" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "DataHolder_reserve" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = static_cast< unsigned int >(val5);
  {
    try {
      (arg1)->reserve(arg2,arg3,arg4,arg5);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_get_nsam(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_get_nsam", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_get_nsam" "', argument " "1"" of type '" "egglib::DataHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::DataHolder const *)arg1)->get_nsam();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_set_nsam(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_set_nsam" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_set_nsam" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_nsam(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_get_nsit_all(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_get_nsit_all", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_get_nsit_all" "', argument " "1"" of type '" "egglib::DataHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::DataHolder const *)arg1)->get_nsit_all();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_get_nsit_sample(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_get_nsit_sample" "', argument " "1"" of type '" "egglib::DataHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_get_nsit_sample" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::DataHolder const *)arg1)->get_nsit_sample(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_set_nsit_all(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_set_nsit_all" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_set_nsit_all" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_nsit_all(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_set_nsit_sample(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_set_nsit_sample", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_set_nsit_sample" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_set_nsit_sample" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "DataHolder_set_nsit_sample" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      (arg1)->set_nsit_sample(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_insert_sites_all(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_insert_sites_all", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_insert_sites_all" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_insert_sites_all" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "DataHolder_insert_sites_all" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      (arg1)->insert_sites_all(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_insert_sites_sample(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_insert_sites_sample", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_insert_sites_sample" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_insert_sites_sample" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "DataHolder_insert_sites_sample" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "DataHolder_insert_sites_sample" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  {
    try {
      (arg1)->insert_sites_sample(arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_get_nlabels(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_get_nlabels" "', argument " "1"" of type '" "egglib::DataHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_get_nlabels" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::DataHolder const *)arg1)->get_nlabels(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_set_nlabels(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_set_nlabels", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_set_nlabels" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_set_nlabels" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "DataHolder_set_nlabels" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      (arg1)->set_nlabels(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_get_sample(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  int result;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_get_sample", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_get_sample" "', argument " "1"" of type '" "egglib::DataHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_get_sample" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "DataHolder_get_sample" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (int)((egglib::DataHolder const *)arg1)->get_sample(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_set_sample(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_set_sample", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_set_sample" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_set_sample" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "DataHolder_set_sample" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "DataHolder_set_sample" "', argument " "4"" of type '" "int""'");
  } 
  arg4 = static_cast< int >(val4);
  {
    try {
      (arg1)->set_sample(arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_get_label(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_get_label", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_get_label" "', argument " "1"" of type '" "egglib::DataHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_get_label" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "DataHolder_get_label" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (char *)((egglib::DataHolder const *)arg1)->get_label(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_set_label(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  char *arg4 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_set_label", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_set_label" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_set_label" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "DataHolder_set_label" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  res4 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "DataHolder_set_label" "', argument " "4"" of type '" "char const *""'");
  }
  arg4 = reinterpret_cast< char * >(buf4);
  {
    try {
      (arg1)->set_label(arg2,arg3,(char const *)arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc4 == SWIG_NEWOBJ) delete[] buf4;
  return resultobj;
fail:
  if (alloc4 == SWIG_NEWOBJ) delete[] buf4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_add_label(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  char *arg3 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_add_label", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_add_label" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_add_label" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "DataHolder_add_label" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = reinterpret_cast< char * >(buf3);
  {
    try {
      (arg1)->add_label(arg2,(char const *)arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_get_name(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_get_name" "', argument " "1"" of type '" "egglib::DataHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_get_name" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (char *)((egglib::DataHolder const *)arg1)->get_name(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_set_name(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  char *arg3 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_set_name", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_set_name" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_set_name" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "DataHolder_set_name" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = reinterpret_cast< char * >(buf3);
  {
    try {
      (arg1)->set_name(arg2,(char const *)arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_name_appendch(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  char arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  char val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_name_appendch", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_name_appendch" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_name_appendch" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_char(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "DataHolder_name_appendch" "', argument " "3"" of type '" "char""'");
  } 
  arg3 = static_cast< char >(val3);
  {
    try {
      (arg1)->name_appendch(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_name_append(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  char *arg3 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_name_append", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_name_append" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_name_append" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "DataHolder_name_append" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = reinterpret_cast< char * >(buf3);
  {
    try {
      (arg1)->name_append(arg2,(char const *)arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_del_sample(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_del_sample" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_del_sample" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->del_sample(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_del_sites_all(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_del_sites_all", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_del_sites_all" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_del_sites_all" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "DataHolder_del_sites_all" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      (arg1)->del_sites_all(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_del_sites_sample(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_del_sites_sample", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_del_sites_sample" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_del_sites_sample" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "DataHolder_del_sites_sample" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "DataHolder_del_sites_sample" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  {
    try {
      (arg1)->del_sites_sample(arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_reset" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_reset" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->reset(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_clear" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_clear" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->clear(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_change_case(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  bool arg2 ;
  int arg3 ;
  int arg4 ;
  int arg5 ;
  egglib::AbstractBaseAlphabet *arg6 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  PyObject *swig_obj[6] ;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_change_case", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_change_case" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_change_case" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "DataHolder_change_case" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "DataHolder_change_case" "', argument " "4"" of type '" "int""'");
  } 
  arg4 = static_cast< int >(val4);
  ecode5 = SWIG_AsVal_int(swig_obj[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "DataHolder_change_case" "', argument " "5"" of type '" "int""'");
  } 
  arg5 = static_cast< int >(val5);
  res6 = SWIG_ConvertPtr(swig_obj[4], &argp6, SWIGTYPE_p_egglib__AbstractBaseAlphabet,  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "DataHolder_change_case" "', argument " "6"" of type '" "egglib::AbstractBaseAlphabet &""'"); 
  }
  if (!argp6) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "DataHolder_change_case" "', argument " "6"" of type '" "egglib::AbstractBaseAlphabet &""'"); 
  }
  arg6 = reinterpret_cast< egglib::AbstractBaseAlphabet * >(argp6);
  {
    try {
      (arg1)->change_case(arg2,arg3,arg4,arg5,*arg6);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_find__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  egglib::VectorInt *arg3 = 0 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int result;
  
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_find" "', argument " "1"" of type '" "egglib::DataHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_find" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_egglib__VectorInt,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "DataHolder_find" "', argument " "3"" of type '" "egglib::VectorInt &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "DataHolder_find" "', argument " "3"" of type '" "egglib::VectorInt &""'"); 
  }
  arg3 = reinterpret_cast< egglib::VectorInt * >(argp3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "DataHolder_find" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "DataHolder_find" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = static_cast< unsigned int >(val5);
  {
    try {
      result = (unsigned int)((egglib::DataHolder const *)arg1)->find(arg2,*arg3,arg4,arg5);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_find__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  egglib::VectorInt *arg3 = 0 ;
  unsigned int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int result;
  
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_find" "', argument " "1"" of type '" "egglib::DataHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_find" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_egglib__VectorInt,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "DataHolder_find" "', argument " "3"" of type '" "egglib::VectorInt &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "DataHolder_find" "', argument " "3"" of type '" "egglib::VectorInt &""'"); 
  }
  arg3 = reinterpret_cast< egglib::VectorInt * >(argp3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "DataHolder_find" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  {
    try {
      result = (unsigned int)((egglib::DataHolder const *)arg1)->find(arg2,*arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_find__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  egglib::VectorInt *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  unsigned int result;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_find" "', argument " "1"" of type '" "egglib::DataHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_find" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_egglib__VectorInt,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "DataHolder_find" "', argument " "3"" of type '" "egglib::VectorInt &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "DataHolder_find" "', argument " "3"" of type '" "egglib::VectorInt &""'"); 
  }
  arg3 = reinterpret_cast< egglib::VectorInt * >(argp3);
  {
    try {
      result = (unsigned int)((egglib::DataHolder const *)arg1)->find(arg2,*arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_find(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[6] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "DataHolder_find", 0, 5, argv+1))) SWIG_fail;
  argv[0] = self;
  if (argc == 3) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__DataHolder, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_egglib__VectorInt, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_DataHolder_find__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__DataHolder, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_egglib__VectorInt, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_unsigned_SS_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_DataHolder_find__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__DataHolder, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_egglib__VectorInt, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_unsigned_SS_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            {
              int res = SWIG_AsVal_unsigned_SS_int(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              return _wrap_DataHolder_find__SWIG_0(self, argc, argv);
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'DataHolder_find'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::DataHolder::find(unsigned int,egglib::VectorInt &,unsigned int,unsigned int) const\n"
    "    egglib::DataHolder::find(unsigned int,egglib::VectorInt &,unsigned int) const\n"
    "    egglib::DataHolder::find(unsigned int,egglib::VectorInt &) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_DataHolder_is_equal(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_is_equal", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_is_equal" "', argument " "1"" of type '" "egglib::DataHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  {
    try {
      result = (bool)((egglib::DataHolder const *)arg1)->is_equal();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_strip_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_strip_clear", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_strip_clear" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  {
    try {
      (arg1)->strip_clear();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_strip_add(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_strip_add" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_strip_add" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      (arg1)->strip_add(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DataHolder_strip(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DataHolder *arg1 = (egglib::DataHolder *) 0 ;
  unsigned int arg2 ;
  bool arg3 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "DataHolder_strip", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DataHolder_strip" "', argument " "1"" of type '" "egglib::DataHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::DataHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "DataHolder_strip" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_bool(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "DataHolder_strip" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  ecode4 = SWIG_AsVal_bool(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "DataHolder_strip" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  {
    try {
      (arg1)->strip(arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_DataHolder) /* defines _wrap_delete_DataHolder_destructor_closure */

SWIGINTERN int _wrap_new_IntersperseAlign(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::IntersperseAlign *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_IntersperseAlign")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_IntersperseAlign", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::IntersperseAlign *)new egglib::IntersperseAlign();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__IntersperseAlign, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_IntersperseAlign(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::IntersperseAlign *arg1 = (egglib::IntersperseAlign *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_IntersperseAlign", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__IntersperseAlign, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_IntersperseAlign" "', argument " "1"" of type '" "egglib::IntersperseAlign *""'"); 
  }
  arg1 = reinterpret_cast< egglib::IntersperseAlign * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_IntersperseAlign_load(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::IntersperseAlign *arg1 = (egglib::IntersperseAlign *) 0 ;
  egglib::DataHolder *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__IntersperseAlign, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "IntersperseAlign_load" "', argument " "1"" of type '" "egglib::IntersperseAlign *""'"); 
  }
  arg1 = reinterpret_cast< egglib::IntersperseAlign * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__DataHolder,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "IntersperseAlign_load" "', argument " "2"" of type '" "egglib::DataHolder &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "IntersperseAlign_load" "', argument " "2"" of type '" "egglib::DataHolder &""'"); 
  }
  arg2 = reinterpret_cast< egglib::DataHolder * >(argp2);
  {
    try {
      (arg1)->load(*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_IntersperseAlign_set_length(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::IntersperseAlign *arg1 = (egglib::IntersperseAlign *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__IntersperseAlign, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "IntersperseAlign_set_length" "', argument " "1"" of type '" "egglib::IntersperseAlign *""'"); 
  }
  arg1 = reinterpret_cast< egglib::IntersperseAlign * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "IntersperseAlign_set_length" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_length(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_IntersperseAlign_set_position(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::IntersperseAlign *arg1 = (egglib::IntersperseAlign *) 0 ;
  unsigned int arg2 ;
  double arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "IntersperseAlign_set_position", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__IntersperseAlign, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "IntersperseAlign_set_position" "', argument " "1"" of type '" "egglib::IntersperseAlign *""'"); 
  }
  arg1 = reinterpret_cast< egglib::IntersperseAlign * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "IntersperseAlign_set_position" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "IntersperseAlign_set_position" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  {
    try {
      (arg1)->set_position(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_IntersperseAlign_set_round_position(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::IntersperseAlign *arg1 = (egglib::IntersperseAlign *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "IntersperseAlign_set_round_position", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__IntersperseAlign, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "IntersperseAlign_set_round_position" "', argument " "1"" of type '" "egglib::IntersperseAlign *""'"); 
  }
  arg1 = reinterpret_cast< egglib::IntersperseAlign * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "IntersperseAlign_set_round_position" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "IntersperseAlign_set_round_position" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      (arg1)->set_round_position(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_IntersperseAlign_get_positions(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::IntersperseAlign *arg1 = (egglib::IntersperseAlign *) 0 ;
  egglib::Coalesce *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__IntersperseAlign, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "IntersperseAlign_get_positions" "', argument " "1"" of type '" "egglib::IntersperseAlign *""'"); 
  }
  arg1 = reinterpret_cast< egglib::IntersperseAlign * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__Coalesce,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "IntersperseAlign_get_positions" "', argument " "2"" of type '" "egglib::Coalesce const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "IntersperseAlign_get_positions" "', argument " "2"" of type '" "egglib::Coalesce const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::Coalesce * >(argp2);
  {
    try {
      (arg1)->get_positions((egglib::Coalesce const &)*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_IntersperseAlign_set_num_alleles(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::IntersperseAlign *arg1 = (egglib::IntersperseAlign *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__IntersperseAlign, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "IntersperseAlign_set_num_alleles" "', argument " "1"" of type '" "egglib::IntersperseAlign *""'"); 
  }
  arg1 = reinterpret_cast< egglib::IntersperseAlign * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "IntersperseAlign_set_num_alleles" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_num_alleles(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_IntersperseAlign_set_allele(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::IntersperseAlign *arg1 = (egglib::IntersperseAlign *) 0 ;
  unsigned int arg2 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "IntersperseAlign_set_allele", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__IntersperseAlign, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "IntersperseAlign_set_allele" "', argument " "1"" of type '" "egglib::IntersperseAlign *""'"); 
  }
  arg1 = reinterpret_cast< egglib::IntersperseAlign * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "IntersperseAlign_set_allele" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "IntersperseAlign_set_allele" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  {
    try {
      (arg1)->set_allele(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_IntersperseAlign_intersperse__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::IntersperseAlign *arg1 = (egglib::IntersperseAlign *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__IntersperseAlign, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "IntersperseAlign_intersperse" "', argument " "1"" of type '" "egglib::IntersperseAlign *""'"); 
  }
  arg1 = reinterpret_cast< egglib::IntersperseAlign * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "IntersperseAlign_intersperse" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->intersperse(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_IntersperseAlign_intersperse__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::IntersperseAlign *arg1 = (egglib::IntersperseAlign *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__IntersperseAlign, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "IntersperseAlign_intersperse" "', argument " "1"" of type '" "egglib::IntersperseAlign *""'"); 
  }
  arg1 = reinterpret_cast< egglib::IntersperseAlign * >(argp1);
  {
    try {
      (arg1)->intersperse();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_IntersperseAlign_intersperse(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "IntersperseAlign_intersperse", 0, 2, argv+1))) SWIG_fail;
  argv[0] = self;
  if (argc == 1) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__IntersperseAlign, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_IntersperseAlign_intersperse__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__IntersperseAlign, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_IntersperseAlign_intersperse__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'IntersperseAlign_intersperse'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::IntersperseAlign::intersperse(bool)\n"
    "    egglib::IntersperseAlign::intersperse()\n");
  return 0;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_IntersperseAlign) /* defines _wrap_delete_IntersperseAlign_destructor_closure */

SWIGINTERN int _wrap_new_StructureHolder(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_StructureHolder")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_StructureHolder", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::StructureHolder *)new egglib::StructureHolder();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__StructureHolder, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_StructureHolder(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_StructureHolder", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_StructureHolder" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_reset", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_reset" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  {
    try {
      (arg1)->reset();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_check_ploidy__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_check_ploidy" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "StructureHolder_check_ploidy" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->check_ploidy(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_check_ploidy__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_check_ploidy" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  {
    try {
      (arg1)->check_ploidy();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_check_ploidy(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "StructureHolder_check_ploidy", 0, 2, argv+1))) SWIG_fail;
  argv[0] = self;
  if (argc == 1) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__StructureHolder, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_StructureHolder_check_ploidy__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__StructureHolder, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_StructureHolder_check_ploidy__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'StructureHolder_check_ploidy'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::StructureHolder::check_ploidy(unsigned int)\n"
    "    egglib::StructureHolder::check_ploidy()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_StructureHolder_get_ploidy(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_get_ploidy", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_get_ploidy" "', argument " "1"" of type '" "egglib::StructureHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::StructureHolder const *)arg1)->get_ploidy();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_get_structure(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  egglib::DataHolder *arg2 = 0 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  bool arg7 ;
  char *arg8 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  bool val7 ;
  int ecode7 = 0 ;
  int res8 ;
  char *buf8 = 0 ;
  int alloc8 = 0 ;
  PyObject *swig_obj[8] ;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_get_structure", 7, 7, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_get_structure" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__DataHolder,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "StructureHolder_get_structure" "', argument " "2"" of type '" "egglib::DataHolder &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "StructureHolder_get_structure" "', argument " "2"" of type '" "egglib::DataHolder &""'"); 
  }
  arg2 = reinterpret_cast< egglib::DataHolder * >(argp2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "StructureHolder_get_structure" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "StructureHolder_get_structure" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "StructureHolder_get_structure" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = static_cast< unsigned int >(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(swig_obj[4], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "StructureHolder_get_structure" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = static_cast< unsigned int >(val6);
  ecode7 = SWIG_AsVal_bool(swig_obj[5], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "StructureHolder_get_structure" "', argument " "7"" of type '" "bool""'");
  } 
  arg7 = static_cast< bool >(val7);
  res8 = SWIG_AsCharPtrAndSize(swig_obj[6], &buf8, NULL, &alloc8);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "StructureHolder_get_structure" "', argument " "8"" of type '" "char const *""'");
  }
  arg8 = reinterpret_cast< char * >(buf8);
  {
    try {
      (arg1)->get_structure(*arg2,arg3,arg4,arg5,arg6,arg7,(char const *)arg8);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc8 == SWIG_NEWOBJ) delete[] buf8;
  return resultobj;
fail:
  if (alloc8 == SWIG_NEWOBJ) delete[] buf8;
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_mk_dummy_structure(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_mk_dummy_structure", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_mk_dummy_structure" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "StructureHolder_mk_dummy_structure" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "StructureHolder_mk_dummy_structure" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      (arg1)->mk_dummy_structure(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_dummy_add_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_dummy_add_pop" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "StructureHolder_dummy_add_pop" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->dummy_add_pop(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_process_ingroup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  unsigned int arg2 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  char *arg5 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  PyObject *swig_obj[5] ;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_process_ingroup", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_process_ingroup" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "StructureHolder_process_ingroup" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "StructureHolder_process_ingroup" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = reinterpret_cast< char * >(buf3);
  res4 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "StructureHolder_process_ingroup" "', argument " "4"" of type '" "char const *""'");
  }
  arg4 = reinterpret_cast< char * >(buf4);
  res5 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "StructureHolder_process_ingroup" "', argument " "5"" of type '" "char const *""'");
  }
  arg5 = reinterpret_cast< char * >(buf5);
  {
    try {
      (arg1)->process_ingroup(arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  if (alloc4 == SWIG_NEWOBJ) delete[] buf4;
  if (alloc5 == SWIG_NEWOBJ) delete[] buf5;
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  if (alloc4 == SWIG_NEWOBJ) delete[] buf4;
  if (alloc5 == SWIG_NEWOBJ) delete[] buf5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_process_outgroup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  unsigned int arg2 ;
  char *arg3 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_process_outgroup", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_process_outgroup" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "StructureHolder_process_outgroup" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "StructureHolder_process_outgroup" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = reinterpret_cast< char * >(buf3);
  {
    try {
      (arg1)->process_outgroup(arg2,(char const *)arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_add_cluster(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::StructureCluster *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_add_cluster" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "StructureHolder_add_cluster" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      result = (egglib::StructureCluster *)(arg1)->add_cluster((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__StructureCluster, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_add_population(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  char *arg2 = (char *) 0 ;
  egglib::StructureCluster *arg3 = (egglib::StructureCluster *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  egglib::StructurePopulation *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_add_population", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_add_population" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "StructureHolder_add_population" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  res3 = SWIG_ConvertPtr(swig_obj[1], &argp3,SWIGTYPE_p_egglib__StructureCluster, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "StructureHolder_add_population" "', argument " "3"" of type '" "egglib::StructureCluster *""'"); 
  }
  arg3 = reinterpret_cast< egglib::StructureCluster * >(argp3);
  {
    try {
      result = (egglib::StructurePopulation *)(arg1)->add_population((char const *)arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__StructurePopulation, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_add_individual_ingroup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  char *arg2 = (char *) 0 ;
  egglib::StructurePopulation *arg3 = (egglib::StructurePopulation *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  egglib::StructureIndiv *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_add_individual_ingroup", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_add_individual_ingroup" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "StructureHolder_add_individual_ingroup" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  res3 = SWIG_ConvertPtr(swig_obj[1], &argp3,SWIGTYPE_p_egglib__StructurePopulation, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "StructureHolder_add_individual_ingroup" "', argument " "3"" of type '" "egglib::StructurePopulation *""'"); 
  }
  arg3 = reinterpret_cast< egglib::StructurePopulation * >(argp3);
  {
    try {
      result = (egglib::StructureIndiv *)(arg1)->add_individual_ingroup((char const *)arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__StructureIndiv, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_add_individual_outgroup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::StructureIndiv *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_add_individual_outgroup" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "StructureHolder_add_individual_outgroup" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      result = (egglib::StructureIndiv *)(arg1)->add_individual_outgroup((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__StructureIndiv, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_add_sample_ingroup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  unsigned int arg2 ;
  egglib::StructureIndiv *arg3 = (egglib::StructureIndiv *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_add_sample_ingroup", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_add_sample_ingroup" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "StructureHolder_add_sample_ingroup" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[1], &argp3,SWIGTYPE_p_egglib__StructureIndiv, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "StructureHolder_add_sample_ingroup" "', argument " "3"" of type '" "egglib::StructureIndiv *""'"); 
  }
  arg3 = reinterpret_cast< egglib::StructureIndiv * >(argp3);
  {
    try {
      (arg1)->add_sample_ingroup(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_add_sample_outgroup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  unsigned int arg2 ;
  egglib::StructureIndiv *arg3 = (egglib::StructureIndiv *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_add_sample_outgroup", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_add_sample_outgroup" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "StructureHolder_add_sample_outgroup" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[1], &argp3,SWIGTYPE_p_egglib__StructureIndiv, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "StructureHolder_add_sample_outgroup" "', argument " "3"" of type '" "egglib::StructureIndiv *""'"); 
  }
  arg3 = reinterpret_cast< egglib::StructureIndiv * >(argp3);
  {
    try {
      (arg1)->add_sample_outgroup(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_num_clust(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_num_clust", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_num_clust" "', argument " "1"" of type '" "egglib::StructureHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::StructureHolder const *)arg1)->num_clust();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_num_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_num_pop", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_num_pop" "', argument " "1"" of type '" "egglib::StructureHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::StructureHolder const *)arg1)->num_pop();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_num_indiv_ingroup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_num_indiv_ingroup", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_num_indiv_ingroup" "', argument " "1"" of type '" "egglib::StructureHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::StructureHolder const *)arg1)->num_indiv_ingroup();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_num_indiv_outgroup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_num_indiv_outgroup", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_num_indiv_outgroup" "', argument " "1"" of type '" "egglib::StructureHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::StructureHolder const *)arg1)->num_indiv_outgroup();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_get_cluster(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::StructureCluster *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_get_cluster" "', argument " "1"" of type '" "egglib::StructureHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "StructureHolder_get_cluster" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::StructureCluster *) &((egglib::StructureHolder const *)arg1)->get_cluster(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__StructureCluster, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_get_population(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::StructurePopulation *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_get_population" "', argument " "1"" of type '" "egglib::StructureHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "StructureHolder_get_population" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::StructurePopulation *) &((egglib::StructureHolder const *)arg1)->get_population(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__StructurePopulation, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_get_indiv_ingroup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::StructureIndiv *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_get_indiv_ingroup" "', argument " "1"" of type '" "egglib::StructureHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "StructureHolder_get_indiv_ingroup" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::StructureIndiv *) &((egglib::StructureHolder const *)arg1)->get_indiv_ingroup(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__StructureIndiv, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_get_indiv_outgroup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::StructureIndiv *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_get_indiv_outgroup" "', argument " "1"" of type '" "egglib::StructureHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "StructureHolder_get_indiv_outgroup" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::StructureIndiv *) &((egglib::StructureHolder const *)arg1)->get_indiv_outgroup(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__StructureIndiv, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_get_ni(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_get_ni", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_get_ni" "', argument " "1"" of type '" "egglib::StructureHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::StructureHolder const *)arg1)->get_ni();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_get_no(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_get_no", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_get_no" "', argument " "1"" of type '" "egglib::StructureHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::StructureHolder const *)arg1)->get_no();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_get_req(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_get_req", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_get_req" "', argument " "1"" of type '" "egglib::StructureHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::StructureHolder const *)arg1)->get_req();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_copy(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  egglib::StructureHolder *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_copy" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__StructureHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "StructureHolder_copy" "', argument " "2"" of type '" "egglib::StructureHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "StructureHolder_copy" "', argument " "2"" of type '" "egglib::StructureHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::StructureHolder * >(argp2);
  {
    try {
      (arg1)->copy((egglib::StructureHolder const &)*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_subset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  egglib::StructureHolder *arg2 = 0 ;
  char *arg3 = (char *) 0 ;
  char *arg4 = (char *) 0 ;
  bool arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_subset", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_subset" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__StructureHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "StructureHolder_subset" "', argument " "2"" of type '" "egglib::StructureHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "StructureHolder_subset" "', argument " "2"" of type '" "egglib::StructureHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::StructureHolder * >(argp2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "StructureHolder_subset" "', argument " "3"" of type '" "char *""'");
  }
  arg3 = reinterpret_cast< char * >(buf3);
  res4 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "StructureHolder_subset" "', argument " "4"" of type '" "char *""'");
  }
  arg4 = reinterpret_cast< char * >(buf4);
  ecode5 = SWIG_AsVal_bool(swig_obj[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "StructureHolder_subset" "', argument " "5"" of type '" "bool""'");
  } 
  arg5 = static_cast< bool >(val5);
  {
    try {
      result = (char *)(arg1)->subset((egglib::StructureHolder const &)*arg2,arg3,arg4,arg5);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  if (alloc4 == SWIG_NEWOBJ) delete[] buf4;
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  if (alloc4 == SWIG_NEWOBJ) delete[] buf4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_shuffle_init(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_shuffle_init" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "StructureHolder_shuffle_init" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      (arg1)->shuffle_init(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_shuffle(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_shuffle", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_shuffle" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  {
    try {
      (arg1)->shuffle();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_shuffle_restore(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_shuffle_restore", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_shuffle_restore" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  {
    try {
      (arg1)->shuffle_restore();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_init_i(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_init_i", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_init_i" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  {
    try {
      result = (unsigned int)(arg1)->init_i();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_init_o(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_init_o", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_init_o" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  {
    try {
      result = (unsigned int)(arg1)->init_o();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_iter_i(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_iter_i", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_iter_i" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  {
    try {
      result = (unsigned int)(arg1)->iter_i();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureHolder_iter_o(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureHolder *arg1 = (egglib::StructureHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureHolder_iter_o", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureHolder_iter_o" "', argument " "1"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureHolder * >(argp1);
  {
    try {
      result = (unsigned int)(arg1)->iter_o();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_StructureHolder) /* defines _wrap_delete_StructureHolder_destructor_closure */

SWIGINTERN int _wrap_new_StructureCluster(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::StructureCluster *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_StructureCluster")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_StructureCluster", 1, 1, swig_obj)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_StructureCluster" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = reinterpret_cast< char * >(buf1);
  {
    try {
      result = (egglib::StructureCluster *)new egglib::StructureCluster((char const *)arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__StructureCluster, SWIG_BUILTIN_INIT |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  return resultobj == Py_None ? -1 : 0;
fail:
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_StructureCluster(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureCluster *arg1 = (egglib::StructureCluster *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_StructureCluster", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureCluster, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_StructureCluster" "', argument " "1"" of type '" "egglib::StructureCluster *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureCluster * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureCluster_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureCluster *arg1 = (egglib::StructureCluster *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureCluster, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureCluster_reset" "', argument " "1"" of type '" "egglib::StructureCluster *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureCluster * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "StructureCluster_reset" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->reset((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureCluster_add_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureCluster *arg1 = (egglib::StructureCluster *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::StructurePopulation *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureCluster, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureCluster_add_pop" "', argument " "1"" of type '" "egglib::StructureCluster *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureCluster * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "StructureCluster_add_pop" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      result = (egglib::StructurePopulation *)(arg1)->add_pop((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__StructurePopulation, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureCluster_num_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureCluster *arg1 = (egglib::StructureCluster *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureCluster_num_pop", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureCluster, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureCluster_num_pop" "', argument " "1"" of type '" "egglib::StructureCluster const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureCluster * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::StructureCluster const *)arg1)->num_pop();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureCluster_get_population(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureCluster *arg1 = (egglib::StructureCluster *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::StructurePopulation *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureCluster, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureCluster_get_population" "', argument " "1"" of type '" "egglib::StructureCluster const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureCluster * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "StructureCluster_get_population" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::StructurePopulation *) &((egglib::StructureCluster const *)arg1)->get_population(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__StructurePopulation, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureCluster_get_label(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureCluster *arg1 = (egglib::StructureCluster *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureCluster_get_label", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureCluster, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureCluster_get_label" "', argument " "1"" of type '" "egglib::StructureCluster const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureCluster * >(argp1);
  {
    try {
      result = (char *)((egglib::StructureCluster const *)arg1)->get_label();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_StructureCluster) /* defines _wrap_delete_StructureCluster_destructor_closure */

SWIGINTERN int _wrap_new_StructurePopulation(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::StructurePopulation *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_StructurePopulation")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_StructurePopulation", 1, 1, swig_obj)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_StructurePopulation" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = reinterpret_cast< char * >(buf1);
  {
    try {
      result = (egglib::StructurePopulation *)new egglib::StructurePopulation((char const *)arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__StructurePopulation, SWIG_BUILTIN_INIT |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  return resultobj == Py_None ? -1 : 0;
fail:
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_StructurePopulation(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructurePopulation *arg1 = (egglib::StructurePopulation *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_StructurePopulation", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructurePopulation, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_StructurePopulation" "', argument " "1"" of type '" "egglib::StructurePopulation *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructurePopulation * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructurePopulation_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructurePopulation *arg1 = (egglib::StructurePopulation *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructurePopulation, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructurePopulation_reset" "', argument " "1"" of type '" "egglib::StructurePopulation *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructurePopulation * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "StructurePopulation_reset" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->reset((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructurePopulation_add_indiv(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructurePopulation *arg1 = (egglib::StructurePopulation *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::StructureIndiv *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructurePopulation, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructurePopulation_add_indiv" "', argument " "1"" of type '" "egglib::StructurePopulation *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructurePopulation * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "StructurePopulation_add_indiv" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      result = (egglib::StructureIndiv *)(arg1)->add_indiv((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__StructureIndiv, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructurePopulation_num_indiv(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructurePopulation *arg1 = (egglib::StructurePopulation *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "StructurePopulation_num_indiv", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructurePopulation, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructurePopulation_num_indiv" "', argument " "1"" of type '" "egglib::StructurePopulation const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructurePopulation * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::StructurePopulation const *)arg1)->num_indiv();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructurePopulation_get_indiv(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructurePopulation *arg1 = (egglib::StructurePopulation *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::StructureIndiv *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructurePopulation, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructurePopulation_get_indiv" "', argument " "1"" of type '" "egglib::StructurePopulation const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructurePopulation * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "StructurePopulation_get_indiv" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::StructureIndiv *) &((egglib::StructurePopulation const *)arg1)->get_indiv(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__StructureIndiv, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructurePopulation_get_label(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructurePopulation *arg1 = (egglib::StructurePopulation *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "StructurePopulation_get_label", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructurePopulation, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructurePopulation_get_label" "', argument " "1"" of type '" "egglib::StructurePopulation const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructurePopulation * >(argp1);
  {
    try {
      result = (char *)((egglib::StructurePopulation const *)arg1)->get_label();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_StructurePopulation) /* defines _wrap_delete_StructurePopulation_destructor_closure */

SWIGINTERN int _wrap_new_StructureIndiv(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::StructureIndiv *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_StructureIndiv")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_StructureIndiv", 1, 1, swig_obj)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_StructureIndiv" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = reinterpret_cast< char * >(buf1);
  {
    try {
      result = (egglib::StructureIndiv *)new egglib::StructureIndiv((char const *)arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__StructureIndiv, SWIG_BUILTIN_INIT |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  return resultobj == Py_None ? -1 : 0;
fail:
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_StructureIndiv(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureIndiv *arg1 = (egglib::StructureIndiv *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_StructureIndiv", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureIndiv, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_StructureIndiv" "', argument " "1"" of type '" "egglib::StructureIndiv *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureIndiv * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureIndiv_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureIndiv *arg1 = (egglib::StructureIndiv *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureIndiv, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureIndiv_reset" "', argument " "1"" of type '" "egglib::StructureIndiv *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureIndiv * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "StructureIndiv_reset" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->reset((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureIndiv_num_samples(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureIndiv *arg1 = (egglib::StructureIndiv *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureIndiv_num_samples", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureIndiv, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureIndiv_num_samples" "', argument " "1"" of type '" "egglib::StructureIndiv const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureIndiv * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::StructureIndiv const *)arg1)->num_samples();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureIndiv_add_sample(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureIndiv *arg1 = (egglib::StructureIndiv *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureIndiv, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureIndiv_add_sample" "', argument " "1"" of type '" "egglib::StructureIndiv *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureIndiv * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "StructureIndiv_add_sample" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->add_sample(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureIndiv_get_sample(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureIndiv *arg1 = (egglib::StructureIndiv *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureIndiv, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureIndiv_get_sample" "', argument " "1"" of type '" "egglib::StructureIndiv const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureIndiv * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "StructureIndiv_get_sample" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::StructureIndiv const *)arg1)->get_sample(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StructureIndiv_get_label(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::StructureIndiv *arg1 = (egglib::StructureIndiv *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "StructureIndiv_get_label", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__StructureIndiv, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StructureIndiv_get_label" "', argument " "1"" of type '" "egglib::StructureIndiv const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::StructureIndiv * >(argp1);
  {
    try {
      result = (char *)((egglib::StructureIndiv const *)arg1)->get_label();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_StructureIndiv) /* defines _wrap_delete_StructureIndiv_destructor_closure */

SWIGINTERN int _wrap_new_FreqSet(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::FreqSet *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_FreqSet")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_FreqSet", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::FreqSet *)new egglib::FreqSet();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__FreqSet, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_FreqSet(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqSet *arg1 = (egglib::FreqSet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_FreqSet", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqSet, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_FreqSet" "', argument " "1"" of type '" "egglib::FreqSet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqSet * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqSet_setup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqSet *arg1 = (egglib::FreqSet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "FreqSet_setup", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqSet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqSet_setup" "', argument " "1"" of type '" "egglib::FreqSet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqSet * >(argp1);
  {
    try {
      (arg1)->setup();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqSet_set_nall(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqSet *arg1 = (egglib::FreqSet *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqSet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqSet_set_nall" "', argument " "1"" of type '" "egglib::FreqSet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqSet * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FreqSet_set_nall" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_nall(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqSet_add_genotypes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqSet *arg1 = (egglib::FreqSet *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqSet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqSet_add_genotypes" "', argument " "1"" of type '" "egglib::FreqSet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqSet * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FreqSet_add_genotypes" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->add_genotypes(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqSet_incr_allele(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqSet *arg1 = (egglib::FreqSet *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "FreqSet_incr_allele", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqSet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqSet_incr_allele" "', argument " "1"" of type '" "egglib::FreqSet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqSet * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FreqSet_incr_allele" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "FreqSet_incr_allele" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      (arg1)->incr_allele(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqSet_incr_genotype(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqSet *arg1 = (egglib::FreqSet *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "FreqSet_incr_genotype", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqSet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqSet_incr_genotype" "', argument " "1"" of type '" "egglib::FreqSet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqSet * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FreqSet_incr_genotype" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "FreqSet_incr_genotype" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      (arg1)->incr_genotype(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqSet_tell_het(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqSet *arg1 = (egglib::FreqSet *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "FreqSet_tell_het", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqSet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqSet_tell_het" "', argument " "1"" of type '" "egglib::FreqSet *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqSet * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FreqSet_tell_het" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "FreqSet_tell_het" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      (arg1)->tell_het(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqSet_num_alleles(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqSet *arg1 = (egglib::FreqSet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "FreqSet_num_alleles", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqSet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqSet_num_alleles" "', argument " "1"" of type '" "egglib::FreqSet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqSet * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::FreqSet const *)arg1)->num_alleles();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqSet_num_alleles_eff(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqSet *arg1 = (egglib::FreqSet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "FreqSet_num_alleles_eff", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqSet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqSet_num_alleles_eff" "', argument " "1"" of type '" "egglib::FreqSet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqSet * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::FreqSet const *)arg1)->num_alleles_eff();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqSet_num_genotypes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqSet *arg1 = (egglib::FreqSet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "FreqSet_num_genotypes", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqSet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqSet_num_genotypes" "', argument " "1"" of type '" "egglib::FreqSet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqSet * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::FreqSet const *)arg1)->num_genotypes();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqSet_num_genotypes_eff(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqSet *arg1 = (egglib::FreqSet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "FreqSet_num_genotypes_eff", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqSet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqSet_num_genotypes_eff" "', argument " "1"" of type '" "egglib::FreqSet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqSet * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::FreqSet const *)arg1)->num_genotypes_eff();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqSet_nseff(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqSet *arg1 = (egglib::FreqSet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "FreqSet_nseff", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqSet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqSet_nseff" "', argument " "1"" of type '" "egglib::FreqSet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqSet * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::FreqSet const *)arg1)->nseff();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqSet_nieff(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqSet *arg1 = (egglib::FreqSet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "FreqSet_nieff", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqSet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqSet_nieff" "', argument " "1"" of type '" "egglib::FreqSet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqSet * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::FreqSet const *)arg1)->nieff();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqSet_frq_all(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqSet *arg1 = (egglib::FreqSet *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqSet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqSet_frq_all" "', argument " "1"" of type '" "egglib::FreqSet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqSet * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FreqSet_frq_all" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::FreqSet const *)arg1)->frq_all(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqSet_frq_gen(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqSet *arg1 = (egglib::FreqSet *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqSet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqSet_frq_gen" "', argument " "1"" of type '" "egglib::FreqSet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqSet * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FreqSet_frq_gen" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::FreqSet const *)arg1)->frq_gen(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqSet_frq_het(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqSet *arg1 = (egglib::FreqSet *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqSet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqSet_frq_het" "', argument " "1"" of type '" "egglib::FreqSet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqSet * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FreqSet_frq_het" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::FreqSet const *)arg1)->frq_het(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqSet_tot_het(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqSet *arg1 = (egglib::FreqSet *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "FreqSet_tot_het", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqSet, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqSet_tot_het" "', argument " "1"" of type '" "egglib::FreqSet const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqSet * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::FreqSet const *)arg1)->tot_het();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_FreqSet) /* defines _wrap_delete_FreqSet_destructor_closure */

SWIGINTERN int _wrap_new_FreqHolder(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_FreqHolder")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_FreqHolder", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::FreqHolder *)new egglib::FreqHolder();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__FreqHolder, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_FreqHolder(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_FreqHolder", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_FreqHolder" "', argument " "1"" of type '" "egglib::FreqHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_setup_structure(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  egglib::StructureHolder *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_setup_structure" "', argument " "1"" of type '" "egglib::FreqHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__StructureHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FreqHolder_setup_structure" "', argument " "2"" of type '" "egglib::StructureHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "FreqHolder_setup_structure" "', argument " "2"" of type '" "egglib::StructureHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::StructureHolder * >(argp2);
  {
    try {
      (arg1)->setup_structure((egglib::StructureHolder const &)*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_setup_raw(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "FreqHolder_setup_raw", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_setup_raw" "', argument " "1"" of type '" "egglib::FreqHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FreqHolder_setup_raw" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "FreqHolder_setup_raw" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "FreqHolder_setup_raw" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  {
    try {
      (arg1)->setup_raw(arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_setup_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  
  if (!SWIG_Python_UnpackTuple(args, "FreqHolder_setup_pop", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_setup_pop" "', argument " "1"" of type '" "egglib::FreqHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FreqHolder_setup_pop" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "FreqHolder_setup_pop" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "FreqHolder_setup_pop" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "FreqHolder_setup_pop" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = static_cast< unsigned int >(val5);
  {
    try {
      (arg1)->setup_pop(arg2,arg3,arg4,arg5);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_set_ngeno(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_set_ngeno" "', argument " "1"" of type '" "egglib::FreqHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FreqHolder_set_ngeno" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_ngeno(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_process_site(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  egglib::SiteHolder *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_process_site" "', argument " "1"" of type '" "egglib::FreqHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__SiteHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FreqHolder_process_site" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "FreqHolder_process_site" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::SiteHolder * >(argp2);
  {
    try {
      (arg1)->process_site((egglib::SiteHolder const &)*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_process_vcf(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  egglib::VcfParser *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_process_vcf" "', argument " "1"" of type '" "egglib::FreqHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__VcfParser,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FreqHolder_process_vcf" "', argument " "2"" of type '" "egglib::VcfParser const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "FreqHolder_process_vcf" "', argument " "2"" of type '" "egglib::VcfParser const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::VcfParser * >(argp2);
  {
    try {
      (arg1)->process_vcf((egglib::VcfParser const &)*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_frq_ingroup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::FreqSet *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "FreqHolder_frq_ingroup", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_frq_ingroup" "', argument " "1"" of type '" "egglib::FreqHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  {
    try {
      result = (egglib::FreqSet *) &((egglib::FreqHolder const *)arg1)->frq_ingroup();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__FreqSet, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_frq_outgroup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::FreqSet *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "FreqHolder_frq_outgroup", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_frq_outgroup" "', argument " "1"" of type '" "egglib::FreqHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  {
    try {
      result = (egglib::FreqSet *) &((egglib::FreqHolder const *)arg1)->frq_outgroup();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__FreqSet, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_num_clusters(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "FreqHolder_num_clusters", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_num_clusters" "', argument " "1"" of type '" "egglib::FreqHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::FreqHolder const *)arg1)->num_clusters();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_num_populations(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "FreqHolder_num_populations", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_num_populations" "', argument " "1"" of type '" "egglib::FreqHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::FreqHolder const *)arg1)->num_populations();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_frq_cluster(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::FreqSet *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_frq_cluster" "', argument " "1"" of type '" "egglib::FreqHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FreqHolder_frq_cluster" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::FreqSet *) &((egglib::FreqHolder const *)arg1)->frq_cluster(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__FreqSet, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_frq_population(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::FreqSet *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_frq_population" "', argument " "1"" of type '" "egglib::FreqHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FreqHolder_frq_population" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::FreqSet *) &((egglib::FreqHolder const *)arg1)->frq_population(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__FreqSet, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_num_alleles(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "FreqHolder_num_alleles", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_num_alleles" "', argument " "1"" of type '" "egglib::FreqHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::FreqHolder const *)arg1)->num_alleles();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_allele(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_allele" "', argument " "1"" of type '" "egglib::FreqHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FreqHolder_allele" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (int)((egglib::FreqHolder const *)arg1)->allele(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_get_allele_index(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_get_allele_index" "', argument " "1"" of type '" "egglib::FreqHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FreqHolder_get_allele_index" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      result = (unsigned int)((egglib::FreqHolder const *)arg1)->get_allele_index(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_ploidy(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "FreqHolder_ploidy", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_ploidy" "', argument " "1"" of type '" "egglib::FreqHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::FreqHolder const *)arg1)->ploidy();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_num_genotypes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "FreqHolder_num_genotypes", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_num_genotypes" "', argument " "1"" of type '" "egglib::FreqHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::FreqHolder const *)arg1)->num_genotypes();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_genotype_het(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_genotype_het" "', argument " "1"" of type '" "egglib::FreqHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FreqHolder_genotype_het" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (bool)((egglib::FreqHolder const *)arg1)->genotype_het(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_genotype(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_genotype" "', argument " "1"" of type '" "egglib::FreqHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FreqHolder_genotype" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (int *)((egglib::FreqHolder const *)arg1)->genotype(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_int, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_genotype_item(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  int result;
  
  if (!SWIG_Python_UnpackTuple(args, "FreqHolder_genotype_item", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_genotype_item" "', argument " "1"" of type '" "egglib::FreqHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FreqHolder_genotype_item" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "FreqHolder_genotype_item" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (int)((egglib::FreqHolder const *)arg1)->genotype_item(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_set_genotype_item(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "FreqHolder_set_genotype_item", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_set_genotype_item" "', argument " "1"" of type '" "egglib::FreqHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FreqHolder_set_genotype_item" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "FreqHolder_set_genotype_item" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "FreqHolder_set_genotype_item" "', argument " "4"" of type '" "int""'");
  } 
  arg4 = static_cast< int >(val4);
  {
    try {
      (arg1)->set_genotype_item(arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_find_allele(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_find_allele" "', argument " "1"" of type '" "egglib::FreqHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FreqHolder_find_allele" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      result = (unsigned int)(arg1)->find_allele(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FreqHolder_cluster_index(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FreqHolder *arg1 = (egglib::FreqHolder *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FreqHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FreqHolder_cluster_index" "', argument " "1"" of type '" "egglib::FreqHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FreqHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FreqHolder_cluster_index" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::FreqHolder const *)arg1)->cluster_index(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_FreqHolder) /* defines _wrap_delete_FreqHolder_destructor_closure */

SWIGINTERN int _wrap_new_AlleleStatus(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::AlleleStatus *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_AlleleStatus")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_AlleleStatus", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::AlleleStatus *)new egglib::AlleleStatus();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__AlleleStatus, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_AlleleStatus(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AlleleStatus *arg1 = (egglib::AlleleStatus *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_AlleleStatus", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AlleleStatus, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_AlleleStatus" "', argument " "1"" of type '" "egglib::AlleleStatus *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AlleleStatus * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AlleleStatus_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AlleleStatus *arg1 = (egglib::AlleleStatus *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "AlleleStatus_reset", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AlleleStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AlleleStatus_reset" "', argument " "1"" of type '" "egglib::AlleleStatus *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AlleleStatus * >(argp1);
  {
    try {
      (arg1)->reset();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AlleleStatus_process(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AlleleStatus *arg1 = (egglib::AlleleStatus *) 0 ;
  egglib::FreqHolder *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AlleleStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AlleleStatus_process" "', argument " "1"" of type '" "egglib::AlleleStatus *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AlleleStatus * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__FreqHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "AlleleStatus_process" "', argument " "2"" of type '" "egglib::FreqHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "AlleleStatus_process" "', argument " "2"" of type '" "egglib::FreqHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::FreqHolder * >(argp2);
  {
    try {
      (arg1)->process((egglib::FreqHolder const &)*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AlleleStatus_total(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AlleleStatus *arg1 = (egglib::AlleleStatus *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "AlleleStatus_total", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AlleleStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AlleleStatus_total" "', argument " "1"" of type '" "egglib::AlleleStatus *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AlleleStatus * >(argp1);
  {
    try {
      (arg1)->total();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AlleleStatus_nsites(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AlleleStatus *arg1 = (egglib::AlleleStatus *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "AlleleStatus_nsites", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AlleleStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AlleleStatus_nsites" "', argument " "1"" of type '" "egglib::AlleleStatus const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AlleleStatus * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::AlleleStatus const *)arg1)->nsites();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AlleleStatus_nsites_o(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AlleleStatus *arg1 = (egglib::AlleleStatus *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "AlleleStatus_nsites_o", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AlleleStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AlleleStatus_nsites_o" "', argument " "1"" of type '" "egglib::AlleleStatus const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AlleleStatus * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::AlleleStatus const *)arg1)->nsites_o();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AlleleStatus_Sp(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AlleleStatus *arg1 = (egglib::AlleleStatus *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "AlleleStatus_Sp", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AlleleStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AlleleStatus_Sp" "', argument " "1"" of type '" "egglib::AlleleStatus const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AlleleStatus * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::AlleleStatus const *)arg1)->Sp();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AlleleStatus_Sp_T1(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AlleleStatus *arg1 = (egglib::AlleleStatus *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "AlleleStatus_Sp_T1", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AlleleStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AlleleStatus_Sp_T1" "', argument " "1"" of type '" "egglib::AlleleStatus const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AlleleStatus * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::AlleleStatus const *)arg1)->Sp_T1();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AlleleStatus_Spd(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AlleleStatus *arg1 = (egglib::AlleleStatus *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "AlleleStatus_Spd", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AlleleStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AlleleStatus_Spd" "', argument " "1"" of type '" "egglib::AlleleStatus const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AlleleStatus * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::AlleleStatus const *)arg1)->Spd();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AlleleStatus_Spd_T1(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AlleleStatus *arg1 = (egglib::AlleleStatus *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "AlleleStatus_Spd_T1", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AlleleStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AlleleStatus_Spd_T1" "', argument " "1"" of type '" "egglib::AlleleStatus const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AlleleStatus * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::AlleleStatus const *)arg1)->Spd_T1();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AlleleStatus_ShP(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AlleleStatus *arg1 = (egglib::AlleleStatus *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "AlleleStatus_ShP", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AlleleStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AlleleStatus_ShP" "', argument " "1"" of type '" "egglib::AlleleStatus const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AlleleStatus * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::AlleleStatus const *)arg1)->ShP();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AlleleStatus_ShP_T1(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AlleleStatus *arg1 = (egglib::AlleleStatus *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "AlleleStatus_ShP_T1", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AlleleStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AlleleStatus_ShP_T1" "', argument " "1"" of type '" "egglib::AlleleStatus const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AlleleStatus * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::AlleleStatus const *)arg1)->ShP_T1();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AlleleStatus_ShA(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AlleleStatus *arg1 = (egglib::AlleleStatus *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "AlleleStatus_ShA", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AlleleStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AlleleStatus_ShA" "', argument " "1"" of type '" "egglib::AlleleStatus const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AlleleStatus * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::AlleleStatus const *)arg1)->ShA();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AlleleStatus_ShA_T1(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AlleleStatus *arg1 = (egglib::AlleleStatus *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "AlleleStatus_ShA_T1", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AlleleStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AlleleStatus_ShA_T1" "', argument " "1"" of type '" "egglib::AlleleStatus const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AlleleStatus * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::AlleleStatus const *)arg1)->ShA_T1();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AlleleStatus_FxA(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AlleleStatus *arg1 = (egglib::AlleleStatus *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "AlleleStatus_FxA", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AlleleStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AlleleStatus_FxA" "', argument " "1"" of type '" "egglib::AlleleStatus const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AlleleStatus * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::AlleleStatus const *)arg1)->FxA();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AlleleStatus_FxA_T1(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AlleleStatus *arg1 = (egglib::AlleleStatus *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "AlleleStatus_FxA_T1", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AlleleStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AlleleStatus_FxA_T1" "', argument " "1"" of type '" "egglib::AlleleStatus const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AlleleStatus * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::AlleleStatus const *)arg1)->FxA_T1();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AlleleStatus_FxD(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AlleleStatus *arg1 = (egglib::AlleleStatus *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "AlleleStatus_FxD", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AlleleStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AlleleStatus_FxD" "', argument " "1"" of type '" "egglib::AlleleStatus const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AlleleStatus * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::AlleleStatus const *)arg1)->FxD();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_AlleleStatus_FxD_T1(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::AlleleStatus *arg1 = (egglib::AlleleStatus *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "AlleleStatus_FxD_T1", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__AlleleStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "AlleleStatus_FxD_T1" "', argument " "1"" of type '" "egglib::AlleleStatus const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::AlleleStatus * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::AlleleStatus const *)arg1)->FxD_T1();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_AlleleStatus) /* defines _wrap_delete_AlleleStatus_destructor_closure */

SWIGINTERN int _wrap_new_GeneticCode__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  unsigned int arg1 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  egglib::GeneticCode *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_GeneticCode" "', argument " "1"" of type '" "unsigned int""'");
  } 
  arg1 = static_cast< unsigned int >(val1);
  {
    try {
      result = (egglib::GeneticCode *)new egglib::GeneticCode(arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__GeneticCode, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_GeneticCode__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::GeneticCode *result = 0 ;
  
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  {
    try {
      result = (egglib::GeneticCode *)new egglib::GeneticCode();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__GeneticCode, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_GeneticCode(PyObject *self, PyObject *args, PyObject *kwargs) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_GeneticCode")) SWIG_fail;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_GeneticCode", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_GeneticCode__SWIG_1(self, argc, argv);
  }
  if (argc == 1) {
    int _v;
    {
      int res = SWIG_AsVal_unsigned_SS_int(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_GeneticCode__SWIG_0(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_GeneticCode'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::GeneticCode::GeneticCode(unsigned int)\n"
    "    egglib::GeneticCode::GeneticCode()\n");
  return -1;
}


SWIGINTERN PyObject *_wrap_GeneticCode_get_code(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::GeneticCode *arg1 = (egglib::GeneticCode *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "GeneticCode_get_code", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GeneticCode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GeneticCode_get_code" "', argument " "1"" of type '" "egglib::GeneticCode const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GeneticCode * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::GeneticCode const *)arg1)->get_code();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GeneticCode_set_code(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::GeneticCode *arg1 = (egglib::GeneticCode *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GeneticCode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GeneticCode_set_code" "', argument " "1"" of type '" "egglib::GeneticCode *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GeneticCode * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "GeneticCode_set_code" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_code(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GeneticCode_start(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::GeneticCode *arg1 = (egglib::GeneticCode *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GeneticCode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GeneticCode_start" "', argument " "1"" of type '" "egglib::GeneticCode const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GeneticCode * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "GeneticCode_start" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      result = (bool)((egglib::GeneticCode const *)arg1)->start(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GeneticCode_is_stop_unsmart(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::GeneticCode *arg1 = (egglib::GeneticCode *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GeneticCode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GeneticCode_is_stop_unsmart" "', argument " "1"" of type '" "egglib::GeneticCode const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GeneticCode * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "GeneticCode_is_stop_unsmart" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      result = (bool)((egglib::GeneticCode const *)arg1)->is_stop_unsmart(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GeneticCode_NSsites__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::GeneticCode *arg1 = (egglib::GeneticCode *) 0 ;
  int arg2 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  double result;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GeneticCode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GeneticCode_NSsites" "', argument " "1"" of type '" "egglib::GeneticCode const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GeneticCode * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "GeneticCode_NSsites" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "GeneticCode_NSsites" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  {
    try {
      result = (double)((egglib::GeneticCode const *)arg1)->NSsites(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GeneticCode_NSsites__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::GeneticCode *arg1 = (egglib::GeneticCode *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  double result;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GeneticCode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GeneticCode_NSsites" "', argument " "1"" of type '" "egglib::GeneticCode const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GeneticCode * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "GeneticCode_NSsites" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      result = (double)((egglib::GeneticCode const *)arg1)->NSsites(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GeneticCode_Ssites__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::GeneticCode *arg1 = (egglib::GeneticCode *) 0 ;
  int arg2 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  double result;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GeneticCode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GeneticCode_Ssites" "', argument " "1"" of type '" "egglib::GeneticCode const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GeneticCode * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "GeneticCode_Ssites" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "GeneticCode_Ssites" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  {
    try {
      result = (double)((egglib::GeneticCode const *)arg1)->Ssites(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GeneticCode_Ssites__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::GeneticCode *arg1 = (egglib::GeneticCode *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  double result;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GeneticCode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GeneticCode_Ssites" "', argument " "1"" of type '" "egglib::GeneticCode const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GeneticCode * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "GeneticCode_Ssites" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      result = (double)((egglib::GeneticCode const *)arg1)->Ssites(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GeneticCode_translate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::GeneticCode *arg1 = (egglib::GeneticCode *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GeneticCode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GeneticCode_translate" "', argument " "1"" of type '" "egglib::GeneticCode *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GeneticCode * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "GeneticCode_translate" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      result = (int)(arg1)->translate(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GeneticCode_name(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::GeneticCode *arg1 = (egglib::GeneticCode *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "GeneticCode_name", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GeneticCode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GeneticCode_name" "', argument " "1"" of type '" "egglib::GeneticCode const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GeneticCode * >(argp1);
  {
    try {
      result = (char *)((egglib::GeneticCode const *)arg1)->name();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GeneticCode_NSsites__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::GeneticCode *arg1 = (egglib::GeneticCode *) 0 ;
  egglib::SiteHolder *arg2 = 0 ;
  unsigned int *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  double result;
  
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GeneticCode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GeneticCode_NSsites" "', argument " "1"" of type '" "egglib::GeneticCode const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GeneticCode * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_egglib__SiteHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "GeneticCode_NSsites" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "GeneticCode_NSsites" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::SiteHolder * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_unsigned_int,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "GeneticCode_NSsites" "', argument " "3"" of type '" "unsigned int &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "GeneticCode_NSsites" "', argument " "3"" of type '" "unsigned int &""'"); 
  }
  arg3 = reinterpret_cast< unsigned int * >(argp3);
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "GeneticCode_NSsites" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  {
    try {
      result = (double)((egglib::GeneticCode const *)arg1)->NSsites((egglib::SiteHolder const &)*arg2,*arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GeneticCode_NSsites__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::GeneticCode *arg1 = (egglib::GeneticCode *) 0 ;
  egglib::SiteHolder *arg2 = 0 ;
  unsigned int *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  double result;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GeneticCode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GeneticCode_NSsites" "', argument " "1"" of type '" "egglib::GeneticCode const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GeneticCode * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_egglib__SiteHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "GeneticCode_NSsites" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "GeneticCode_NSsites" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::SiteHolder * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_unsigned_int,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "GeneticCode_NSsites" "', argument " "3"" of type '" "unsigned int &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "GeneticCode_NSsites" "', argument " "3"" of type '" "unsigned int &""'"); 
  }
  arg3 = reinterpret_cast< unsigned int * >(argp3);
  {
    try {
      result = (double)((egglib::GeneticCode const *)arg1)->NSsites((egglib::SiteHolder const &)*arg2,*arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GeneticCode_NSsites(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "GeneticCode_NSsites", 0, 4, argv+1))) SWIG_fail;
  argv[0] = self;
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__GeneticCode, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_GeneticCode_NSsites__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__GeneticCode, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_egglib__SiteHolder, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_unsigned_int, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_GeneticCode_NSsites__SWIG_3(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__GeneticCode, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_GeneticCode_NSsites__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__GeneticCode, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_egglib__SiteHolder, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_unsigned_int, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_GeneticCode_NSsites__SWIG_2(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'GeneticCode_NSsites'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::GeneticCode::NSsites(int,bool) const\n"
    "    egglib::GeneticCode::NSsites(int) const\n"
    "    egglib::GeneticCode::NSsites(egglib::SiteHolder const &,unsigned int &,bool) const\n"
    "    egglib::GeneticCode::NSsites(egglib::SiteHolder const &,unsigned int &) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_GeneticCode_Ssites__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::GeneticCode *arg1 = (egglib::GeneticCode *) 0 ;
  egglib::SiteHolder *arg2 = 0 ;
  unsigned int *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  double result;
  
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GeneticCode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GeneticCode_Ssites" "', argument " "1"" of type '" "egglib::GeneticCode const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GeneticCode * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_egglib__SiteHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "GeneticCode_Ssites" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "GeneticCode_Ssites" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::SiteHolder * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_unsigned_int,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "GeneticCode_Ssites" "', argument " "3"" of type '" "unsigned int &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "GeneticCode_Ssites" "', argument " "3"" of type '" "unsigned int &""'"); 
  }
  arg3 = reinterpret_cast< unsigned int * >(argp3);
  ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "GeneticCode_Ssites" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  {
    try {
      result = (double)((egglib::GeneticCode const *)arg1)->Ssites((egglib::SiteHolder const &)*arg2,*arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GeneticCode_Ssites__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::GeneticCode *arg1 = (egglib::GeneticCode *) 0 ;
  egglib::SiteHolder *arg2 = 0 ;
  unsigned int *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  double result;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GeneticCode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GeneticCode_Ssites" "', argument " "1"" of type '" "egglib::GeneticCode const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GeneticCode * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_egglib__SiteHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "GeneticCode_Ssites" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "GeneticCode_Ssites" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::SiteHolder * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_unsigned_int,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "GeneticCode_Ssites" "', argument " "3"" of type '" "unsigned int &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "GeneticCode_Ssites" "', argument " "3"" of type '" "unsigned int &""'"); 
  }
  arg3 = reinterpret_cast< unsigned int * >(argp3);
  {
    try {
      result = (double)((egglib::GeneticCode const *)arg1)->Ssites((egglib::SiteHolder const &)*arg2,*arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GeneticCode_Ssites(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "GeneticCode_Ssites", 0, 4, argv+1))) SWIG_fail;
  argv[0] = self;
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__GeneticCode, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_GeneticCode_Ssites__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__GeneticCode, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_egglib__SiteHolder, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_unsigned_int, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_GeneticCode_Ssites__SWIG_3(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__GeneticCode, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_GeneticCode_Ssites__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__GeneticCode, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_egglib__SiteHolder, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_unsigned_int, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_GeneticCode_Ssites__SWIG_2(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'GeneticCode_Ssites'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::GeneticCode::Ssites(int,bool) const\n"
    "    egglib::GeneticCode::Ssites(int) const\n"
    "    egglib::GeneticCode::Ssites(egglib::SiteHolder const &,unsigned int &,bool) const\n"
    "    egglib::GeneticCode::Ssites(egglib::SiteHolder const &,unsigned int &) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_GeneticCode_ndiff(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  int arg1 ;
  int arg2 ;
  int val1 ;
  int ecode1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "GeneticCode_ndiff", 2, 2, swig_obj)) SWIG_fail;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "GeneticCode_ndiff" "', argument " "1"" of type '" "int""'");
  } 
  arg1 = static_cast< int >(val1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "GeneticCode_ndiff" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      result = (unsigned int)egglib::GeneticCode::ndiff(arg1,arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GeneticCode_diff1(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  int arg1 ;
  int arg2 ;
  int val1 ;
  int ecode1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "GeneticCode_diff1", 2, 2, swig_obj)) SWIG_fail;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "GeneticCode_diff1" "', argument " "1"" of type '" "int""'");
  } 
  arg1 = static_cast< int >(val1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "GeneticCode_diff1" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      result = (bool)egglib::GeneticCode::diff1(arg1,arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GeneticCode_diff2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  int arg1 ;
  int arg2 ;
  int val1 ;
  int ecode1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "GeneticCode_diff2", 2, 2, swig_obj)) SWIG_fail;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "GeneticCode_diff2" "', argument " "1"" of type '" "int""'");
  } 
  arg1 = static_cast< int >(val1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "GeneticCode_diff2" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      result = (bool)egglib::GeneticCode::diff2(arg1,arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GeneticCode_diff3(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  unsigned int arg1 ;
  unsigned int arg2 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "GeneticCode_diff3", 2, 2, swig_obj)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "GeneticCode_diff3" "', argument " "1"" of type '" "unsigned int""'");
  } 
  arg1 = static_cast< unsigned int >(val1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "GeneticCode_diff3" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (bool)egglib::GeneticCode::diff3(arg1,arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GeneticCode_num_codes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "GeneticCode_num_codes", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (unsigned int)egglib::GeneticCode::num_codes();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_GeneticCode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::GeneticCode *arg1 = (egglib::GeneticCode *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_GeneticCode", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GeneticCode, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_GeneticCode" "', argument " "1"" of type '" "egglib::GeneticCode *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GeneticCode * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_GeneticCode) /* defines _wrap_delete_GeneticCode_destructor_closure */

SWIGINTERN int _wrap_new_SiteDiversity(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_SiteDiversity")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_SiteDiversity", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::SiteDiversity *)new egglib::SiteDiversity();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__SiteDiversity, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_SiteDiversity(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_SiteDiversity", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_SiteDiversity" "', argument " "1"" of type '" "egglib::SiteDiversity *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_toggle_off(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_toggle_off", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_toggle_off" "', argument " "1"" of type '" "egglib::SiteDiversity *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      (arg1)->toggle_off();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_toggle_fstats_diplo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_toggle_fstats_diplo", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_toggle_fstats_diplo" "', argument " "1"" of type '" "egglib::SiteDiversity *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      (arg1)->toggle_fstats_diplo();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_toggle_fstats_haplo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_toggle_fstats_haplo", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_toggle_fstats_haplo" "', argument " "1"" of type '" "egglib::SiteDiversity *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      (arg1)->toggle_fstats_haplo();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_toggle_fstats_hier(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_toggle_fstats_hier", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_toggle_fstats_hier" "', argument " "1"" of type '" "egglib::SiteDiversity *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      (arg1)->toggle_fstats_hier();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_toggle_hstats(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_toggle_hstats", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_toggle_hstats" "', argument " "1"" of type '" "egglib::SiteDiversity *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      (arg1)->toggle_hstats();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_set_maf(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_set_maf" "', argument " "1"" of type '" "egglib::SiteDiversity *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SiteDiversity_set_maf" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->set_maf(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_process(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  egglib::FreqHolder *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_process" "', argument " "1"" of type '" "egglib::SiteDiversity *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__FreqHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SiteDiversity_process" "', argument " "2"" of type '" "egglib::FreqHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SiteDiversity_process" "', argument " "2"" of type '" "egglib::FreqHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::FreqHolder * >(argp2);
  {
    try {
      result = (unsigned int)(arg1)->process((egglib::FreqHolder const &)*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_average(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_average", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_average" "', argument " "1"" of type '" "egglib::SiteDiversity *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (unsigned int)(arg1)->average();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_ns(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_ns", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_ns" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->ns();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_nso(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_nso", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_nso" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->nso();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_k(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_k", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_k" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::SiteDiversity const *)arg1)->k();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_npop_eff1(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_npop_eff1", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_npop_eff1" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::SiteDiversity const *)arg1)->npop_eff1();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_npop_eff2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_npop_eff2", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_npop_eff2" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::SiteDiversity const *)arg1)->npop_eff2();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_npop_eff3(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_npop_eff3", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_npop_eff3" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::SiteDiversity const *)arg1)->npop_eff3();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_nclu_eff(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_nclu_eff", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_nclu_eff" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::SiteDiversity const *)arg1)->nclu_eff();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_Aglob(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_Aglob", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_Aglob" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->Aglob();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_Aing(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_Aing", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_Aing" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->Aing();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_Aout(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_Aout", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_Aout" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->Aout();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_global_allele(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_global_allele" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SiteDiversity_global_allele" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (int)((egglib::SiteDiversity const *)arg1)->global_allele(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_S(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_S", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_S" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->S();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_Sd(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_Sd", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_Sd" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->Sd();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_R(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_R", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_R" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->R();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_He(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_He", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_He" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->He();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_Ho(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_Ho", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_Ho" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->Ho();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_He_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_He_pop" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SiteDiversity_He_pop" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->He_pop(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_pairdiff_inter(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_pairdiff_inter", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_pairdiff_inter" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SiteDiversity_pairdiff_inter" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "SiteDiversity_pairdiff_inter" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->pairdiff_inter(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_orientable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_orientable", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_orientable" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (bool)((egglib::SiteDiversity const *)arg1)->orientable();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_derived(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_derived" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SiteDiversity_derived" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->derived(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_a(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_a", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_a" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->a();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_b(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_b", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_b" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->b();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_c(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_c", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_c" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->c();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_a0(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_a0", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_a0" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->a0();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_b1(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_b1", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_b1" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->b1();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_b2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_b2", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_b2" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->b2();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_c0(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_c0", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_c0" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->c0();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_n(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_n", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_n" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->n();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_d(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_d", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_d" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->d();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_Hst(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_Hst", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_Hst" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->Hst();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_Gst(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_Gst", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_Gst" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->Gst();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_Gste(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_Gste", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_Gste" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->Gste();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_D(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_D", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_D" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->D();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_thetaIAM(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_thetaIAM", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_thetaIAM" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->thetaIAM();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_thetaSMM(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_thetaSMM", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_thetaSMM" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->thetaSMM();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_maf(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_maf", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_maf" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->maf();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_maf_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_maf_pop" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SiteDiversity_maf_pop" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::SiteDiversity const *)arg1)->maf_pop(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_num_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_num_pop", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_num_pop" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::SiteDiversity const *)arg1)->num_pop();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_reset", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_reset" "', argument " "1"" of type '" "egglib::SiteDiversity *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      (arg1)->reset();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_nsites1(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_nsites1", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_nsites1" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::SiteDiversity const *)arg1)->nsites1();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_nsites2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_nsites2", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_nsites2" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::SiteDiversity const *)arg1)->nsites2();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_nsites3(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_nsites3", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_nsites3" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::SiteDiversity const *)arg1)->nsites3();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_nsites4(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_nsites4", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_nsites4" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::SiteDiversity const *)arg1)->nsites4();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_nsites5(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_nsites5", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_nsites5" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::SiteDiversity const *)arg1)->nsites5();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_nsites6(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_nsites6", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_nsites6" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::SiteDiversity const *)arg1)->nsites6();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_nsites7(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_nsites7", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_nsites7" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::SiteDiversity const *)arg1)->nsites7();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_nsites8(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_nsites8", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_nsites8" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::SiteDiversity const *)arg1)->nsites8();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_nsites9(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_nsites9", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_nsites9" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::SiteDiversity const *)arg1)->nsites9();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_nsites10(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_nsites10", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_nsites10" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::SiteDiversity const *)arg1)->nsites10();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_nsites11(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_nsites11", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_nsites11" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::SiteDiversity const *)arg1)->nsites11();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteDiversity_nsites12(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteDiversity *arg1 = (egglib::SiteDiversity *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteDiversity_nsites12", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteDiversity, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteDiversity_nsites12" "', argument " "1"" of type '" "egglib::SiteDiversity const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteDiversity * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::SiteDiversity const *)arg1)->nsites12();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_SiteDiversity) /* defines _wrap_delete_SiteDiversity_destructor_closure */

SWIGINTERN int _wrap_new_SiteHolder(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::SiteHolder *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_SiteHolder")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_SiteHolder", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::SiteHolder *)new egglib::SiteHolder();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__SiteHolder, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_SiteHolder(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteHolder *arg1 = (egglib::SiteHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_SiteHolder", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteHolder, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_SiteHolder" "', argument " "1"" of type '" "egglib::SiteHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteHolder * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteHolder_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteHolder *arg1 = (egglib::SiteHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteHolder_reset", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteHolder_reset" "', argument " "1"" of type '" "egglib::SiteHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteHolder * >(argp1);
  {
    try {
      (arg1)->reset();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteHolder_get_position(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteHolder *arg1 = (egglib::SiteHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteHolder_get_position", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteHolder_get_position" "', argument " "1"" of type '" "egglib::SiteHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteHolder * >(argp1);
  {
    try {
      result = (double)((egglib::SiteHolder const *)arg1)->get_position();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteHolder_set_position(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteHolder *arg1 = (egglib::SiteHolder *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteHolder_set_position" "', argument " "1"" of type '" "egglib::SiteHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteHolder * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SiteHolder_set_position" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->set_position(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteHolder_add(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteHolder *arg1 = (egglib::SiteHolder *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteHolder_add" "', argument " "1"" of type '" "egglib::SiteHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SiteHolder_add" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->add(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteHolder_set_sample(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteHolder *arg1 = (egglib::SiteHolder *) 0 ;
  unsigned int arg2 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteHolder_set_sample", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteHolder_set_sample" "', argument " "1"" of type '" "egglib::SiteHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SiteHolder_set_sample" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "SiteHolder_set_sample" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = static_cast< int >(val3);
  {
    try {
      (arg1)->set_sample(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteHolder_get_ns(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteHolder *arg1 = (egglib::SiteHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteHolder_get_ns", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteHolder_get_ns" "', argument " "1"" of type '" "egglib::SiteHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteHolder * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::SiteHolder const *)arg1)->get_ns();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteHolder_get_sample(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteHolder *arg1 = (egglib::SiteHolder *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteHolder_get_sample" "', argument " "1"" of type '" "egglib::SiteHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SiteHolder_get_sample" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (int)((egglib::SiteHolder const *)arg1)->get_sample(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteHolder_del_sample(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteHolder *arg1 = (egglib::SiteHolder *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteHolder_del_sample" "', argument " "1"" of type '" "egglib::SiteHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteHolder * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SiteHolder_del_sample" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->del_sample(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteHolder_get_missing(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteHolder *arg1 = (egglib::SiteHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteHolder_get_missing", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteHolder_get_missing" "', argument " "1"" of type '" "egglib::SiteHolder const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteHolder * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::SiteHolder const *)arg1)->get_missing();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteHolder_process_align__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::SiteHolder *arg1 = (egglib::SiteHolder *) 0 ;
  egglib::DataHolder *arg2 = 0 ;
  unsigned int arg3 ;
  egglib::StructureHolder *arg4 = (egglib::StructureHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  unsigned int result;
  
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteHolder_process_align" "', argument " "1"" of type '" "egglib::SiteHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteHolder * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_egglib__DataHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SiteHolder_process_align" "', argument " "2"" of type '" "egglib::DataHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SiteHolder_process_align" "', argument " "2"" of type '" "egglib::DataHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::DataHolder * >(argp2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "SiteHolder_process_align" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "SiteHolder_process_align" "', argument " "4"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg4 = reinterpret_cast< egglib::StructureHolder * >(argp4);
  {
    try {
      result = (unsigned int)(arg1)->process_align((egglib::DataHolder const &)*arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteHolder_process_align__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::SiteHolder *arg1 = (egglib::SiteHolder *) 0 ;
  egglib::DataHolder *arg2 = 0 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int result;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteHolder_process_align" "', argument " "1"" of type '" "egglib::SiteHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteHolder * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_egglib__DataHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SiteHolder_process_align" "', argument " "2"" of type '" "egglib::DataHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SiteHolder_process_align" "', argument " "2"" of type '" "egglib::DataHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::DataHolder * >(argp2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "SiteHolder_process_align" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (unsigned int)(arg1)->process_align((egglib::DataHolder const &)*arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SiteHolder_process_align(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "SiteHolder_process_align", 0, 4, argv+1))) SWIG_fail;
  argv[0] = self;
  if (argc == 3) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__SiteHolder, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_egglib__DataHolder, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_SiteHolder_process_align__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__SiteHolder, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_egglib__DataHolder, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_p_egglib__StructureHolder, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_SiteHolder_process_align__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'SiteHolder_process_align'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::SiteHolder::process_align(egglib::DataHolder const &,unsigned int,egglib::StructureHolder *)\n"
    "    egglib::SiteHolder::process_align(egglib::DataHolder const &,unsigned int)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_SiteHolder_process_vcf(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::SiteHolder *arg1 = (egglib::SiteHolder *) 0 ;
  egglib::VcfParser *arg2 = 0 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SiteHolder_process_vcf", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__SiteHolder, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SiteHolder_process_vcf" "', argument " "1"" of type '" "egglib::SiteHolder *""'"); 
  }
  arg1 = reinterpret_cast< egglib::SiteHolder * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__VcfParser,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SiteHolder_process_vcf" "', argument " "2"" of type '" "egglib::VcfParser &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SiteHolder_process_vcf" "', argument " "2"" of type '" "egglib::VcfParser &""'"); 
  }
  arg2 = reinterpret_cast< egglib::VcfParser * >(argp2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "SiteHolder_process_vcf" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "SiteHolder_process_vcf" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  {
    try {
      result = (unsigned int)(arg1)->process_vcf(*arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_SiteHolder) /* defines _wrap_delete_SiteHolder_destructor_closure */

SWIGINTERN int _wrap_new_Genotypes(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::Genotypes *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_Genotypes")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_Genotypes", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::Genotypes *)new egglib::Genotypes();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Genotypes, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_Genotypes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Genotypes *arg1 = (egglib::Genotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_Genotypes", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Genotypes, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Genotypes" "', argument " "1"" of type '" "egglib::Genotypes *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Genotypes * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Genotypes_process(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Genotypes *arg1 = (egglib::Genotypes *) 0 ;
  egglib::SiteHolder *arg2 = 0 ;
  egglib::StructureHolder *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Genotypes_process", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Genotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Genotypes_process" "', argument " "1"" of type '" "egglib::Genotypes *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Genotypes * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__SiteHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Genotypes_process" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Genotypes_process" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::SiteHolder * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[1], &argp3, SWIGTYPE_p_egglib__StructureHolder,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Genotypes_process" "', argument " "3"" of type '" "egglib::StructureHolder &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Genotypes_process" "', argument " "3"" of type '" "egglib::StructureHolder &""'"); 
  }
  arg3 = reinterpret_cast< egglib::StructureHolder * >(argp3);
  {
    try {
      (arg1)->process((egglib::SiteHolder const &)*arg2,*arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Genotypes_ploidy(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Genotypes *arg1 = (egglib::Genotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Genotypes_ploidy", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Genotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Genotypes_ploidy" "', argument " "1"" of type '" "egglib::Genotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Genotypes * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Genotypes const *)arg1)->ploidy();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Genotypes_site(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Genotypes *arg1 = (egglib::Genotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::SiteHolder *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "Genotypes_site", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Genotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Genotypes_site" "', argument " "1"" of type '" "egglib::Genotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Genotypes * >(argp1);
  {
    try {
      result = (egglib::SiteHolder *) &((egglib::Genotypes const *)arg1)->site();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__SiteHolder, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Genotypes_heter(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Genotypes *arg1 = (egglib::Genotypes *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Genotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Genotypes_heter" "', argument " "1"" of type '" "egglib::Genotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Genotypes * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Genotypes_heter" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (bool)((egglib::Genotypes const *)arg1)->heter(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Genotypes_genot(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Genotypes *arg1 = (egglib::Genotypes *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Genotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Genotypes_genot" "', argument " "1"" of type '" "egglib::Genotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Genotypes * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Genotypes_genot" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (int *)((egglib::Genotypes const *)arg1)->genot(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_int, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Genotypes_num_genotypes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Genotypes *arg1 = (egglib::Genotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Genotypes_num_genotypes", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Genotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Genotypes_num_genotypes" "', argument " "1"" of type '" "egglib::Genotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Genotypes * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Genotypes const *)arg1)->num_genotypes();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_Genotypes) /* defines _wrap_delete_Genotypes_destructor_closure */

SWIGINTERN int _wrap_new_Diversity1(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_Diversity1")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_Diversity1", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::Diversity1 *)new egglib::Diversity1();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Diversity1, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_Diversity1(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_Diversity1", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Diversity1" "', argument " "1"" of type '" "egglib::Diversity1 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_load(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  egglib::FreqHolder *arg2 = 0 ;
  egglib::SiteDiversity *arg3 = 0 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_load", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_load" "', argument " "1"" of type '" "egglib::Diversity1 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__FreqHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Diversity1_load" "', argument " "2"" of type '" "egglib::FreqHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Diversity1_load" "', argument " "2"" of type '" "egglib::FreqHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::FreqHolder * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[1], &argp3, SWIGTYPE_p_egglib__SiteDiversity,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Diversity1_load" "', argument " "3"" of type '" "egglib::SiteDiversity const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Diversity1_load" "', argument " "3"" of type '" "egglib::SiteDiversity const &""'"); 
  }
  arg3 = reinterpret_cast< egglib::SiteDiversity * >(argp3);
  ecode4 = SWIG_AsVal_double(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Diversity1_load" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->load((egglib::FreqHolder const &)*arg2,(egglib::SiteDiversity const &)*arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_compute(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_compute", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_compute" "', argument " "1"" of type '" "egglib::Diversity1 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (unsigned int)(arg1)->compute();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_reset_stats(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_reset_stats", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_reset_stats" "', argument " "1"" of type '" "egglib::Diversity1 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      (arg1)->reset_stats();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_toggle_off(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_toggle_off", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_toggle_off" "', argument " "1"" of type '" "egglib::Diversity1 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      (arg1)->toggle_off();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_toggle_ori_site(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_toggle_ori_site", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_toggle_ori_site" "', argument " "1"" of type '" "egglib::Diversity1 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      (arg1)->toggle_ori_site();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_toggle_ori_div(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_toggle_ori_div", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_toggle_ori_div" "', argument " "1"" of type '" "egglib::Diversity1 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      (arg1)->toggle_ori_div();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_toggle_basic(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_toggle_basic", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_toggle_basic" "', argument " "1"" of type '" "egglib::Diversity1 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      (arg1)->toggle_basic();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_toggle_site_lists(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_toggle_site_lists", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_toggle_site_lists" "', argument " "1"" of type '" "egglib::Diversity1 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      (arg1)->toggle_site_lists();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_set_option_multiple(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_set_option_multiple" "', argument " "1"" of type '" "egglib::Diversity1 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Diversity1_set_option_multiple" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->set_option_multiple(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_set_option_ns_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_set_option_ns_set" "', argument " "1"" of type '" "egglib::Diversity1 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Diversity1_set_option_ns_set" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_option_ns_set(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_lt(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_lt", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_lt" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Diversity1 const *)arg1)->lt();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_ls(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_ls", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_ls" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Diversity1 const *)arg1)->ls();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_lso(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_lso", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_lso" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Diversity1 const *)arg1)->lso();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_S(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_S", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_S" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Diversity1 const *)arg1)->S();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_Ss(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_Ss", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_Ss" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Diversity1 const *)arg1)->Ss();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_So(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_So", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_So" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Diversity1 const *)arg1)->So();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_Sso(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_Sso", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_Sso" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Diversity1 const *)arg1)->Sso();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_nsingld(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_nsingld", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_nsingld" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Diversity1 const *)arg1)->nsingld();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_site(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_site" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Diversity1_site" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::Diversity1 const *)arg1)->site(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_site_o(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_site_o" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Diversity1_site_o" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::Diversity1 const *)arg1)->site_o(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_singl(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_singl" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Diversity1_singl" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::Diversity1 const *)arg1)->singl(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_singl_o(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_singl_o" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Diversity1_singl_o" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::Diversity1 const *)arg1)->singl_o(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_nall(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_nall" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Diversity1_nall" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::Diversity1 const *)arg1)->nall(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_frq(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_frq", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_frq" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Diversity1_frq" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Diversity1_frq" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (unsigned int)((egglib::Diversity1 const *)arg1)->frq(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_frqp(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_frqp", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_frqp" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Diversity1_frqp" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Diversity1_frqp" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Diversity1_frqp" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  {
    try {
      result = (unsigned int)((egglib::Diversity1 const *)arg1)->frqp(arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_eta(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_eta", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_eta" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Diversity1 const *)arg1)->eta();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_etao(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_etao", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_etao" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Diversity1 const *)arg1)->etao();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_D(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_D", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_D" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity1 const *)arg1)->D();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_Deta(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_Deta", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_Deta" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity1 const *)arg1)->Deta();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_Dstar(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_Dstar", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_Dstar" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity1 const *)arg1)->Dstar();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_Fstar(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_Fstar", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_Fstar" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity1 const *)arg1)->Fstar();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_thetaW(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_thetaW", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_thetaW" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity1 const *)arg1)->thetaW();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_Pi(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_Pi", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_Pi" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity1 const *)arg1)->Pi();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_Dxy(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_Dxy", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_Dxy" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity1 const *)arg1)->Dxy();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_Da(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_Da", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_Da" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity1 const *)arg1)->Da();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_nseff(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_nseff", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_nseff" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity1 const *)arg1)->nseff();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_nsmax(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_nsmax", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_nsmax" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Diversity1 const *)arg1)->nsmax();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_nseffo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_nseffo", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_nseffo" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity1 const *)arg1)->nseffo();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_nsmaxo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_nsmaxo", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_nsmaxo" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Diversity1 const *)arg1)->nsmaxo();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_thetaH(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_thetaH", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_thetaH" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity1 const *)arg1)->thetaH();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_thetaPi(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_thetaPi", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_thetaPi" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity1 const *)arg1)->thetaPi();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_thetaL(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_thetaL", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_thetaL" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity1 const *)arg1)->thetaL();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_Hns(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_Hns", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_Hns" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity1 const *)arg1)->Hns();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_Hsd(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_Hsd", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_Hsd" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity1 const *)arg1)->Hsd();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_E(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_E", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_E" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity1 const *)arg1)->E();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_Dfl(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_Dfl", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_Dfl" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity1 const *)arg1)->Dfl();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_F(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_F", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_F" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity1 const *)arg1)->F();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_pM(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_pM", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_pM" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity1 const *)arg1)->pM();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity1_nM(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity1 *arg1 = (egglib::Diversity1 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity1_nM", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity1, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity1_nM" "', argument " "1"" of type '" "egglib::Diversity1 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity1 * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Diversity1 const *)arg1)->nM();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_Diversity1) /* defines _wrap_delete_Diversity1_destructor_closure */

SWIGINTERN int _wrap_new_Diversity2(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_Diversity2")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_Diversity2", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::Diversity2 *)new egglib::Diversity2();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Diversity2, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_Diversity2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_Diversity2", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Diversity2" "', argument " "1"" of type '" "egglib::Diversity2 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_reset", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_reset" "', argument " "1"" of type '" "egglib::Diversity2 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      (arg1)->reset();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_reset_stats(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_reset_stats", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_reset_stats" "', argument " "1"" of type '" "egglib::Diversity2 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      (arg1)->reset_stats();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_set_structure(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  egglib::StructureHolder *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_set_structure" "', argument " "1"" of type '" "egglib::Diversity2 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__StructureHolder,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Diversity2_set_structure" "', argument " "2"" of type '" "egglib::StructureHolder &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Diversity2_set_structure" "', argument " "2"" of type '" "egglib::StructureHolder &""'"); 
  }
  arg2 = reinterpret_cast< egglib::StructureHolder * >(argp2);
  {
    try {
      (arg1)->set_structure(*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_load(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  egglib::SiteHolder *arg2 = 0 ;
  egglib::SiteDiversity *arg3 = 0 ;
  egglib::FreqHolder *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_load", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_load" "', argument " "1"" of type '" "egglib::Diversity2 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__SiteHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Diversity2_load" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Diversity2_load" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::SiteHolder * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[1], &argp3, SWIGTYPE_p_egglib__SiteDiversity,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Diversity2_load" "', argument " "3"" of type '" "egglib::SiteDiversity const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Diversity2_load" "', argument " "3"" of type '" "egglib::SiteDiversity const &""'"); 
  }
  arg3 = reinterpret_cast< egglib::SiteDiversity * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[2], &argp4, SWIGTYPE_p_egglib__FreqHolder,  0  | 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Diversity2_load" "', argument " "4"" of type '" "egglib::FreqHolder const &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Diversity2_load" "', argument " "4"" of type '" "egglib::FreqHolder const &""'"); 
  }
  arg4 = reinterpret_cast< egglib::FreqHolder * >(argp4);
  {
    try {
      (arg1)->load((egglib::SiteHolder const &)*arg2,(egglib::SiteDiversity const &)*arg3,(egglib::FreqHolder const &)*arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_num_sites(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_num_sites", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_num_sites" "', argument " "1"" of type '" "egglib::Diversity2 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Diversity2 const *)arg1)->num_sites();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_num_orientable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_num_orientable", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_num_orientable" "', argument " "1"" of type '" "egglib::Diversity2 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Diversity2 const *)arg1)->num_orientable();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_num_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_num_clear", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_num_clear" "', argument " "1"" of type '" "egglib::Diversity2 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Diversity2 const *)arg1)->num_clear();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_num_samples(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_num_samples", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_num_samples" "', argument " "1"" of type '" "egglib::Diversity2 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Diversity2 const *)arg1)->num_samples();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_toggle_off(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_toggle_off", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_toggle_off" "', argument " "1"" of type '" "egglib::Diversity2 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      (arg1)->toggle_off();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_toggle_singletons(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_toggle_singletons", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_toggle_singletons" "', argument " "1"" of type '" "egglib::Diversity2 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      (arg1)->toggle_singletons();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_toggle_partitions(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_toggle_partitions", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_toggle_partitions" "', argument " "1"" of type '" "egglib::Diversity2 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      (arg1)->toggle_partitions();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_set_option_multiple(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_set_option_multiple" "', argument " "1"" of type '" "egglib::Diversity2 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Diversity2_set_option_multiple" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->set_option_multiple(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_compute(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_compute", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_compute" "', argument " "1"" of type '" "egglib::Diversity2 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      result = (unsigned int)(arg1)->compute();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_k(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_k", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_k" "', argument " "1"" of type '" "egglib::Diversity2 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity2 const *)arg1)->k();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_ko(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_ko", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_ko" "', argument " "1"" of type '" "egglib::Diversity2 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity2 const *)arg1)->ko();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_R2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_R2", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_R2" "', argument " "1"" of type '" "egglib::Diversity2 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity2 const *)arg1)->R2();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_R3(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_R3", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_R3" "', argument " "1"" of type '" "egglib::Diversity2 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity2 const *)arg1)->R3();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_R4(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_R4", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_R4" "', argument " "1"" of type '" "egglib::Diversity2 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity2 const *)arg1)->R4();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_Ch(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_Ch", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_Ch" "', argument " "1"" of type '" "egglib::Diversity2 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity2 const *)arg1)->Ch();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_R2E(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_R2E", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_R2E" "', argument " "1"" of type '" "egglib::Diversity2 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity2 const *)arg1)->R2E();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_R3E(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_R3E", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_R3E" "', argument " "1"" of type '" "egglib::Diversity2 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity2 const *)arg1)->R3E();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_R4E(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_R4E", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_R4E" "', argument " "1"" of type '" "egglib::Diversity2 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity2 const *)arg1)->R4E();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_ChE(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_ChE", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_ChE" "', argument " "1"" of type '" "egglib::Diversity2 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity2 const *)arg1)->ChE();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_B(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_B", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_B" "', argument " "1"" of type '" "egglib::Diversity2 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity2 const *)arg1)->B();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diversity2_Q(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Diversity2 *arg1 = (egglib::Diversity2 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Diversity2_Q", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Diversity2, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diversity2_Q" "', argument " "1"" of type '" "egglib::Diversity2 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Diversity2 * >(argp1);
  {
    try {
      result = (double)((egglib::Diversity2 const *)arg1)->Q();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_Diversity2) /* defines _wrap_delete_Diversity2_destructor_closure */

SWIGINTERN int _wrap_new_Haplotypes(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_Haplotypes")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_Haplotypes", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::Haplotypes *)new egglib::Haplotypes();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Haplotypes, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_Haplotypes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_Haplotypes", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Haplotypes" "', argument " "1"" of type '" "egglib::Haplotypes *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_setup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  egglib::StructureHolder *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_setup" "', argument " "1"" of type '" "egglib::Haplotypes *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__StructureHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Haplotypes_setup" "', argument " "2"" of type '" "egglib::StructureHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Haplotypes_setup" "', argument " "2"" of type '" "egglib::StructureHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::StructureHolder * >(argp2);
  {
    try {
      (arg1)->setup((egglib::StructureHolder const &)*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_set_structure(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  egglib::StructureHolder *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_set_structure" "', argument " "1"" of type '" "egglib::Haplotypes *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__StructureHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Haplotypes_set_structure" "', argument " "2"" of type '" "egglib::StructureHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Haplotypes_set_structure" "', argument " "2"" of type '" "egglib::StructureHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::StructureHolder * >(argp2);
  {
    try {
      (arg1)->set_structure((egglib::StructureHolder const &)*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_reset_stats(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_reset_stats", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_reset_stats" "', argument " "1"" of type '" "egglib::Haplotypes *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  {
    try {
      (arg1)->reset_stats();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_load(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  egglib::Genotypes *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_load" "', argument " "1"" of type '" "egglib::Haplotypes *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__Genotypes,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Haplotypes_load" "', argument " "2"" of type '" "egglib::Genotypes const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Haplotypes_load" "', argument " "2"" of type '" "egglib::Genotypes const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::Genotypes * >(argp2);
  {
    try {
      (arg1)->load((egglib::Genotypes const &)*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_prepare_impute(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_prepare_impute" "', argument " "1"" of type '" "egglib::Haplotypes *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Haplotypes_prepare_impute" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->prepare_impute(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_resolve(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  egglib::Genotypes *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_resolve" "', argument " "1"" of type '" "egglib::Haplotypes *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__Genotypes,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Haplotypes_resolve" "', argument " "2"" of type '" "egglib::Genotypes const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Haplotypes_resolve" "', argument " "2"" of type '" "egglib::Genotypes const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::Genotypes * >(argp2);
  {
    try {
      (arg1)->resolve((egglib::Genotypes const &)*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_cp_haplotypes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_cp_haplotypes", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_cp_haplotypes" "', argument " "1"" of type '" "egglib::Haplotypes *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  {
    try {
      (arg1)->cp_haplotypes();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_impute(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_impute", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_impute" "', argument " "1"" of type '" "egglib::Haplotypes *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  {
    try {
      (arg1)->impute();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_cp_dist(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_cp_dist", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_cp_dist" "', argument " "1"" of type '" "egglib::Haplotypes *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  {
    try {
      (arg1)->cp_dist();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_cp_stats(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_cp_stats", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_cp_stats" "', argument " "1"" of type '" "egglib::Haplotypes *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  {
    try {
      result = (unsigned int)(arg1)->cp_stats();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_n_ing(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_n_ing", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_n_ing" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Haplotypes const *)arg1)->n_ing();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_n_otg(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_n_otg", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_n_otg" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Haplotypes const *)arg1)->n_otg();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_n_sam(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_n_sam", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_n_sam" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Haplotypes const *)arg1)->n_sam();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_nt_hapl(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_nt_hapl", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_nt_hapl" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Haplotypes const *)arg1)->nt_hapl();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_hapl(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_hapl", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_hapl" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Haplotypes_hapl" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Haplotypes_hapl" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (unsigned int)((egglib::Haplotypes const *)arg1)->hapl(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_map_sample(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_map_sample" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Haplotypes_map_sample" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::Haplotypes const *)arg1)->map_sample(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_n_sites(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_n_sites", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_n_sites" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Haplotypes const *)arg1)->n_sites();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_ne_ing(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_ne_ing", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_ne_ing" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Haplotypes const *)arg1)->ne_ing();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_ne_otg(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_ne_otg", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_ne_otg" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Haplotypes const *)arg1)->ne_otg();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_n_missing(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_n_missing" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Haplotypes_n_missing" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::Haplotypes const *)arg1)->n_missing(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_n_mis(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_n_mis", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_n_mis" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Haplotypes const *)arg1)->n_mis();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_mis_idx(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_mis_idx" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Haplotypes_mis_idx" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::Haplotypes const *)arg1)->mis_idx(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_ni_hapl(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_ni_hapl", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_ni_hapl" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Haplotypes const *)arg1)->ni_hapl();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_ng_hapl(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_ng_hapl", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_ng_hapl" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Haplotypes const *)arg1)->ng_hapl();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_freq_i(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_freq_i" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Haplotypes_freq_i" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::Haplotypes const *)arg1)->freq_i(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_freq_o(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_freq_o" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Haplotypes_freq_o" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::Haplotypes const *)arg1)->freq_o(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_n_potential(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_n_potential" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Haplotypes_n_potential" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::Haplotypes const *)arg1)->n_potential(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_potential(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_potential", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_potential" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Haplotypes_potential" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Haplotypes_potential" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (unsigned int)((egglib::Haplotypes const *)arg1)->potential(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_dist(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_dist", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_dist" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Haplotypes_dist" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Haplotypes_dist" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (unsigned int)((egglib::Haplotypes const *)arg1)->dist(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_Fst(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_Fst", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_Fst" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  {
    try {
      result = (double)((egglib::Haplotypes const *)arg1)->Fst();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_Kst(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_Kst", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_Kst" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  {
    try {
      result = (double)((egglib::Haplotypes const *)arg1)->Kst();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_Snn(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_Snn", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_Snn" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  {
    try {
      result = (double)((egglib::Haplotypes const *)arg1)->Snn();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_n_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_n_pop", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_n_pop" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Haplotypes const *)arg1)->n_pop();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_ne_pop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_ne_pop", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_ne_pop" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Haplotypes const *)arg1)->ne_pop();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_get_site(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  egglib::SiteHolder *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_get_site" "', argument " "1"" of type '" "egglib::Haplotypes *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__SiteHolder,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Haplotypes_get_site" "', argument " "2"" of type '" "egglib::SiteHolder &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Haplotypes_get_site" "', argument " "2"" of type '" "egglib::SiteHolder &""'"); 
  }
  arg2 = reinterpret_cast< egglib::SiteHolder * >(argp2);
  {
    try {
      (arg1)->get_site(*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Haplotypes_invalid(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Haplotypes *arg1 = (egglib::Haplotypes *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "Haplotypes_invalid", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Haplotypes, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Haplotypes_invalid" "', argument " "1"" of type '" "egglib::Haplotypes const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Haplotypes * >(argp1);
  {
    try {
      result = (bool)((egglib::Haplotypes const *)arg1)->invalid();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_Haplotypes) /* defines _wrap_delete_Haplotypes_destructor_closure */

SWIGINTERN int _wrap_new_PairwiseLD(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::PairwiseLD *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_PairwiseLD")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_PairwiseLD", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::PairwiseLD *)new egglib::PairwiseLD();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__PairwiseLD, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_PairwiseLD(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::PairwiseLD *arg1 = (egglib::PairwiseLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_PairwiseLD", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__PairwiseLD, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_PairwiseLD" "', argument " "1"" of type '" "egglib::PairwiseLD *""'"); 
  }
  arg1 = reinterpret_cast< egglib::PairwiseLD * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseLD_process__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::PairwiseLD *arg1 = (egglib::PairwiseLD *) 0 ;
  egglib::SiteHolder *arg2 = 0 ;
  egglib::SiteHolder *arg3 = 0 ;
  egglib::FreqHolder *arg4 = 0 ;
  egglib::FreqHolder *arg5 = 0 ;
  egglib::StructureHolder *arg6 = 0 ;
  unsigned int arg7 ;
  double arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  double val8 ;
  int ecode8 = 0 ;
  bool result;
  
  if ((nobjs < 8) || (nobjs > 8)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__PairwiseLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseLD_process" "', argument " "1"" of type '" "egglib::PairwiseLD *""'"); 
  }
  arg1 = reinterpret_cast< egglib::PairwiseLD * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_egglib__SiteHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PairwiseLD_process" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PairwiseLD_process" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::SiteHolder * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_egglib__SiteHolder,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "PairwiseLD_process" "', argument " "3"" of type '" "egglib::SiteHolder const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PairwiseLD_process" "', argument " "3"" of type '" "egglib::SiteHolder const &""'"); 
  }
  arg3 = reinterpret_cast< egglib::SiteHolder * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_egglib__FreqHolder,  0  | 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "PairwiseLD_process" "', argument " "4"" of type '" "egglib::FreqHolder const &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PairwiseLD_process" "', argument " "4"" of type '" "egglib::FreqHolder const &""'"); 
  }
  arg4 = reinterpret_cast< egglib::FreqHolder * >(argp4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5, SWIGTYPE_p_egglib__FreqHolder,  0  | 0);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "PairwiseLD_process" "', argument " "5"" of type '" "egglib::FreqHolder const &""'"); 
  }
  if (!argp5) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PairwiseLD_process" "', argument " "5"" of type '" "egglib::FreqHolder const &""'"); 
  }
  arg5 = reinterpret_cast< egglib::FreqHolder * >(argp5);
  res6 = SWIG_ConvertPtr(swig_obj[5], &argp6, SWIGTYPE_p_egglib__StructureHolder,  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "PairwiseLD_process" "', argument " "6"" of type '" "egglib::StructureHolder &""'"); 
  }
  if (!argp6) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PairwiseLD_process" "', argument " "6"" of type '" "egglib::StructureHolder &""'"); 
  }
  arg6 = reinterpret_cast< egglib::StructureHolder * >(argp6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(swig_obj[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "PairwiseLD_process" "', argument " "7"" of type '" "unsigned int""'");
  } 
  arg7 = static_cast< unsigned int >(val7);
  ecode8 = SWIG_AsVal_double(swig_obj[7], &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "PairwiseLD_process" "', argument " "8"" of type '" "double""'");
  } 
  arg8 = static_cast< double >(val8);
  {
    try {
      result = (bool)(arg1)->process((egglib::SiteHolder const &)*arg2,(egglib::SiteHolder const &)*arg3,(egglib::FreqHolder const &)*arg4,(egglib::FreqHolder const &)*arg5,*arg6,arg7,arg8);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseLD_process__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::PairwiseLD *arg1 = (egglib::PairwiseLD *) 0 ;
  egglib::SiteHolder *arg2 = 0 ;
  egglib::SiteHolder *arg3 = 0 ;
  egglib::FreqHolder *arg4 = 0 ;
  egglib::FreqHolder *arg5 = 0 ;
  egglib::StructureHolder *arg6 = 0 ;
  unsigned int arg7 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  bool result;
  
  if ((nobjs < 7) || (nobjs > 7)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__PairwiseLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseLD_process" "', argument " "1"" of type '" "egglib::PairwiseLD *""'"); 
  }
  arg1 = reinterpret_cast< egglib::PairwiseLD * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_egglib__SiteHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PairwiseLD_process" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PairwiseLD_process" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::SiteHolder * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_egglib__SiteHolder,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "PairwiseLD_process" "', argument " "3"" of type '" "egglib::SiteHolder const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PairwiseLD_process" "', argument " "3"" of type '" "egglib::SiteHolder const &""'"); 
  }
  arg3 = reinterpret_cast< egglib::SiteHolder * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_egglib__FreqHolder,  0  | 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "PairwiseLD_process" "', argument " "4"" of type '" "egglib::FreqHolder const &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PairwiseLD_process" "', argument " "4"" of type '" "egglib::FreqHolder const &""'"); 
  }
  arg4 = reinterpret_cast< egglib::FreqHolder * >(argp4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5, SWIGTYPE_p_egglib__FreqHolder,  0  | 0);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "PairwiseLD_process" "', argument " "5"" of type '" "egglib::FreqHolder const &""'"); 
  }
  if (!argp5) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PairwiseLD_process" "', argument " "5"" of type '" "egglib::FreqHolder const &""'"); 
  }
  arg5 = reinterpret_cast< egglib::FreqHolder * >(argp5);
  res6 = SWIG_ConvertPtr(swig_obj[5], &argp6, SWIGTYPE_p_egglib__StructureHolder,  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "PairwiseLD_process" "', argument " "6"" of type '" "egglib::StructureHolder &""'"); 
  }
  if (!argp6) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PairwiseLD_process" "', argument " "6"" of type '" "egglib::StructureHolder &""'"); 
  }
  arg6 = reinterpret_cast< egglib::StructureHolder * >(argp6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(swig_obj[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "PairwiseLD_process" "', argument " "7"" of type '" "unsigned int""'");
  } 
  arg7 = static_cast< unsigned int >(val7);
  {
    try {
      result = (bool)(arg1)->process((egglib::SiteHolder const &)*arg2,(egglib::SiteHolder const &)*arg3,(egglib::FreqHolder const &)*arg4,(egglib::FreqHolder const &)*arg5,*arg6,arg7);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseLD_process__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::PairwiseLD *arg1 = (egglib::PairwiseLD *) 0 ;
  egglib::SiteHolder *arg2 = 0 ;
  egglib::SiteHolder *arg3 = 0 ;
  egglib::FreqHolder *arg4 = 0 ;
  egglib::FreqHolder *arg5 = 0 ;
  egglib::StructureHolder *arg6 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  bool result;
  
  if ((nobjs < 6) || (nobjs > 6)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__PairwiseLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseLD_process" "', argument " "1"" of type '" "egglib::PairwiseLD *""'"); 
  }
  arg1 = reinterpret_cast< egglib::PairwiseLD * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_egglib__SiteHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PairwiseLD_process" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PairwiseLD_process" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::SiteHolder * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_egglib__SiteHolder,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "PairwiseLD_process" "', argument " "3"" of type '" "egglib::SiteHolder const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PairwiseLD_process" "', argument " "3"" of type '" "egglib::SiteHolder const &""'"); 
  }
  arg3 = reinterpret_cast< egglib::SiteHolder * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_egglib__FreqHolder,  0  | 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "PairwiseLD_process" "', argument " "4"" of type '" "egglib::FreqHolder const &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PairwiseLD_process" "', argument " "4"" of type '" "egglib::FreqHolder const &""'"); 
  }
  arg4 = reinterpret_cast< egglib::FreqHolder * >(argp4);
  res5 = SWIG_ConvertPtr(swig_obj[4], &argp5, SWIGTYPE_p_egglib__FreqHolder,  0  | 0);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "PairwiseLD_process" "', argument " "5"" of type '" "egglib::FreqHolder const &""'"); 
  }
  if (!argp5) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PairwiseLD_process" "', argument " "5"" of type '" "egglib::FreqHolder const &""'"); 
  }
  arg5 = reinterpret_cast< egglib::FreqHolder * >(argp5);
  res6 = SWIG_ConvertPtr(swig_obj[5], &argp6, SWIGTYPE_p_egglib__StructureHolder,  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "PairwiseLD_process" "', argument " "6"" of type '" "egglib::StructureHolder &""'"); 
  }
  if (!argp6) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PairwiseLD_process" "', argument " "6"" of type '" "egglib::StructureHolder &""'"); 
  }
  arg6 = reinterpret_cast< egglib::StructureHolder * >(argp6);
  {
    try {
      result = (bool)(arg1)->process((egglib::SiteHolder const &)*arg2,(egglib::SiteHolder const &)*arg3,(egglib::FreqHolder const &)*arg4,(egglib::FreqHolder const &)*arg5,*arg6);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseLD_process(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[9] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "PairwiseLD_process", 0, 8, argv+1))) SWIG_fail;
  argv[0] = self;
  if (argc == 6) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__PairwiseLD, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_egglib__SiteHolder, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_egglib__SiteHolder, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_ConvertPtr(argv[3], 0, SWIGTYPE_p_egglib__FreqHolder, SWIG_POINTER_NO_NULL | 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            int res = SWIG_ConvertPtr(argv[4], 0, SWIGTYPE_p_egglib__FreqHolder, SWIG_POINTER_NO_NULL | 0);
            _v = SWIG_CheckState(res);
            if (_v) {
              void *vptr = 0;
              int res = SWIG_ConvertPtr(argv[5], &vptr, SWIGTYPE_p_egglib__StructureHolder, SWIG_POINTER_NO_NULL);
              _v = SWIG_CheckState(res);
              if (_v) {
                return _wrap_PairwiseLD_process__SWIG_2(self, argc, argv);
              }
            }
          }
        }
      }
    }
  }
  if (argc == 7) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__PairwiseLD, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_egglib__SiteHolder, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_egglib__SiteHolder, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_ConvertPtr(argv[3], 0, SWIGTYPE_p_egglib__FreqHolder, SWIG_POINTER_NO_NULL | 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            int res = SWIG_ConvertPtr(argv[4], 0, SWIGTYPE_p_egglib__FreqHolder, SWIG_POINTER_NO_NULL | 0);
            _v = SWIG_CheckState(res);
            if (_v) {
              void *vptr = 0;
              int res = SWIG_ConvertPtr(argv[5], &vptr, SWIGTYPE_p_egglib__StructureHolder, SWIG_POINTER_NO_NULL);
              _v = SWIG_CheckState(res);
              if (_v) {
                {
                  int res = SWIG_AsVal_unsigned_SS_int(argv[6], NULL);
                  _v = SWIG_CheckState(res);
                }
                if (_v) {
                  return _wrap_PairwiseLD_process__SWIG_1(self, argc, argv);
                }
              }
            }
          }
        }
      }
    }
  }
  if (argc == 8) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__PairwiseLD, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_egglib__SiteHolder, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_ConvertPtr(argv[2], 0, SWIGTYPE_p_egglib__SiteHolder, SWIG_POINTER_NO_NULL | 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_ConvertPtr(argv[3], 0, SWIGTYPE_p_egglib__FreqHolder, SWIG_POINTER_NO_NULL | 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            int res = SWIG_ConvertPtr(argv[4], 0, SWIGTYPE_p_egglib__FreqHolder, SWIG_POINTER_NO_NULL | 0);
            _v = SWIG_CheckState(res);
            if (_v) {
              void *vptr = 0;
              int res = SWIG_ConvertPtr(argv[5], &vptr, SWIGTYPE_p_egglib__StructureHolder, SWIG_POINTER_NO_NULL);
              _v = SWIG_CheckState(res);
              if (_v) {
                {
                  int res = SWIG_AsVal_unsigned_SS_int(argv[6], NULL);
                  _v = SWIG_CheckState(res);
                }
                if (_v) {
                  {
                    int res = SWIG_AsVal_double(argv[7], NULL);
                    _v = SWIG_CheckState(res);
                  }
                  if (_v) {
                    return _wrap_PairwiseLD_process__SWIG_0(self, argc, argv);
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'PairwiseLD_process'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::PairwiseLD::process(egglib::SiteHolder const &,egglib::SiteHolder const &,egglib::FreqHolder const &,egglib::FreqHolder const &,egglib::StructureHolder &,unsigned int,double)\n"
    "    egglib::PairwiseLD::process(egglib::SiteHolder const &,egglib::SiteHolder const &,egglib::FreqHolder const &,egglib::FreqHolder const &,egglib::StructureHolder &,unsigned int)\n"
    "    egglib::PairwiseLD::process(egglib::SiteHolder const &,egglib::SiteHolder const &,egglib::FreqHolder const &,egglib::FreqHolder const &,egglib::StructureHolder &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_PairwiseLD_num_alleles1(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::PairwiseLD *arg1 = (egglib::PairwiseLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "PairwiseLD_num_alleles1", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__PairwiseLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseLD_num_alleles1" "', argument " "1"" of type '" "egglib::PairwiseLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::PairwiseLD * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::PairwiseLD const *)arg1)->num_alleles1();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseLD_num_alleles2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::PairwiseLD *arg1 = (egglib::PairwiseLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "PairwiseLD_num_alleles2", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__PairwiseLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseLD_num_alleles2" "', argument " "1"" of type '" "egglib::PairwiseLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::PairwiseLD * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::PairwiseLD const *)arg1)->num_alleles2();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseLD_index1(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::PairwiseLD *arg1 = (egglib::PairwiseLD *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__PairwiseLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseLD_index1" "', argument " "1"" of type '" "egglib::PairwiseLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::PairwiseLD * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseLD_index1" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::PairwiseLD const *)arg1)->index1(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseLD_index2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::PairwiseLD *arg1 = (egglib::PairwiseLD *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__PairwiseLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseLD_index2" "', argument " "1"" of type '" "egglib::PairwiseLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::PairwiseLD * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseLD_index2" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::PairwiseLD const *)arg1)->index2(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseLD_freq1(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::PairwiseLD *arg1 = (egglib::PairwiseLD *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__PairwiseLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseLD_freq1" "', argument " "1"" of type '" "egglib::PairwiseLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::PairwiseLD * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseLD_freq1" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::PairwiseLD const *)arg1)->freq1(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseLD_freq2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::PairwiseLD *arg1 = (egglib::PairwiseLD *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__PairwiseLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseLD_freq2" "', argument " "1"" of type '" "egglib::PairwiseLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::PairwiseLD * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseLD_freq2" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::PairwiseLD const *)arg1)->freq2(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseLD_freq(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::PairwiseLD *arg1 = (egglib::PairwiseLD *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "PairwiseLD_freq", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__PairwiseLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseLD_freq" "', argument " "1"" of type '" "egglib::PairwiseLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::PairwiseLD * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseLD_freq" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "PairwiseLD_freq" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (unsigned int)((egglib::PairwiseLD const *)arg1)->freq(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseLD_nsam(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::PairwiseLD *arg1 = (egglib::PairwiseLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "PairwiseLD_nsam", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__PairwiseLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseLD_nsam" "', argument " "1"" of type '" "egglib::PairwiseLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::PairwiseLD * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::PairwiseLD const *)arg1)->nsam();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseLD_compute(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::PairwiseLD *arg1 = (egglib::PairwiseLD *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "PairwiseLD_compute", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__PairwiseLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseLD_compute" "', argument " "1"" of type '" "egglib::PairwiseLD *""'"); 
  }
  arg1 = reinterpret_cast< egglib::PairwiseLD * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseLD_compute" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "PairwiseLD_compute" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      (arg1)->compute(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseLD_D(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::PairwiseLD *arg1 = (egglib::PairwiseLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "PairwiseLD_D", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__PairwiseLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseLD_D" "', argument " "1"" of type '" "egglib::PairwiseLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::PairwiseLD * >(argp1);
  {
    try {
      result = (double)((egglib::PairwiseLD const *)arg1)->D();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseLD_Dp(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::PairwiseLD *arg1 = (egglib::PairwiseLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "PairwiseLD_Dp", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__PairwiseLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseLD_Dp" "', argument " "1"" of type '" "egglib::PairwiseLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::PairwiseLD * >(argp1);
  {
    try {
      result = (double)((egglib::PairwiseLD const *)arg1)->Dp();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseLD_r(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::PairwiseLD *arg1 = (egglib::PairwiseLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "PairwiseLD_r", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__PairwiseLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseLD_r" "', argument " "1"" of type '" "egglib::PairwiseLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::PairwiseLD * >(argp1);
  {
    try {
      result = (double)((egglib::PairwiseLD const *)arg1)->r();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseLD_rsq(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::PairwiseLD *arg1 = (egglib::PairwiseLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "PairwiseLD_rsq", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__PairwiseLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseLD_rsq" "', argument " "1"" of type '" "egglib::PairwiseLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::PairwiseLD * >(argp1);
  {
    try {
      result = (double)((egglib::PairwiseLD const *)arg1)->rsq();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseLD_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::PairwiseLD *arg1 = (egglib::PairwiseLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "PairwiseLD_reset", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__PairwiseLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseLD_reset" "', argument " "1"" of type '" "egglib::PairwiseLD *""'"); 
  }
  arg1 = reinterpret_cast< egglib::PairwiseLD * >(argp1);
  {
    try {
      (arg1)->reset();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_PairwiseLD) /* defines _wrap_delete_PairwiseLD_destructor_closure */

SWIGINTERN int _wrap_new_MatrixLD(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_MatrixLD")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_MatrixLD", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::MatrixLD *)new egglib::MatrixLD();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__MatrixLD, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_MatrixLD(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_MatrixLD", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_MatrixLD" "', argument " "1"" of type '" "egglib::MatrixLD *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "MatrixLD_reset", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_reset" "', argument " "1"" of type '" "egglib::MatrixLD *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  {
    try {
      (arg1)->reset();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_set_structure(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  egglib::StructureHolder *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_set_structure" "', argument " "1"" of type '" "egglib::MatrixLD *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__StructureHolder,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "MatrixLD_set_structure" "', argument " "2"" of type '" "egglib::StructureHolder &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "MatrixLD_set_structure" "', argument " "2"" of type '" "egglib::StructureHolder &""'"); 
  }
  arg2 = reinterpret_cast< egglib::StructureHolder * >(argp2);
  {
    try {
      (arg1)->set_structure(*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_load(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  egglib::Genotypes *arg2 = 0 ;
  double arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "MatrixLD_load", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_load" "', argument " "1"" of type '" "egglib::MatrixLD *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__Genotypes,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "MatrixLD_load" "', argument " "2"" of type '" "egglib::Genotypes const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "MatrixLD_load" "', argument " "2"" of type '" "egglib::Genotypes const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::Genotypes * >(argp2);
  ecode3 = SWIG_AsVal_double(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "MatrixLD_load" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  {
    try {
      (arg1)->load((egglib::Genotypes const &)*arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_computeLD__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  unsigned int arg2 ;
  double arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_computeLD" "', argument " "1"" of type '" "egglib::MatrixLD *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "MatrixLD_computeLD" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "MatrixLD_computeLD" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  {
    try {
      (arg1)->computeLD(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_computeLD__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_computeLD" "', argument " "1"" of type '" "egglib::MatrixLD *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "MatrixLD_computeLD" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->computeLD(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_computeLD__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_computeLD" "', argument " "1"" of type '" "egglib::MatrixLD *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  {
    try {
      (arg1)->computeLD();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_computeLD(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "MatrixLD_computeLD", 0, 3, argv+1))) SWIG_fail;
  argv[0] = self;
  if (argc == 1) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__MatrixLD, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_MatrixLD_computeLD__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__MatrixLD, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_MatrixLD_computeLD__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__MatrixLD, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_double(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_MatrixLD_computeLD__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'MatrixLD_computeLD'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::MatrixLD::computeLD(unsigned int,double)\n"
    "    egglib::MatrixLD::computeLD(unsigned int)\n"
    "    egglib::MatrixLD::computeLD()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_MatrixLD_num_tot(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "MatrixLD_num_tot", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_num_tot" "', argument " "1"" of type '" "egglib::MatrixLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::MatrixLD const *)arg1)->num_tot();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_num_pairs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "MatrixLD_num_pairs", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_num_pairs" "', argument " "1"" of type '" "egglib::MatrixLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::MatrixLD const *)arg1)->num_pairs();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_num_alleles(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "MatrixLD_num_alleles", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_num_alleles" "', argument " "1"" of type '" "egglib::MatrixLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::MatrixLD const *)arg1)->num_alleles();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_pairLD(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::PairwiseLD *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_pairLD" "', argument " "1"" of type '" "egglib::MatrixLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "MatrixLD_pairLD" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::PairwiseLD *) &((egglib::MatrixLD const *)arg1)->pairLD(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__PairwiseLD, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_index1(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_index1" "', argument " "1"" of type '" "egglib::MatrixLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "MatrixLD_index1" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::MatrixLD const *)arg1)->index1(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_index2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_index2" "', argument " "1"" of type '" "egglib::MatrixLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "MatrixLD_index2" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::MatrixLD const *)arg1)->index2(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_distance(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_distance" "', argument " "1"" of type '" "egglib::MatrixLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "MatrixLD_distance" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::MatrixLD const *)arg1)->distance(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_computeStats__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  egglib::MatrixLD::MultiAllelic arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_computeStats" "', argument " "1"" of type '" "egglib::MatrixLD *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "MatrixLD_computeStats" "', argument " "2"" of type '" "egglib::MatrixLD::MultiAllelic""'");
  } 
  arg2 = static_cast< egglib::MatrixLD::MultiAllelic >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "MatrixLD_computeStats" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      (arg1)->computeStats(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_computeStats__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  egglib::MatrixLD::MultiAllelic arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_computeStats" "', argument " "1"" of type '" "egglib::MatrixLD *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "MatrixLD_computeStats" "', argument " "2"" of type '" "egglib::MatrixLD::MultiAllelic""'");
  } 
  arg2 = static_cast< egglib::MatrixLD::MultiAllelic >(val2);
  {
    try {
      (arg1)->computeStats(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_computeStats__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_computeStats" "', argument " "1"" of type '" "egglib::MatrixLD *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  {
    try {
      (arg1)->computeStats();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_computeStats(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "MatrixLD_computeStats", 0, 3, argv+1))) SWIG_fail;
  argv[0] = self;
  if (argc == 1) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__MatrixLD, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_MatrixLD_computeStats__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__MatrixLD, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_MatrixLD_computeStats__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__MatrixLD, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_MatrixLD_computeStats__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'MatrixLD_computeStats'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::MatrixLD::computeStats(egglib::MatrixLD::MultiAllelic,unsigned int)\n"
    "    egglib::MatrixLD::computeStats(egglib::MatrixLD::MultiAllelic)\n"
    "    egglib::MatrixLD::computeStats()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_MatrixLD_num_allele_pairs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "MatrixLD_num_allele_pairs", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_num_allele_pairs" "', argument " "1"" of type '" "egglib::MatrixLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::MatrixLD const *)arg1)->num_allele_pairs();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_num_allele_pairs_adj(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "MatrixLD_num_allele_pairs_adj", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_num_allele_pairs_adj" "', argument " "1"" of type '" "egglib::MatrixLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::MatrixLD const *)arg1)->num_allele_pairs_adj();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_ZnS(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "MatrixLD_ZnS", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_ZnS" "', argument " "1"" of type '" "egglib::MatrixLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  {
    try {
      result = (double)((egglib::MatrixLD const *)arg1)->ZnS();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_ZnS_star1(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "MatrixLD_ZnS_star1", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_ZnS_star1" "', argument " "1"" of type '" "egglib::MatrixLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  {
    try {
      result = (double)((egglib::MatrixLD const *)arg1)->ZnS_star1();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_ZnS_star2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "MatrixLD_ZnS_star2", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_ZnS_star2" "', argument " "1"" of type '" "egglib::MatrixLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  {
    try {
      result = (double)((egglib::MatrixLD const *)arg1)->ZnS_star2();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_Za(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "MatrixLD_Za", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_Za" "', argument " "1"" of type '" "egglib::MatrixLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  {
    try {
      result = (double)((egglib::MatrixLD const *)arg1)->Za();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_ZZ(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "MatrixLD_ZZ", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_ZZ" "', argument " "1"" of type '" "egglib::MatrixLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  {
    try {
      result = (double)((egglib::MatrixLD const *)arg1)->ZZ();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_computeRmin__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_computeRmin" "', argument " "1"" of type '" "egglib::MatrixLD *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "MatrixLD_computeRmin" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->computeRmin(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_computeRmin__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_computeRmin" "', argument " "1"" of type '" "egglib::MatrixLD *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  {
    try {
      (arg1)->computeRmin();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_computeRmin(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "MatrixLD_computeRmin", 0, 2, argv+1))) SWIG_fail;
  argv[0] = self;
  if (argc == 1) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__MatrixLD, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_MatrixLD_computeRmin__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__MatrixLD, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_MatrixLD_computeRmin__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'MatrixLD_computeRmin'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::MatrixLD::computeRmin(bool)\n"
    "    egglib::MatrixLD::computeRmin()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_MatrixLD_Rmin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "MatrixLD_Rmin", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_Rmin" "', argument " "1"" of type '" "egglib::MatrixLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::MatrixLD const *)arg1)->Rmin();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_Rmin_num_sites(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "MatrixLD_Rmin_num_sites", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_Rmin_num_sites" "', argument " "1"" of type '" "egglib::MatrixLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::MatrixLD const *)arg1)->Rmin_num_sites();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_Rmin_left(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_Rmin_left" "', argument " "1"" of type '" "egglib::MatrixLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "MatrixLD_Rmin_left" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::MatrixLD const *)arg1)->Rmin_left(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_Rmin_right(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_Rmin_right" "', argument " "1"" of type '" "egglib::MatrixLD const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "MatrixLD_Rmin_right" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::MatrixLD const *)arg1)->Rmin_right(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_toggle_off(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "MatrixLD_toggle_off", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_toggle_off" "', argument " "1"" of type '" "egglib::MatrixLD *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  {
    try {
      (arg1)->toggle_off();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_toggle_Rmin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "MatrixLD_toggle_Rmin", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_toggle_Rmin" "', argument " "1"" of type '" "egglib::MatrixLD *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  {
    try {
      (arg1)->toggle_Rmin();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_toggle_stats(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "MatrixLD_toggle_stats", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_toggle_stats" "', argument " "1"" of type '" "egglib::MatrixLD *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  {
    try {
      (arg1)->toggle_stats();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_MatrixLD_process(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::MatrixLD *arg1 = (egglib::MatrixLD *) 0 ;
  unsigned int arg2 ;
  double arg3 ;
  egglib::MatrixLD::MultiAllelic arg4 ;
  unsigned int arg5 ;
  bool arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  bool val6 ;
  int ecode6 = 0 ;
  PyObject *swig_obj[6] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "MatrixLD_process", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__MatrixLD, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "MatrixLD_process" "', argument " "1"" of type '" "egglib::MatrixLD *""'"); 
  }
  arg1 = reinterpret_cast< egglib::MatrixLD * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "MatrixLD_process" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "MatrixLD_process" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "MatrixLD_process" "', argument " "4"" of type '" "egglib::MatrixLD::MultiAllelic""'");
  } 
  arg4 = static_cast< egglib::MatrixLD::MultiAllelic >(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "MatrixLD_process" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = static_cast< unsigned int >(val5);
  ecode6 = SWIG_AsVal_bool(swig_obj[4], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "MatrixLD_process" "', argument " "6"" of type '" "bool""'");
  } 
  arg6 = static_cast< bool >(val6);
  {
    try {
      result = (unsigned int)(arg1)->process(arg2,arg3,arg4,arg5,arg6);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_MatrixLD) /* defines _wrap_delete_MatrixLD_destructor_closure */

SWIGINTERN int _wrap_new_ParalogPi(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::ParalogPi *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_ParalogPi")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_ParalogPi", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::ParalogPi *)new egglib::ParalogPi();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__ParalogPi, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_ParalogPi(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ParalogPi *arg1 = (egglib::ParalogPi *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_ParalogPi", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ParalogPi, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_ParalogPi" "', argument " "1"" of type '" "egglib::ParalogPi *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ParalogPi * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ParalogPi_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ParalogPi *arg1 = (egglib::ParalogPi *) 0 ;
  egglib::StructureHolder *arg2 = 0 ;
  egglib::StructureHolder *arg3 = 0 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "ParalogPi_reset", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ParalogPi, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ParalogPi_reset" "', argument " "1"" of type '" "egglib::ParalogPi *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ParalogPi * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__StructureHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "ParalogPi_reset" "', argument " "2"" of type '" "egglib::StructureHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "ParalogPi_reset" "', argument " "2"" of type '" "egglib::StructureHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::StructureHolder * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[1], &argp3, SWIGTYPE_p_egglib__StructureHolder,  0  | 0);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "ParalogPi_reset" "', argument " "3"" of type '" "egglib::StructureHolder const &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "ParalogPi_reset" "', argument " "3"" of type '" "egglib::StructureHolder const &""'"); 
  }
  arg3 = reinterpret_cast< egglib::StructureHolder * >(argp3);
  ecode4 = SWIG_AsVal_double(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "ParalogPi_reset" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->reset((egglib::StructureHolder const &)*arg2,(egglib::StructureHolder const &)*arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ParalogPi_load(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ParalogPi *arg1 = (egglib::ParalogPi *) 0 ;
  egglib::SiteHolder *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ParalogPi, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ParalogPi_load" "', argument " "1"" of type '" "egglib::ParalogPi *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ParalogPi * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__SiteHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "ParalogPi_load" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "ParalogPi_load" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::SiteHolder * >(argp2);
  {
    try {
      (arg1)->load((egglib::SiteHolder const &)*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ParalogPi_num_sites_tot(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ParalogPi *arg1 = (egglib::ParalogPi *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "ParalogPi_num_sites_tot", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ParalogPi, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ParalogPi_num_sites_tot" "', argument " "1"" of type '" "egglib::ParalogPi const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ParalogPi * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::ParalogPi const *)arg1)->num_sites_tot();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ParalogPi_num_sites_paralog(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ParalogPi *arg1 = (egglib::ParalogPi *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ParalogPi, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ParalogPi_num_sites_paralog" "', argument " "1"" of type '" "egglib::ParalogPi const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ParalogPi * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "ParalogPi_num_sites_paralog" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::ParalogPi const *)arg1)->num_sites_paralog(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ParalogPi_num_sites_pair(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ParalogPi *arg1 = (egglib::ParalogPi *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "ParalogPi_num_sites_pair", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ParalogPi, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ParalogPi_num_sites_pair" "', argument " "1"" of type '" "egglib::ParalogPi const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ParalogPi * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "ParalogPi_num_sites_pair" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "ParalogPi_num_sites_pair" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (unsigned int)((egglib::ParalogPi const *)arg1)->num_sites_pair(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ParalogPi_num_samples(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ParalogPi *arg1 = (egglib::ParalogPi *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "ParalogPi_num_samples", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ParalogPi, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ParalogPi_num_samples" "', argument " "1"" of type '" "egglib::ParalogPi const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ParalogPi * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::ParalogPi const *)arg1)->num_samples();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ParalogPi_num_paralogs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ParalogPi *arg1 = (egglib::ParalogPi *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "ParalogPi_num_paralogs", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ParalogPi, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ParalogPi_num_paralogs" "', argument " "1"" of type '" "egglib::ParalogPi const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ParalogPi * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::ParalogPi const *)arg1)->num_paralogs();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ParalogPi_Piw(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ParalogPi *arg1 = (egglib::ParalogPi *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ParalogPi, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ParalogPi_Piw" "', argument " "1"" of type '" "egglib::ParalogPi const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ParalogPi * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "ParalogPi_Piw" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::ParalogPi const *)arg1)->Piw(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ParalogPi_Pib(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ParalogPi *arg1 = (egglib::ParalogPi *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "ParalogPi_Pib", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ParalogPi, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ParalogPi_Pib" "', argument " "1"" of type '" "egglib::ParalogPi const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ParalogPi * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "ParalogPi_Pib" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "ParalogPi_Pib" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (double)((egglib::ParalogPi const *)arg1)->Pib(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_ParalogPi) /* defines _wrap_delete_ParalogPi_destructor_closure */

SWIGINTERN int _wrap_new_Rd(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::Rd *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_Rd")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_Rd", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::Rd *)new egglib::Rd();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Rd, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_Rd(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Rd *arg1 = (egglib::Rd *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_Rd", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Rd, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Rd" "', argument " "1"" of type '" "egglib::Rd *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Rd * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rd_load(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Rd *arg1 = (egglib::Rd *) 0 ;
  egglib::SiteHolder *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Rd, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rd_load" "', argument " "1"" of type '" "egglib::Rd *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Rd * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__SiteHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rd_load" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Rd_load" "', argument " "2"" of type '" "egglib::SiteHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::SiteHolder * >(argp2);
  {
    try {
      (arg1)->load((egglib::SiteHolder const &)*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rd_configure(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Rd *arg1 = (egglib::Rd *) 0 ;
  egglib::StructureHolder *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Rd, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rd_configure" "', argument " "1"" of type '" "egglib::Rd *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Rd * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__StructureHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Rd_configure" "', argument " "2"" of type '" "egglib::StructureHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Rd_configure" "', argument " "2"" of type '" "egglib::StructureHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::StructureHolder * >(argp2);
  {
    try {
      (arg1)->configure((egglib::StructureHolder const &)*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rd_compute(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Rd *arg1 = (egglib::Rd *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Rd_compute", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Rd, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rd_compute" "', argument " "1"" of type '" "egglib::Rd *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Rd * >(argp1);
  {
    try {
      result = (double)(arg1)->compute();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rd_num_loci(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Rd *arg1 = (egglib::Rd *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Rd_num_loci", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Rd, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rd_num_loci" "', argument " "1"" of type '" "egglib::Rd const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Rd * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Rd const *)arg1)->num_loci();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rd_ploidy(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Rd *arg1 = (egglib::Rd *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Rd_ploidy", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Rd, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rd_ploidy" "', argument " "1"" of type '" "egglib::Rd const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Rd * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Rd const *)arg1)->ploidy();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rd_num_indiv(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Rd *arg1 = (egglib::Rd *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Rd_num_indiv", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Rd, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rd_num_indiv" "', argument " "1"" of type '" "egglib::Rd const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Rd * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Rd const *)arg1)->num_indiv();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rd_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Rd *arg1 = (egglib::Rd *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Rd_reset", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Rd, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rd_reset" "', argument " "1"" of type '" "egglib::Rd *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Rd * >(argp1);
  {
    try {
      (arg1)->reset();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Rd_reset_stats(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Rd *arg1 = (egglib::Rd *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Rd_reset_stats", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Rd, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Rd_reset_stats" "', argument " "1"" of type '" "egglib::Rd *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Rd * >(argp1);
  {
    try {
      (arg1)->reset_stats();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_Rd) /* defines _wrap_delete_Rd_destructor_closure */

SWIGINTERN int _wrap_new_Migration(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  unsigned int arg1 ;
  double arg2 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::Migration *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_Migration")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_Migration", 2, 2, swig_obj)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Migration" "', argument " "1"" of type '" "unsigned int""'");
  } 
  arg1 = static_cast< unsigned int >(val1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Migration" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      result = (egglib::Migration *)new egglib::Migration(arg1,arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Migration, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_Migration(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Migration *arg1 = (egglib::Migration *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_Migration", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Migration, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Migration" "', argument " "1"" of type '" "egglib::Migration *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Migration * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Migration_n(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Migration *arg1 = (egglib::Migration *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Migration_n", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Migration, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Migration_n" "', argument " "1"" of type '" "egglib::Migration const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Migration * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Migration const *)arg1)->n();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Migration_set_all(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Migration *arg1 = (egglib::Migration *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Migration, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Migration_set_all" "', argument " "1"" of type '" "egglib::Migration *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Migration * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Migration_set_all" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->set_all(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Migration_set_all_R(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Migration *arg1 = (egglib::Migration *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Migration, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Migration_set_all_R" "', argument " "1"" of type '" "egglib::Migration *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Migration * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Migration_set_all_R" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->set_all_R(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Migration_set_row(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Migration *arg1 = (egglib::Migration *) 0 ;
  unsigned int arg2 ;
  double arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Migration_set_row", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Migration, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Migration_set_row" "', argument " "1"" of type '" "egglib::Migration *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Migration * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Migration_set_row" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Migration_set_row" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  {
    try {
      (arg1)->set_row(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Migration_set_row_R(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Migration *arg1 = (egglib::Migration *) 0 ;
  unsigned int arg2 ;
  double arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Migration_set_row_R", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Migration, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Migration_set_row_R" "', argument " "1"" of type '" "egglib::Migration *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Migration * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Migration_set_row_R" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Migration_set_row_R" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  {
    try {
      (arg1)->set_row_R(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Migration_set_pair(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Migration *arg1 = (egglib::Migration *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Migration_set_pair", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Migration, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Migration_set_pair" "', argument " "1"" of type '" "egglib::Migration *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Migration * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Migration_set_pair" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Migration_set_pair" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_double(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Migration_set_pair" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->set_pair(arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Migration_set_pair_R(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Migration *arg1 = (egglib::Migration *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Migration_set_pair_R", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Migration, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Migration_set_pair_R" "', argument " "1"" of type '" "egglib::Migration *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Migration * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Migration_set_pair_R" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Migration_set_pair_R" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_double(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Migration_set_pair_R" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->set_pair_R(arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Migration_get_row(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Migration *arg1 = (egglib::Migration *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Migration, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Migration_get_row" "', argument " "1"" of type '" "egglib::Migration const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Migration * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Migration_get_row" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::Migration const *)arg1)->get_row(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Migration_get_pair(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Migration *arg1 = (egglib::Migration *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Migration_get_pair", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Migration, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Migration_get_pair" "', argument " "1"" of type '" "egglib::Migration const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Migration * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Migration_get_pair" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Migration_get_pair" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (double)((egglib::Migration const *)arg1)->get_pair(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Migration_restore(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Migration *arg1 = (egglib::Migration *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Migration_restore", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Migration, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Migration_restore" "', argument " "1"" of type '" "egglib::Migration *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Migration * >(argp1);
  {
    try {
      (arg1)->restore();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_Migration) /* defines _wrap_delete_Migration_destructor_closure */

SWIGINTERN int _wrap_new_Event(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::Event::Type arg1 ;
  double arg2 ;
  int val1 ;
  int ecode1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::Event *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_Event")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_Event", 2, 2, swig_obj)) SWIG_fail;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Event" "', argument " "1"" of type '" "egglib::Event::Type""'");
  } 
  arg1 = static_cast< egglib::Event::Type >(val1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Event" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      result = (egglib::Event *)new egglib::Event(arg1,arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Event, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_Event_copy(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Event *arg1 = (egglib::Event *) 0 ;
  egglib::Event *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Event_copy" "', argument " "1"" of type '" "egglib::Event *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Event * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__Event,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Event_copy" "', argument " "2"" of type '" "egglib::Event const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Event_copy" "', argument " "2"" of type '" "egglib::Event const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::Event * >(argp2);
  {
    try {
      (arg1)->copy((egglib::Event const &)*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Event_event_type(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Event *arg1 = (egglib::Event *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::Event::Type result;
  
  if (!SWIG_Python_UnpackTuple(args, "Event_event_type", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Event_event_type" "', argument " "1"" of type '" "egglib::Event const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Event * >(argp1);
  {
    try {
      result = (egglib::Event::Type)((egglib::Event const *)arg1)->event_type();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Event_date(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Event *arg1 = (egglib::Event *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Event_date", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Event_date" "', argument " "1"" of type '" "egglib::Event const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Event * >(argp1);
  {
    try {
      result = (double)((egglib::Event const *)arg1)->date();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Event_get_param(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Event *arg1 = (egglib::Event *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Event_get_param", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Event_get_param" "', argument " "1"" of type '" "egglib::Event const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Event * >(argp1);
  {
    try {
      result = (double)((egglib::Event const *)arg1)->get_param();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Event_get_index(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Event *arg1 = (egglib::Event *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Event_get_index", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Event_get_index" "', argument " "1"" of type '" "egglib::Event const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Event * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Event const *)arg1)->get_index();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Event_get_dest(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Event *arg1 = (egglib::Event *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Event_get_dest", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Event_get_dest" "', argument " "1"" of type '" "egglib::Event const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Event * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Event const *)arg1)->get_dest();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Event_get_number1(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Event *arg1 = (egglib::Event *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Event_get_number1", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Event_get_number1" "', argument " "1"" of type '" "egglib::Event const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Event * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Event const *)arg1)->get_number1();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Event_get_number2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Event *arg1 = (egglib::Event *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Event_get_number2", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Event_get_number2" "', argument " "1"" of type '" "egglib::Event const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Event * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Event const *)arg1)->get_number2();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Event_perform(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Event *arg1 = (egglib::Event *) 0 ;
  egglib::Params *arg2 = (egglib::Params *) 0 ;
  egglib::Coalesce *arg3 = (egglib::Coalesce *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Event_perform", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Event_perform" "', argument " "1"" of type '" "egglib::Event *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Event * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Event_perform" "', argument " "2"" of type '" "egglib::Params *""'"); 
  }
  arg2 = reinterpret_cast< egglib::Params * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[1], &argp3,SWIGTYPE_p_egglib__Coalesce, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Event_perform" "', argument " "3"" of type '" "egglib::Coalesce *""'"); 
  }
  arg3 = reinterpret_cast< egglib::Coalesce * >(argp3);
  {
    try {
      (arg1)->perform(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Event_insert(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Event *arg1 = (egglib::Event *) 0 ;
  egglib::Event *arg2 = (egglib::Event *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Event_insert" "', argument " "1"" of type '" "egglib::Event *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Event * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Event_insert" "', argument " "2"" of type '" "egglib::Event *""'"); 
  }
  arg2 = reinterpret_cast< egglib::Event * >(argp2);
  {
    try {
      (arg1)->insert(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Event_disconnect(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Event *arg1 = (egglib::Event *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Event_disconnect", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Event_disconnect" "', argument " "1"" of type '" "egglib::Event *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Event * >(argp1);
  {
    try {
      (arg1)->disconnect();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Event_next(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Event *arg1 = (egglib::Event *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::Event *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "Event_next", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Event_next" "', argument " "1"" of type '" "egglib::Event *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Event * >(argp1);
  {
    try {
      result = (egglib::Event *)(arg1)->next();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Event, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Event_prev(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Event *arg1 = (egglib::Event *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::Event *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "Event_prev", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Event_prev" "', argument " "1"" of type '" "egglib::Event *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Event * >(argp1);
  {
    try {
      result = (egglib::Event *)(arg1)->prev();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Event, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Event_move(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Event *arg1 = (egglib::Event *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Event_move" "', argument " "1"" of type '" "egglib::Event *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Event * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Event_move" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->move(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Event_set_index(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Event *arg1 = (egglib::Event *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Event_set_index" "', argument " "1"" of type '" "egglib::Event *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Event * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Event_set_index" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_index(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Event_set_param(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Event *arg1 = (egglib::Event *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Event_set_param" "', argument " "1"" of type '" "egglib::Event *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Event * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Event_set_param" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->set_param(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Event_set_dest(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Event *arg1 = (egglib::Event *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Event_set_dest" "', argument " "1"" of type '" "egglib::Event *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Event * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Event_set_dest" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_dest(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Event_set_number1(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Event *arg1 = (egglib::Event *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Event_set_number1" "', argument " "1"" of type '" "egglib::Event *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Event * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Event_set_number1" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_number1(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Event_set_number2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Event *arg1 = (egglib::Event *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Event_set_number2" "', argument " "1"" of type '" "egglib::Event *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Event * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Event_set_number2" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_number2(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Event_get_label(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Event *arg1 = (egglib::Event *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "Event_get_label", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Event_get_label" "', argument " "1"" of type '" "egglib::Event const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Event * >(argp1);
  {
    try {
      result = (char *)((egglib::Event const *)arg1)->get_label();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Event_set_label(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Event *arg1 = (egglib::Event *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Event_set_label" "', argument " "1"" of type '" "egglib::Event *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Event * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Event_set_label" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->set_label((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Event(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Event *arg1 = (egglib::Event *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_Event", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Event, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Event" "', argument " "1"" of type '" "egglib::Event *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Event * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_Event) /* defines _wrap_delete_Event_destructor_closure */

SWIGINTERN int _wrap_new_Params__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::Params *result = 0 ;
  
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  {
    try {
      result = (egglib::Params *)new egglib::Params();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Params, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_Params__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  unsigned int arg1 ;
  double arg2 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  egglib::Params *result = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Params" "', argument " "1"" of type '" "unsigned int""'");
  } 
  arg1 = static_cast< unsigned int >(val1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Params" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      result = (egglib::Params *)new egglib::Params(arg1,arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Params, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_Params__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  unsigned int arg1 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  egglib::Params *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Params" "', argument " "1"" of type '" "unsigned int""'");
  } 
  arg1 = static_cast< unsigned int >(val1);
  {
    try {
      result = (egglib::Params *)new egglib::Params(arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Params, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_Params(PyObject *self, PyObject *args, PyObject *kwargs) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_Params")) SWIG_fail;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_Params", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_Params__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v;
    {
      int res = SWIG_AsVal_unsigned_SS_int(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_Params__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v;
    {
      int res = SWIG_AsVal_unsigned_SS_int(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      {
        int res = SWIG_AsVal_double(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_new_Params__SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_Params'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::Params::Params()\n"
    "    egglib::Params::Params(unsigned int,double)\n"
    "    egglib::Params::Params(unsigned int)\n");
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_Params(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_Params", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Params" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_summary(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_summary", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_summary" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      result = ((egglib::Params const *)arg1)->summary();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_k(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_k", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_k" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Params const *)arg1)->k();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_get_K(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_get_K", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_get_K" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Params const *)arg1)->get_K();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_set_K(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_set_K" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_set_K" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_K(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_get_max_iter(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned long result;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_get_max_iter", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_get_max_iter" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      result = (unsigned long)((egglib::Params const *)arg1)->get_max_iter();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_set_max_iter(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned long arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_set_max_iter" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_set_max_iter" "', argument " "2"" of type '" "unsigned long""'");
  } 
  arg2 = static_cast< unsigned long >(val2);
  {
    try {
      (arg1)->set_max_iter(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_get_random_start_allele(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_get_random_start_allele", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_get_random_start_allele" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      result = (bool)((egglib::Params const *)arg1)->get_random_start_allele();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_set_random_start_allele(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_set_random_start_allele" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_set_random_start_allele" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->set_random_start_allele(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_set_R(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_set_R" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_set_R" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->set_R(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_R_R(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_R_R" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_R_R" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->R_R(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_get_R(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_get_R", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_get_R" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      result = (double)((egglib::Params const *)arg1)->get_R();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_set_theta(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_set_theta" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_set_theta" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->set_theta(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_get_theta(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_get_theta", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_get_theta" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      result = (double)((egglib::Params const *)arg1)->get_theta();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_set_fixed(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_set_fixed" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_set_fixed" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_fixed(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_get_fixed(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_get_fixed", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_get_fixed" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Params const *)arg1)->get_fixed();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_set_mutmodel(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  egglib::Params::MutationModel arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_set_mutmodel" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_set_mutmodel" "', argument " "2"" of type '" "egglib::Params::MutationModel""'");
  } 
  arg2 = static_cast< egglib::Params::MutationModel >(val2);
  {
    try {
      (arg1)->set_mutmodel(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_get_mutmodel(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::Params::MutationModel result;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_get_mutmodel", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_get_mutmodel" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      result = (egglib::Params::MutationModel)((egglib::Params const *)arg1)->get_mutmodel();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_set_TPMproba(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_set_TPMproba" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_set_TPMproba" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->set_TPMproba(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_get_TPMproba(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_get_TPMproba", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_get_TPMproba" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      result = (double)((egglib::Params const *)arg1)->get_TPMproba();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_set_TPMparam(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_set_TPMparam" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_set_TPMparam" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->set_TPMparam(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_get_TPMparam(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_get_TPMparam", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_get_TPMparam" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      result = (double)((egglib::Params const *)arg1)->get_TPMparam();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_set_n1(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_set_n1", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_set_n1" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_set_n1" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Params_set_n1" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      (arg1)->set_n1(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_get_n1(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_get_n1" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_get_n1" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::Params const *)arg1)->get_n1(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_set_n2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_set_n2", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_set_n2" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_set_n2" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Params_set_n2" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      (arg1)->set_n2(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_get_n2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_get_n2" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_get_n2" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::Params const *)arg1)->get_n2(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_set_N(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  double arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_set_N", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_set_N" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_set_N" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Params_set_N" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  {
    try {
      (arg1)->set_N(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_N_R(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  double arg3 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_N_R", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_N_R" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_N_R" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Params_N_R" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  ecode4 = SWIG_AsVal_double(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Params_N_R" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->N_R(arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_get_N(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_get_N" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_get_N" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::Params const *)arg1)->get_N(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_set_G(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  double arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_set_G", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_set_G" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_set_G" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Params_set_G" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  {
    try {
      (arg1)->set_G(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_G_R(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  double arg3 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_G_R", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_G_R" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_G_R" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Params_G_R" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  ecode4 = SWIG_AsVal_double(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Params_G_R" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->G_R(arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_get_G(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_get_G" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_get_G" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::Params const *)arg1)->get_G(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_set_s(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  double arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_set_s", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_set_s" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_set_s" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Params_set_s" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  {
    try {
      (arg1)->set_s(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_s_R(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  double arg3 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_s_R", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_s_R" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_s_R" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Params_s_R" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  ecode4 = SWIG_AsVal_double(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Params_s_R" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->s_R(arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_get_s(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_get_s" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_get_s" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::Params const *)arg1)->get_s(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_get_L(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_get_L", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_get_L" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Params const *)arg1)->get_L();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_set_L(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_set_L" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_set_L" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_L(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_autoSitePos(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_autoSitePos", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_autoSitePos" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      (arg1)->autoSitePos();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_set_sitePos(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  double arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_set_sitePos", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_set_sitePos" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_set_sitePos" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Params_set_sitePos" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  {
    try {
      (arg1)->set_sitePos(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_get_sitePos(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_get_sitePos" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_get_sitePos" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::Params const *)arg1)->get_sitePos(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_set_siteW(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  double arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_set_siteW", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_set_siteW" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_set_siteW" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Params_set_siteW" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  {
    try {
      (arg1)->set_siteW(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_get_siteW(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_get_siteW" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_get_siteW" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::Params const *)arg1)->get_siteW(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_totalSiteW(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_totalSiteW", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_totalSiteW" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      result = (double)((egglib::Params const *)arg1)->totalSiteW();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_M(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::Migration *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_M", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_M" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      result = (egglib::Migration *) &(arg1)->M();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Migration, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_set_transW_matrix(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_set_transW_matrix" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_set_transW_matrix" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->set_transW_matrix(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_get_transW_matrix(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_get_transW_matrix", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_get_transW_matrix" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      result = (bool)((egglib::Params const *)arg1)->get_transW_matrix();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_set_transW_pair(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_set_transW_pair", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_set_transW_pair" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_set_transW_pair" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Params_set_transW_pair" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_double(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Params_set_transW_pair" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->set_transW_pair(arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_get_transW_pair(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_get_transW_pair", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_get_transW_pair" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_get_transW_pair" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Params_get_transW_pair" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (double)((egglib::Params const *)arg1)->get_transW_pair(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_get_transW_row(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_get_transW_row" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_get_transW_row" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::Params const *)arg1)->get_transW_row(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_addChange(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  egglib::Event *arg2 = (egglib::Event *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_addChange" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2,SWIGTYPE_p_egglib__Event, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Params_addChange" "', argument " "2"" of type '" "egglib::Event *""'"); 
  }
  arg2 = reinterpret_cast< egglib::Event * >(argp2);
  {
    try {
      (arg1)->addChange(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_clearChanges(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_clearChanges", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_clearChanges" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      (arg1)->clearChanges();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_numChanges(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_numChanges", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_numChanges" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Params const *)arg1)->numChanges();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_firstChange(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::Event *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_firstChange", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_firstChange" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      result = (egglib::Event *)(arg1)->firstChange();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Event, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_nextChangeDate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_nextChangeDate", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_nextChangeDate" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      result = (double)((egglib::Params const *)arg1)->nextChangeDate();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_nextChangeDo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  egglib::Coalesce *arg2 = (egglib::Coalesce *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_nextChangeDo" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2,SWIGTYPE_p_egglib__Coalesce, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Params_nextChangeDo" "', argument " "2"" of type '" "egglib::Coalesce *""'"); 
  }
  arg2 = reinterpret_cast< egglib::Coalesce * >(argp2);
  {
    try {
      (arg1)->nextChangeDo(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_nDSChanges(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_nDSChanges", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_nDSChanges" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Params const *)arg1)->nDSChanges();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_lastChange(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_lastChange" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Params_lastChange" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::Params const *)arg1)->lastChange(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_restore(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_restore", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_restore" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      (arg1)->restore();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_validate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_validate", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_validate" "', argument " "1"" of type '" "egglib::Params const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      ((egglib::Params const *)arg1)->validate();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Params_get_nsam(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Params *arg1 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Params_get_nsam", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Params_get_nsam" "', argument " "1"" of type '" "egglib::Params *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Params * >(argp1);
  {
    try {
      result = (unsigned int)(arg1)->get_nsam();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_Params) /* defines _wrap_delete_Params_destructor_closure */

SWIGINTERN int _wrap_new_Node__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::Node *result = 0 ;
  
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  {
    try {
      result = (egglib::Node *)new egglib::Node();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Node, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_Node__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::Node *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  egglib::Node *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_egglib__Node,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Node" "', argument " "1"" of type '" "egglib::Node const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Node" "', argument " "1"" of type '" "egglib::Node const &""'"); 
  }
  arg1 = reinterpret_cast< egglib::Node * >(argp1);
  {
    try {
      result = (egglib::Node *)new egglib::Node((egglib::Node const &)*arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Node, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_Node(PyObject *self, PyObject *args, PyObject *kwargs) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_Node")) SWIG_fail;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_Node", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_Node__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_egglib__Node, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_Node__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_Node'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::Node::Node()\n"
    "    egglib::Node::Node(egglib::Node const &)\n");
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_Node(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Node *arg1 = (egglib::Node *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_Node", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Node, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Node" "', argument " "1"" of type '" "egglib::Node *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Node * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Node_set_internal(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Node *arg1 = (egglib::Node *) 0 ;
  double arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Node_set_internal", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Node, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Node_set_internal" "', argument " "1"" of type '" "egglib::Node *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Node * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Node_set_internal" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Node_set_internal" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Node_set_internal" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  {
    try {
      (arg1)->set_internal(arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Node_set_terminal(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Node *arg1 = (egglib::Node *) 0 ;
  double arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Node_set_terminal", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Node, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Node_set_terminal" "', argument " "1"" of type '" "egglib::Node *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Node * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Node_set_terminal" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Node_set_terminal" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      (arg1)->set_terminal(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Node_is_terminal(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Node *arg1 = (egglib::Node *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "Node_is_terminal", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Node, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Node_is_terminal" "', argument " "1"" of type '" "egglib::Node const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Node * >(argp1);
  {
    try {
      result = (bool)((egglib::Node const *)arg1)->is_terminal();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Node_label(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Node *arg1 = (egglib::Node *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Node_label", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Node, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Node_label" "', argument " "1"" of type '" "egglib::Node const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Node * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Node const *)arg1)->label();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Node_son1(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Node *arg1 = (egglib::Node *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Node_son1", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Node, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Node_son1" "', argument " "1"" of type '" "egglib::Node const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Node * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Node const *)arg1)->son1();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Node_son2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Node *arg1 = (egglib::Node *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Node_son2", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Node, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Node_son2" "', argument " "1"" of type '" "egglib::Node const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Node * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Node const *)arg1)->son2();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Node_date(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Node *arg1 = (egglib::Node *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Node_date", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Node, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Node_date" "', argument " "1"" of type '" "egglib::Node const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Node * >(argp1);
  {
    try {
      result = (double)((egglib::Node const *)arg1)->date();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Node_get_L(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Node *arg1 = (egglib::Node *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Node_get_L", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Node, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Node_get_L" "', argument " "1"" of type '" "egglib::Node const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Node * >(argp1);
  {
    try {
      result = (double)((egglib::Node const *)arg1)->get_L();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Node_set_L(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Node *arg1 = (egglib::Node *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Node, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Node_set_L" "', argument " "1"" of type '" "egglib::Node *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Node * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Node_set_L" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->set_L(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Node_nmut(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Node *arg1 = (egglib::Node *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Node_nmut", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Node, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Node_nmut" "', argument " "1"" of type '" "egglib::Node const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Node * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Node const *)arg1)->nmut();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Node_mutationSite(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Node *arg1 = (egglib::Node *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Node, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Node_mutationSite" "', argument " "1"" of type '" "egglib::Node const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Node * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Node_mutationSite" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::Node const *)arg1)->mutationSite(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Node_mutationPos(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Node *arg1 = (egglib::Node *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Node, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Node_mutationPos" "', argument " "1"" of type '" "egglib::Node const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Node * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Node_mutationPos" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::Node const *)arg1)->mutationPos(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Node_addMutation(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Node *arg1 = (egglib::Node *) 0 ;
  unsigned int arg2 ;
  double arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Node_addMutation", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Node, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Node_addMutation" "', argument " "1"" of type '" "egglib::Node *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Node * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Node_addMutation" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Node_addMutation" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  {
    try {
      (arg1)->addMutation(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Node_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Node *arg1 = (egglib::Node *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Node_reset", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Node, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Node_reset" "', argument " "1"" of type '" "egglib::Node *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Node * >(argp1);
  {
    try {
      (arg1)->reset();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Node_clear_mutations(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Node *arg1 = (egglib::Node *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Node_clear_mutations", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Node, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Node_clear_mutations" "', argument " "1"" of type '" "egglib::Node *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Node * >(argp1);
  {
    try {
      (arg1)->clear_mutations();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_Node) /* defines _wrap_delete_Node_destructor_closure */

SWIGINTERN int _wrap_new_Tree__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  unsigned int arg1 ;
  double arg2 ;
  double arg3 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  egglib::Tree *result = 0 ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Tree" "', argument " "1"" of type '" "unsigned int""'");
  } 
  arg1 = static_cast< unsigned int >(val1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Tree" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_Tree" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  {
    try {
      result = (egglib::Tree *)new egglib::Tree(arg1,arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Tree, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_Tree__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::Tree *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  egglib::Tree *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_egglib__Tree,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Tree" "', argument " "1"" of type '" "egglib::Tree const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Tree" "', argument " "1"" of type '" "egglib::Tree const &""'"); 
  }
  arg1 = reinterpret_cast< egglib::Tree * >(argp1);
  {
    try {
      result = (egglib::Tree *)new egglib::Tree((egglib::Tree const &)*arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Tree, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_Tree(PyObject *self, PyObject *args, PyObject *kwargs) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_Tree")) SWIG_fail;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_Tree", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_egglib__Tree, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_Tree__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 3) {
    int _v;
    {
      int res = SWIG_AsVal_unsigned_SS_int(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      {
        int res = SWIG_AsVal_double(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_double(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_new_Tree__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_Tree'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::Tree::Tree(unsigned int,double,double)\n"
    "    egglib::Tree::Tree(egglib::Tree const &)\n");
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_Tree(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Tree *arg1 = (egglib::Tree *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_Tree", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Tree, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Tree" "', argument " "1"" of type '" "egglib::Tree *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Tree * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tree_recomb(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Tree *arg1 = (egglib::Tree *) 0 ;
  double arg2 ;
  egglib::Tree *arg3 = (egglib::Tree *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Tree_recomb", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Tree, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tree_recomb" "', argument " "1"" of type '" "egglib::Tree *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Tree * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tree_recomb" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[1], &argp3,SWIGTYPE_p_egglib__Tree, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Tree_recomb" "', argument " "3"" of type '" "egglib::Tree *""'"); 
  }
  arg3 = reinterpret_cast< egglib::Tree * >(argp3);
  {
    try {
      (arg1)->recomb(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tree_coal(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Tree *arg1 = (egglib::Tree *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Tree_coal", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Tree, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tree_coal" "', argument " "1"" of type '" "egglib::Tree *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Tree * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tree_coal" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Tree_coal" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_double(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Tree_coal" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      result = (unsigned int)(arg1)->coal(arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tree_mutate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Tree *arg1 = (egglib::Tree *) 0 ;
  unsigned int arg2 ;
  egglib::DataHolder *arg3 = 0 ;
  egglib::Params *arg4 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject *swig_obj[4] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Tree_mutate", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Tree, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tree_mutate" "', argument " "1"" of type '" "egglib::Tree *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Tree * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tree_mutate" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[1], &argp3, SWIGTYPE_p_egglib__DataHolder,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Tree_mutate" "', argument " "3"" of type '" "egglib::DataHolder &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Tree_mutate" "', argument " "3"" of type '" "egglib::DataHolder &""'"); 
  }
  arg3 = reinterpret_cast< egglib::DataHolder * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[2], &argp4,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Tree_mutate" "', argument " "4"" of type '" "egglib::Params const *""'"); 
  }
  arg4 = reinterpret_cast< egglib::Params * >(argp4);
  {
    try {
      result = (unsigned int)(arg1)->mutate(arg2,*arg3,(egglib::Params const *)arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tree_start(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Tree *arg1 = (egglib::Tree *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Tree_start", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Tree, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tree_start" "', argument " "1"" of type '" "egglib::Tree const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Tree * >(argp1);
  {
    try {
      result = (double)((egglib::Tree const *)arg1)->start();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tree_stop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Tree *arg1 = (egglib::Tree *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Tree_stop", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Tree, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tree_stop" "', argument " "1"" of type '" "egglib::Tree const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Tree * >(argp1);
  {
    try {
      result = (double)((egglib::Tree const *)arg1)->stop();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tree_cov(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Tree *arg1 = (egglib::Tree *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Tree_cov", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Tree, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tree_cov" "', argument " "1"" of type '" "egglib::Tree const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Tree * >(argp1);
  {
    try {
      result = (double)((egglib::Tree const *)arg1)->cov();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tree_L(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Tree *arg1 = (egglib::Tree *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Tree_L", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Tree, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tree_L" "', argument " "1"" of type '" "egglib::Tree const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Tree * >(argp1);
  {
    try {
      result = (double)((egglib::Tree const *)arg1)->L();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tree_root(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Tree *arg1 = (egglib::Tree *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::Node *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "Tree_root", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Tree, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tree_root" "', argument " "1"" of type '" "egglib::Tree const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Tree * >(argp1);
  {
    try {
      result = (egglib::Node *)((egglib::Tree const *)arg1)->root();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Node, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tree_nnodes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Tree *arg1 = (egglib::Tree *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Tree_nnodes", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Tree, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tree_nnodes" "', argument " "1"" of type '" "egglib::Tree const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Tree * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Tree const *)arg1)->nnodes();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tree_node(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Tree *arg1 = (egglib::Tree *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::Node *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Tree, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tree_node" "', argument " "1"" of type '" "egglib::Tree const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Tree * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tree_node" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::Node *)((egglib::Tree const *)arg1)->node(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Node, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tree_addNode(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Tree *arg1 = (egglib::Tree *) 0 ;
  double arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Tree_addNode", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Tree, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tree_addNode" "', argument " "1"" of type '" "egglib::Tree *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Tree * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tree_addNode" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Tree_addNode" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (unsigned int)(arg1)->addNode(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tree_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Tree *arg1 = (egglib::Tree *) 0 ;
  unsigned int arg2 ;
  double arg3 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Tree_reset", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Tree, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tree_reset" "', argument " "1"" of type '" "egglib::Tree *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Tree * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tree_reset" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Tree_reset" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  ecode4 = SWIG_AsVal_double(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Tree_reset" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->reset(arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tree_clear_mutations(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Tree *arg1 = (egglib::Tree *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Tree_clear_mutations", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Tree, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tree_clear_mutations" "', argument " "1"" of type '" "egglib::Tree *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Tree * >(argp1);
  {
    try {
      (arg1)->clear_mutations();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_Tree) /* defines _wrap_delete_Tree_destructor_closure */

SWIGINTERN int _wrap_new_Coalesce(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::Coalesce *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_Coalesce")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_Coalesce", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::Coalesce *)new egglib::Coalesce();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Coalesce, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_Coalesce(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Coalesce *arg1 = (egglib::Coalesce *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_Coalesce", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Coalesce, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Coalesce" "', argument " "1"" of type '" "egglib::Coalesce *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Coalesce * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Coalesce_simul__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::Coalesce *arg1 = (egglib::Coalesce *) 0 ;
  egglib::Params *arg2 = (egglib::Params *) 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Coalesce, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Coalesce_simul" "', argument " "1"" of type '" "egglib::Coalesce *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Coalesce * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Coalesce_simul" "', argument " "2"" of type '" "egglib::Params *""'"); 
  }
  arg2 = reinterpret_cast< egglib::Params * >(argp2);
  ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Coalesce_simul" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  {
    try {
      (arg1)->simul(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Coalesce_simul__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::Coalesce *arg1 = (egglib::Coalesce *) 0 ;
  egglib::Params *arg2 = (egglib::Params *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Coalesce, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Coalesce_simul" "', argument " "1"" of type '" "egglib::Coalesce *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Coalesce * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_egglib__Params, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Coalesce_simul" "', argument " "2"" of type '" "egglib::Params *""'"); 
  }
  arg2 = reinterpret_cast< egglib::Params * >(argp2);
  {
    try {
      (arg1)->simul(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Coalesce_simul(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Coalesce_simul", 0, 3, argv+1))) SWIG_fail;
  argv[0] = self;
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__Coalesce, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_egglib__Params, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_Coalesce_simul__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__Coalesce, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_egglib__Params, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Coalesce_simul__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Coalesce_simul'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::Coalesce::simul(egglib::Params *,bool)\n"
    "    egglib::Coalesce::simul(egglib::Params *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Coalesce_mutate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Coalesce *arg1 = (egglib::Coalesce *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Coalesce_mutate", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Coalesce, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Coalesce_mutate" "', argument " "1"" of type '" "egglib::Coalesce *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Coalesce * >(argp1);
  {
    try {
      (arg1)->mutate();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Coalesce_coalescence(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Coalesce *arg1 = (egglib::Coalesce *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Coalesce_coalescence", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Coalesce, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Coalesce_coalescence" "', argument " "1"" of type '" "egglib::Coalesce *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Coalesce * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Coalesce_coalescence" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Coalesce_coalescence" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Coalesce_coalescence" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  {
    try {
      (arg1)->coalescence(arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Coalesce_migrate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Coalesce *arg1 = (egglib::Coalesce *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Coalesce_migrate", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Coalesce, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Coalesce_migrate" "', argument " "1"" of type '" "egglib::Coalesce *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Coalesce * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Coalesce_migrate" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Coalesce_migrate" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Coalesce_migrate" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  {
    try {
      (arg1)->migrate(arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Coalesce_admixt(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Coalesce *arg1 = (egglib::Coalesce *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Coalesce_admixt", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Coalesce, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Coalesce_admixt" "', argument " "1"" of type '" "egglib::Coalesce *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Coalesce * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Coalesce_admixt" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Coalesce_admixt" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_double(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Coalesce_admixt" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->admixt(arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Coalesce_bottleneck(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Coalesce *arg1 = (egglib::Coalesce *) 0 ;
  unsigned int arg2 ;
  double arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Coalesce_bottleneck", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Coalesce, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Coalesce_bottleneck" "', argument " "1"" of type '" "egglib::Coalesce *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Coalesce * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Coalesce_bottleneck" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Coalesce_bottleneck" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  {
    try {
      (arg1)->bottleneck(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Coalesce_number_of_trees(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Coalesce *arg1 = (egglib::Coalesce *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Coalesce_number_of_trees", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Coalesce, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Coalesce_number_of_trees" "', argument " "1"" of type '" "egglib::Coalesce const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Coalesce * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Coalesce const *)arg1)->number_of_trees();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Coalesce_delayedSample(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Coalesce *arg1 = (egglib::Coalesce *) 0 ;
  double arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Coalesce_delayedSample", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Coalesce, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Coalesce_delayedSample" "', argument " "1"" of type '" "egglib::Coalesce *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Coalesce * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Coalesce_delayedSample" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Coalesce_delayedSample" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Coalesce_delayedSample" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Coalesce_delayedSample" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = static_cast< unsigned int >(val5);
  {
    try {
      (arg1)->delayedSample(arg2,arg3,arg4,arg5);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Coalesce_tree(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Coalesce *arg1 = (egglib::Coalesce *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::Tree *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Coalesce, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Coalesce_tree" "', argument " "1"" of type '" "egglib::Coalesce const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Coalesce * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Coalesce_tree" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::Tree *)((egglib::Coalesce const *)arg1)->tree(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Tree, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Coalesce_data(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Coalesce *arg1 = (egglib::Coalesce *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::DataHolder *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "Coalesce_data", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Coalesce, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Coalesce_data" "', argument " "1"" of type '" "egglib::Coalesce const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Coalesce * >(argp1);
  {
    try {
      result = (egglib::DataHolder *)((egglib::Coalesce const *)arg1)->data();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Coalesce_tree_start(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Coalesce *arg1 = (egglib::Coalesce *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Coalesce, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Coalesce_tree_start" "', argument " "1"" of type '" "egglib::Coalesce const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Coalesce * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Coalesce_tree_start" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::Coalesce const *)arg1)->tree_start(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Coalesce_tree_stop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Coalesce *arg1 = (egglib::Coalesce *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Coalesce, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Coalesce_tree_stop" "', argument " "1"" of type '" "egglib::Coalesce const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Coalesce * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Coalesce_tree_stop" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::Coalesce const *)arg1)->tree_stop(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Coalesce_site_position(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Coalesce *arg1 = (egglib::Coalesce *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Coalesce, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Coalesce_site_position" "', argument " "1"" of type '" "egglib::Coalesce const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Coalesce * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Coalesce_site_position" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::Coalesce const *)arg1)->site_position(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_Coalesce) /* defines _wrap_delete_Coalesce_destructor_closure */

SWIGINTERN int _wrap_new_EHH(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::EHH *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_EHH")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_EHH", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::EHH *)new egglib::EHH();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__EHH, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_EHH(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_EHH", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_EHH" "', argument " "1"" of type '" "egglib::EHH *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_set_core(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  egglib::SiteHolder *arg2 = (egglib::SiteHolder *) 0 ;
  egglib::StructureHolder *arg3 = (egglib::StructureHolder *) 0 ;
  double arg4 ;
  double arg5 ;
  double arg6 ;
  double arg7 ;
  unsigned int arg8 ;
  unsigned int arg9 ;
  bool arg10 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  double val4 ;
  int ecode4 = 0 ;
  double val5 ;
  int ecode5 = 0 ;
  double val6 ;
  int ecode6 = 0 ;
  double val7 ;
  int ecode7 = 0 ;
  unsigned int val8 ;
  int ecode8 = 0 ;
  unsigned int val9 ;
  int ecode9 = 0 ;
  bool val10 ;
  int ecode10 = 0 ;
  PyObject *swig_obj[10] ;
  
  if (!SWIG_Python_UnpackTuple(args, "EHH_set_core", 9, 9, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_set_core" "', argument " "1"" of type '" "egglib::EHH *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2,SWIGTYPE_p_egglib__SiteHolder, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "EHH_set_core" "', argument " "2"" of type '" "egglib::SiteHolder const *""'"); 
  }
  arg2 = reinterpret_cast< egglib::SiteHolder * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[1], &argp3,SWIGTYPE_p_egglib__StructureHolder, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "EHH_set_core" "', argument " "3"" of type '" "egglib::StructureHolder *""'"); 
  }
  arg3 = reinterpret_cast< egglib::StructureHolder * >(argp3);
  ecode4 = SWIG_AsVal_double(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "EHH_set_core" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  ecode5 = SWIG_AsVal_double(swig_obj[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "EHH_set_core" "', argument " "5"" of type '" "double""'");
  } 
  arg5 = static_cast< double >(val5);
  ecode6 = SWIG_AsVal_double(swig_obj[4], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "EHH_set_core" "', argument " "6"" of type '" "double""'");
  } 
  arg6 = static_cast< double >(val6);
  ecode7 = SWIG_AsVal_double(swig_obj[5], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "EHH_set_core" "', argument " "7"" of type '" "double""'");
  } 
  arg7 = static_cast< double >(val7);
  ecode8 = SWIG_AsVal_unsigned_SS_int(swig_obj[6], &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "EHH_set_core" "', argument " "8"" of type '" "unsigned int""'");
  } 
  arg8 = static_cast< unsigned int >(val8);
  ecode9 = SWIG_AsVal_unsigned_SS_int(swig_obj[7], &val9);
  if (!SWIG_IsOK(ecode9)) {
    SWIG_exception_fail(SWIG_ArgError(ecode9), "in method '" "EHH_set_core" "', argument " "9"" of type '" "unsigned int""'");
  } 
  arg9 = static_cast< unsigned int >(val9);
  ecode10 = SWIG_AsVal_bool(swig_obj[8], &val10);
  if (!SWIG_IsOK(ecode10)) {
    SWIG_exception_fail(SWIG_ArgError(ecode10), "in method '" "EHH_set_core" "', argument " "10"" of type '" "bool""'");
  } 
  arg10 = static_cast< bool >(val10);
  {
    try {
      (arg1)->set_core((egglib::SiteHolder const *)arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_load_distant(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  egglib::SiteHolder *arg2 = (egglib::SiteHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_load_distant" "', argument " "1"" of type '" "egglib::EHH *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2,SWIGTYPE_p_egglib__SiteHolder, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "EHH_load_distant" "', argument " "2"" of type '" "egglib::SiteHolder const *""'"); 
  }
  arg2 = reinterpret_cast< egglib::SiteHolder * >(argp2);
  {
    try {
      (arg1)->load_distant((egglib::SiteHolder const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_K_core(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "EHH_K_core", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_K_core" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::EHH const *)arg1)->K_core();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_K_cur(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "EHH_K_cur", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_K_cur" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::EHH const *)arg1)->K_cur();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_num_avail_tot(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "EHH_num_avail_tot", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_num_avail_tot" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::EHH const *)arg1)->num_avail_tot();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_num_avail_core(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_num_avail_core" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EHH_num_avail_core" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::EHH const *)arg1)->num_avail_core(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_num_avail_cur(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_num_avail_cur" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EHH_num_avail_cur" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::EHH const *)arg1)->num_avail_cur(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_EHHS(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "EHH_EHHS", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_EHHS" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  {
    try {
      result = (double)((egglib::EHH const *)arg1)->EHHS();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_EHHG(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "EHH_EHHG", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_EHHG" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  {
    try {
      result = (double)((egglib::EHH const *)arg1)->EHHG();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_EHHi(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_EHHi" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EHH_EHHi" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::EHH const *)arg1)->EHHi(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_EHHc(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_EHHc" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EHH_EHHc" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::EHH const *)arg1)->EHHc(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_rEHH(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_rEHH" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EHH_rEHH" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::EHH const *)arg1)->rEHH(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_iES(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "EHH_iES", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_iES" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  {
    try {
      result = (double)((egglib::EHH const *)arg1)->iES();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_iEG(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "EHH_iEG", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_iEG" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  {
    try {
      result = (double)((egglib::EHH const *)arg1)->iEG();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_iHH(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_iHH" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EHH_iHH" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::EHH const *)arg1)->iHH(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_iHHc(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_iHHc" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EHH_iHHc" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::EHH const *)arg1)->iHHc(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_iHS(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_iHS" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EHH_iHS" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::EHH const *)arg1)->iHS(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_num_EHH_done(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "EHH_num_EHH_done", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_num_EHH_done" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::EHH const *)arg1)->num_EHH_done();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_num_EHHc_done(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "EHH_num_EHHc_done", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_num_EHHc_done" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::EHH const *)arg1)->num_EHHc_done();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_flag_EHHS_done(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "EHH_flag_EHHS_done", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_flag_EHHS_done" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  {
    try {
      result = (bool)((egglib::EHH const *)arg1)->flag_EHHS_done();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_flag_EHHG_done(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "EHH_flag_EHHG_done", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_flag_EHHG_done" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  {
    try {
      result = (bool)((egglib::EHH const *)arg1)->flag_EHHG_done();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_dEHH(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_dEHH" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EHH_dEHH" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::EHH const *)arg1)->dEHH(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_dEHHc(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_dEHHc" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EHH_dEHHc" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::EHH const *)arg1)->dEHHc(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_dEHH_max(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "EHH_dEHH_max", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_dEHH_max" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  {
    try {
      result = (double)((egglib::EHH const *)arg1)->dEHH_max();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_dEHH_mean(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "EHH_dEHH_mean", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_dEHH_mean" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  {
    try {
      result = (double)((egglib::EHH const *)arg1)->dEHH_mean();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_dEHHS(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "EHH_dEHHS", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_dEHHS" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  {
    try {
      result = (double)((egglib::EHH const *)arg1)->dEHHS();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EHH_dEHHG(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::EHH *arg1 = (egglib::EHH *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "EHH_dEHHG", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__EHH, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EHH_dEHHG" "', argument " "1"" of type '" "egglib::EHH const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::EHH * >(argp1);
  {
    try {
      result = (double)((egglib::EHH const *)arg1)->dEHHG();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_EHH) /* defines _wrap_delete_EHH_destructor_closure */

SWIGINTERN int _wrap_new_BaseFormatter(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::BaseFormatter *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_BaseFormatter")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_BaseFormatter", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::BaseFormatter *)new egglib::BaseFormatter();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__BaseFormatter, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_BaseFormatter(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::BaseFormatter *arg1 = (egglib::BaseFormatter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_BaseFormatter", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__BaseFormatter, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_BaseFormatter" "', argument " "1"" of type '" "egglib::BaseFormatter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::BaseFormatter * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BaseFormatter_open_file(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::BaseFormatter *arg1 = (egglib::BaseFormatter *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__BaseFormatter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BaseFormatter_open_file" "', argument " "1"" of type '" "egglib::BaseFormatter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::BaseFormatter * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "BaseFormatter_open_file" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      result = (bool)(arg1)->open_file((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_BaseFormatter_to_str(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::BaseFormatter *arg1 = (egglib::BaseFormatter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "BaseFormatter_to_str", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__BaseFormatter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BaseFormatter_to_str" "', argument " "1"" of type '" "egglib::BaseFormatter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::BaseFormatter * >(argp1);
  {
    try {
      (arg1)->to_str();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BaseFormatter_get_str(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::BaseFormatter *arg1 = (egglib::BaseFormatter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!SWIG_Python_UnpackTuple(args, "BaseFormatter_get_str", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__BaseFormatter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BaseFormatter_get_str" "', argument " "1"" of type '" "egglib::BaseFormatter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::BaseFormatter * >(argp1);
  {
    try {
      result = (arg1)->get_str();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BaseFormatter_to_cout(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::BaseFormatter *arg1 = (egglib::BaseFormatter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "BaseFormatter_to_cout", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__BaseFormatter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BaseFormatter_to_cout" "', argument " "1"" of type '" "egglib::BaseFormatter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::BaseFormatter * >(argp1);
  {
    try {
      (arg1)->to_cout();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BaseFormatter_write(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::BaseFormatter *arg1 = (egglib::BaseFormatter *) 0 ;
  char *arg2 = (char *) 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "BaseFormatter_write", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__BaseFormatter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BaseFormatter_write" "', argument " "1"" of type '" "egglib::BaseFormatter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::BaseFormatter * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "BaseFormatter_write" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  ecode3 = SWIG_AsVal_bool(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "BaseFormatter_write" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  {
    try {
      (arg1)->write((char const *)arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_BaseFormatter_flush(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::BaseFormatter *arg1 = (egglib::BaseFormatter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "BaseFormatter_flush", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__BaseFormatter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BaseFormatter_flush" "', argument " "1"" of type '" "egglib::BaseFormatter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::BaseFormatter * >(argp1);
  {
    try {
      (arg1)->flush();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BaseFormatter_close(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::BaseFormatter *arg1 = (egglib::BaseFormatter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "BaseFormatter_close", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__BaseFormatter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BaseFormatter_close" "', argument " "1"" of type '" "egglib::BaseFormatter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::BaseFormatter * >(argp1);
  {
    try {
      (arg1)->close();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_BaseFormatter) /* defines _wrap_delete_BaseFormatter_destructor_closure */

SWIGINTERN int _wrap_new_Export(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::Export *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_Export")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_Export", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::Export *)new egglib::Export();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Export, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_Export(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Export *arg1 = (egglib::Export *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_Export", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Export, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Export" "', argument " "1"" of type '" "egglib::Export *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Export * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Export_newick(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Export *arg1 = (egglib::Export *) 0 ;
  egglib::Tree *arg2 = 0 ;
  bool arg3 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Export_newick", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Export, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Export_newick" "', argument " "1"" of type '" "egglib::Export *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Export * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__Tree,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Export_newick" "', argument " "2"" of type '" "egglib::Tree const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Export_newick" "', argument " "2"" of type '" "egglib::Tree const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::Tree * >(argp2);
  ecode3 = SWIG_AsVal_bool(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Export_newick" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  ecode4 = SWIG_AsVal_bool(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Export_newick" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  {
    try {
      (arg1)->newick((egglib::Tree const &)*arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Export_ms_num_positions(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Export *arg1 = (egglib::Export *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Export, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Export_ms_num_positions" "', argument " "1"" of type '" "egglib::Export *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Export * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Export_ms_num_positions" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->ms_num_positions(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Export_ms_position(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Export *arg1 = (egglib::Export *) 0 ;
  unsigned int arg2 ;
  double arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Export_ms_position", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Export, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Export_ms_position" "', argument " "1"" of type '" "egglib::Export *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Export * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Export_ms_position" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Export_ms_position" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  {
    try {
      (arg1)->ms_position(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Export_ms_auto_positions(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Export *arg1 = (egglib::Export *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Export, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Export_ms_auto_positions" "', argument " "1"" of type '" "egglib::Export *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Export * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Export_ms_auto_positions" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->ms_auto_positions(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Export_ms(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Export *arg1 = (egglib::Export *) 0 ;
  egglib::DataHolder *arg2 = 0 ;
  bool arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Export_ms", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Export, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Export_ms" "', argument " "1"" of type '" "egglib::Export *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Export * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__DataHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Export_ms" "', argument " "2"" of type '" "egglib::DataHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Export_ms" "', argument " "2"" of type '" "egglib::DataHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::DataHolder * >(argp2);
  ecode3 = SWIG_AsVal_bool(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Export_ms" "', argument " "3"" of type '" "bool""'");
  } 
  arg3 = static_cast< bool >(val3);
  {
    try {
      (arg1)->ms((egglib::DataHolder const &)*arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_Export) /* defines _wrap_delete_Export_destructor_closure */

SWIGINTERN int _wrap_new_FastaFormatter(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::FastaFormatter *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_FastaFormatter")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_FastaFormatter", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::FastaFormatter *)new egglib::FastaFormatter();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__FastaFormatter, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_FastaFormatter(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FastaFormatter *arg1 = (egglib::FastaFormatter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_FastaFormatter", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaFormatter, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_FastaFormatter" "', argument " "1"" of type '" "egglib::FastaFormatter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaFormatter * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaFormatter_defaults(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FastaFormatter *arg1 = (egglib::FastaFormatter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "FastaFormatter_defaults", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaFormatter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaFormatter_defaults" "', argument " "1"" of type '" "egglib::FastaFormatter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaFormatter * >(argp1);
  {
    try {
      (arg1)->defaults();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaFormatter_set_first__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::FastaFormatter *arg1 = (egglib::FastaFormatter *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaFormatter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaFormatter_set_first" "', argument " "1"" of type '" "egglib::FastaFormatter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaFormatter * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FastaFormatter_set_first" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_first(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaFormatter_set_first__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::FastaFormatter *arg1 = (egglib::FastaFormatter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaFormatter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaFormatter_set_first" "', argument " "1"" of type '" "egglib::FastaFormatter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaFormatter * >(argp1);
  {
    try {
      (arg1)->set_first();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaFormatter_set_first(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "FastaFormatter_set_first", 0, 2, argv+1))) SWIG_fail;
  argv[0] = self;
  if (argc == 1) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__FastaFormatter, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_FastaFormatter_set_first__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__FastaFormatter, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_FastaFormatter_set_first__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'FastaFormatter_set_first'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::FastaFormatter::set_first(unsigned int)\n"
    "    egglib::FastaFormatter::set_first()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_FastaFormatter_set_last__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::FastaFormatter *arg1 = (egglib::FastaFormatter *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaFormatter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaFormatter_set_last" "', argument " "1"" of type '" "egglib::FastaFormatter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaFormatter * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FastaFormatter_set_last" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_last(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaFormatter_set_last__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::FastaFormatter *arg1 = (egglib::FastaFormatter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaFormatter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaFormatter_set_last" "', argument " "1"" of type '" "egglib::FastaFormatter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaFormatter * >(argp1);
  {
    try {
      (arg1)->set_last();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaFormatter_set_last(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "FastaFormatter_set_last", 0, 2, argv+1))) SWIG_fail;
  argv[0] = self;
  if (argc == 1) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__FastaFormatter, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_FastaFormatter_set_last__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__FastaFormatter, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_FastaFormatter_set_last__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'FastaFormatter_set_last'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::FastaFormatter::set_last(unsigned int)\n"
    "    egglib::FastaFormatter::set_last()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_FastaFormatter_set_labels__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::FastaFormatter *arg1 = (egglib::FastaFormatter *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaFormatter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaFormatter_set_labels" "', argument " "1"" of type '" "egglib::FastaFormatter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaFormatter * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FastaFormatter_set_labels" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->set_labels(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaFormatter_set_labels__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::FastaFormatter *arg1 = (egglib::FastaFormatter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaFormatter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaFormatter_set_labels" "', argument " "1"" of type '" "egglib::FastaFormatter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaFormatter * >(argp1);
  {
    try {
      (arg1)->set_labels();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaFormatter_set_labels(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "FastaFormatter_set_labels", 0, 2, argv+1))) SWIG_fail;
  argv[0] = self;
  if (argc == 1) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__FastaFormatter, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_FastaFormatter_set_labels__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__FastaFormatter, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_FastaFormatter_set_labels__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'FastaFormatter_set_labels'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::FastaFormatter::set_labels(bool)\n"
    "    egglib::FastaFormatter::set_labels()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_FastaFormatter_set_linelength__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::FastaFormatter *arg1 = (egglib::FastaFormatter *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaFormatter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaFormatter_set_linelength" "', argument " "1"" of type '" "egglib::FastaFormatter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaFormatter * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FastaFormatter_set_linelength" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_linelength(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaFormatter_set_linelength__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::FastaFormatter *arg1 = (egglib::FastaFormatter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaFormatter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaFormatter_set_linelength" "', argument " "1"" of type '" "egglib::FastaFormatter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaFormatter * >(argp1);
  {
    try {
      (arg1)->set_linelength();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaFormatter_set_linelength(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "FastaFormatter_set_linelength", 0, 2, argv+1))) SWIG_fail;
  argv[0] = self;
  if (argc == 1) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__FastaFormatter, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_FastaFormatter_set_linelength__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__FastaFormatter, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_FastaFormatter_set_linelength__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'FastaFormatter_set_linelength'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::FastaFormatter::set_linelength(unsigned int)\n"
    "    egglib::FastaFormatter::set_linelength()\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_FastaFormatter_write(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FastaFormatter *arg1 = (egglib::FastaFormatter *) 0 ;
  egglib::DataHolder *arg2 = 0 ;
  egglib::AbstractBaseAlphabet *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "FastaFormatter_write", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaFormatter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaFormatter_write" "', argument " "1"" of type '" "egglib::FastaFormatter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaFormatter * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__DataHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FastaFormatter_write" "', argument " "2"" of type '" "egglib::DataHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "FastaFormatter_write" "', argument " "2"" of type '" "egglib::DataHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::DataHolder * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[1], &argp3, SWIGTYPE_p_egglib__AbstractBaseAlphabet,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FastaFormatter_write" "', argument " "3"" of type '" "egglib::AbstractBaseAlphabet &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "FastaFormatter_write" "', argument " "3"" of type '" "egglib::AbstractBaseAlphabet &""'"); 
  }
  arg3 = reinterpret_cast< egglib::AbstractBaseAlphabet * >(argp3);
  {
    try {
      (arg1)->write((egglib::DataHolder const &)*arg2,*arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaFormatter_write_string(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FastaFormatter *arg1 = (egglib::FastaFormatter *) 0 ;
  egglib::DataHolder *arg2 = 0 ;
  egglib::AbstractBaseAlphabet *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  std::string result;
  
  if (!SWIG_Python_UnpackTuple(args, "FastaFormatter_write_string", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaFormatter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaFormatter_write_string" "', argument " "1"" of type '" "egglib::FastaFormatter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaFormatter * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__DataHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FastaFormatter_write_string" "', argument " "2"" of type '" "egglib::DataHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "FastaFormatter_write_string" "', argument " "2"" of type '" "egglib::DataHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::DataHolder * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[1], &argp3, SWIGTYPE_p_egglib__AbstractBaseAlphabet,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FastaFormatter_write_string" "', argument " "3"" of type '" "egglib::AbstractBaseAlphabet &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "FastaFormatter_write_string" "', argument " "3"" of type '" "egglib::AbstractBaseAlphabet &""'"); 
  }
  arg3 = reinterpret_cast< egglib::AbstractBaseAlphabet * >(argp3);
  {
    try {
      result = (arg1)->write_string((egglib::DataHolder const &)*arg2,*arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_FastaFormatter) /* defines _wrap_delete_FastaFormatter_destructor_closure */

SWIGINTERN int _wrap_new_FastaParser(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::FastaParser *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_FastaParser")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_FastaParser", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::FastaParser *)new egglib::FastaParser();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__FastaParser, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_FastaParser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FastaParser *arg1 = (egglib::FastaParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_FastaParser", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaParser, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_FastaParser" "', argument " "1"" of type '" "egglib::FastaParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaParser * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaParser_reserve(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FastaParser *arg1 = (egglib::FastaParser *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  
  if (!SWIG_Python_UnpackTuple(args, "FastaParser_reserve", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaParser_reserve" "', argument " "1"" of type '" "egglib::FastaParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FastaParser_reserve" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "FastaParser_reserve" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "FastaParser_reserve" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "FastaParser_reserve" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = static_cast< unsigned int >(val5);
  {
    try {
      (arg1)->reserve(arg2,arg3,arg4,arg5);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaParser_open_file__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::FastaParser *arg1 = (egglib::FastaParser *) 0 ;
  char *arg2 = (char *) 0 ;
  egglib::FiniteAlphabet< char > *arg3 = 0 ;
  unsigned int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaParser_open_file" "', argument " "1"" of type '" "egglib::FastaParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaParser * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FastaParser_open_file" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_egglib__FiniteAlphabetT_char_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FastaParser_open_file" "', argument " "3"" of type '" "egglib::FiniteAlphabet< char > &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "FastaParser_open_file" "', argument " "3"" of type '" "egglib::FiniteAlphabet< char > &""'"); 
  }
  arg3 = reinterpret_cast< egglib::FiniteAlphabet< char > * >(argp3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "FastaParser_open_file" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  {
    try {
      (arg1)->open_file((char const *)arg2,*arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaParser_open_file__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::FastaParser *arg1 = (egglib::FastaParser *) 0 ;
  char *arg2 = (char *) 0 ;
  egglib::FiniteAlphabet< char > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaParser_open_file" "', argument " "1"" of type '" "egglib::FastaParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaParser * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FastaParser_open_file" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_egglib__FiniteAlphabetT_char_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FastaParser_open_file" "', argument " "3"" of type '" "egglib::FiniteAlphabet< char > &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "FastaParser_open_file" "', argument " "3"" of type '" "egglib::FiniteAlphabet< char > &""'"); 
  }
  arg3 = reinterpret_cast< egglib::FiniteAlphabet< char > * >(argp3);
  {
    try {
      (arg1)->open_file((char const *)arg2,*arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaParser_open_file(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "FastaParser_open_file", 0, 4, argv+1))) SWIG_fail;
  argv[0] = self;
  if (argc == 3) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__FastaParser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_egglib__FiniteAlphabetT_char_t, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_FastaParser_open_file__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__FastaParser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_egglib__FiniteAlphabetT_char_t, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_unsigned_SS_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_FastaParser_open_file__SWIG_0(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'FastaParser_open_file'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::FastaParser::open_file(char const *,egglib::FiniteAlphabet< char > &,unsigned int)\n"
    "    egglib::FastaParser::open_file(char const *,egglib::FiniteAlphabet< char > &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_FastaParser_set_string(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FastaParser *arg1 = (egglib::FastaParser *) 0 ;
  char *arg2 = (char *) 0 ;
  egglib::FiniteAlphabet< char > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "FastaParser_set_string", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaParser_set_string" "', argument " "1"" of type '" "egglib::FastaParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaParser * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FastaParser_set_string" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  res3 = SWIG_ConvertPtr(swig_obj[1], &argp3, SWIGTYPE_p_egglib__FiniteAlphabetT_char_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FastaParser_set_string" "', argument " "3"" of type '" "egglib::FiniteAlphabet< char > &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "FastaParser_set_string" "', argument " "3"" of type '" "egglib::FiniteAlphabet< char > &""'"); 
  }
  arg3 = reinterpret_cast< egglib::FiniteAlphabet< char > * >(argp3);
  {
    try {
      (arg1)->set_string((char const *)arg2,*arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaParser_read_sequence__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::FastaParser *arg1 = (egglib::FastaParser *) 0 ;
  bool arg2 ;
  egglib::DataHolder *arg3 = (egglib::DataHolder *) 0 ;
  char arg4 ;
  char arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  char val4 ;
  int ecode4 = 0 ;
  char val5 ;
  int ecode5 = 0 ;
  
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaParser_read_sequence" "', argument " "1"" of type '" "egglib::FastaParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaParser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FastaParser_read_sequence" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FastaParser_read_sequence" "', argument " "3"" of type '" "egglib::DataHolder *""'"); 
  }
  arg3 = reinterpret_cast< egglib::DataHolder * >(argp3);
  ecode4 = SWIG_AsVal_char(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "FastaParser_read_sequence" "', argument " "4"" of type '" "char""'");
  } 
  arg4 = static_cast< char >(val4);
  ecode5 = SWIG_AsVal_char(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "FastaParser_read_sequence" "', argument " "5"" of type '" "char""'");
  } 
  arg5 = static_cast< char >(val5);
  {
    try {
      (arg1)->read_sequence(arg2,arg3,arg4,arg5);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaParser_read_sequence__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::FastaParser *arg1 = (egglib::FastaParser *) 0 ;
  bool arg2 ;
  egglib::DataHolder *arg3 = (egglib::DataHolder *) 0 ;
  char arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  char val4 ;
  int ecode4 = 0 ;
  
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaParser_read_sequence" "', argument " "1"" of type '" "egglib::FastaParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaParser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FastaParser_read_sequence" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FastaParser_read_sequence" "', argument " "3"" of type '" "egglib::DataHolder *""'"); 
  }
  arg3 = reinterpret_cast< egglib::DataHolder * >(argp3);
  ecode4 = SWIG_AsVal_char(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "FastaParser_read_sequence" "', argument " "4"" of type '" "char""'");
  } 
  arg4 = static_cast< char >(val4);
  {
    try {
      (arg1)->read_sequence(arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaParser_read_sequence__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::FastaParser *arg1 = (egglib::FastaParser *) 0 ;
  bool arg2 ;
  egglib::DataHolder *arg3 = (egglib::DataHolder *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaParser_read_sequence" "', argument " "1"" of type '" "egglib::FastaParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaParser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FastaParser_read_sequence" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FastaParser_read_sequence" "', argument " "3"" of type '" "egglib::DataHolder *""'"); 
  }
  arg3 = reinterpret_cast< egglib::DataHolder * >(argp3);
  {
    try {
      (arg1)->read_sequence(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaParser_read_sequence__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::FastaParser *arg1 = (egglib::FastaParser *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaParser_read_sequence" "', argument " "1"" of type '" "egglib::FastaParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaParser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FastaParser_read_sequence" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->read_sequence(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaParser_read_sequence(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[6] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "FastaParser_read_sequence", 0, 5, argv+1))) SWIG_fail;
  argv[0] = self;
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__FastaParser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_FastaParser_read_sequence__SWIG_3(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__FastaParser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_egglib__DataHolder, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_FastaParser_read_sequence__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__FastaParser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_egglib__DataHolder, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_char(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_FastaParser_read_sequence__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__FastaParser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_egglib__DataHolder, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_char(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            {
              int res = SWIG_AsVal_char(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              return _wrap_FastaParser_read_sequence__SWIG_0(self, argc, argv);
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'FastaParser_read_sequence'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::FastaParser::read_sequence(bool,egglib::DataHolder *,char,char)\n"
    "    egglib::FastaParser::read_sequence(bool,egglib::DataHolder *,char)\n"
    "    egglib::FastaParser::read_sequence(bool,egglib::DataHolder *)\n"
    "    egglib::FastaParser::read_sequence(bool)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_FastaParser_read_all__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::FastaParser *arg1 = (egglib::FastaParser *) 0 ;
  bool arg2 ;
  egglib::DataHolder *arg3 = 0 ;
  char arg4 ;
  char arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  char val4 ;
  int ecode4 = 0 ;
  char val5 ;
  int ecode5 = 0 ;
  
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaParser_read_all" "', argument " "1"" of type '" "egglib::FastaParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaParser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FastaParser_read_all" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_egglib__DataHolder,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FastaParser_read_all" "', argument " "3"" of type '" "egglib::DataHolder &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "FastaParser_read_all" "', argument " "3"" of type '" "egglib::DataHolder &""'"); 
  }
  arg3 = reinterpret_cast< egglib::DataHolder * >(argp3);
  ecode4 = SWIG_AsVal_char(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "FastaParser_read_all" "', argument " "4"" of type '" "char""'");
  } 
  arg4 = static_cast< char >(val4);
  ecode5 = SWIG_AsVal_char(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "FastaParser_read_all" "', argument " "5"" of type '" "char""'");
  } 
  arg5 = static_cast< char >(val5);
  {
    try {
      (arg1)->read_all(arg2,*arg3,arg4,arg5);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaParser_read_all__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::FastaParser *arg1 = (egglib::FastaParser *) 0 ;
  bool arg2 ;
  egglib::DataHolder *arg3 = 0 ;
  char arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  char val4 ;
  int ecode4 = 0 ;
  
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaParser_read_all" "', argument " "1"" of type '" "egglib::FastaParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaParser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FastaParser_read_all" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_egglib__DataHolder,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FastaParser_read_all" "', argument " "3"" of type '" "egglib::DataHolder &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "FastaParser_read_all" "', argument " "3"" of type '" "egglib::DataHolder &""'"); 
  }
  arg3 = reinterpret_cast< egglib::DataHolder * >(argp3);
  ecode4 = SWIG_AsVal_char(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "FastaParser_read_all" "', argument " "4"" of type '" "char""'");
  } 
  arg4 = static_cast< char >(val4);
  {
    try {
      (arg1)->read_all(arg2,*arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaParser_read_all__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::FastaParser *arg1 = (egglib::FastaParser *) 0 ;
  bool arg2 ;
  egglib::DataHolder *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaParser_read_all" "', argument " "1"" of type '" "egglib::FastaParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaParser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FastaParser_read_all" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_egglib__DataHolder,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "FastaParser_read_all" "', argument " "3"" of type '" "egglib::DataHolder &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "FastaParser_read_all" "', argument " "3"" of type '" "egglib::DataHolder &""'"); 
  }
  arg3 = reinterpret_cast< egglib::DataHolder * >(argp3);
  {
    try {
      (arg1)->read_all(arg2,*arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaParser_read_all(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[6] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "FastaParser_read_all", 0, 5, argv+1))) SWIG_fail;
  argv[0] = self;
  if (argc == 3) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__FastaParser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_egglib__DataHolder, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_FastaParser_read_all__SWIG_2(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__FastaParser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_egglib__DataHolder, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_char(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_FastaParser_read_all__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__FastaParser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_p_egglib__DataHolder, SWIG_POINTER_NO_NULL);
        _v = SWIG_CheckState(res);
        if (_v) {
          {
            int res = SWIG_AsVal_char(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            {
              int res = SWIG_AsVal_char(argv[4], NULL);
              _v = SWIG_CheckState(res);
            }
            if (_v) {
              return _wrap_FastaParser_read_all__SWIG_0(self, argc, argv);
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'FastaParser_read_all'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::FastaParser::read_all(bool,egglib::DataHolder &,char,char)\n"
    "    egglib::FastaParser::read_all(bool,egglib::DataHolder &,char)\n"
    "    egglib::FastaParser::read_all(bool,egglib::DataHolder &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_FastaParser_close(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FastaParser *arg1 = (egglib::FastaParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "FastaParser_close", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaParser_close" "', argument " "1"" of type '" "egglib::FastaParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaParser * >(argp1);
  {
    try {
      (arg1)->close();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaParser_good(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FastaParser *arg1 = (egglib::FastaParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "FastaParser_good", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaParser_good" "', argument " "1"" of type '" "egglib::FastaParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaParser * >(argp1);
  {
    try {
      result = (bool)((egglib::FastaParser const *)arg1)->good();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaParser_name(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FastaParser *arg1 = (egglib::FastaParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "FastaParser_name", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaParser_name" "', argument " "1"" of type '" "egglib::FastaParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaParser * >(argp1);
  {
    try {
      result = (char *)((egglib::FastaParser const *)arg1)->name();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaParser_ls(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FastaParser *arg1 = (egglib::FastaParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "FastaParser_ls", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaParser_ls" "', argument " "1"" of type '" "egglib::FastaParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::FastaParser const *)arg1)->ls();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaParser_ch(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FastaParser *arg1 = (egglib::FastaParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaParser_ch" "', argument " "1"" of type '" "egglib::FastaParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FastaParser_ch" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (char)((egglib::FastaParser const *)arg1)->ch(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_char(static_cast< char >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaParser_nlabels(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FastaParser *arg1 = (egglib::FastaParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "FastaParser_nlabels", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaParser_nlabels" "', argument " "1"" of type '" "egglib::FastaParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::FastaParser const *)arg1)->nlabels();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaParser_label(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FastaParser *arg1 = (egglib::FastaParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaParser_label" "', argument " "1"" of type '" "egglib::FastaParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "FastaParser_label" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (char *)((egglib::FastaParser const *)arg1)->label(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FastaParser_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::FastaParser *arg1 = (egglib::FastaParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "FastaParser_clear", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__FastaParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FastaParser_clear" "', argument " "1"" of type '" "egglib::FastaParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::FastaParser * >(argp1);
  {
    try {
      (arg1)->clear();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_FastaParser) /* defines _wrap_delete_FastaParser_destructor_closure */

SWIGINTERN int _wrap_new_Feature(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::Feature *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_Feature")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_Feature", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::Feature *)new egglib::Feature();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Feature, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_Feature(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_Feature", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Feature" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_reset", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_reset" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      (arg1)->reset();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_clear", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_clear" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      (arg1)->clear();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_seqid(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_get_seqid", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_seqid" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      result = (char *)((egglib::Feature const *)arg1)->get_seqid();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_seqid(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_seqid" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Feature_set_seqid" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->set_seqid((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_source(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_get_source", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_source" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      result = (char *)((egglib::Feature const *)arg1)->get_source();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_source(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_source" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Feature_set_source" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->set_source((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_type(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_get_type", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_type" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      result = (char *)((egglib::Feature const *)arg1)->get_type();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_type(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_type" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Feature_set_type" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->set_type((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_num_fragments(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_get_num_fragments", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_num_fragments" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Feature const *)arg1)->get_num_fragments();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_num_fragments(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_num_fragments" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_num_fragments" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_num_fragments(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_start(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_start" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_get_start" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::Feature const *)arg1)->get_start(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_start(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_set_start", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_start" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_start" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Feature_set_start" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      (arg1)->set_start(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_end" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_get_end" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::Feature const *)arg1)->get_end(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_set_end", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_end" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_end" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Feature_set_end" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      (arg1)->set_end(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_score(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_get_score", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_score" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      result = (double)((egglib::Feature const *)arg1)->get_score();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_score(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_score" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_score" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->set_score(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_strand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::Feature::STRAND result;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_get_strand", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_strand" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      result = (egglib::Feature::STRAND)((egglib::Feature const *)arg1)->get_strand();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_strand(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  egglib::Feature::STRAND arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_strand" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_strand" "', argument " "2"" of type '" "egglib::Feature::STRAND""'");
  } 
  arg2 = static_cast< egglib::Feature::STRAND >(val2);
  {
    try {
      (arg1)->set_strand(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_phase(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::Feature::PHASE result;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_get_phase", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_phase" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      result = (egglib::Feature::PHASE)((egglib::Feature const *)arg1)->get_phase();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_phase(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  egglib::Feature::PHASE arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_phase" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_phase" "', argument " "2"" of type '" "egglib::Feature::PHASE""'");
  } 
  arg2 = static_cast< egglib::Feature::PHASE >(val2);
  {
    try {
      (arg1)->set_phase(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_ID(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_get_ID", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_ID" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      result = (char *)((egglib::Feature const *)arg1)->get_ID();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_ID(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_ID" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Feature_set_ID" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->set_ID((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_Name(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_get_Name", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_Name" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      result = (char *)((egglib::Feature const *)arg1)->get_Name();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_Name(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_Name" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Feature_set_Name" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->set_Name((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_num_Alias(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_get_num_Alias", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_num_Alias" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Feature const *)arg1)->get_num_Alias();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_num_Alias(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_num_Alias" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_num_Alias" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_num_Alias(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_Alias(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_Alias" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_get_Alias" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (char *)((egglib::Feature const *)arg1)->get_Alias(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_Alias(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  char *arg3 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_set_Alias", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_Alias" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_Alias" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Feature_set_Alias" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = reinterpret_cast< char * >(buf3);
  {
    try {
      (arg1)->set_Alias(arg2,(char const *)arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_num_Parent(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_get_num_Parent", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_num_Parent" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Feature const *)arg1)->get_num_Parent();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_num_Parent(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_num_Parent" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_num_Parent" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_num_Parent(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_Parent(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_Parent" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_get_Parent" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (char *)((egglib::Feature const *)arg1)->get_Parent(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_Parent(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  char *arg3 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_set_Parent", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_Parent" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_Parent" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Feature_set_Parent" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = reinterpret_cast< char * >(buf3);
  {
    try {
      (arg1)->set_Parent(arg2,(char const *)arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_Target(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_get_Target", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_Target" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      result = (char *)((egglib::Feature const *)arg1)->get_Target();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_Target(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_Target" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Feature_set_Target" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->set_Target((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_Gap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_get_Gap", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_Gap" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      result = (char *)((egglib::Feature const *)arg1)->get_Gap();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_Gap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_Gap" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Feature_set_Gap" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->set_Gap((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_Derives_from(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_get_Derives_from", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_Derives_from" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      result = (char *)((egglib::Feature const *)arg1)->get_Derives_from();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_Derives_from(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_Derives_from" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Feature_set_Derives_from" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->set_Derives_from((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_num_Note(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_get_num_Note", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_num_Note" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Feature const *)arg1)->get_num_Note();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_num_Note(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_num_Note" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_num_Note" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_num_Note(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_Note(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_Note" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_get_Note" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (char *)((egglib::Feature const *)arg1)->get_Note(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_Note(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  char *arg3 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_set_Note", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_Note" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_Note" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Feature_set_Note" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = reinterpret_cast< char * >(buf3);
  {
    try {
      (arg1)->set_Note(arg2,(char const *)arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_num_Dbxref(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_get_num_Dbxref", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_num_Dbxref" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Feature const *)arg1)->get_num_Dbxref();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_num_Dbxref(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_num_Dbxref" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_num_Dbxref" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_num_Dbxref(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_Dbxref(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_Dbxref" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_get_Dbxref" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (char *)((egglib::Feature const *)arg1)->get_Dbxref(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_Dbxref(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  char *arg3 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_set_Dbxref", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_Dbxref" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_Dbxref" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Feature_set_Dbxref" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = reinterpret_cast< char * >(buf3);
  {
    try {
      (arg1)->set_Dbxref(arg2,(char const *)arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_num_Ontology_term(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_get_num_Ontology_term", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_num_Ontology_term" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Feature const *)arg1)->get_num_Ontology_term();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_num_Ontology_term(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_num_Ontology_term" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_num_Ontology_term" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_num_Ontology_term(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_Ontology_term(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_Ontology_term" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_get_Ontology_term" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (char *)((egglib::Feature const *)arg1)->get_Ontology_term(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_Ontology_term(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  char *arg3 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_set_Ontology_term", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_Ontology_term" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_Ontology_term" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Feature_set_Ontology_term" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = reinterpret_cast< char * >(buf3);
  {
    try {
      (arg1)->set_Ontology_term(arg2,(char const *)arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_Is_circular(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_get_Is_circular", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_Is_circular" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      result = (bool)((egglib::Feature const *)arg1)->get_Is_circular();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_Is_circular(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_Is_circular" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_Is_circular" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->set_Is_circular(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_num_attributes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_get_num_attributes", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_num_attributes" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Feature const *)arg1)->get_num_attributes();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_num_attributes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_num_attributes" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_num_attributes" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_num_attributes(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_num_items_attribute(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_num_items_attribute" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_get_num_items_attribute" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::Feature const *)arg1)->get_num_items_attribute(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_num_items_attribute(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_set_num_items_attribute", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_num_items_attribute" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_num_items_attribute" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Feature_set_num_items_attribute" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      (arg1)->set_num_items_attribute(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_attribute_key(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_attribute_key" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_get_attribute_key" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (char *)((egglib::Feature const *)arg1)->get_attribute_key(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_attribute_key(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  char *arg3 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_set_attribute_key", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_attribute_key" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_attribute_key" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Feature_set_attribute_key" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = reinterpret_cast< char * >(buf3);
  {
    try {
      (arg1)->set_attribute_key(arg2,(char const *)arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_attribute_value(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_get_attribute_value", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_attribute_value" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_get_attribute_value" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Feature_get_attribute_value" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (char *)((egglib::Feature const *)arg1)->get_attribute_value(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_attribute_value(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  char *arg4 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_set_attribute_value", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_attribute_value" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_attribute_value" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Feature_set_attribute_value" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  res4 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Feature_set_attribute_value" "', argument " "4"" of type '" "char const *""'");
  }
  arg4 = reinterpret_cast< char * >(buf4);
  {
    try {
      (arg1)->set_attribute_value(arg2,arg3,(char const *)arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc4 == SWIG_NEWOBJ) delete[] buf4;
  return resultobj;
fail:
  if (alloc4 == SWIG_NEWOBJ) delete[] buf4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_num_parents(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_get_num_parents", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_num_parents" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Feature const *)arg1)->get_num_parents();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_num_parents(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_num_parents" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_num_parents" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_num_parents(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_parent(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::Feature *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_parent" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_get_parent" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::Feature *)((egglib::Feature const *)arg1)->get_parent(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Feature, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_parent(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  egglib::Feature *arg3 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_set_parent", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_parent" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_parent" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[1], &argp3,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Feature_set_parent" "', argument " "3"" of type '" "egglib::Feature *""'"); 
  }
  arg3 = reinterpret_cast< egglib::Feature * >(argp3);
  {
    try {
      (arg1)->set_parent(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_num_parts(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_get_num_parts", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_num_parts" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::Feature const *)arg1)->get_num_parts();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_num_parts(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_num_parts" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_num_parts" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_num_parts(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_get_part(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::Feature *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_get_part" "', argument " "1"" of type '" "egglib::Feature const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_get_part" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::Feature *)((egglib::Feature const *)arg1)->get_part(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Feature, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Feature_set_part(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::Feature *arg1 = (egglib::Feature *) 0 ;
  unsigned int arg2 ;
  egglib::Feature *arg3 = (egglib::Feature *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Feature_set_part", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Feature_set_part" "', argument " "1"" of type '" "egglib::Feature *""'"); 
  }
  arg1 = reinterpret_cast< egglib::Feature * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Feature_set_part" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[1], &argp3,SWIGTYPE_p_egglib__Feature, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Feature_set_part" "', argument " "3"" of type '" "egglib::Feature *""'"); 
  }
  arg3 = reinterpret_cast< egglib::Feature * >(argp3);
  {
    try {
      (arg1)->set_part(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_Feature) /* defines _wrap_delete_Feature_destructor_closure */

SWIGINTERN int _wrap_new_GFF3(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::GFF3 *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_GFF3")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_GFF3", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::GFF3 *)new egglib::GFF3();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__GFF3, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_GFF3(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::GFF3 *arg1 = (egglib::GFF3 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_GFF3", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GFF3, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_GFF3" "', argument " "1"" of type '" "egglib::GFF3 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GFF3 * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GFF3_liberal(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::GFF3 *arg1 = (egglib::GFF3 *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GFF3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GFF3_liberal" "', argument " "1"" of type '" "egglib::GFF3 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GFF3 * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "GFF3_liberal" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->liberal(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GFF3_parse__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::GFF3 *arg1 = (egglib::GFF3 *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GFF3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GFF3_parse" "', argument " "1"" of type '" "egglib::GFF3 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GFF3 * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "GFF3_parse" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->parse((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_GFF3_parse__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::GFF3 *arg1 = (egglib::GFF3 *) 0 ;
  std::istream *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GFF3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GFF3_parse" "', argument " "1"" of type '" "egglib::GFF3 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GFF3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_std__istream,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "GFF3_parse" "', argument " "2"" of type '" "std::istream &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "GFF3_parse" "', argument " "2"" of type '" "std::istream &""'"); 
  }
  arg2 = reinterpret_cast< std::istream * >(argp2);
  {
    try {
      (arg1)->parse(*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GFF3_parse(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "GFF3_parse", 0, 2, argv+1))) SWIG_fail;
  argv[0] = self;
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__GFF3, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_std__istream, SWIG_POINTER_NO_NULL);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_GFF3_parse__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__GFF3, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_GFF3_parse__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'GFF3_parse'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::GFF3::parse(char const *)\n"
    "    egglib::GFF3::parse(std::istream &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_GFF3_parse_string(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::GFF3 *arg1 = (egglib::GFF3 *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GFF3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GFF3_parse_string" "', argument " "1"" of type '" "egglib::GFF3 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GFF3 * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_std__string,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "GFF3_parse_string" "', argument " "2"" of type '" "std::string &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "GFF3_parse_string" "', argument " "2"" of type '" "std::string &""'"); 
  }
  arg2 = reinterpret_cast< std::string * >(argp2);
  {
    try {
      (arg1)->parse_string(*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GFF3_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::GFF3 *arg1 = (egglib::GFF3 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "GFF3_reset", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GFF3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GFF3_reset" "', argument " "1"" of type '" "egglib::GFF3 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GFF3 * >(argp1);
  {
    try {
      (arg1)->reset();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GFF3_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::GFF3 *arg1 = (egglib::GFF3 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "GFF3_clear", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GFF3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GFF3_clear" "', argument " "1"" of type '" "egglib::GFF3 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GFF3 * >(argp1);
  {
    try {
      (arg1)->clear();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GFF3_num_metadata(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::GFF3 *arg1 = (egglib::GFF3 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "GFF3_num_metadata", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GFF3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GFF3_num_metadata" "', argument " "1"" of type '" "egglib::GFF3 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GFF3 * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::GFF3 const *)arg1)->num_metadata();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GFF3_metadata_key(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::GFF3 *arg1 = (egglib::GFF3 *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GFF3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GFF3_metadata_key" "', argument " "1"" of type '" "egglib::GFF3 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GFF3 * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "GFF3_metadata_key" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (char *)((egglib::GFF3 const *)arg1)->metadata_key(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GFF3_metadata_value(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::GFF3 *arg1 = (egglib::GFF3 *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GFF3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GFF3_metadata_value" "', argument " "1"" of type '" "egglib::GFF3 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GFF3 * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "GFF3_metadata_value" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (char *)((egglib::GFF3 const *)arg1)->metadata_value(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GFF3_num_genes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::GFF3 *arg1 = (egglib::GFF3 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "GFF3_num_genes", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GFF3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GFF3_num_genes" "', argument " "1"" of type '" "egglib::GFF3 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GFF3 * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::GFF3 const *)arg1)->num_genes();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GFF3_gene(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::GFF3 *arg1 = (egglib::GFF3 *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::Feature *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GFF3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GFF3_gene" "', argument " "1"" of type '" "egglib::GFF3 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GFF3 * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "GFF3_gene" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::Feature *) &(arg1)->gene(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Feature, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GFF3_num_features(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::GFF3 *arg1 = (egglib::GFF3 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "GFF3_num_features", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GFF3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GFF3_num_features" "', argument " "1"" of type '" "egglib::GFF3 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GFF3 * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::GFF3 const *)arg1)->num_features();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GFF3_feature(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::GFF3 *arg1 = (egglib::GFF3 *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::Feature *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GFF3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GFF3_feature" "', argument " "1"" of type '" "egglib::GFF3 *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GFF3 * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "GFF3_feature" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::Feature *) &(arg1)->feature(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__Feature, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_GFF3_sequences(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::GFF3 *arg1 = (egglib::GFF3 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::DataHolder *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "GFF3_sequences", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__GFF3, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "GFF3_sequences" "', argument " "1"" of type '" "egglib::GFF3 const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::GFF3 * >(argp1);
  {
    try {
      result = (egglib::DataHolder *) &((egglib::GFF3 const *)arg1)->sequences();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__DataHolder, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_GFF3) /* defines _wrap_delete_GFF3_destructor_closure */

SWIGINTERN int Swig_var_NUM_ALTERNATE_set(PyObject *) {
  SWIG_Error(SWIG_AttributeError,"Variable NUM_ALTERNATE is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_NUM_ALTERNATE_get(void) {
  PyObject *pyobj = 0;
  PyObject *self = 0;
  
  (void)self;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(egglib::vcf::NUM_ALTERNATE));
  return pyobj;
}


SWIGINTERN int Swig_var_NUM_GENOTYPES_set(PyObject *) {
  SWIG_Error(SWIG_AttributeError,"Variable NUM_GENOTYPES is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_NUM_GENOTYPES_get(void) {
  PyObject *pyobj = 0;
  PyObject *self = 0;
  
  (void)self;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(egglib::vcf::NUM_GENOTYPES));
  return pyobj;
}


SWIGINTERN int Swig_var_NUM_POSSIBLE_ALLELES_set(PyObject *) {
  SWIG_Error(SWIG_AttributeError,"Variable NUM_POSSIBLE_ALLELES is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_NUM_POSSIBLE_ALLELES_get(void) {
  PyObject *pyobj = 0;
  PyObject *self = 0;
  
  (void)self;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(egglib::vcf::NUM_POSSIBLE_ALLELES));
  return pyobj;
}


SWIGINTERN int _wrap_new_VcfFilter__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::vcf::Filter *result = 0 ;
  
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  {
    try {
      result = (egglib::vcf::Filter *)new egglib::vcf::Filter();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Filter, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_VcfFilter__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  egglib::vcf::Filter *result = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_VcfFilter" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = reinterpret_cast< char * >(buf1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_VcfFilter" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      result = (egglib::vcf::Filter *)new egglib::vcf::Filter((char const *)arg1,(char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Filter, SWIG_BUILTIN_INIT |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj == Py_None ? -1 : 0;
fail:
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return -1;
}


SWIGINTERN int _wrap_new_VcfFilter__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::vcf::Filter *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  egglib::vcf::Filter *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_egglib__vcf__Filter,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_VcfFilter" "', argument " "1"" of type '" "egglib::vcf::Filter const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_VcfFilter" "', argument " "1"" of type '" "egglib::vcf::Filter const &""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Filter * >(argp1);
  {
    try {
      result = (egglib::vcf::Filter *)new egglib::vcf::Filter((egglib::vcf::Filter const &)*arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Filter, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_VcfFilter(PyObject *self, PyObject *args, PyObject *kwargs) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_VcfFilter")) SWIG_fail;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_VcfFilter", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_VcfFilter__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_egglib__vcf__Filter, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_VcfFilter__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v;
    int res = SWIG_AsCharPtrAndSize(argv[0], 0, NULL, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_VcfFilter__SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_VcfFilter'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::vcf::Filter::Filter()\n"
    "    egglib::vcf::Filter::Filter(char const *,char const *)\n"
    "    egglib::vcf::Filter::Filter(egglib::vcf::Filter const &)\n");
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_VcfFilter(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Filter *arg1 = (egglib::vcf::Filter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_VcfFilter", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Filter, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_VcfFilter" "', argument " "1"" of type '" "egglib::vcf::Filter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Filter * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfFilter_update(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Filter *arg1 = (egglib::vcf::Filter *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfFilter_update", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Filter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfFilter_update" "', argument " "1"" of type '" "egglib::vcf::Filter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Filter * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfFilter_update" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VcfFilter_update" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = reinterpret_cast< char * >(buf3);
  {
    try {
      (arg1)->update((char const *)arg2,(char const *)arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfFilter_get_ID(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Filter *arg1 = (egglib::vcf::Filter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfFilter_get_ID", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Filter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfFilter_get_ID" "', argument " "1"" of type '" "egglib::vcf::Filter const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Filter * >(argp1);
  {
    try {
      result = (char *)((egglib::vcf::Filter const *)arg1)->get_ID();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfFilter_set_ID(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Filter *arg1 = (egglib::vcf::Filter *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Filter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfFilter_set_ID" "', argument " "1"" of type '" "egglib::vcf::Filter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Filter * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfFilter_set_ID" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->set_ID((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfFilter_get_description(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Filter *arg1 = (egglib::vcf::Filter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfFilter_get_description", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Filter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfFilter_get_description" "', argument " "1"" of type '" "egglib::vcf::Filter const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Filter * >(argp1);
  {
    try {
      result = (char *)((egglib::vcf::Filter const *)arg1)->get_description();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfFilter_set_description(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Filter *arg1 = (egglib::vcf::Filter *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Filter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfFilter_set_description" "', argument " "1"" of type '" "egglib::vcf::Filter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Filter * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfFilter_set_description" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->set_description((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfFilter_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Filter *arg1 = (egglib::vcf::Filter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfFilter_clear", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Filter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfFilter_clear" "', argument " "1"" of type '" "egglib::vcf::Filter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Filter * >(argp1);
  {
    try {
      (arg1)->clear();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfFilter_set_extra(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Filter *arg1 = (egglib::vcf::Filter *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfFilter_set_extra", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Filter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfFilter_set_extra" "', argument " "1"" of type '" "egglib::vcf::Filter *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Filter * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfFilter_set_extra" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VcfFilter_set_extra" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = reinterpret_cast< char * >(buf3);
  {
    try {
      (arg1)->set_extra((char const *)arg2,(char const *)arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfFilter_get_extra_key(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Filter *arg1 = (egglib::vcf::Filter *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Filter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfFilter_get_extra_key" "', argument " "1"" of type '" "egglib::vcf::Filter const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Filter * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfFilter_get_extra_key" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (char *)((egglib::vcf::Filter const *)arg1)->get_extra_key(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfFilter_get_extra_value(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Filter *arg1 = (egglib::vcf::Filter *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Filter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfFilter_get_extra_value" "', argument " "1"" of type '" "egglib::vcf::Filter const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Filter * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfFilter_get_extra_value" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (char *)((egglib::vcf::Filter const *)arg1)->get_extra_value(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfFilter_get_num_extra(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Filter *arg1 = (egglib::vcf::Filter *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfFilter_get_num_extra", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Filter, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfFilter_get_num_extra" "', argument " "1"" of type '" "egglib::vcf::Filter const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Filter * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::vcf::Filter const *)arg1)->get_num_extra();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_VcfFilter) /* defines _wrap_delete_VcfFilter_destructor_closure */

SWIGINTERN int _wrap_new_Info__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::vcf::Info *result = 0 ;
  
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  {
    try {
      result = (egglib::vcf::Info *)new egglib::vcf::Info();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Info, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_Info__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  unsigned int arg2 ;
  egglib::vcf::Info::Type arg3 ;
  char *arg4 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  egglib::vcf::Info *result = 0 ;
  
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Info" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = reinterpret_cast< char * >(buf1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Info" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_Info" "', argument " "3"" of type '" "egglib::vcf::Info::Type""'");
  } 
  arg3 = static_cast< egglib::vcf::Info::Type >(val3);
  res4 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_Info" "', argument " "4"" of type '" "char const *""'");
  }
  arg4 = reinterpret_cast< char * >(buf4);
  {
    try {
      result = (egglib::vcf::Info *)new egglib::vcf::Info((char const *)arg1,arg2,arg3,(char const *)arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Info, SWIG_BUILTIN_INIT |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  if (alloc4 == SWIG_NEWOBJ) delete[] buf4;
  return resultobj == Py_None ? -1 : 0;
fail:
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  if (alloc4 == SWIG_NEWOBJ) delete[] buf4;
  return -1;
}


SWIGINTERN int _wrap_new_Info__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::vcf::Info *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  egglib::vcf::Info *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_egglib__vcf__Info,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Info" "', argument " "1"" of type '" "egglib::vcf::Info const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Info" "', argument " "1"" of type '" "egglib::vcf::Info const &""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Info * >(argp1);
  {
    try {
      result = (egglib::vcf::Info *)new egglib::vcf::Info((egglib::vcf::Info const &)*arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Info, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_Info(PyObject *self, PyObject *args, PyObject *kwargs) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_Info")) SWIG_fail;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_Info", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_Info__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_egglib__vcf__Info, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_Info__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 4) {
    int _v;
    int res = SWIG_AsCharPtrAndSize(argv[0], 0, NULL, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsCharPtrAndSize(argv[3], 0, NULL, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_new_Info__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_Info'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::vcf::Info::Info()\n"
    "    egglib::vcf::Info::Info(char const *,unsigned int,egglib::vcf::Info::Type,char const *)\n"
    "    egglib::vcf::Info::Info(egglib::vcf::Info const &)\n");
  return -1;
}


SWIGINTERN PyObject *_wrap_Info_update(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Info *arg1 = (egglib::vcf::Info *) 0 ;
  char *arg2 = (char *) 0 ;
  unsigned int arg3 ;
  egglib::vcf::Info::Type arg4 ;
  char *arg5 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  PyObject *swig_obj[5] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Info_update", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Info, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Info_update" "', argument " "1"" of type '" "egglib::vcf::Info *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Info * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Info_update" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Info_update" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Info_update" "', argument " "4"" of type '" "egglib::vcf::Info::Type""'");
  } 
  arg4 = static_cast< egglib::vcf::Info::Type >(val4);
  res5 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Info_update" "', argument " "5"" of type '" "char const *""'");
  }
  arg5 = reinterpret_cast< char * >(buf5);
  {
    try {
      (arg1)->update((char const *)arg2,arg3,arg4,(char const *)arg5);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc5 == SWIG_NEWOBJ) delete[] buf5;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc5 == SWIG_NEWOBJ) delete[] buf5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Info_get_number(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Info *arg1 = (egglib::vcf::Info *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "Info_get_number", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Info, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Info_get_number" "', argument " "1"" of type '" "egglib::vcf::Info const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Info * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::vcf::Info const *)arg1)->get_number();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Info_set_number(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Info *arg1 = (egglib::vcf::Info *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Info, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Info_set_number" "', argument " "1"" of type '" "egglib::vcf::Info *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Info * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Info_set_number" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_number(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Info_get_type(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Info *arg1 = (egglib::vcf::Info *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::vcf::Info::Type result;
  
  if (!SWIG_Python_UnpackTuple(args, "Info_get_type", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Info, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Info_get_type" "', argument " "1"" of type '" "egglib::vcf::Info const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Info * >(argp1);
  {
    try {
      result = (egglib::vcf::Info::Type)((egglib::vcf::Info const *)arg1)->get_type();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Info_set_type(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Info *arg1 = (egglib::vcf::Info *) 0 ;
  egglib::vcf::Info::Type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Info, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Info_set_type" "', argument " "1"" of type '" "egglib::vcf::Info *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Info * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Info_set_type" "', argument " "2"" of type '" "egglib::vcf::Info::Type""'");
  } 
  arg2 = static_cast< egglib::vcf::Info::Type >(val2);
  {
    try {
      (arg1)->set_type(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Info(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Info *arg1 = (egglib::vcf::Info *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_Info", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Info, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Info" "', argument " "1"" of type '" "egglib::vcf::Info *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Info * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_Info) /* defines _wrap_delete_Info_destructor_closure */

SWIGINTERN int _wrap_new_Format__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::vcf::Format *result = 0 ;
  
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  {
    try {
      result = (egglib::vcf::Format *)new egglib::vcf::Format();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Format, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_Format__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  unsigned int arg2 ;
  egglib::vcf::Info::Type arg3 ;
  char *arg4 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  egglib::vcf::Format *result = 0 ;
  
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Format" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = reinterpret_cast< char * >(buf1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Format" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_Format" "', argument " "3"" of type '" "egglib::vcf::Info::Type""'");
  } 
  arg3 = static_cast< egglib::vcf::Info::Type >(val3);
  res4 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_Format" "', argument " "4"" of type '" "char const *""'");
  }
  arg4 = reinterpret_cast< char * >(buf4);
  {
    try {
      result = (egglib::vcf::Format *)new egglib::vcf::Format((char const *)arg1,arg2,arg3,(char const *)arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Format, SWIG_BUILTIN_INIT |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  if (alloc4 == SWIG_NEWOBJ) delete[] buf4;
  return resultobj == Py_None ? -1 : 0;
fail:
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  if (alloc4 == SWIG_NEWOBJ) delete[] buf4;
  return -1;
}


SWIGINTERN int _wrap_new_Format__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::vcf::Format *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  egglib::vcf::Format *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_egglib__vcf__Format,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Format" "', argument " "1"" of type '" "egglib::vcf::Format const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Format" "', argument " "1"" of type '" "egglib::vcf::Format const &""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Format * >(argp1);
  {
    try {
      result = (egglib::vcf::Format *)new egglib::vcf::Format((egglib::vcf::Format const &)*arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Format, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_Format__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::vcf::Info *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  egglib::vcf::Format *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_egglib__vcf__Info,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Format" "', argument " "1"" of type '" "egglib::vcf::Info const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Format" "', argument " "1"" of type '" "egglib::vcf::Info const &""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Info * >(argp1);
  {
    try {
      result = (egglib::vcf::Format *)new egglib::vcf::Format((egglib::vcf::Info const &)*arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Format, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_Format(PyObject *self, PyObject *args, PyObject *kwargs) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_Format")) SWIG_fail;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_Format", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_Format__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_egglib__vcf__Format, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_Format__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_egglib__vcf__Info, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_Format__SWIG_3(self, argc, argv);
    }
  }
  if (argc == 4) {
    int _v;
    int res = SWIG_AsCharPtrAndSize(argv[0], 0, NULL, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsCharPtrAndSize(argv[3], 0, NULL, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_new_Format__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_Format'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::vcf::Format::Format()\n"
    "    egglib::vcf::Format::Format(char const *,unsigned int,egglib::vcf::Info::Type,char const *)\n"
    "    egglib::vcf::Format::Format(egglib::vcf::Format const &)\n"
    "    egglib::vcf::Format::Format(egglib::vcf::Info const &)\n");
  return -1;
}


SWIGINTERN PyObject *_wrap_Format_update(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Format *arg1 = (egglib::vcf::Format *) 0 ;
  char *arg2 = (char *) 0 ;
  unsigned int arg3 ;
  egglib::vcf::Info::Type arg4 ;
  char *arg5 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  PyObject *swig_obj[5] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Format_update", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Format, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Format_update" "', argument " "1"" of type '" "egglib::vcf::Format *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Format * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Format_update" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Format_update" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Format_update" "', argument " "4"" of type '" "egglib::vcf::Info::Type""'");
  } 
  arg4 = static_cast< egglib::vcf::Info::Type >(val4);
  res5 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Format_update" "', argument " "5"" of type '" "char const *""'");
  }
  arg5 = reinterpret_cast< char * >(buf5);
  {
    try {
      (arg1)->update((char const *)arg2,arg3,arg4,(char const *)arg5);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc5 == SWIG_NEWOBJ) delete[] buf5;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc5 == SWIG_NEWOBJ) delete[] buf5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Format_get_type(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Format *arg1 = (egglib::vcf::Format *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::vcf::Info::Type result;
  
  if (!SWIG_Python_UnpackTuple(args, "Format_get_type", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Format, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Format_get_type" "', argument " "1"" of type '" "egglib::vcf::Format const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Format * >(argp1);
  {
    try {
      result = (egglib::vcf::Info::Type)((egglib::vcf::Format const *)arg1)->get_type();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Format_set_type(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Format *arg1 = (egglib::vcf::Format *) 0 ;
  egglib::vcf::Info::Type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Format, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Format_set_type" "', argument " "1"" of type '" "egglib::vcf::Format *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Format * >(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Format_set_type" "', argument " "2"" of type '" "egglib::vcf::Info::Type""'");
  } 
  arg2 = static_cast< egglib::vcf::Info::Type >(val2);
  {
    try {
      (arg1)->set_type(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Format(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Format *arg1 = (egglib::vcf::Format *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_Format", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Format, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Format" "', argument " "1"" of type '" "egglib::vcf::Format *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Format * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_Format) /* defines _wrap_delete_Format_destructor_closure */

SWIGINTERN int _wrap_new_Meta__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::vcf::Meta *result = 0 ;
  
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  {
    try {
      result = (egglib::vcf::Meta *)new egglib::vcf::Meta();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Meta, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_Meta__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  egglib::vcf::Meta *result = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Meta" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = reinterpret_cast< char * >(buf1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_Meta" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      result = (egglib::vcf::Meta *)new egglib::vcf::Meta((char const *)arg1,(char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Meta, SWIG_BUILTIN_INIT |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj == Py_None ? -1 : 0;
fail:
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return -1;
}


SWIGINTERN int _wrap_new_Meta__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::vcf::Meta *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  egglib::vcf::Meta *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_egglib__vcf__Meta,  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Meta" "', argument " "1"" of type '" "egglib::vcf::Meta &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Meta" "', argument " "1"" of type '" "egglib::vcf::Meta &""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Meta * >(argp1);
  {
    try {
      result = (egglib::vcf::Meta *)new egglib::vcf::Meta(*arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Meta, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_Meta(PyObject *self, PyObject *args, PyObject *kwargs) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_Meta")) SWIG_fail;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_Meta", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_Meta__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__vcf__Meta, SWIG_POINTER_NO_NULL);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_Meta__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v;
    int res = SWIG_AsCharPtrAndSize(argv[0], 0, NULL, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_Meta__SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_Meta'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::vcf::Meta::Meta()\n"
    "    egglib::vcf::Meta::Meta(char const *,char const *)\n"
    "    egglib::vcf::Meta::Meta(egglib::vcf::Meta &)\n");
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_Meta(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Meta *arg1 = (egglib::vcf::Meta *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_Meta", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Meta, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Meta" "', argument " "1"" of type '" "egglib::vcf::Meta *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Meta * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Meta_update(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Meta *arg1 = (egglib::vcf::Meta *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Meta_update", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Meta, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Meta_update" "', argument " "1"" of type '" "egglib::vcf::Meta *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Meta * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Meta_update" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Meta_update" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = reinterpret_cast< char * >(buf3);
  {
    try {
      (arg1)->update((char const *)arg2,(char const *)arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Meta_set_key(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Meta *arg1 = (egglib::vcf::Meta *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Meta, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Meta_set_key" "', argument " "1"" of type '" "egglib::vcf::Meta *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Meta * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Meta_set_key" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->set_key((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Meta_set_value(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Meta *arg1 = (egglib::vcf::Meta *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Meta, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Meta_set_value" "', argument " "1"" of type '" "egglib::vcf::Meta *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Meta * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Meta_set_value" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->set_value((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Meta_get_key(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Meta *arg1 = (egglib::vcf::Meta *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "Meta_get_key", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Meta, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Meta_get_key" "', argument " "1"" of type '" "egglib::vcf::Meta const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Meta * >(argp1);
  {
    try {
      result = (char *)((egglib::vcf::Meta const *)arg1)->get_key();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Meta_get_value(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Meta *arg1 = (egglib::vcf::Meta *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "Meta_get_value", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Meta, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Meta_get_value" "', argument " "1"" of type '" "egglib::vcf::Meta const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Meta * >(argp1);
  {
    try {
      result = (char *)((egglib::vcf::Meta const *)arg1)->get_value();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Meta_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Meta *arg1 = (egglib::vcf::Meta *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "Meta_clear", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Meta, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Meta_clear" "', argument " "1"" of type '" "egglib::vcf::Meta *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Meta * >(argp1);
  {
    try {
      (arg1)->clear();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_Meta) /* defines _wrap_delete_Meta_destructor_closure */

SWIGINTERN int _wrap_new_Alt__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::vcf::Alt *result = 0 ;
  
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  {
    try {
      result = (egglib::vcf::Alt *)new egglib::vcf::Alt();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Alt, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_Alt__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  char *arg2 = (char *) 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  egglib::vcf::Alt *result = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Alt" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = reinterpret_cast< char * >(buf1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_Alt" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      result = (egglib::vcf::Alt *)new egglib::vcf::Alt((char const *)arg1,(char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Alt, SWIG_BUILTIN_INIT |  0 );
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj == Py_None ? -1 : 0;
fail:
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return -1;
}


SWIGINTERN int _wrap_new_Alt__SWIG_2(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::vcf::Alt *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  egglib::vcf::Alt *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_egglib__vcf__Alt,  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Alt" "', argument " "1"" of type '" "egglib::vcf::Alt &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Alt" "', argument " "1"" of type '" "egglib::vcf::Alt &""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Alt * >(argp1);
  {
    try {
      result = (egglib::vcf::Alt *)new egglib::vcf::Alt(*arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Alt, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_Alt__SWIG_3(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::vcf::Filter *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  egglib::vcf::Alt *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_egglib__vcf__Filter,  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Alt" "', argument " "1"" of type '" "egglib::vcf::Filter &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Alt" "', argument " "1"" of type '" "egglib::vcf::Filter &""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Filter * >(argp1);
  {
    try {
      result = (egglib::vcf::Alt *)new egglib::vcf::Alt(*arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Alt, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_Alt(PyObject *self, PyObject *args, PyObject *kwargs) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_Alt")) SWIG_fail;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_Alt", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_Alt__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__vcf__Alt, SWIG_POINTER_NO_NULL);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_Alt__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__vcf__Filter, SWIG_POINTER_NO_NULL);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_Alt__SWIG_3(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v;
    int res = SWIG_AsCharPtrAndSize(argv[0], 0, NULL, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_Alt__SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_Alt'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::vcf::Alt::Alt()\n"
    "    egglib::vcf::Alt::Alt(char const *,char const *)\n"
    "    egglib::vcf::Alt::Alt(egglib::vcf::Alt &)\n"
    "    egglib::vcf::Alt::Alt(egglib::vcf::Filter &)\n");
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_Alt(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::Alt *arg1 = (egglib::vcf::Alt *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_Alt", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__Alt, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Alt" "', argument " "1"" of type '" "egglib::vcf::Alt *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::Alt * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_Alt) /* defines _wrap_delete_Alt_destructor_closure */

SWIGINTERN int _wrap_new_FlagInfo__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::vcf::FlagInfo *result = 0 ;
  
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  {
    try {
      result = (egglib::vcf::FlagInfo *)new egglib::vcf::FlagInfo();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__FlagInfo, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_FlagInfo__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::vcf::FlagInfo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  egglib::vcf::FlagInfo *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_egglib__vcf__FlagInfo,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_FlagInfo" "', argument " "1"" of type '" "egglib::vcf::FlagInfo const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_FlagInfo" "', argument " "1"" of type '" "egglib::vcf::FlagInfo const &""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::FlagInfo * >(argp1);
  {
    try {
      result = (egglib::vcf::FlagInfo *)new egglib::vcf::FlagInfo((egglib::vcf::FlagInfo const &)*arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__FlagInfo, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_FlagInfo(PyObject *self, PyObject *args, PyObject *kwargs) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_FlagInfo")) SWIG_fail;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_FlagInfo", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_FlagInfo__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_egglib__vcf__FlagInfo, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_FlagInfo__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_FlagInfo'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::vcf::FlagInfo::FlagInfo()\n"
    "    egglib::vcf::FlagInfo::FlagInfo(egglib::vcf::FlagInfo const &)\n");
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_FlagInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::FlagInfo *arg1 = (egglib::vcf::FlagInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_FlagInfo", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__FlagInfo, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_FlagInfo" "', argument " "1"" of type '" "egglib::vcf::FlagInfo *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::FlagInfo * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_FlagInfo_set_ID(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::FlagInfo *arg1 = (egglib::vcf::FlagInfo *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__FlagInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FlagInfo_set_ID" "', argument " "1"" of type '" "egglib::vcf::FlagInfo *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::FlagInfo * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "FlagInfo_set_ID" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->set_ID((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_FlagInfo_get_ID(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::FlagInfo *arg1 = (egglib::vcf::FlagInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "FlagInfo_get_ID", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__FlagInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "FlagInfo_get_ID" "', argument " "1"" of type '" "egglib::vcf::FlagInfo const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::FlagInfo * >(argp1);
  {
    try {
      result = (char *)((egglib::vcf::FlagInfo const *)arg1)->get_ID();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_FlagInfo) /* defines _wrap_delete_FlagInfo_destructor_closure */

SWIGINTERN int _wrap_new_TypeInfoString__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< char * > *result = 0 ;
  
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  {
    try {
      result = (egglib::vcf::TypeInfo< char * > *)new egglib::vcf::TypeInfo< char * >();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__TypeInfoT_char_p_t, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_TypeInfoString__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< char * > *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  egglib::vcf::TypeInfo< char * > *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_egglib__vcf__TypeInfoT_char_p_t,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_TypeInfoString" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< char * > const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_TypeInfoString" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< char * > const &""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< char * > * >(argp1);
  {
    try {
      result = (egglib::vcf::TypeInfo< char * > *)new egglib::vcf::TypeInfo< char * >((egglib::vcf::TypeInfo< char * > const &)*arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__TypeInfoT_char_p_t, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_TypeInfoString(PyObject *self, PyObject *args, PyObject *kwargs) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_TypeInfoString")) SWIG_fail;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_TypeInfoString", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_TypeInfoString__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_egglib__vcf__TypeInfoT_char_p_t, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_TypeInfoString__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_TypeInfoString'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::vcf::TypeInfo< char * >::TypeInfo()\n"
    "    egglib::vcf::TypeInfo< char * >::TypeInfo(egglib::vcf::TypeInfo< char * > const &)\n");
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_TypeInfoString(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< char * > *arg1 = (egglib::vcf::TypeInfo< char * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_TypeInfoString", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_char_p_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TypeInfoString" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< char * > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< char * > * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TypeInfoString_num_items(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< char * > *arg1 = (egglib::vcf::TypeInfo< char * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "TypeInfoString_num_items", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_char_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TypeInfoString_num_items" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< char * > const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< char * > * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::vcf::TypeInfo< char * > const *)arg1)->num_items();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TypeInfoString_item(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< char * > *arg1 = (egglib::vcf::TypeInfo< char * > *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char **result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_char_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TypeInfoString_item" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< char * > const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< char * > * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TypeInfoString_item" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (char **) &((egglib::vcf::TypeInfo< char * > const *)arg1)->item(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(*result), SWIGTYPE_p_char, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TypeInfoString_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< char * > *arg1 = (egglib::vcf::TypeInfo< char * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "TypeInfoString_reset", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_char_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TypeInfoString_reset" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< char * > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< char * > * >(argp1);
  {
    try {
      (arg1)->reset();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TypeInfoString_get_expected_number(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< char * > *arg1 = (egglib::vcf::TypeInfo< char * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "TypeInfoString_get_expected_number", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_char_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TypeInfoString_get_expected_number" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< char * > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< char * > * >(argp1);
  {
    try {
      result = (unsigned int)(arg1)->get_expected_number();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TypeInfoString_set_expected_number(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< char * > *arg1 = (egglib::vcf::TypeInfo< char * > *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_char_p_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TypeInfoString_set_expected_number" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< char * > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< char * > * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TypeInfoString_set_expected_number" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_expected_number(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_TypeInfoString) /* defines _wrap_delete_TypeInfoString_destructor_closure */

SWIGINTERN int _wrap_new_TypeInfoInt__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< int > *result = 0 ;
  
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  {
    try {
      result = (egglib::vcf::TypeInfo< int > *)new egglib::vcf::TypeInfo< int >();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__TypeInfoT_int_t, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_TypeInfoInt__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< int > *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  egglib::vcf::TypeInfo< int > *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_egglib__vcf__TypeInfoT_int_t,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_TypeInfoInt" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< int > const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_TypeInfoInt" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< int > const &""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< int > * >(argp1);
  {
    try {
      result = (egglib::vcf::TypeInfo< int > *)new egglib::vcf::TypeInfo< int >((egglib::vcf::TypeInfo< int > const &)*arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__TypeInfoT_int_t, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_TypeInfoInt(PyObject *self, PyObject *args, PyObject *kwargs) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_TypeInfoInt")) SWIG_fail;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_TypeInfoInt", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_TypeInfoInt__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_egglib__vcf__TypeInfoT_int_t, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_TypeInfoInt__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_TypeInfoInt'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::vcf::TypeInfo< int >::TypeInfo()\n"
    "    egglib::vcf::TypeInfo< int >::TypeInfo(egglib::vcf::TypeInfo< int > const &)\n");
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_TypeInfoInt(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< int > *arg1 = (egglib::vcf::TypeInfo< int > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_TypeInfoInt", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_int_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TypeInfoInt" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< int > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< int > * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TypeInfoInt_num_items(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< int > *arg1 = (egglib::vcf::TypeInfo< int > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "TypeInfoInt_num_items", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TypeInfoInt_num_items" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< int > const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< int > * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::vcf::TypeInfo< int > const *)arg1)->num_items();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TypeInfoInt_item(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< int > *arg1 = (egglib::vcf::TypeInfo< int > *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TypeInfoInt_item" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< int > const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< int > * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TypeInfoInt_item" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (int *) &((egglib::vcf::TypeInfo< int > const *)arg1)->item(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TypeInfoInt_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< int > *arg1 = (egglib::vcf::TypeInfo< int > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "TypeInfoInt_reset", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TypeInfoInt_reset" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< int > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< int > * >(argp1);
  {
    try {
      (arg1)->reset();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TypeInfoInt_get_expected_number(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< int > *arg1 = (egglib::vcf::TypeInfo< int > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "TypeInfoInt_get_expected_number", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TypeInfoInt_get_expected_number" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< int > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< int > * >(argp1);
  {
    try {
      result = (unsigned int)(arg1)->get_expected_number();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TypeInfoInt_set_expected_number(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< int > *arg1 = (egglib::vcf::TypeInfo< int > *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_int_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TypeInfoInt_set_expected_number" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< int > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< int > * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TypeInfoInt_set_expected_number" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_expected_number(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_TypeInfoInt) /* defines _wrap_delete_TypeInfoInt_destructor_closure */

SWIGINTERN int _wrap_new_TypeInfoChar__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< char > *result = 0 ;
  
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  {
    try {
      result = (egglib::vcf::TypeInfo< char > *)new egglib::vcf::TypeInfo< char >();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__TypeInfoT_char_t, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_TypeInfoChar__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< char > *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  egglib::vcf::TypeInfo< char > *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_egglib__vcf__TypeInfoT_char_t,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_TypeInfoChar" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< char > const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_TypeInfoChar" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< char > const &""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< char > * >(argp1);
  {
    try {
      result = (egglib::vcf::TypeInfo< char > *)new egglib::vcf::TypeInfo< char >((egglib::vcf::TypeInfo< char > const &)*arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__TypeInfoT_char_t, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_TypeInfoChar(PyObject *self, PyObject *args, PyObject *kwargs) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_TypeInfoChar")) SWIG_fail;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_TypeInfoChar", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_TypeInfoChar__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_egglib__vcf__TypeInfoT_char_t, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_TypeInfoChar__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_TypeInfoChar'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::vcf::TypeInfo< char >::TypeInfo()\n"
    "    egglib::vcf::TypeInfo< char >::TypeInfo(egglib::vcf::TypeInfo< char > const &)\n");
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_TypeInfoChar(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< char > *arg1 = (egglib::vcf::TypeInfo< char > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_TypeInfoChar", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_char_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TypeInfoChar" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< char > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< char > * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TypeInfoChar_num_items(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< char > *arg1 = (egglib::vcf::TypeInfo< char > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "TypeInfoChar_num_items", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_char_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TypeInfoChar_num_items" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< char > const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< char > * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::vcf::TypeInfo< char > const *)arg1)->num_items();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TypeInfoChar_item(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< char > *arg1 = (egglib::vcf::TypeInfo< char > *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_char_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TypeInfoChar_item" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< char > const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< char > * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TypeInfoChar_item" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (char *) &((egglib::vcf::TypeInfo< char > const *)arg1)->item(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_char(static_cast< char >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TypeInfoChar_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< char > *arg1 = (egglib::vcf::TypeInfo< char > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "TypeInfoChar_reset", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_char_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TypeInfoChar_reset" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< char > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< char > * >(argp1);
  {
    try {
      (arg1)->reset();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TypeInfoChar_get_expected_number(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< char > *arg1 = (egglib::vcf::TypeInfo< char > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "TypeInfoChar_get_expected_number", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_char_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TypeInfoChar_get_expected_number" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< char > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< char > * >(argp1);
  {
    try {
      result = (unsigned int)(arg1)->get_expected_number();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TypeInfoChar_set_expected_number(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< char > *arg1 = (egglib::vcf::TypeInfo< char > *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_char_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TypeInfoChar_set_expected_number" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< char > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< char > * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TypeInfoChar_set_expected_number" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_expected_number(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_TypeInfoChar) /* defines _wrap_delete_TypeInfoChar_destructor_closure */

SWIGINTERN int _wrap_new_TypeInfoDouble__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< double > *result = 0 ;
  
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  {
    try {
      result = (egglib::vcf::TypeInfo< double > *)new egglib::vcf::TypeInfo< double >();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__TypeInfoT_double_t, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_TypeInfoDouble__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< double > *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  egglib::vcf::TypeInfo< double > *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_egglib__vcf__TypeInfoT_double_t,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_TypeInfoDouble" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< double > const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_TypeInfoDouble" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< double > const &""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< double > * >(argp1);
  {
    try {
      result = (egglib::vcf::TypeInfo< double > *)new egglib::vcf::TypeInfo< double >((egglib::vcf::TypeInfo< double > const &)*arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__TypeInfoT_double_t, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_TypeInfoDouble(PyObject *self, PyObject *args, PyObject *kwargs) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_TypeInfoDouble")) SWIG_fail;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_TypeInfoDouble", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_TypeInfoDouble__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_egglib__vcf__TypeInfoT_double_t, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_TypeInfoDouble__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_TypeInfoDouble'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::vcf::TypeInfo< double >::TypeInfo()\n"
    "    egglib::vcf::TypeInfo< double >::TypeInfo(egglib::vcf::TypeInfo< double > const &)\n");
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_TypeInfoDouble(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< double > *arg1 = (egglib::vcf::TypeInfo< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_TypeInfoDouble", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_double_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TypeInfoDouble" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< double > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< double > * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TypeInfoDouble_num_items(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< double > *arg1 = (egglib::vcf::TypeInfo< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "TypeInfoDouble_num_items", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TypeInfoDouble_num_items" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< double > const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< double > * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::vcf::TypeInfo< double > const *)arg1)->num_items();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TypeInfoDouble_item(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< double > *arg1 = (egglib::vcf::TypeInfo< double > *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TypeInfoDouble_item" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< double > const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< double > * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TypeInfoDouble_item" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double *) &((egglib::vcf::TypeInfo< double > const *)arg1)->item(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TypeInfoDouble_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< double > *arg1 = (egglib::vcf::TypeInfo< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "TypeInfoDouble_reset", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TypeInfoDouble_reset" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< double > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< double > * >(argp1);
  {
    try {
      (arg1)->reset();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TypeInfoDouble_get_expected_number(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< double > *arg1 = (egglib::vcf::TypeInfo< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "TypeInfoDouble_get_expected_number", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TypeInfoDouble_get_expected_number" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< double > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< double > * >(argp1);
  {
    try {
      result = (unsigned int)(arg1)->get_expected_number();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TypeInfoDouble_set_expected_number(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::TypeInfo< double > *arg1 = (egglib::vcf::TypeInfo< double > *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__TypeInfoT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TypeInfoDouble_set_expected_number" "', argument " "1"" of type '" "egglib::vcf::TypeInfo< double > *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::TypeInfo< double > * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TypeInfoDouble_set_expected_number" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_expected_number(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_TypeInfoDouble) /* defines _wrap_delete_TypeInfoDouble_destructor_closure */

SWIGINTERN int _wrap_new_StringInfo__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::vcf::StringInfo *result = 0 ;
  
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  {
    try {
      result = (egglib::vcf::StringInfo *)new egglib::vcf::StringInfo();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__StringInfo, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_StringInfo__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::vcf::StringInfo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  egglib::vcf::StringInfo *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_egglib__vcf__StringInfo,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_StringInfo" "', argument " "1"" of type '" "egglib::vcf::StringInfo const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_StringInfo" "', argument " "1"" of type '" "egglib::vcf::StringInfo const &""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::StringInfo * >(argp1);
  {
    try {
      result = (egglib::vcf::StringInfo *)new egglib::vcf::StringInfo((egglib::vcf::StringInfo const &)*arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__StringInfo, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_StringInfo(PyObject *self, PyObject *args, PyObject *kwargs) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_StringInfo")) SWIG_fail;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_StringInfo", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_StringInfo__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_egglib__vcf__StringInfo, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_StringInfo__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_StringInfo'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::vcf::StringInfo::StringInfo()\n"
    "    egglib::vcf::StringInfo::StringInfo(egglib::vcf::StringInfo const &)\n");
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_StringInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::StringInfo *arg1 = (egglib::vcf::StringInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_StringInfo", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__StringInfo, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_StringInfo" "', argument " "1"" of type '" "egglib::vcf::StringInfo *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::StringInfo * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StringInfo_item(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::StringInfo *arg1 = (egglib::vcf::StringInfo *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__StringInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StringInfo_item" "', argument " "1"" of type '" "egglib::vcf::StringInfo *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::StringInfo * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "StringInfo_item" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (char *)(arg1)->item(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_StringInfo_change(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::StringInfo *arg1 = (egglib::vcf::StringInfo *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  char arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  char val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "StringInfo_change", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__StringInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "StringInfo_change" "', argument " "1"" of type '" "egglib::vcf::StringInfo *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::StringInfo * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "StringInfo_change" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "StringInfo_change" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_char(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "StringInfo_change" "', argument " "4"" of type '" "char""'");
  } 
  arg4 = static_cast< char >(val4);
  {
    try {
      (arg1)->change(arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_StringInfo) /* defines _wrap_delete_StringInfo_destructor_closure */

SWIGINTERN int _wrap_new_SampleInfo__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::vcf::SampleInfo *result = 0 ;
  
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  {
    try {
      result = (egglib::vcf::SampleInfo *)new egglib::vcf::SampleInfo();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__SampleInfo, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_SampleInfo__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::vcf::SampleInfo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  egglib::vcf::SampleInfo *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_egglib__vcf__SampleInfo,  0  | 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_SampleInfo" "', argument " "1"" of type '" "egglib::vcf::SampleInfo const &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_SampleInfo" "', argument " "1"" of type '" "egglib::vcf::SampleInfo const &""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::SampleInfo * >(argp1);
  {
    try {
      result = (egglib::vcf::SampleInfo *)new egglib::vcf::SampleInfo((egglib::vcf::SampleInfo const &)*arg1);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__SampleInfo, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN int _wrap_new_SampleInfo(PyObject *self, PyObject *args, PyObject *kwargs) {
  Py_ssize_t argc;
  PyObject *argv[2] = {
    0
  };
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_SampleInfo")) SWIG_fail;
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_SampleInfo", 0, 1, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_SampleInfo__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v;
    int res = SWIG_ConvertPtr(argv[0], 0, SWIGTYPE_p_egglib__vcf__SampleInfo, SWIG_POINTER_NO_NULL | 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_SampleInfo__SWIG_1(self, argc, argv);
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_SampleInfo'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::vcf::SampleInfo::SampleInfo()\n"
    "    egglib::vcf::SampleInfo::SampleInfo(egglib::vcf::SampleInfo const &)\n");
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_SampleInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::SampleInfo *arg1 = (egglib::vcf::SampleInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_SampleInfo", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__SampleInfo, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_SampleInfo" "', argument " "1"" of type '" "egglib::vcf::SampleInfo *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::SampleInfo * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SampleInfo_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::SampleInfo *arg1 = (egglib::vcf::SampleInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "SampleInfo_reset", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__SampleInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SampleInfo_reset" "', argument " "1"" of type '" "egglib::vcf::SampleInfo *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::SampleInfo * >(argp1);
  {
    try {
      (arg1)->reset();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SampleInfo_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::SampleInfo *arg1 = (egglib::vcf::SampleInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "SampleInfo_clear", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__SampleInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SampleInfo_clear" "', argument " "1"" of type '" "egglib::vcf::SampleInfo *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::SampleInfo * >(argp1);
  {
    try {
      (arg1)->clear();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SampleInfo_num_IntegerEntries(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::SampleInfo *arg1 = (egglib::vcf::SampleInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SampleInfo_num_IntegerEntries", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__SampleInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SampleInfo_num_IntegerEntries" "', argument " "1"" of type '" "egglib::vcf::SampleInfo const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::SampleInfo * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::vcf::SampleInfo const *)arg1)->num_IntegerEntries();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SampleInfo_num_IntegerItems(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::SampleInfo *arg1 = (egglib::vcf::SampleInfo *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__SampleInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SampleInfo_num_IntegerItems" "', argument " "1"" of type '" "egglib::vcf::SampleInfo const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::SampleInfo * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SampleInfo_num_IntegerItems" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::vcf::SampleInfo const *)arg1)->num_IntegerItems(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SampleInfo_IntegerItem(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::SampleInfo *arg1 = (egglib::vcf::SampleInfo *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SampleInfo_IntegerItem", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__SampleInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SampleInfo_IntegerItem" "', argument " "1"" of type '" "egglib::vcf::SampleInfo const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::SampleInfo * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SampleInfo_IntegerItem" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "SampleInfo_IntegerItem" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (int)((egglib::vcf::SampleInfo const *)arg1)->IntegerItem(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SampleInfo_num_FloatEntries(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::SampleInfo *arg1 = (egglib::vcf::SampleInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SampleInfo_num_FloatEntries", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__SampleInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SampleInfo_num_FloatEntries" "', argument " "1"" of type '" "egglib::vcf::SampleInfo const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::SampleInfo * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::vcf::SampleInfo const *)arg1)->num_FloatEntries();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SampleInfo_num_FloatItems(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::SampleInfo *arg1 = (egglib::vcf::SampleInfo *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__SampleInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SampleInfo_num_FloatItems" "', argument " "1"" of type '" "egglib::vcf::SampleInfo const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::SampleInfo * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SampleInfo_num_FloatItems" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::vcf::SampleInfo const *)arg1)->num_FloatItems(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SampleInfo_FloatItem(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::SampleInfo *arg1 = (egglib::vcf::SampleInfo *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "SampleInfo_FloatItem", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__SampleInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SampleInfo_FloatItem" "', argument " "1"" of type '" "egglib::vcf::SampleInfo const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::SampleInfo * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SampleInfo_FloatItem" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "SampleInfo_FloatItem" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (double)((egglib::vcf::SampleInfo const *)arg1)->FloatItem(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SampleInfo_num_CharacterEntries(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::SampleInfo *arg1 = (egglib::vcf::SampleInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SampleInfo_num_CharacterEntries", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__SampleInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SampleInfo_num_CharacterEntries" "', argument " "1"" of type '" "egglib::vcf::SampleInfo const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::SampleInfo * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::vcf::SampleInfo const *)arg1)->num_CharacterEntries();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SampleInfo_num_CharacterItems(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::SampleInfo *arg1 = (egglib::vcf::SampleInfo *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__SampleInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SampleInfo_num_CharacterItems" "', argument " "1"" of type '" "egglib::vcf::SampleInfo const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::SampleInfo * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SampleInfo_num_CharacterItems" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::vcf::SampleInfo const *)arg1)->num_CharacterItems(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SampleInfo_CharacterItem(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::SampleInfo *arg1 = (egglib::vcf::SampleInfo *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  char result;
  
  if (!SWIG_Python_UnpackTuple(args, "SampleInfo_CharacterItem", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__SampleInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SampleInfo_CharacterItem" "', argument " "1"" of type '" "egglib::vcf::SampleInfo const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::SampleInfo * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SampleInfo_CharacterItem" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "SampleInfo_CharacterItem" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (char)((egglib::vcf::SampleInfo const *)arg1)->CharacterItem(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_char(static_cast< char >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SampleInfo_num_StringEntries(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::SampleInfo *arg1 = (egglib::vcf::SampleInfo *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "SampleInfo_num_StringEntries", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__SampleInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SampleInfo_num_StringEntries" "', argument " "1"" of type '" "egglib::vcf::SampleInfo const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::SampleInfo * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::vcf::SampleInfo const *)arg1)->num_StringEntries();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SampleInfo_num_StringItems(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::SampleInfo *arg1 = (egglib::vcf::SampleInfo *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__SampleInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SampleInfo_num_StringItems" "', argument " "1"" of type '" "egglib::vcf::SampleInfo const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::SampleInfo * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SampleInfo_num_StringItems" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::vcf::SampleInfo const *)arg1)->num_StringItems(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SampleInfo_StringItem(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::vcf::SampleInfo *arg1 = (egglib::vcf::SampleInfo *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "SampleInfo_StringItem", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__vcf__SampleInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SampleInfo_StringItem" "', argument " "1"" of type '" "egglib::vcf::SampleInfo const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::vcf::SampleInfo * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SampleInfo_StringItem" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "SampleInfo_StringItem" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (char *)((egglib::vcf::SampleInfo const *)arg1)->StringItem(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_SampleInfo) /* defines _wrap_delete_SampleInfo_destructor_closure */

SWIGINTERN int _wrap_new_VcfParser(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::VcfParser *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_VcfParser")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_VcfParser", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::VcfParser *)new egglib::VcfParser();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__VcfParser, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_VcfParser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_VcfParser", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_VcfParser" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_open_file(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_open_file" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfParser_open_file" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->open_file((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_set_stream(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  std::istream *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_set_stream" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_std__istream,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfParser_set_stream" "', argument " "2"" of type '" "std::istream &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VcfParser_set_stream" "', argument " "2"" of type '" "std::istream &""'"); 
  }
  arg2 = reinterpret_cast< std::istream * >(argp2);
  {
    try {
      (arg1)->set_stream(*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_read_header(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_read_header" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfParser_read_header" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->read_header((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_set_threshold_PL(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_set_threshold_PL" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_set_threshold_PL" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_threshold_PL(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_set_threshold_GL(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_set_threshold_GL" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_set_threshold_GL" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->set_threshold_GL(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_get_threshold_PL(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_get_threshold_PL", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_get_threshold_PL" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->get_threshold_PL();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_get_threshold_GL(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_get_threshold_GL", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_get_threshold_GL" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->get_threshold_GL();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_allow_X(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_allow_X" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_allow_X" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->allow_X(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_allow_gap(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_allow_gap" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_allow_gap" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->allow_gap(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_read(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_read", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_read" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      (arg1)->read();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_readline(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_readline" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfParser_readline" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->readline((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_get_currline(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned long result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_get_currline", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_get_currline" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned long)((egglib::VcfParser const *)arg1)->get_currline();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_good(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_good", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_good" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (bool)((egglib::VcfParser const *)arg1)->good();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_has_data(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_has_data", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_has_data" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (bool)((egglib::VcfParser const *)arg1)->has_data();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_reset", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_reset" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      (arg1)->reset();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_reset_variant(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_reset_variant", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_reset_variant" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      (arg1)->reset_variant();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_clear(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_clear", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_clear" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      (arg1)->clear();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_file_format(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_file_format", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_file_format" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (char *)((egglib::VcfParser const *)arg1)->file_format();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_num_filter(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_num_filter", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_num_filter" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->num_filter();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_get_filter(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::vcf::Filter *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_get_filter" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_get_filter" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::vcf::Filter *)((egglib::VcfParser const *)arg1)->get_filter(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Filter, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_num_info(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_num_info", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_num_info" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->num_info();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_get_info(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::vcf::Info *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_get_info" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_get_info" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::vcf::Info *)((egglib::VcfParser const *)arg1)->get_info(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Info, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_num_format(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_num_format", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_num_format" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->num_format();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_get_format(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::vcf::Format *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_get_format" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_get_format" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::vcf::Format *)((egglib::VcfParser const *)arg1)->get_format(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Format, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_num_meta(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_num_meta", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_num_meta" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->num_meta();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_get_meta(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::vcf::Meta *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_get_meta" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_get_meta" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::vcf::Meta *)((egglib::VcfParser const *)arg1)->get_meta(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Meta, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_num_alt(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_num_alt", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_num_alt" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->num_alt();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_get_alt(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::vcf::Alt *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_get_alt" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_get_alt" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::vcf::Alt *)((egglib::VcfParser const *)arg1)->get_alt(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Alt, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_add_filter(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_add_filter", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_add_filter" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfParser_add_filter" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VcfParser_add_filter" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = reinterpret_cast< char * >(buf3);
  {
    try {
      (arg1)->add_filter((char const *)arg2,(char const *)arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_add_alt(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_add_alt", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_add_alt" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfParser_add_alt" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VcfParser_add_alt" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = reinterpret_cast< char * >(buf3);
  {
    try {
      (arg1)->add_alt((char const *)arg2,(char const *)arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_add_info(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  char *arg2 = (char *) 0 ;
  unsigned int arg3 ;
  egglib::vcf::Info::Type arg4 ;
  char *arg5 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  PyObject *swig_obj[5] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_add_info", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_add_info" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfParser_add_info" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VcfParser_add_info" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "VcfParser_add_info" "', argument " "4"" of type '" "egglib::vcf::Info::Type""'");
  } 
  arg4 = static_cast< egglib::vcf::Info::Type >(val4);
  res5 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "VcfParser_add_info" "', argument " "5"" of type '" "char const *""'");
  }
  arg5 = reinterpret_cast< char * >(buf5);
  {
    try {
      (arg1)->add_info((char const *)arg2,arg3,arg4,(char const *)arg5);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc5 == SWIG_NEWOBJ) delete[] buf5;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc5 == SWIG_NEWOBJ) delete[] buf5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_add_format(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  char *arg2 = (char *) 0 ;
  unsigned int arg3 ;
  egglib::vcf::Info::Type arg4 ;
  char *arg5 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  PyObject *swig_obj[5] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_add_format", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_add_format" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfParser_add_format" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VcfParser_add_format" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "VcfParser_add_format" "', argument " "4"" of type '" "egglib::vcf::Info::Type""'");
  } 
  arg4 = static_cast< egglib::vcf::Info::Type >(val4);
  res5 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "VcfParser_add_format" "', argument " "5"" of type '" "char const *""'");
  }
  arg5 = reinterpret_cast< char * >(buf5);
  {
    try {
      (arg1)->add_format((char const *)arg2,arg3,arg4,(char const *)arg5);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc5 == SWIG_NEWOBJ) delete[] buf5;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc5 == SWIG_NEWOBJ) delete[] buf5;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_add_meta(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  char *arg2 = (char *) 0 ;
  char *arg3 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_add_meta", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_add_meta" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfParser_add_meta" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VcfParser_add_meta" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = reinterpret_cast< char * >(buf3);
  {
    try {
      (arg1)->add_meta((char const *)arg2,(char const *)arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_num_samples(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_num_samples", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_num_samples" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->num_samples();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_get_sample(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_get_sample" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_get_sample" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (char *)((egglib::VcfParser const *)arg1)->get_sample(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_find_filter(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::vcf::Filter *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_find_filter" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfParser_find_filter" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      result = (egglib::vcf::Filter *)(arg1)->find_filter((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Filter, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_find_format(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::vcf::Format *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_find_format" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfParser_find_format" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      result = (egglib::vcf::Format *)(arg1)->find_format((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Format, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_find_info(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::vcf::Info *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_find_info" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfParser_find_info" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      result = (egglib::vcf::Info *)(arg1)->find_info((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Info, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_find_meta(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::vcf::Meta *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_find_meta" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfParser_find_meta" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      result = (egglib::vcf::Meta *)(arg1)->find_meta((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Meta, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_find_alt(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::vcf::Alt *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_find_alt" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfParser_find_alt" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      result = (egglib::vcf::Alt *)(arg1)->find_alt((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Alt, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_chromosome(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_chromosome", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_chromosome" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (char *)((egglib::VcfParser const *)arg1)->chromosome();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_position(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned long result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_position", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_position" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned long)((egglib::VcfParser const *)arg1)->position();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_num_ID(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_num_ID", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_num_ID" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->num_ID();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_ID(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_ID" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_ID" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (char *)((egglib::VcfParser const *)arg1)->ID(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_len_reference(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_len_reference", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_len_reference" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->len_reference();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_reference(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_reference", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_reference" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (char *)((egglib::VcfParser const *)arg1)->reference();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_num_alternate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_num_alternate", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_num_alternate" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->num_alternate();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_alternate_type(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::vcf::AltType result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_alternate_type" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_alternate_type" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::vcf::AltType)((egglib::VcfParser const *)arg1)->alternate_type(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_alternate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_alternate" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_alternate" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (char *)((egglib::VcfParser const *)arg1)->alternate(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_quality(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_quality", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_quality" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (double)((egglib::VcfParser const *)arg1)->quality();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_num_failed_tests(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_num_failed_tests", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_num_failed_tests" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->num_failed_tests();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_failed_test(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_failed_test" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_failed_test" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (char *)((egglib::VcfParser const *)arg1)->failed_test(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_num_FlagInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_num_FlagInfo", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_num_FlagInfo" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->num_FlagInfo();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_FlagInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::vcf::FlagInfo result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_FlagInfo" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_FlagInfo" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = ((egglib::VcfParser const *)arg1)->FlagInfo(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj((new egglib::vcf::FlagInfo(static_cast< const egglib::vcf::FlagInfo& >(result))), SWIGTYPE_p_egglib__vcf__FlagInfo, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_num_IntegerInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_num_IntegerInfo", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_num_IntegerInfo" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->num_IntegerInfo();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_IntegerInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::vcf::TypeInfo< int > *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_IntegerInfo" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_IntegerInfo" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::vcf::TypeInfo< int > *) &((egglib::VcfParser const *)arg1)->IntegerInfo(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__TypeInfoT_int_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_num_FloatInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_num_FloatInfo", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_num_FloatInfo" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->num_FloatInfo();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_FloatInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::vcf::TypeInfo< double > *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_FloatInfo" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_FloatInfo" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::vcf::TypeInfo< double > *) &((egglib::VcfParser const *)arg1)->FloatInfo(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__TypeInfoT_double_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_num_CharacterInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_num_CharacterInfo", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_num_CharacterInfo" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->num_CharacterInfo();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_CharacterInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::vcf::TypeInfo< char > *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_CharacterInfo" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_CharacterInfo" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::vcf::TypeInfo< char > *) &((egglib::VcfParser const *)arg1)->CharacterInfo(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__TypeInfoT_char_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_num_StringInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_num_StringInfo", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_num_StringInfo" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->num_StringInfo();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_StringInfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::vcf::StringInfo *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_StringInfo" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_StringInfo" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::vcf::StringInfo *) &((egglib::VcfParser const *)arg1)->StringInfo(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__StringInfo, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_num_fields(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_num_fields", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_num_fields" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->num_fields();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_field(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::vcf::Format *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_field" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_field" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::vcf::Format *) &((egglib::VcfParser const *)arg1)->field(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__Format, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_field_index(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_field_index" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfParser_field_index" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->field_index((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_field_rank__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int result;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_field_rank" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_field_rank" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->field_rank(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_field_rank__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  unsigned int result;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_field_rank" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfParser_field_rank" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->field_rank((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_field_rank(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VcfParser_field_rank", 0, 2, argv+1))) SWIG_fail;
  argv[0] = self;
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__VcfParser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VcfParser_field_rank__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__VcfParser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_VcfParser_field_rank__SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VcfParser_field_rank'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::VcfParser::field_rank(unsigned int) const\n"
    "    egglib::VcfParser::field_rank(char const *) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VcfParser_sample_info(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::vcf::SampleInfo *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_sample_info" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_sample_info" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::vcf::SampleInfo *) &((egglib::VcfParser const *)arg1)->sample_info(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__vcf__SampleInfo, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_has_AN(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_has_AN", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_has_AN" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (bool)((egglib::VcfParser const *)arg1)->has_AN();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_AN(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_AN", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_AN" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->AN();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_has_AA(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_has_AA", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_has_AA" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (bool)((egglib::VcfParser const *)arg1)->has_AA();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_AA_string(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_AA_string", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_AA_string" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (char *)((egglib::VcfParser const *)arg1)->AA_string();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_AA_index(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_AA_index", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_AA_index" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->AA_index();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_has_AC(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_has_AC", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_has_AC" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (bool)((egglib::VcfParser const *)arg1)->has_AC();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_num_AC(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_num_AC", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_num_AC" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->num_AC();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_AC(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_AC" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_AC" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->AC(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_has_AF(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_has_AF", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_has_AF" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (bool)((egglib::VcfParser const *)arg1)->has_AF();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_num_AF(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_num_AF", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_num_AF" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->num_AF();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_AF(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_AF" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_AF" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((egglib::VcfParser const *)arg1)->AF(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_has_GT(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_has_GT", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_has_GT" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (bool)((egglib::VcfParser const *)arg1)->has_GT();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_ploidy(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_ploidy", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_ploidy" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->ploidy();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_num_genotypes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_num_genotypes", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_num_genotypes" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->num_genotypes();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_GT_phased__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  bool result;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_GT_phased" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_GT_phased" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (bool)((egglib::VcfParser const *)arg1)->GT_phased(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_GT_phased__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool result;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_GT_phased" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (bool)((egglib::VcfParser const *)arg1)->GT_phased();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_GT_phased(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "VcfParser_GT_phased", 0, 2, argv+1))) SWIG_fail;
  argv[0] = self;
  if (argc == 1) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__VcfParser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_VcfParser_GT_phased__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_egglib__VcfParser, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_VcfParser_GT_phased__SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'VcfParser_GT_phased'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    egglib::VcfParser::GT_phased(unsigned int) const\n"
    "    egglib::VcfParser::GT_phased() const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_VcfParser_GT(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_GT", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_GT" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_GT" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VcfParser_GT" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->GT(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_has_PL(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_has_PL", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_has_PL" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (bool)((egglib::VcfParser const *)arg1)->has_PL();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_has_GL(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_has_GL", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_has_GL" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (bool)((egglib::VcfParser const *)arg1)->has_GL();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_PL(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_PL", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_PL" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_PL" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VcfParser_PL" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->PL(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_GL(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_GL", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_GL" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfParser_GL" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VcfParser_GL" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (double)((egglib::VcfParser const *)arg1)->GL(arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_unread(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_unread", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_unread" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      (arg1)->unread();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_get_index(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::VcfIndex *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_get_index", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_get_index" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (egglib::VcfIndex *) &(arg1)->get_index();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__VcfIndex, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_rewind(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_rewind", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_rewind" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      (arg1)->rewind();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_type_alleles(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfParser_type_alleles", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_type_alleles" "', argument " "1"" of type '" "egglib::VcfParser const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfParser const *)arg1)->type_alleles();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfParser_set_alleles(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = (egglib::VcfParser *) 0 ;
  egglib::StringAlphabet *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfParser_set_alleles" "', argument " "1"" of type '" "egglib::VcfParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__StringAlphabet,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfParser_set_alleles" "', argument " "2"" of type '" "egglib::StringAlphabet &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VcfParser_set_alleles" "', argument " "2"" of type '" "egglib::StringAlphabet &""'"); 
  }
  arg2 = reinterpret_cast< egglib::StringAlphabet * >(argp2);
  {
    try {
      (arg1)->set_alleles(*arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_VcfParser) /* defines _wrap_delete_VcfParser_destructor_closure */

SWIGINTERN PyObject *_wrap_VcfIndex_go_to(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfIndex *arg1 = (egglib::VcfIndex *) 0 ;
  char *arg2 = (char *) 0 ;
  unsigned long arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  unsigned long val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfIndex_go_to", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfIndex, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfIndex_go_to" "', argument " "1"" of type '" "egglib::VcfIndex *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfIndex * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfIndex_go_to" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  ecode3 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VcfIndex_go_to" "', argument " "3"" of type '" "unsigned long""'");
  } 
  arg3 = static_cast< unsigned long >(val3);
  {
    try {
      (arg1)->go_to((char const *)arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfIndex_num(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfIndex *arg1 = (egglib::VcfIndex *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned long result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfIndex_num", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfIndex, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfIndex_num" "', argument " "1"" of type '" "egglib::VcfIndex const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfIndex * >(argp1);
  {
    try {
      result = (unsigned long)((egglib::VcfIndex const *)arg1)->num();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfIndex_has_data(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfIndex *arg1 = (egglib::VcfIndex *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfIndex_has_data", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfIndex, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfIndex_has_data" "', argument " "1"" of type '" "egglib::VcfIndex const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfIndex * >(argp1);
  {
    try {
      result = (bool)((egglib::VcfIndex const *)arg1)->has_data();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfIndex_load_data(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfIndex *arg1 = (egglib::VcfIndex *) 0 ;
  egglib::VcfParser *arg2 = 0 ;
  char *arg3 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfIndex_load_data", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfIndex, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfIndex_load_data" "', argument " "1"" of type '" "egglib::VcfIndex *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfIndex * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__VcfParser,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfIndex_load_data" "', argument " "2"" of type '" "egglib::VcfParser &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VcfIndex_load_data" "', argument " "2"" of type '" "egglib::VcfParser &""'"); 
  }
  arg2 = reinterpret_cast< egglib::VcfParser * >(argp2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VcfIndex_load_data" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = reinterpret_cast< char * >(buf3);
  {
    try {
      (arg1)->load_data(*arg2,(char const *)arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) delete[] buf3;
  return NULL;
}


SWIGINTERN int _wrap_new_VcfIndex(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::VcfIndex *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_VcfIndex")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_VcfIndex", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::VcfIndex *)new egglib::VcfIndex();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__VcfIndex, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_VcfIndex(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfIndex *arg1 = (egglib::VcfIndex *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_VcfIndex", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfIndex, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_VcfIndex" "', argument " "1"" of type '" "egglib::VcfIndex *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfIndex * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_VcfIndex) /* defines _wrap_delete_VcfIndex_destructor_closure */

SWIGINTERN PyObject *_wrap_WSite_site(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::WSite *arg1 = (egglib::WSite *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::SiteHolder *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "WSite_site", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__WSite, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "WSite_site" "', argument " "1"" of type '" "egglib::WSite *""'"); 
  }
  arg1 = reinterpret_cast< egglib::WSite * >(argp1);
  {
    try {
      result = (egglib::SiteHolder *) &(arg1)->site();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__SiteHolder, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_WSite_alphabet(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::WSite *arg1 = (egglib::WSite *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::StringAlphabet *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "WSite_alphabet", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__WSite, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "WSite_alphabet" "', argument " "1"" of type '" "egglib::WSite *""'"); 
  }
  arg1 = reinterpret_cast< egglib::WSite * >(argp1);
  {
    try {
      result = (egglib::StringAlphabet *)(arg1)->alphabet();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__StringAlphabet, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_WSite_get_pos(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::WSite *arg1 = (egglib::WSite *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "WSite_get_pos", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__WSite, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "WSite_get_pos" "', argument " "1"" of type '" "egglib::WSite const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::WSite * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::WSite const *)arg1)->get_pos();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_WSite_next(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::WSite *arg1 = (egglib::WSite *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::WSite *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "WSite_next", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__WSite, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "WSite_next" "', argument " "1"" of type '" "egglib::WSite *""'"); 
  }
  arg1 = reinterpret_cast< egglib::WSite * >(argp1);
  {
    try {
      result = (egglib::WSite *)(arg1)->next();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__WSite, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_WSite_prev(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::WSite *arg1 = (egglib::WSite *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::WSite *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "WSite_prev", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__WSite, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "WSite_prev" "', argument " "1"" of type '" "egglib::WSite *""'"); 
  }
  arg1 = reinterpret_cast< egglib::WSite * >(argp1);
  {
    try {
      result = (egglib::WSite *)(arg1)->prev();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__WSite, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_WSite(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::WSite *arg1 = (egglib::WSite *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_WSite", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__WSite, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_WSite" "', argument " "1"" of type '" "egglib::WSite *""'"); 
  }
  arg1 = reinterpret_cast< egglib::WSite * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_WSite) /* defines _wrap_delete_WSite_destructor_closure */

SWIGINTERN PyObject *_wrap_delete_VcfWindowBase(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowBase *arg1 = (egglib::VcfWindowBase *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_VcfWindowBase", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowBase, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_VcfWindowBase" "', argument " "1"" of type '" "egglib::VcfWindowBase *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowBase * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfWindowBase_setup_base(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowBase *arg1 = (egglib::VcfWindowBase *) 0 ;
  egglib::VcfParser *arg2 = 0 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfWindowBase_setup_base", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowBase, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfWindowBase_setup_base" "', argument " "1"" of type '" "egglib::VcfWindowBase *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowBase * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__VcfParser,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfWindowBase_setup_base" "', argument " "2"" of type '" "egglib::VcfParser &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VcfWindowBase_setup_base" "', argument " "2"" of type '" "egglib::VcfParser &""'"); 
  }
  arg2 = reinterpret_cast< egglib::VcfParser * >(argp2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VcfWindowBase_setup_base" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "VcfWindowBase_setup_base" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  {
    try {
      (arg1)->setup_base(*arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfWindowBase_chromosome(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowBase *arg1 = (egglib::VcfWindowBase *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfWindowBase_chromosome", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowBase, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfWindowBase_chromosome" "', argument " "1"" of type '" "egglib::VcfWindowBase const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowBase * >(argp1);
  {
    try {
      result = (char *)((egglib::VcfWindowBase const *)arg1)->chromosome();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfWindowBase_num_samples(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowBase *arg1 = (egglib::VcfWindowBase *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfWindowBase_num_samples", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowBase, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfWindowBase_num_samples" "', argument " "1"" of type '" "egglib::VcfWindowBase const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowBase * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfWindowBase const *)arg1)->num_samples();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfWindowBase_num_sites(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowBase *arg1 = (egglib::VcfWindowBase *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfWindowBase_num_sites", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowBase, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfWindowBase_num_sites" "', argument " "1"" of type '" "egglib::VcfWindowBase const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowBase * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::VcfWindowBase const *)arg1)->num_sites();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfWindowBase_win_start(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowBase *arg1 = (egglib::VcfWindowBase *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned long result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfWindowBase_win_start", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowBase, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfWindowBase_win_start" "', argument " "1"" of type '" "egglib::VcfWindowBase const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowBase * >(argp1);
  {
    try {
      result = (unsigned long)((egglib::VcfWindowBase const *)arg1)->win_start();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfWindowBase_win_stop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowBase *arg1 = (egglib::VcfWindowBase *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned long result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfWindowBase_win_stop", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowBase, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfWindowBase_win_stop" "', argument " "1"" of type '" "egglib::VcfWindowBase const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowBase * >(argp1);
  {
    try {
      result = (unsigned long)((egglib::VcfWindowBase const *)arg1)->win_stop();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfWindowBase_first_site(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowBase *arg1 = (egglib::VcfWindowBase *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::WSite *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfWindowBase_first_site", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowBase, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfWindowBase_first_site" "', argument " "1"" of type '" "egglib::VcfWindowBase const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowBase * >(argp1);
  {
    try {
      result = (egglib::WSite *)((egglib::VcfWindowBase const *)arg1)->first_site();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__WSite, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfWindowBase_last_site(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowBase *arg1 = (egglib::VcfWindowBase *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  egglib::WSite *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfWindowBase_last_site", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowBase, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfWindowBase_last_site" "', argument " "1"" of type '" "egglib::VcfWindowBase const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowBase * >(argp1);
  {
    try {
      result = (egglib::WSite *)((egglib::VcfWindowBase const *)arg1)->last_site();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__WSite, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfWindowBase_good(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowBase *arg1 = (egglib::VcfWindowBase *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfWindowBase_good", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowBase, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfWindowBase_good" "', argument " "1"" of type '" "egglib::VcfWindowBase const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowBase * >(argp1);
  {
    try {
      result = (bool)((egglib::VcfWindowBase const *)arg1)->good();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfWindowBase_get_site(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowBase *arg1 = (egglib::VcfWindowBase *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  egglib::WSite *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowBase, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfWindowBase_get_site" "', argument " "1"" of type '" "egglib::VcfWindowBase const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowBase * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "VcfWindowBase_get_site" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (egglib::WSite *)((egglib::VcfWindowBase const *)arg1)->get_site(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__WSite, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfWindowBase_next_window(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowBase *arg1 = (egglib::VcfWindowBase *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfWindowBase_next_window", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowBase, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfWindowBase_next_window" "', argument " "1"" of type '" "egglib::VcfWindowBase *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowBase * >(argp1);
  {
    try {
      (arg1)->next_window();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_VcfWindowBase) /* defines _wrap_delete_VcfWindowBase_destructor_closure */

SWIGINTERN PyObject *_wrap_delete_VcfWindowSliderBase(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowSliderBase *arg1 = (egglib::VcfWindowSliderBase *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_VcfWindowSliderBase", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowSliderBase, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_VcfWindowSliderBase" "', argument " "1"" of type '" "egglib::VcfWindowSliderBase *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowSliderBase * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfWindowSliderBase_setup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowSliderBase *arg1 = (egglib::VcfWindowSliderBase *) 0 ;
  egglib::VcfParser *arg2 = 0 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned long arg5 ;
  unsigned long arg6 ;
  unsigned int arg7 ;
  unsigned int arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned long val5 ;
  int ecode5 = 0 ;
  unsigned long val6 ;
  int ecode6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  unsigned int val8 ;
  int ecode8 = 0 ;
  PyObject *swig_obj[8] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfWindowSliderBase_setup", 7, 7, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowSliderBase, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfWindowSliderBase_setup" "', argument " "1"" of type '" "egglib::VcfWindowSliderBase *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowSliderBase * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__VcfParser,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfWindowSliderBase_setup" "', argument " "2"" of type '" "egglib::VcfParser &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VcfWindowSliderBase_setup" "', argument " "2"" of type '" "egglib::VcfParser &""'"); 
  }
  arg2 = reinterpret_cast< egglib::VcfParser * >(argp2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VcfWindowSliderBase_setup" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "VcfWindowSliderBase_setup" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_long(swig_obj[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "VcfWindowSliderBase_setup" "', argument " "5"" of type '" "unsigned long""'");
  } 
  arg5 = static_cast< unsigned long >(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_long(swig_obj[4], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "VcfWindowSliderBase_setup" "', argument " "6"" of type '" "unsigned long""'");
  } 
  arg6 = static_cast< unsigned long >(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(swig_obj[5], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "VcfWindowSliderBase_setup" "', argument " "7"" of type '" "unsigned int""'");
  } 
  arg7 = static_cast< unsigned int >(val7);
  ecode8 = SWIG_AsVal_unsigned_SS_int(swig_obj[6], &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "VcfWindowSliderBase_setup" "', argument " "8"" of type '" "unsigned int""'");
  } 
  arg8 = static_cast< unsigned int >(val8);
  {
    try {
      (arg1)->setup(*arg2,arg3,arg4,arg5,arg6,arg7,arg8);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfWindowSliderBase_next_window(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowSliderBase *arg1 = (egglib::VcfWindowSliderBase *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfWindowSliderBase_next_window", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowSliderBase, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfWindowSliderBase_next_window" "', argument " "1"" of type '" "egglib::VcfWindowSliderBase *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowSliderBase * >(argp1);
  {
    try {
      (arg1)->next_window();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_VcfWindowSliderBase) /* defines _wrap_delete_VcfWindowSliderBase_destructor_closure */

SWIGINTERN int _wrap_new_VcfWindowSlider(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::VcfWindowSlider *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_VcfWindowSlider")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_VcfWindowSlider", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::VcfWindowSlider *)new egglib::VcfWindowSlider();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__VcfWindowSlider, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_VcfWindowSlider_setup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowSlider *arg1 = (egglib::VcfWindowSlider *) 0 ;
  egglib::VcfParser *arg2 = 0 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned long arg5 ;
  unsigned long arg6 ;
  unsigned int arg7 ;
  unsigned int arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned long val5 ;
  int ecode5 = 0 ;
  unsigned long val6 ;
  int ecode6 = 0 ;
  unsigned int val7 ;
  int ecode7 = 0 ;
  unsigned int val8 ;
  int ecode8 = 0 ;
  PyObject *swig_obj[8] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfWindowSlider_setup", 7, 7, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowSlider, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfWindowSlider_setup" "', argument " "1"" of type '" "egglib::VcfWindowSlider *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowSlider * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__VcfParser,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfWindowSlider_setup" "', argument " "2"" of type '" "egglib::VcfParser &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VcfWindowSlider_setup" "', argument " "2"" of type '" "egglib::VcfParser &""'"); 
  }
  arg2 = reinterpret_cast< egglib::VcfParser * >(argp2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "VcfWindowSlider_setup" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "VcfWindowSlider_setup" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_long(swig_obj[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "VcfWindowSlider_setup" "', argument " "5"" of type '" "unsigned long""'");
  } 
  arg5 = static_cast< unsigned long >(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_long(swig_obj[4], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "VcfWindowSlider_setup" "', argument " "6"" of type '" "unsigned long""'");
  } 
  arg6 = static_cast< unsigned long >(val6);
  ecode7 = SWIG_AsVal_unsigned_SS_int(swig_obj[5], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "VcfWindowSlider_setup" "', argument " "7"" of type '" "unsigned int""'");
  } 
  arg7 = static_cast< unsigned int >(val7);
  ecode8 = SWIG_AsVal_unsigned_SS_int(swig_obj[6], &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "VcfWindowSlider_setup" "', argument " "8"" of type '" "unsigned int""'");
  } 
  arg8 = static_cast< unsigned int >(val8);
  {
    try {
      (arg1)->setup(*arg2,arg3,arg4,arg5,arg6,arg7,arg8);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_VcfWindowSlider(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowSlider *arg1 = (egglib::VcfWindowSlider *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_VcfWindowSlider", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowSlider, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_VcfWindowSlider" "', argument " "1"" of type '" "egglib::VcfWindowSlider *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowSlider * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfWindowSlider_next_window(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowSlider *arg1 = (egglib::VcfWindowSlider *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfWindowSlider_next_window", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowSlider, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfWindowSlider_next_window" "', argument " "1"" of type '" "egglib::VcfWindowSlider *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowSlider * >(argp1);
  {
    try {
      (arg1)->next_window();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_VcfWindowSlider) /* defines _wrap_delete_VcfWindowSlider_destructor_closure */

SWIGINTERN int _wrap_new_VcfWindowSliderPerSite(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::VcfWindowSliderPerSite *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_VcfWindowSliderPerSite")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_VcfWindowSliderPerSite", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::VcfWindowSliderPerSite *)new egglib::VcfWindowSliderPerSite();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__VcfWindowSliderPerSite, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_VcfWindowSliderPerSite(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowSliderPerSite *arg1 = (egglib::VcfWindowSliderPerSite *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_VcfWindowSliderPerSite", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowSliderPerSite, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_VcfWindowSliderPerSite" "', argument " "1"" of type '" "egglib::VcfWindowSliderPerSite *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowSliderPerSite * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfWindowSliderPerSite_next_window(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowSliderPerSite *arg1 = (egglib::VcfWindowSliderPerSite *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfWindowSliderPerSite_next_window", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowSliderPerSite, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfWindowSliderPerSite_next_window" "', argument " "1"" of type '" "egglib::VcfWindowSliderPerSite *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowSliderPerSite * >(argp1);
  {
    try {
      (arg1)->next_window();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_VcfWindowSliderPerSite) /* defines _wrap_delete_VcfWindowSliderPerSite_destructor_closure */

SWIGINTERN int _wrap_new_VcfWindowBED(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::VcfWindowBED *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_VcfWindowBED")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_VcfWindowBED", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::VcfWindowBED *)new egglib::VcfWindowBED();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__VcfWindowBED, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_VcfWindowBED(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowBED *arg1 = (egglib::VcfWindowBED *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_VcfWindowBED", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowBED, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_VcfWindowBED" "', argument " "1"" of type '" "egglib::VcfWindowBED *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowBED * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfWindowBED_setup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowBED *arg1 = (egglib::VcfWindowBED *) 0 ;
  egglib::VcfParser *arg2 = 0 ;
  egglib::BedParser *arg3 = 0 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfWindowBED_setup", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowBED, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfWindowBED_setup" "', argument " "1"" of type '" "egglib::VcfWindowBED *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowBED * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__VcfParser,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "VcfWindowBED_setup" "', argument " "2"" of type '" "egglib::VcfParser &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VcfWindowBED_setup" "', argument " "2"" of type '" "egglib::VcfParser &""'"); 
  }
  arg2 = reinterpret_cast< egglib::VcfParser * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[1], &argp3, SWIGTYPE_p_egglib__BedParser,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "VcfWindowBED_setup" "', argument " "3"" of type '" "egglib::BedParser &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "VcfWindowBED_setup" "', argument " "3"" of type '" "egglib::BedParser &""'"); 
  }
  arg3 = reinterpret_cast< egglib::BedParser * >(argp3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "VcfWindowBED_setup" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(swig_obj[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "VcfWindowBED_setup" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = static_cast< unsigned int >(val5);
  {
    try {
      (arg1)->setup(*arg2,*arg3,arg4,arg5);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_VcfWindowBED_next_window(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfWindowBED *arg1 = (egglib::VcfWindowBED *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "VcfWindowBED_next_window", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__VcfWindowBED, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "VcfWindowBED_next_window" "', argument " "1"" of type '" "egglib::VcfWindowBED *""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfWindowBED * >(argp1);
  {
    try {
      (arg1)->next_window();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_VcfWindowBED) /* defines _wrap_delete_VcfWindowBED_destructor_closure */

SWIGINTERN int _wrap_new_BedParser(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::BedParser *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_BedParser")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_BedParser", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::BedParser *)new egglib::BedParser();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__BedParser, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_BedParser(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::BedParser *arg1 = (egglib::BedParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_BedParser", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__BedParser, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_BedParser" "', argument " "1"" of type '" "egglib::BedParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::BedParser * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BedParser_get_bed_file(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::BedParser *arg1 = (egglib::BedParser *) 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__BedParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BedParser_get_bed_file" "', argument " "1"" of type '" "egglib::BedParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::BedParser * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "BedParser_get_bed_file" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      (arg1)->get_bed_file((char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_BedParser_n_bed_data(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::BedParser *arg1 = (egglib::BedParser *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "BedParser_n_bed_data", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__BedParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BedParser_n_bed_data" "', argument " "1"" of type '" "egglib::BedParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::BedParser * >(argp1);
  {
    try {
      result = (unsigned int)(arg1)->n_bed_data();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BedParser_get_chrom(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::BedParser *arg1 = (egglib::BedParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__BedParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BedParser_get_chrom" "', argument " "1"" of type '" "egglib::BedParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::BedParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "BedParser_get_chrom" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (char *)(arg1)->get_chrom(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BedParser_get_start(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::BedParser *arg1 = (egglib::BedParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned long result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__BedParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BedParser_get_start" "', argument " "1"" of type '" "egglib::BedParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::BedParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "BedParser_get_start" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned long)(arg1)->get_start(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BedParser_get_end(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::BedParser *arg1 = (egglib::BedParser *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  unsigned long result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__BedParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BedParser_get_end" "', argument " "1"" of type '" "egglib::BedParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::BedParser * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "BedParser_get_end" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (unsigned long)(arg1)->get_end(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_BedParser_append(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::BedParser *arg1 = (egglib::BedParser *) 0 ;
  char *arg2 = (char *) 0 ;
  unsigned long arg3 ;
  unsigned long arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  unsigned long val3 ;
  int ecode3 = 0 ;
  unsigned long val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "BedParser_append", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__BedParser, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "BedParser_append" "', argument " "1"" of type '" "egglib::BedParser *""'"); 
  }
  arg1 = reinterpret_cast< egglib::BedParser * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "BedParser_append" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  ecode3 = SWIG_AsVal_unsigned_SS_long(swig_obj[1], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "BedParser_append" "', argument " "3"" of type '" "unsigned long""'");
  } 
  arg3 = static_cast< unsigned long >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_long(swig_obj[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "BedParser_append" "', argument " "4"" of type '" "unsigned long""'");
  } 
  arg4 = static_cast< unsigned long >(val4);
  {
    try {
      (arg1)->append((char const *)arg2,arg3,arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_BedParser) /* defines _wrap_delete_BedParser_destructor_closure */

SWIGINTERN int _wrap_new_ComputeV(PyObject *self, PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  egglib::ComputeV *result = 0 ;
  
  if (!SWIG_Python_CheckNoKeywords(kwargs, "new_ComputeV")) SWIG_fail;
  if (!SWIG_Python_UnpackTuple(args, "new_ComputeV", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::ComputeV *)new egglib::ComputeV();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__ComputeV, SWIG_BUILTIN_INIT |  0 );
  return resultobj == Py_None ? -1 : 0;
fail:
  return -1;
}


SWIGINTERN PyObject *_wrap_delete_ComputeV(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ComputeV *arg1 = (egglib::ComputeV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "delete_ComputeV", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ComputeV, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_ComputeV" "', argument " "1"" of type '" "egglib::ComputeV *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ComputeV * >(argp1);
  {
    try {
      delete arg1;
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ComputeV_reset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ComputeV *arg1 = (egglib::ComputeV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!SWIG_Python_UnpackTuple(args, "ComputeV_reset", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ComputeV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ComputeV_reset" "', argument " "1"" of type '" "egglib::ComputeV *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ComputeV * >(argp1);
  {
    try {
      (arg1)->reset();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ComputeV_compute(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ComputeV *arg1 = (egglib::ComputeV *) 0 ;
  egglib::FreqHolder *arg2 = 0 ;
  egglib::AbstractTemplateAlphabet< int > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  if (!SWIG_Python_UnpackTuple(args, "ComputeV_compute", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ComputeV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ComputeV_compute" "', argument " "1"" of type '" "egglib::ComputeV *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ComputeV * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[0], &argp2, SWIGTYPE_p_egglib__FreqHolder,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "ComputeV_compute" "', argument " "2"" of type '" "egglib::FreqHolder const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "ComputeV_compute" "', argument " "2"" of type '" "egglib::FreqHolder const &""'"); 
  }
  arg2 = reinterpret_cast< egglib::FreqHolder * >(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[1], &argp3, SWIGTYPE_p_egglib__AbstractTemplateAlphabetT_int_t,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "ComputeV_compute" "', argument " "3"" of type '" "egglib::AbstractTemplateAlphabet< int > &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "ComputeV_compute" "', argument " "3"" of type '" "egglib::AbstractTemplateAlphabet< int > &""'"); 
  }
  arg3 = reinterpret_cast< egglib::AbstractTemplateAlphabet< int > * >(argp3);
  {
    try {
      result = (bool)(arg1)->compute((egglib::FreqHolder const &)*arg2,*arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ComputeV_average_V(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ComputeV *arg1 = (egglib::ComputeV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "ComputeV_average_V", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ComputeV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ComputeV_average_V" "', argument " "1"" of type '" "egglib::ComputeV const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ComputeV * >(argp1);
  {
    try {
      result = (double)((egglib::ComputeV const *)arg1)->average_V();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ComputeV_average_Ar(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ComputeV *arg1 = (egglib::ComputeV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "ComputeV_average_Ar", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ComputeV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ComputeV_average_Ar" "', argument " "1"" of type '" "egglib::ComputeV const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ComputeV * >(argp1);
  {
    try {
      result = (double)((egglib::ComputeV const *)arg1)->average_Ar();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ComputeV_average_M(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ComputeV *arg1 = (egglib::ComputeV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "ComputeV_average_M", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ComputeV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ComputeV_average_M" "', argument " "1"" of type '" "egglib::ComputeV const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ComputeV * >(argp1);
  {
    try {
      result = (double)((egglib::ComputeV const *)arg1)->average_M();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ComputeV_average_Rst(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ComputeV *arg1 = (egglib::ComputeV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "ComputeV_average_Rst", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ComputeV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ComputeV_average_Rst" "', argument " "1"" of type '" "egglib::ComputeV const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ComputeV * >(argp1);
  {
    try {
      result = (double)((egglib::ComputeV const *)arg1)->average_Rst();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ComputeV_curr_V(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ComputeV *arg1 = (egglib::ComputeV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "ComputeV_curr_V", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ComputeV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ComputeV_curr_V" "', argument " "1"" of type '" "egglib::ComputeV const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ComputeV * >(argp1);
  {
    try {
      result = (double)((egglib::ComputeV const *)arg1)->curr_V();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ComputeV_curr_Ar(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ComputeV *arg1 = (egglib::ComputeV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "ComputeV_curr_Ar", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ComputeV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ComputeV_curr_Ar" "', argument " "1"" of type '" "egglib::ComputeV const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ComputeV * >(argp1);
  {
    try {
      result = (double)((egglib::ComputeV const *)arg1)->curr_Ar();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ComputeV_curr_M(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ComputeV *arg1 = (egglib::ComputeV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "ComputeV_curr_M", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ComputeV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ComputeV_curr_M" "', argument " "1"" of type '" "egglib::ComputeV const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ComputeV * >(argp1);
  {
    try {
      result = (double)((egglib::ComputeV const *)arg1)->curr_M();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ComputeV_curr_Rst(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ComputeV *arg1 = (egglib::ComputeV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "ComputeV_curr_Rst", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ComputeV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ComputeV_curr_Rst" "', argument " "1"" of type '" "egglib::ComputeV const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ComputeV * >(argp1);
  {
    try {
      result = (double)((egglib::ComputeV const *)arg1)->curr_Rst();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ComputeV_num_sites(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ComputeV *arg1 = (egglib::ComputeV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "ComputeV_num_sites", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ComputeV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ComputeV_num_sites" "', argument " "1"" of type '" "egglib::ComputeV const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ComputeV * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::ComputeV const *)arg1)->num_sites();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ComputeV_num_sites_m(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ComputeV *arg1 = (egglib::ComputeV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "ComputeV_num_sites_m", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ComputeV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ComputeV_num_sites_m" "', argument " "1"" of type '" "egglib::ComputeV const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ComputeV * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::ComputeV const *)arg1)->num_sites_m();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ComputeV_num_sites_rst(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ComputeV *arg1 = (egglib::ComputeV *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!SWIG_Python_UnpackTuple(args, "ComputeV_num_sites_rst", 0, 0, 0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ComputeV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ComputeV_num_sites_rst" "', argument " "1"" of type '" "egglib::ComputeV const *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ComputeV * >(argp1);
  {
    try {
      result = (unsigned int)((egglib::ComputeV const *)arg1)->num_sites_rst();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ComputeV_set_maf(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::ComputeV *arg1 = (egglib::ComputeV *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_egglib__ComputeV, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ComputeV_set_maf" "', argument " "1"" of type '" "egglib::ComputeV *""'"); 
  }
  arg1 = reinterpret_cast< egglib::ComputeV * >(argp1);
  ecode2 = SWIG_AsVal_double(swig_obj[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "ComputeV_set_maf" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->set_maf(arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGPY_DESTRUCTOR_CLOSURE(_wrap_delete_ComputeV) /* defines _wrap_delete_ComputeV_destructor_closure */

SWIGINTERN int Swig_var_MISSINGDATA_set(PyObject *) {
  SWIG_Error(SWIG_AttributeError,"Variable MISSINGDATA is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_MISSINGDATA_get(void) {
  PyObject *pyobj = 0;
  PyObject *self = 0;
  
  (void)self;
  pyobj = SWIG_From_int(static_cast< int >(egglib::MISSINGDATA));
  return pyobj;
}


SWIGINTERN int Swig_var_MAX_ALLELE_RANGE_set(PyObject *) {
  SWIG_Error(SWIG_AttributeError,"Variable MAX_ALLELE_RANGE is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_MAX_ALLELE_RANGE_get(void) {
  PyObject *pyobj = 0;
  PyObject *self = 0;
  
  (void)self;
  pyobj = SWIG_From_int(static_cast< int >(egglib::MAX_ALLELE_RANGE));
  return pyobj;
}


SWIGINTERN int Swig_var_UNDEF_set(PyObject *) {
  SWIG_Error(SWIG_AttributeError,"Variable UNDEF is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_UNDEF_get(void) {
  PyObject *pyobj = 0;
  PyObject *self = 0;
  
  (void)self;
  pyobj = SWIG_From_double(static_cast< double >(egglib::UNDEF));
  return pyobj;
}


SWIGINTERN int Swig_var_UNKNOWN_set(PyObject *) {
  SWIG_Error(SWIG_AttributeError,"Variable UNKNOWN is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_UNKNOWN_get(void) {
  PyObject *pyobj = 0;
  PyObject *self = 0;
  
  (void)self;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(egglib::UNKNOWN));
  return pyobj;
}


SWIGINTERN int Swig_var_MAX_set(PyObject *) {
  SWIG_Error(SWIG_AttributeError,"Variable MAX is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_MAX_get(void) {
  PyObject *pyobj = 0;
  PyObject *self = 0;
  
  (void)self;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(egglib::MAX));
  return pyobj;
}


SWIGINTERN int Swig_var_MAXCHAR_set(PyObject *) {
  SWIG_Error(SWIG_AttributeError,"Variable MAXCHAR is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_MAXCHAR_get(void) {
  PyObject *pyobj = 0;
  PyObject *self = 0;
  
  (void)self;
  pyobj = SWIG_From_char(static_cast< char >(egglib::MAXCHAR));
  return pyobj;
}


SWIGINTERN int Swig_var_OUTGROUP_set(PyObject *) {
  SWIG_Error(SWIG_AttributeError,"Variable OUTGROUP is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_OUTGROUP_get(void) {
  PyObject *pyobj = 0;
  PyObject *self = 0;
  
  (void)self;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(egglib::OUTGROUP));
  return pyobj;
}


SWIGINTERN int Swig_var_MISSING_set(PyObject *) {
  SWIG_Error(SWIG_AttributeError,"Variable MISSING is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_MISSING_get(void) {
  PyObject *pyobj = 0;
  PyObject *self = 0;
  
  (void)self;
  pyobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(egglib::MISSING));
  return pyobj;
}


SWIGINTERN int Swig_var_BEFORE_set(PyObject *) {
  SWIG_Error(SWIG_AttributeError,"Variable BEFORE is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_BEFORE_get(void) {
  PyObject *pyobj = 0;
  PyObject *self = 0;
  
  (void)self;
  pyobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(egglib::BEFORE));
  return pyobj;
}


SWIGINTERN int Swig_var_FIRST_set(PyObject *) {
  SWIG_Error(SWIG_AttributeError,"Variable FIRST is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_FIRST_get(void) {
  PyObject *pyobj = 0;
  PyObject *self = 0;
  
  (void)self;
  pyobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(egglib::FIRST));
  return pyobj;
}


SWIGINTERN int Swig_var_LAST_set(PyObject *) {
  SWIG_Error(SWIG_AttributeError,"Variable LAST is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_LAST_get(void) {
  PyObject *pyobj = 0;
  PyObject *self = 0;
  
  (void)self;
  pyobj = SWIG_From_unsigned_SS_long(static_cast< unsigned long >(egglib::LAST));
  return pyobj;
}


SWIGINTERN PyObject *_wrap_get_static_DNAAlphabet(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::DNAAlphabet *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "get_static_DNAAlphabet", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::DNAAlphabet *) &egglib::get_static_DNAAlphabet();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__DNAAlphabet, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_get_static_CodonAlphabet(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::CodonAlphabet *result = 0 ;
  
  if (!SWIG_Python_UnpackTuple(args, "get_static_CodonAlphabet", 0, 0, 0)) SWIG_fail;
  {
    try {
      result = (egglib::CodonAlphabet *) &egglib::get_static_CodonAlphabet();
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_egglib__CodonAlphabet, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN int Swig_var_DEBUG_FLAG_set(PyObject *) {
  SWIG_Error(SWIG_AttributeError,"Variable DEBUG_FLAG is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_DEBUG_FLAG_get(void) {
  PyObject *pyobj = 0;
  PyObject *self = 0;
  
  (void)self;
  pyobj = SWIG_From_int(static_cast< int >(egglib::DEBUG_FLAG));
  return pyobj;
}


SWIGINTERN PyObject *_wrap_read_fasta_file(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = (char *) 0 ;
  bool arg2 ;
  egglib::DataHolder *arg3 = 0 ;
  egglib::FiniteAlphabet< char > *arg4 = 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "read_fasta_file", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "read_fasta_file" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = reinterpret_cast< char * >(buf1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "read_fasta_file" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_egglib__DataHolder,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "read_fasta_file" "', argument " "3"" of type '" "egglib::DataHolder &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "read_fasta_file" "', argument " "3"" of type '" "egglib::DataHolder &""'"); 
  }
  arg3 = reinterpret_cast< egglib::DataHolder * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_egglib__FiniteAlphabetT_char_t,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "read_fasta_file" "', argument " "4"" of type '" "egglib::FiniteAlphabet< char > &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "read_fasta_file" "', argument " "4"" of type '" "egglib::FiniteAlphabet< char > &""'"); 
  }
  arg4 = reinterpret_cast< egglib::FiniteAlphabet< char > * >(argp4);
  {
    try {
      egglib::read_fasta_file((char const *)arg1,arg2,*arg3,*arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) delete[] buf1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_read_fasta_string(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  std::string arg1 ;
  bool arg2 ;
  egglib::DataHolder *arg3 = 0 ;
  egglib::FiniteAlphabet< char > *arg4 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject *swig_obj[4] ;
  
  if (!SWIG_Python_UnpackTuple(args, "read_fasta_string", 4, 4, swig_obj)) SWIG_fail;
  {
    std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "read_fasta_string" "', argument " "1"" of type '" "std::string const""'"); 
    }
    arg1 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "read_fasta_string" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3, SWIGTYPE_p_egglib__DataHolder,  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "read_fasta_string" "', argument " "3"" of type '" "egglib::DataHolder &""'"); 
  }
  if (!argp3) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "read_fasta_string" "', argument " "3"" of type '" "egglib::DataHolder &""'"); 
  }
  arg3 = reinterpret_cast< egglib::DataHolder * >(argp3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4, SWIGTYPE_p_egglib__FiniteAlphabetT_char_t,  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "read_fasta_string" "', argument " "4"" of type '" "egglib::FiniteAlphabet< char > &""'"); 
  }
  if (!argp4) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "read_fasta_string" "', argument " "4"" of type '" "egglib::FiniteAlphabet< char > &""'"); 
  }
  arg4 = reinterpret_cast< egglib::FiniteAlphabet< char > * >(argp4);
  {
    try {
      egglib::read_fasta_string(arg1,arg2,*arg3,*arg4);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Fs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  unsigned int arg1 ;
  unsigned int arg2 ;
  double arg3 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  double result;
  
  if (!SWIG_Python_UnpackTuple(args, "Fs", 3, 3, swig_obj)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Fs" "', argument " "1"" of type '" "unsigned int""'");
  } 
  arg1 = static_cast< unsigned int >(val1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Fs" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Fs" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  {
    try {
      result = (double)egglib::Fs(arg1,arg2,arg3);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_make_vcf_index(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  egglib::VcfParser *arg1 = 0 ;
  char *arg2 = (char *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!SWIG_Python_UnpackTuple(args, "make_vcf_index", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1, SWIGTYPE_p_egglib__VcfParser,  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "make_vcf_index" "', argument " "1"" of type '" "egglib::VcfParser &""'"); 
  }
  if (!argp1) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "make_vcf_index" "', argument " "1"" of type '" "egglib::VcfParser &""'"); 
  }
  arg1 = reinterpret_cast< egglib::VcfParser * >(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "make_vcf_index" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = reinterpret_cast< char * >(buf2);
  {
    try {
      egglib::make_vcf_index(*arg1,(char const *)arg2);
    }
    catch (const egglib::EggArgumentValueError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggFormatError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggInvalidAlleleError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidCharacterError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggMemoryError& e) {
      SWIG_exception(SWIG_MemoryError, e.what());
    }
    catch (const egglib::EggOpenFileError& e) {
      SWIG_exception(SWIG_IOError, e.what());
    }
    catch (const egglib::EggRuntimeError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggIndexError& e) {
      SWIG_exception(SWIG_IndexError, e.what());
    }
    catch (const egglib::EggUnalignedError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggNonHierarchicalStructure& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggPloidyError& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidChromosomeIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggInvalidPositionIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggAlphabetError<int>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<char>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggAlphabetError<const char *>& e) {
      SWIG_exception(SWIG_ValueError, e.what());
    }
    catch (const egglib::EggInvalidLineIdxError& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const egglib::EggException& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
    catch (const std::exception& e) {
      SWIG_exception(SWIG_RuntimeError, e.what());
    }
  }
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) delete[] buf2;
  return NULL;
}


static PyMethodDef SwigMethods[] = {
	 { "SWIG_PyInstanceMethod_New", SWIG_PyInstanceMethod_New, METH_O, NULL},
	 { "GeneticCode_ndiff", _wrap_GeneticCode_ndiff, METH_VARARGS, "GeneticCode_ndiff(codon1, codon2) -> unsigned int"},
	 { "GeneticCode_diff1", _wrap_GeneticCode_diff1, METH_VARARGS, "GeneticCode_diff1(codon1, codon2) -> bool"},
	 { "GeneticCode_diff2", _wrap_GeneticCode_diff2, METH_VARARGS, "GeneticCode_diff2(codon1, codon2) -> bool"},
	 { "GeneticCode_diff3", _wrap_GeneticCode_diff3, METH_VARARGS, "GeneticCode_diff3(codon1, codon2) -> bool"},
	 { "GeneticCode_num_codes", _wrap_GeneticCode_num_codes, METH_NOARGS, "GeneticCode_num_codes() -> unsigned int"},
	 { "get_static_DNAAlphabet", _wrap_get_static_DNAAlphabet, METH_NOARGS, "get_static_DNAAlphabet() -> DNAAlphabet"},
	 { "get_static_CodonAlphabet", _wrap_get_static_CodonAlphabet, METH_NOARGS, "get_static_CodonAlphabet() -> CodonAlphabet"},
	 { "read_fasta_file", _wrap_read_fasta_file, METH_VARARGS, "read_fasta_file(fname, labels, dest, arg4)"},
	 { "read_fasta_string", _wrap_read_fasta_string, METH_VARARGS, "read_fasta_string(str, labels, dest, arg4)"},
	 { "Fs", _wrap_Fs, METH_VARARGS, "Fs(n, K, pi) -> double"},
	 { "make_vcf_index", _wrap_make_vcf_index, METH_VARARGS, "make_vcf_index(VCF, output)"},
	 { NULL, NULL, 0, NULL }
};

static PyMethodDef SwigMethods_proxydocs[] = {
	 { NULL, NULL, 0, NULL }
};

static SwigPyGetSet VectorInt___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__VectorInt_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"\n"
		"VectorInt\n"
		"VectorInt\n"
		"", &VectorInt___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__VectorInt_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__VectorInt_methods[] = {
  { "set_num_values", _wrap_VectorInt_set_num_values, METH_O, "set_num_values(self, n)" },
  { "get_num_values", _wrap_VectorInt_get_num_values, METH_NOARGS, "get_num_values(self) -> unsigned int" },
  { "set_item", _wrap_VectorInt_set_item, METH_VARARGS, "set_item(self, i, value)" },
  { "get_item", _wrap_VectorInt_get_item, METH_O, "get_item(self, i) -> int" },
  { "clear", _wrap_VectorInt_clear, METH_NOARGS, "clear(self)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__VectorInt_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.VectorInt",                   /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_VectorInt_destructor_closure,/* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__VectorInt_type.as_number,             /* tp_as_number */
    &SwigPyBuiltin__egglib__VectorInt_type.as_sequence,           /* tp_as_sequence */
    &SwigPyBuiltin__egglib__VectorInt_type.as_mapping,            /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__VectorInt_type.as_buffer,             /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::VectorInt",                      /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__VectorInt_richcompare,                 /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__VectorInt_methods, /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__VectorInt_getset,  /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_VectorInt,                      /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__VectorInt_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__VectorInt_type};

static SwigPyGetSet AbstractBaseAlphabet___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__AbstractBaseAlphabet_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"lock", &AbstractBaseAlphabet___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__AbstractBaseAlphabet_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__AbstractBaseAlphabet_methods[] = {
  { "lock", _wrap_AbstractBaseAlphabet_lock, METH_NOARGS, "lock(self)" },
  { "is_locked", _wrap_AbstractBaseAlphabet_is_locked, METH_NOARGS, "is_locked(self) -> bool" },
  { "set_name", _wrap_AbstractBaseAlphabet_set_name, METH_O, "set_name(self, name)" },
  { "get_name", _wrap_AbstractBaseAlphabet_get_name, METH_NOARGS, "get_name(self) -> char const *" },
  { "num_exploitable", _wrap_AbstractBaseAlphabet_num_exploitable, METH_NOARGS, "num_exploitable(self) -> unsigned int" },
  { "num_missing", _wrap_AbstractBaseAlphabet_num_missing, METH_NOARGS, "num_missing(self) -> unsigned int" },
  { "get_type", _wrap_AbstractBaseAlphabet_get_type, METH_NOARGS, "get_type(self) -> char const *" },
  { "set_type", _wrap_AbstractBaseAlphabet_set_type, METH_O, "set_type(self, arg2)" },
  { "case_insensitive", _wrap_AbstractBaseAlphabet_case_insensitive, METH_NOARGS, "case_insensitive(self) -> bool" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__AbstractBaseAlphabet_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.AbstractBaseAlphabet",        /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_AbstractBaseAlphabet_destructor_closure,         /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__AbstractBaseAlphabet_type.as_number,  /* tp_as_number */
    &SwigPyBuiltin__egglib__AbstractBaseAlphabet_type.as_sequence,/* tp_as_sequence */
    &SwigPyBuiltin__egglib__AbstractBaseAlphabet_type.as_mapping, /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__AbstractBaseAlphabet_type.as_buffer,  /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::AbstractBaseAlphabet",           /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__AbstractBaseAlphabet_richcompare,      /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__AbstractBaseAlphabet_methods,          /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__AbstractBaseAlphabet_getset,           /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    SwigPyBuiltin_BadInit,                    /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__AbstractBaseAlphabet_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__AbstractBaseAlphabet_type};

static SwigPyGetSet IntBaseAlphabet___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_int_t_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"get_value", &IntBaseAlphabet___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_int_t_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_int_t_methods[] = {
  { "get_value", _wrap_IntBaseAlphabet_get_value, METH_O, "get_value(self, code) -> int const" },
  { "get_code", _wrap_IntBaseAlphabet_get_code, METH_O, "get_code(self, value) -> int" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_int_t_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.IntBaseAlphabet",             /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_IntBaseAlphabet_destructor_closure,              /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_int_t_type.as_number,/* tp_as_number */
    &SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_int_t_type.as_sequence,/* tp_as_sequence */
    &SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_int_t_type.as_mapping,/* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_int_t_type.as_buffer,/* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::AbstractTemplateAlphabet< int >",/* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_int_t_richcompare,/* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_int_t_methods,/* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_int_t_getset,/* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    SwigPyBuiltin_BadInit,                    /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_int_t_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_int_t_type};

static SwigPyGetSet CharBaseAlphabet___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_t_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"get_value", &CharBaseAlphabet___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_t_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_t_methods[] = {
  { "get_value", _wrap_CharBaseAlphabet_get_value, METH_O, "get_value(self, code) -> char const" },
  { "get_code", _wrap_CharBaseAlphabet_get_code, METH_O, "get_code(self, value) -> int" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_t_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.CharBaseAlphabet",            /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_CharBaseAlphabet_destructor_closure,             /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_t_type.as_number,/* tp_as_number */
    &SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_t_type.as_sequence,/* tp_as_sequence */
    &SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_t_type.as_mapping,/* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_t_type.as_buffer,/* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::AbstractTemplateAlphabet< char >",                   /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_t_richcompare,/* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_t_methods,/* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_t_getset,/* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    SwigPyBuiltin_BadInit,                    /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_t_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_t_type};

static SwigPyGetSet CStringBaseAlphabet___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_p_t_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"get_value", &CStringBaseAlphabet___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_p_t_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_p_t_methods[] = {
  { "get_value", _wrap_CStringBaseAlphabet_get_value, METH_O, "get_value(self, code) -> char *const" },
  { "get_code", _wrap_CStringBaseAlphabet_get_code, METH_O, "get_code(self, value) -> int" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_p_t_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.CStringBaseAlphabet",         /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_CStringBaseAlphabet_destructor_closure,          /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_p_t_type.as_number,/* tp_as_number */
    &SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_p_t_type.as_sequence,/* tp_as_sequence */
    &SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_p_t_type.as_mapping,/* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_p_t_type.as_buffer,/* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::AbstractTemplateAlphabet< char * >",                 /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_p_t_richcompare,/* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_p_t_methods,/* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_p_t_getset,/* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    SwigPyBuiltin_BadInit,                    /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_p_t_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_p_t_type};

static SwigPyGetSet IntAlphabet___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__FiniteAlphabetT_int_t_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"FiniteAlphabet<(int)>", &IntAlphabet___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__FiniteAlphabetT_int_t_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__FiniteAlphabetT_int_t_methods[] = {
  { "get_value", _wrap_IntAlphabet_get_value, METH_O, "get_value(self, code) -> int const" },
  { "get_code", _wrap_IntAlphabet_get_code, METH_O, "get_code(self, value) -> int" },
  { "add_exploitable", _wrap_IntAlphabet_add_exploitable, METH_O, "add_exploitable(self, value)" },
  { "add_missing", _wrap_IntAlphabet_add_missing, METH_O, "add_missing(self, value)" },
  { "num_exploitable", _wrap_IntAlphabet_num_exploitable, METH_NOARGS, "num_exploitable(self) -> unsigned int" },
  { "num_missing", _wrap_IntAlphabet_num_missing, METH_NOARGS, "num_missing(self) -> unsigned int" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__FiniteAlphabetT_int_t_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.IntAlphabet",                 /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_IntAlphabet_destructor_closure,                  /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__FiniteAlphabetT_int_t_type.as_number, /* tp_as_number */
    &SwigPyBuiltin__egglib__FiniteAlphabetT_int_t_type.as_sequence,/* tp_as_sequence */
    &SwigPyBuiltin__egglib__FiniteAlphabetT_int_t_type.as_mapping,/* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__FiniteAlphabetT_int_t_type.as_buffer, /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::FiniteAlphabet< int >",          /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__FiniteAlphabetT_int_t_richcompare,     /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__FiniteAlphabetT_int_t_methods,         /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__FiniteAlphabetT_int_t_getset,          /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_IntAlphabet,                    /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__FiniteAlphabetT_int_t_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__FiniteAlphabetT_int_t_type};

static SwigPyGetSet CharAlphabet___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__FiniteAlphabetT_char_t_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"FiniteAlphabet<(char)>", &CharAlphabet___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__FiniteAlphabetT_char_t_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__FiniteAlphabetT_char_t_methods[] = {
  { "get_value", _wrap_CharAlphabet_get_value, METH_O, "get_value(self, code) -> char const" },
  { "get_code", _wrap_CharAlphabet_get_code, METH_O, "get_code(self, value) -> int" },
  { "add_exploitable", _wrap_CharAlphabet_add_exploitable, METH_O, "add_exploitable(self, value)" },
  { "add_missing", _wrap_CharAlphabet_add_missing, METH_O, "add_missing(self, value)" },
  { "num_exploitable", _wrap_CharAlphabet_num_exploitable, METH_NOARGS, "num_exploitable(self) -> unsigned int" },
  { "num_missing", _wrap_CharAlphabet_num_missing, METH_NOARGS, "num_missing(self) -> unsigned int" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__FiniteAlphabetT_char_t_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.CharAlphabet",                /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_CharAlphabet_destructor_closure,                 /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__FiniteAlphabetT_char_t_type.as_number,/* tp_as_number */
    &SwigPyBuiltin__egglib__FiniteAlphabetT_char_t_type.as_sequence,/* tp_as_sequence */
    &SwigPyBuiltin__egglib__FiniteAlphabetT_char_t_type.as_mapping,/* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__FiniteAlphabetT_char_t_type.as_buffer,/* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::FiniteAlphabet< char >",         /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__FiniteAlphabetT_char_t_richcompare,    /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__FiniteAlphabetT_char_t_methods,        /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__FiniteAlphabetT_char_t_getset,         /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_CharAlphabet,                   /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__FiniteAlphabetT_char_t_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__FiniteAlphabetT_char_t_type};

static SwigPyGetSet CStringAlphabet___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__FiniteAlphabetT_char_p_t_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"FiniteAlphabet<(p.char)>", &CStringAlphabet___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__FiniteAlphabetT_char_p_t_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__FiniteAlphabetT_char_p_t_methods[] = {
  { "get_value", _wrap_CStringAlphabet_get_value, METH_O, "get_value(self, code) -> char *const" },
  { "get_code", _wrap_CStringAlphabet_get_code, METH_O, "get_code(self, value) -> int" },
  { "add_exploitable", _wrap_CStringAlphabet_add_exploitable, METH_O, "add_exploitable(self, value)" },
  { "add_missing", _wrap_CStringAlphabet_add_missing, METH_O, "add_missing(self, value)" },
  { "num_exploitable", _wrap_CStringAlphabet_num_exploitable, METH_NOARGS, "num_exploitable(self) -> unsigned int" },
  { "num_missing", _wrap_CStringAlphabet_num_missing, METH_NOARGS, "num_missing(self) -> unsigned int" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__FiniteAlphabetT_char_p_t_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.CStringAlphabet",             /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_CStringAlphabet_destructor_closure,              /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__FiniteAlphabetT_char_p_t_type.as_number,/* tp_as_number */
    &SwigPyBuiltin__egglib__FiniteAlphabetT_char_p_t_type.as_sequence,/* tp_as_sequence */
    &SwigPyBuiltin__egglib__FiniteAlphabetT_char_p_t_type.as_mapping,/* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__FiniteAlphabetT_char_p_t_type.as_buffer,/* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::FiniteAlphabet< char * >",       /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__FiniteAlphabetT_char_p_t_richcompare,  /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__FiniteAlphabetT_char_p_t_methods,      /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__FiniteAlphabetT_char_p_t_getset,       /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_CStringAlphabet,                /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__FiniteAlphabetT_char_p_t_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__FiniteAlphabetT_char_p_t_type};

static SwigPyGetSet CaseInsensitiveCharAlphabet___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__CaseInsensitiveCharAlphabet_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"CaseInsensitiveCharAlphabet", &CaseInsensitiveCharAlphabet___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__CaseInsensitiveCharAlphabet_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__CaseInsensitiveCharAlphabet_methods[] = {
  { "add_exploitable", _wrap_CaseInsensitiveCharAlphabet_add_exploitable, METH_O, "add_exploitable(self, value)" },
  { "add_missing", _wrap_CaseInsensitiveCharAlphabet_add_missing, METH_O, "add_missing(self, value)" },
  { "case_insensitive", _wrap_CaseInsensitiveCharAlphabet_case_insensitive, METH_NOARGS, "case_insensitive(self) -> bool" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__CaseInsensitiveCharAlphabet_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.CaseInsensitiveCharAlphabet", /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_CaseInsensitiveCharAlphabet_destructor_closure,  /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__CaseInsensitiveCharAlphabet_type.as_number,/* tp_as_number */
    &SwigPyBuiltin__egglib__CaseInsensitiveCharAlphabet_type.as_sequence,/* tp_as_sequence */
    &SwigPyBuiltin__egglib__CaseInsensitiveCharAlphabet_type.as_mapping,/* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__CaseInsensitiveCharAlphabet_type.as_buffer,/* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::CaseInsensitiveCharAlphabet",    /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__CaseInsensitiveCharAlphabet_richcompare,/* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__CaseInsensitiveCharAlphabet_methods,   /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__CaseInsensitiveCharAlphabet_getset,    /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_CaseInsensitiveCharAlphabet,    /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__CaseInsensitiveCharAlphabet_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__CaseInsensitiveCharAlphabet_type};

static SwigPyGetSet DNAAlphabet___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__DNAAlphabet_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"DNAAlphabet", &DNAAlphabet___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__DNAAlphabet_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__DNAAlphabet_methods[] = {
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__DNAAlphabet_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.DNAAlphabet",                 /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_DNAAlphabet_destructor_closure,                  /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__DNAAlphabet_type.as_number,           /* tp_as_number */
    &SwigPyBuiltin__egglib__DNAAlphabet_type.as_sequence,         /* tp_as_sequence */
    &SwigPyBuiltin__egglib__DNAAlphabet_type.as_mapping,          /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__DNAAlphabet_type.as_buffer,           /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::DNAAlphabet",                    /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__DNAAlphabet_richcompare,               /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__DNAAlphabet_methods,                   /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__DNAAlphabet_getset,/* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_DNAAlphabet,                    /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__DNAAlphabet_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__DNAAlphabet_type};

static SwigPyGetSet StringAlphabet___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__StringAlphabet_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"StringAlphabet", &StringAlphabet___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__StringAlphabet_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__StringAlphabet_methods[] = {
  { "get_code", _wrap_StringAlphabet_get_code, METH_O, "get_code(self, value) -> int" },
  { "add_exploitable", _wrap_StringAlphabet_add_exploitable, METH_O, "add_exploitable(self, value)" },
  { "add_missing", _wrap_StringAlphabet_add_missing, METH_O, "add_missing(self, value)" },
  { "longest_length", _wrap_StringAlphabet_longest_length, METH_NOARGS, "longest_length(self) -> unsigned int" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__StringAlphabet_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.StringAlphabet",              /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_StringAlphabet_destructor_closure,               /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__StringAlphabet_type.as_number,        /* tp_as_number */
    &SwigPyBuiltin__egglib__StringAlphabet_type.as_sequence,      /* tp_as_sequence */
    &SwigPyBuiltin__egglib__StringAlphabet_type.as_mapping,       /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__StringAlphabet_type.as_buffer,        /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::StringAlphabet",                 /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__StringAlphabet_richcompare,            /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__StringAlphabet_methods,                /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__StringAlphabet_getset,                 /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_StringAlphabet,                 /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__StringAlphabet_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__StringAlphabet_type};

static SwigPyGetSet CodonAlphabet___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__CodonAlphabet_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"CodonAlphabet", &CodonAlphabet___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__CodonAlphabet_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__CodonAlphabet_methods[] = {
  { "get_code", _wrap_CodonAlphabet_get_code, METH_O, "get_code(self, value) -> int" },
  { "get_code_from_bases", _wrap_CodonAlphabet_get_code_from_bases, METH_VARARGS, "get_code_from_bases(self, arg2, arg3, arg4) -> int" },
  { "num_exploitable", _wrap_CodonAlphabet_num_exploitable, METH_NOARGS, "num_exploitable(self) -> unsigned int" },
  { "num_missing", _wrap_CodonAlphabet_num_missing, METH_NOARGS, "num_missing(self) -> unsigned int" },
  { "add_exploitable", _wrap_CodonAlphabet_add_exploitable, METH_O, "add_exploitable(self, value)" },
  { "add_missing", _wrap_CodonAlphabet_add_missing, METH_O, "add_missing(self, value)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__CodonAlphabet_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.CodonAlphabet",               /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_CodonAlphabet_destructor_closure,                /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__CodonAlphabet_type.as_number,         /* tp_as_number */
    &SwigPyBuiltin__egglib__CodonAlphabet_type.as_sequence,       /* tp_as_sequence */
    &SwigPyBuiltin__egglib__CodonAlphabet_type.as_mapping,        /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__CodonAlphabet_type.as_buffer,         /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::CodonAlphabet",                  /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__CodonAlphabet_richcompare,             /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__CodonAlphabet_methods,                 /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__CodonAlphabet_getset,                  /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_CodonAlphabet,                  /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__CodonAlphabet_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__CodonAlphabet_type};

static SwigPyGetSet CaseInsensitiveStringAlphabet___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__CaseInsensitiveStringAlphabet_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"CaseInsensitiveStringAlphabet", &CaseInsensitiveStringAlphabet___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__CaseInsensitiveStringAlphabet_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__CaseInsensitiveStringAlphabet_methods[] = {
  { "case_insensitive", _wrap_CaseInsensitiveStringAlphabet_case_insensitive, METH_NOARGS, "case_insensitive(self) -> bool" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__CaseInsensitiveStringAlphabet_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.CaseInsensitiveStringAlphabet",                   /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_CaseInsensitiveStringAlphabet_destructor_closure,/* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__CaseInsensitiveStringAlphabet_type.as_number,/* tp_as_number */
    &SwigPyBuiltin__egglib__CaseInsensitiveStringAlphabet_type.as_sequence,/* tp_as_sequence */
    &SwigPyBuiltin__egglib__CaseInsensitiveStringAlphabet_type.as_mapping,/* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__CaseInsensitiveStringAlphabet_type.as_buffer,/* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::CaseInsensitiveStringAlphabet",  /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__CaseInsensitiveStringAlphabet_richcompare,/* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__CaseInsensitiveStringAlphabet_methods, /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__CaseInsensitiveStringAlphabet_getset,  /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_CaseInsensitiveStringAlphabet,  /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__CaseInsensitiveStringAlphabet_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__CaseInsensitiveStringAlphabet_type};

static SwigPyGetSet CustomStringAlphabet___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__CustomStringAlphabet_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"CustomStringAlphabet", &CustomStringAlphabet___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__CustomStringAlphabet_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__CustomStringAlphabet_methods[] = {
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__CustomStringAlphabet_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.CustomStringAlphabet",        /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_CustomStringAlphabet_destructor_closure,         /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__CustomStringAlphabet_type.as_number,  /* tp_as_number */
    &SwigPyBuiltin__egglib__CustomStringAlphabet_type.as_sequence,/* tp_as_sequence */
    &SwigPyBuiltin__egglib__CustomStringAlphabet_type.as_mapping, /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__CustomStringAlphabet_type.as_buffer,  /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::CustomStringAlphabet",           /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__CustomStringAlphabet_richcompare,      /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__CustomStringAlphabet_methods,          /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__CustomStringAlphabet_getset,           /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_CustomStringAlphabet,           /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__CustomStringAlphabet_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__CustomStringAlphabet_type};

static SwigPyGetSet RangeAlphabet___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__RangeAlphabet_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"RangeAlphabet", &RangeAlphabet___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__RangeAlphabet_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__RangeAlphabet_methods[] = {
  { "get_code", _wrap_RangeAlphabet_get_code, METH_O, "get_code(self, arg2) -> int" },
  { "get_value", _wrap_RangeAlphabet_get_value, METH_O, "get_value(self, code) -> int const" },
  { "first_exploitable", _wrap_RangeAlphabet_first_exploitable, METH_NOARGS, "first_exploitable(self) -> int" },
  { "end_exploitable", _wrap_RangeAlphabet_end_exploitable, METH_NOARGS, "end_exploitable(self) -> int" },
  { "first_missing", _wrap_RangeAlphabet_first_missing, METH_NOARGS, "first_missing(self) -> int" },
  { "end_missing", _wrap_RangeAlphabet_end_missing, METH_NOARGS, "end_missing(self) -> int" },
  { "set_exploitable", _wrap_RangeAlphabet_set_exploitable, METH_VARARGS, "set_exploitable(self, arg2, arg3)" },
  { "set_missing", _wrap_RangeAlphabet_set_missing, METH_VARARGS, "set_missing(self, arg2, arg3)" },
  { "min_value", _wrap_RangeAlphabet_min_value, METH_NOARGS, "min_value(self) -> int" },
  { "max_value", _wrap_RangeAlphabet_max_value, METH_NOARGS, "max_value(self) -> int" },
  { "num_exploitable", _wrap_RangeAlphabet_num_exploitable, METH_NOARGS, "num_exploitable(self) -> unsigned int" },
  { "num_missing", _wrap_RangeAlphabet_num_missing, METH_NOARGS, "num_missing(self) -> unsigned int" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__RangeAlphabet_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.RangeAlphabet",               /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_RangeAlphabet_destructor_closure,                /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__RangeAlphabet_type.as_number,         /* tp_as_number */
    &SwigPyBuiltin__egglib__RangeAlphabet_type.as_sequence,       /* tp_as_sequence */
    &SwigPyBuiltin__egglib__RangeAlphabet_type.as_mapping,        /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__RangeAlphabet_type.as_buffer,         /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::RangeAlphabet",                  /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__RangeAlphabet_richcompare,             /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__RangeAlphabet_methods,                 /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__RangeAlphabet_getset,                  /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_RangeAlphabet,                  /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__RangeAlphabet_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__RangeAlphabet_type};

static SwigPyGetSet DataHolder___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__DataHolder_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"\n"
		"DataHolder\n"
		"DataHolder\n"
		"", &DataHolder___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__DataHolder_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__DataHolder_methods[] = {
  { "valid_phyml_names", _wrap_DataHolder_valid_phyml_names, METH_NOARGS, "valid_phyml_names(self) -> bool" },
  { "set_is_matrix", _wrap_DataHolder_set_is_matrix, METH_O, "set_is_matrix(self, flag)" },
  { "get_is_matrix", _wrap_DataHolder_get_is_matrix, METH_NOARGS, "get_is_matrix(self) -> bool" },
  { "reserve", _wrap_DataHolder_reserve, METH_VARARGS, "reserve(self, ns, ln, ng, ls)" },
  { "get_nsam", _wrap_DataHolder_get_nsam, METH_NOARGS, "get_nsam(self) -> unsigned int" },
  { "set_nsam", _wrap_DataHolder_set_nsam, METH_O, "set_nsam(self, nsam)" },
  { "get_nsit_all", _wrap_DataHolder_get_nsit_all, METH_NOARGS, "get_nsit_all(self) -> unsigned int" },
  { "get_nsit_sample", _wrap_DataHolder_get_nsit_sample, METH_O, "get_nsit_sample(self, sam) -> unsigned int" },
  { "set_nsit_all", _wrap_DataHolder_set_nsit_all, METH_O, "set_nsit_all(self, nsit)" },
  { "set_nsit_sample", _wrap_DataHolder_set_nsit_sample, METH_VARARGS, "set_nsit_sample(self, sam, nsit)" },
  { "insert_sites_all", _wrap_DataHolder_insert_sites_all, METH_VARARGS, "insert_sites_all(self, pos, num)" },
  { "insert_sites_sample", _wrap_DataHolder_insert_sites_sample, METH_VARARGS, "insert_sites_sample(self, sam, pos, num)" },
  { "get_nlabels", _wrap_DataHolder_get_nlabels, METH_O, "get_nlabels(self, arg2) -> unsigned int" },
  { "set_nlabels", _wrap_DataHolder_set_nlabels, METH_VARARGS, "set_nlabels(self, arg2, arg3)" },
  { "get_sample", _wrap_DataHolder_get_sample, METH_VARARGS, "get_sample(self, sam, sit) -> int" },
  { "set_sample", _wrap_DataHolder_set_sample, METH_VARARGS, "set_sample(self, sam, sit, value)" },
  { "get_label", _wrap_DataHolder_get_label, METH_VARARGS, "get_label(self, sam, lvl) -> char const *" },
  { "set_label", _wrap_DataHolder_set_label, METH_VARARGS, "set_label(self, sam, lvl, label)" },
  { "add_label", _wrap_DataHolder_add_label, METH_VARARGS, "add_label(self, sam, label)" },
  { "get_name", _wrap_DataHolder_get_name, METH_O, "get_name(self, sam) -> char const *" },
  { "set_name", _wrap_DataHolder_set_name, METH_VARARGS, "set_name(self, sam, name)" },
  { "name_appendch", _wrap_DataHolder_name_appendch, METH_VARARGS, "name_appendch(self, sam, ch)" },
  { "name_append", _wrap_DataHolder_name_append, METH_VARARGS, "name_append(self, sam, ch)" },
  { "del_sample", _wrap_DataHolder_del_sample, METH_O, "del_sample(self, sam)" },
  { "del_sites_all", _wrap_DataHolder_del_sites_all, METH_VARARGS, "del_sites_all(self, start, stop)" },
  { "del_sites_sample", _wrap_DataHolder_del_sites_sample, METH_VARARGS, "del_sites_sample(self, sam, start, stop)" },
  { "reset", _wrap_DataHolder_reset, METH_O, "reset(self, is_matrix)" },
  { "clear", _wrap_DataHolder_clear, METH_O, "clear(self, is_matrix)" },
  { "change_case", _wrap_DataHolder_change_case, METH_VARARGS, "change_case(self, lower, index_i, start_i, stop_i, alph)" },
  { "find", _wrap_DataHolder_find, METH_VARARGS, "find(self, sam, values, start=0, stop=MAX) -> unsigned int" },
  { "is_equal", _wrap_DataHolder_is_equal, METH_NOARGS, "is_equal(self) -> bool" },
  { "strip_clear", _wrap_DataHolder_strip_clear, METH_NOARGS, "strip_clear(self)" },
  { "strip_add", _wrap_DataHolder_strip_add, METH_O, "strip_add(self, arg2)" },
  { "strip", _wrap_DataHolder_strip, METH_VARARGS, "strip(self, index, left, right)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__DataHolder_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.DataHolder",                  /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_DataHolder_destructor_closure,                   /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__DataHolder_type.as_number,            /* tp_as_number */
    &SwigPyBuiltin__egglib__DataHolder_type.as_sequence,          /* tp_as_sequence */
    &SwigPyBuiltin__egglib__DataHolder_type.as_mapping,           /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__DataHolder_type.as_buffer,            /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::DataHolder",                     /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__DataHolder_richcompare,                /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__DataHolder_methods,/* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__DataHolder_getset, /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_DataHolder,                     /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__DataHolder_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__DataHolder_type};

static SwigPyGetSet IntersperseAlign___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__IntersperseAlign_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"IntersperseAlign", &IntersperseAlign___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__IntersperseAlign_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__IntersperseAlign_methods[] = {
  { "load", _wrap_IntersperseAlign_load, METH_O, "load(self, data)" },
  { "set_length", _wrap_IntersperseAlign_set_length, METH_O, "set_length(self, length)" },
  { "set_position", _wrap_IntersperseAlign_set_position, METH_VARARGS, "set_position(self, index, position)" },
  { "set_round_position", _wrap_IntersperseAlign_set_round_position, METH_VARARGS, "set_round_position(self, index, position)" },
  { "get_positions", _wrap_IntersperseAlign_get_positions, METH_O, "get_positions(self, coalesce)" },
  { "set_num_alleles", _wrap_IntersperseAlign_set_num_alleles, METH_O, "set_num_alleles(self, num)" },
  { "set_allele", _wrap_IntersperseAlign_set_allele, METH_VARARGS, "set_allele(self, index, allele)" },
  { "intersperse", _wrap_IntersperseAlign_intersperse, METH_VARARGS, "intersperse(self, round_positions=True)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__IntersperseAlign_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.IntersperseAlign",            /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_IntersperseAlign_destructor_closure,             /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__IntersperseAlign_type.as_number,      /* tp_as_number */
    &SwigPyBuiltin__egglib__IntersperseAlign_type.as_sequence,    /* tp_as_sequence */
    &SwigPyBuiltin__egglib__IntersperseAlign_type.as_mapping,     /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__IntersperseAlign_type.as_buffer,      /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::IntersperseAlign",               /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__IntersperseAlign_richcompare,          /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__IntersperseAlign_methods,              /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__IntersperseAlign_getset,               /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_IntersperseAlign,               /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__IntersperseAlign_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__IntersperseAlign_type};

static SwigPyGetSet StructureHolder___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__StructureHolder_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"StructureHolder", &StructureHolder___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__StructureHolder_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__StructureHolder_methods[] = {
  { "reset", _wrap_StructureHolder_reset, METH_NOARGS, "reset(self)" },
  { "check_ploidy", _wrap_StructureHolder_check_ploidy, METH_VARARGS, "check_ploidy(self, value=UNKNOWN)" },
  { "get_ploidy", _wrap_StructureHolder_get_ploidy, METH_NOARGS, "get_ploidy(self) -> unsigned int" },
  { "get_structure", _wrap_StructureHolder_get_structure, METH_VARARGS, "get_structure(self, data, lvl_clust, lvl_pop, lvl_indiv, ploidy, skip_outgroup, outgroup_label)" },
  { "mk_dummy_structure", _wrap_StructureHolder_mk_dummy_structure, METH_VARARGS, "mk_dummy_structure(self, arg2, arg3)" },
  { "dummy_add_pop", _wrap_StructureHolder_dummy_add_pop, METH_O, "dummy_add_pop(self, arg2)" },
  { "process_ingroup", _wrap_StructureHolder_process_ingroup, METH_VARARGS, "process_ingroup(self, idx, lbl_clust, lbl_pop, lbl_indiv)" },
  { "process_outgroup", _wrap_StructureHolder_process_outgroup, METH_VARARGS, "process_outgroup(self, idx, lbl_indiv)" },
  { "add_cluster", _wrap_StructureHolder_add_cluster, METH_O, "add_cluster(self, label) -> StructureCluster" },
  { "add_population", _wrap_StructureHolder_add_population, METH_VARARGS, "add_population(self, label, cluster) -> StructurePopulation" },
  { "add_individual_ingroup", _wrap_StructureHolder_add_individual_ingroup, METH_VARARGS, "add_individual_ingroup(self, label, population) -> StructureIndiv" },
  { "add_individual_outgroup", _wrap_StructureHolder_add_individual_outgroup, METH_O, "add_individual_outgroup(self, label) -> StructureIndiv" },
  { "add_sample_ingroup", _wrap_StructureHolder_add_sample_ingroup, METH_VARARGS, "add_sample_ingroup(self, sam_idx, indiv)" },
  { "add_sample_outgroup", _wrap_StructureHolder_add_sample_outgroup, METH_VARARGS, "add_sample_outgroup(self, sam_idx, indiv)" },
  { "num_clust", _wrap_StructureHolder_num_clust, METH_NOARGS, "num_clust(self) -> unsigned int" },
  { "num_pop", _wrap_StructureHolder_num_pop, METH_NOARGS, "num_pop(self) -> unsigned int" },
  { "num_indiv_ingroup", _wrap_StructureHolder_num_indiv_ingroup, METH_NOARGS, "num_indiv_ingroup(self) -> unsigned int" },
  { "num_indiv_outgroup", _wrap_StructureHolder_num_indiv_outgroup, METH_NOARGS, "num_indiv_outgroup(self) -> unsigned int" },
  { "get_cluster", _wrap_StructureHolder_get_cluster, METH_O, "get_cluster(self, idx) -> StructureCluster" },
  { "get_population", _wrap_StructureHolder_get_population, METH_O, "get_population(self, idx) -> StructurePopulation" },
  { "get_indiv_ingroup", _wrap_StructureHolder_get_indiv_ingroup, METH_O, "get_indiv_ingroup(self, idx) -> StructureIndiv" },
  { "get_indiv_outgroup", _wrap_StructureHolder_get_indiv_outgroup, METH_O, "get_indiv_outgroup(self, idx) -> StructureIndiv" },
  { "get_ni", _wrap_StructureHolder_get_ni, METH_NOARGS, "get_ni(self) -> unsigned int" },
  { "get_no", _wrap_StructureHolder_get_no, METH_NOARGS, "get_no(self) -> unsigned int" },
  { "get_req", _wrap_StructureHolder_get_req, METH_NOARGS, "get_req(self) -> unsigned int" },
  { "copy", _wrap_StructureHolder_copy, METH_O, "copy(self, source)" },
  { "subset", _wrap_StructureHolder_subset, METH_VARARGS, "subset(self, source, popstring, cluststring, outgroup) -> char const *" },
  { "shuffle_init", _wrap_StructureHolder_shuffle_init, METH_O, "shuffle_init(self, mode)" },
  { "shuffle", _wrap_StructureHolder_shuffle, METH_NOARGS, "shuffle(self)" },
  { "shuffle_restore", _wrap_StructureHolder_shuffle_restore, METH_NOARGS, "shuffle_restore(self)" },
  { "init_i", _wrap_StructureHolder_init_i, METH_NOARGS, "init_i(self) -> unsigned int" },
  { "init_o", _wrap_StructureHolder_init_o, METH_NOARGS, "init_o(self) -> unsigned int" },
  { "iter_i", _wrap_StructureHolder_iter_i, METH_NOARGS, "iter_i(self) -> unsigned int" },
  { "iter_o", _wrap_StructureHolder_iter_o, METH_NOARGS, "iter_o(self) -> unsigned int" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__StructureHolder_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.StructureHolder",             /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_StructureHolder_destructor_closure,              /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__StructureHolder_type.as_number,       /* tp_as_number */
    &SwigPyBuiltin__egglib__StructureHolder_type.as_sequence,     /* tp_as_sequence */
    &SwigPyBuiltin__egglib__StructureHolder_type.as_mapping,      /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__StructureHolder_type.as_buffer,       /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::StructureHolder",                /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__StructureHolder_richcompare,           /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__StructureHolder_methods,               /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__StructureHolder_getset,                /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_StructureHolder,                /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__StructureHolder_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__StructureHolder_type};

static SwigPyGetSet StructureCluster___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__StructureCluster_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"StructureCluster", &StructureCluster___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__StructureCluster_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__StructureCluster_methods[] = {
  { "reset", _wrap_StructureCluster_reset, METH_O, "reset(self, label)" },
  { "add_pop", _wrap_StructureCluster_add_pop, METH_O, "add_pop(self, label) -> StructurePopulation" },
  { "num_pop", _wrap_StructureCluster_num_pop, METH_NOARGS, "num_pop(self) -> unsigned int" },
  { "get_population", _wrap_StructureCluster_get_population, METH_O, "get_population(self, idx) -> StructurePopulation" },
  { "get_label", _wrap_StructureCluster_get_label, METH_NOARGS, "get_label(self) -> char const *" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__StructureCluster_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.StructureCluster",            /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_StructureCluster_destructor_closure,             /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__StructureCluster_type.as_number,      /* tp_as_number */
    &SwigPyBuiltin__egglib__StructureCluster_type.as_sequence,    /* tp_as_sequence */
    &SwigPyBuiltin__egglib__StructureCluster_type.as_mapping,     /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__StructureCluster_type.as_buffer,      /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::StructureCluster",               /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__StructureCluster_richcompare,          /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__StructureCluster_methods,              /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__StructureCluster_getset,               /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_StructureCluster,               /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__StructureCluster_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__StructureCluster_type};

static SwigPyGetSet StructurePopulation___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__StructurePopulation_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"StructurePopulation", &StructurePopulation___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__StructurePopulation_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__StructurePopulation_methods[] = {
  { "reset", _wrap_StructurePopulation_reset, METH_O, "reset(self, label)" },
  { "add_indiv", _wrap_StructurePopulation_add_indiv, METH_O, "add_indiv(self, label) -> StructureIndiv" },
  { "num_indiv", _wrap_StructurePopulation_num_indiv, METH_NOARGS, "num_indiv(self) -> unsigned int" },
  { "get_indiv", _wrap_StructurePopulation_get_indiv, METH_O, "get_indiv(self, idx) -> StructureIndiv" },
  { "get_label", _wrap_StructurePopulation_get_label, METH_NOARGS, "get_label(self) -> char const *" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__StructurePopulation_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.StructurePopulation",         /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_StructurePopulation_destructor_closure,          /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__StructurePopulation_type.as_number,   /* tp_as_number */
    &SwigPyBuiltin__egglib__StructurePopulation_type.as_sequence, /* tp_as_sequence */
    &SwigPyBuiltin__egglib__StructurePopulation_type.as_mapping,  /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__StructurePopulation_type.as_buffer,   /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::StructurePopulation",            /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__StructurePopulation_richcompare,       /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__StructurePopulation_methods,           /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__StructurePopulation_getset,            /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_StructurePopulation,            /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__StructurePopulation_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__StructurePopulation_type};

static SwigPyGetSet StructureIndiv___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__StructureIndiv_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"StructureIndiv", &StructureIndiv___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__StructureIndiv_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__StructureIndiv_methods[] = {
  { "reset", _wrap_StructureIndiv_reset, METH_O, "reset(self, label)" },
  { "num_samples", _wrap_StructureIndiv_num_samples, METH_NOARGS, "num_samples(self) -> unsigned int" },
  { "add_sample", _wrap_StructureIndiv_add_sample, METH_O, "add_sample(self, index)" },
  { "get_sample", _wrap_StructureIndiv_get_sample, METH_O, "get_sample(self, idx) -> unsigned int" },
  { "get_label", _wrap_StructureIndiv_get_label, METH_NOARGS, "get_label(self) -> char const *" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__StructureIndiv_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.StructureIndiv",              /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_StructureIndiv_destructor_closure,               /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__StructureIndiv_type.as_number,        /* tp_as_number */
    &SwigPyBuiltin__egglib__StructureIndiv_type.as_sequence,      /* tp_as_sequence */
    &SwigPyBuiltin__egglib__StructureIndiv_type.as_mapping,       /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__StructureIndiv_type.as_buffer,        /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::StructureIndiv",                 /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__StructureIndiv_richcompare,            /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__StructureIndiv_methods,                /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__StructureIndiv_getset,                 /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_StructureIndiv,                 /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__StructureIndiv_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__StructureIndiv_type};

static SwigPyGetSet FreqSet___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__FreqSet_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"FreqSet", &FreqSet___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__FreqSet_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__FreqSet_methods[] = {
  { "setup", _wrap_FreqSet_setup, METH_NOARGS, "setup(self)" },
  { "set_nall", _wrap_FreqSet_set_nall, METH_O, "set_nall(self, arg2)" },
  { "add_genotypes", _wrap_FreqSet_add_genotypes, METH_O, "add_genotypes(self, num)" },
  { "incr_allele", _wrap_FreqSet_incr_allele, METH_VARARGS, "incr_allele(self, all_idx, num)" },
  { "incr_genotype", _wrap_FreqSet_incr_genotype, METH_VARARGS, "incr_genotype(self, gen_idx, num)" },
  { "tell_het", _wrap_FreqSet_tell_het, METH_VARARGS, "tell_het(self, i, a)" },
  { "num_alleles", _wrap_FreqSet_num_alleles, METH_NOARGS, "num_alleles(self) -> unsigned int" },
  { "num_alleles_eff", _wrap_FreqSet_num_alleles_eff, METH_NOARGS, "num_alleles_eff(self) -> unsigned int" },
  { "num_genotypes", _wrap_FreqSet_num_genotypes, METH_NOARGS, "num_genotypes(self) -> unsigned int" },
  { "num_genotypes_eff", _wrap_FreqSet_num_genotypes_eff, METH_NOARGS, "num_genotypes_eff(self) -> unsigned int" },
  { "nseff", _wrap_FreqSet_nseff, METH_NOARGS, "nseff(self) -> unsigned int" },
  { "nieff", _wrap_FreqSet_nieff, METH_NOARGS, "nieff(self) -> unsigned int" },
  { "frq_all", _wrap_FreqSet_frq_all, METH_O, "frq_all(self, arg2) -> unsigned int" },
  { "frq_gen", _wrap_FreqSet_frq_gen, METH_O, "frq_gen(self, arg2) -> unsigned int" },
  { "frq_het", _wrap_FreqSet_frq_het, METH_O, "frq_het(self, arg2) -> unsigned int" },
  { "tot_het", _wrap_FreqSet_tot_het, METH_NOARGS, "tot_het(self) -> unsigned int" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__FreqSet_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.FreqSet",                     /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_FreqSet_destructor_closure,  /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__FreqSet_type.as_number,               /* tp_as_number */
    &SwigPyBuiltin__egglib__FreqSet_type.as_sequence,             /* tp_as_sequence */
    &SwigPyBuiltin__egglib__FreqSet_type.as_mapping,              /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__FreqSet_type.as_buffer,               /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::FreqSet",                        /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__FreqSet_richcompare,                   /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__FreqSet_methods,   /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__FreqSet_getset,    /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_FreqSet,                        /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__FreqSet_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__FreqSet_type};

static SwigPyGetSet FreqHolder___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__FreqHolder_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"FreqHolder", &FreqHolder___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__FreqHolder_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__FreqHolder_methods[] = {
  { "setup_structure", _wrap_FreqHolder_setup_structure, METH_O, "setup_structure(self, structure)" },
  { "setup_raw", _wrap_FreqHolder_setup_raw, METH_VARARGS, "setup_raw(self, nc, np, ploidy)" },
  { "setup_pop", _wrap_FreqHolder_setup_pop, METH_VARARGS, "setup_pop(self, i, clu_idx, rel_idx, ns)" },
  { "set_ngeno", _wrap_FreqHolder_set_ngeno, METH_O, "set_ngeno(self, ng)" },
  { "process_site", _wrap_FreqHolder_process_site, METH_O, "process_site(self, site)" },
  { "process_vcf", _wrap_FreqHolder_process_vcf, METH_O, "process_vcf(self, vcf)" },
  { "frq_ingroup", _wrap_FreqHolder_frq_ingroup, METH_NOARGS, "frq_ingroup(self) -> FreqSet" },
  { "frq_outgroup", _wrap_FreqHolder_frq_outgroup, METH_NOARGS, "frq_outgroup(self) -> FreqSet" },
  { "num_clusters", _wrap_FreqHolder_num_clusters, METH_NOARGS, "num_clusters(self) -> unsigned int" },
  { "num_populations", _wrap_FreqHolder_num_populations, METH_NOARGS, "num_populations(self) -> unsigned int" },
  { "frq_cluster", _wrap_FreqHolder_frq_cluster, METH_O, "frq_cluster(self, arg2) -> FreqSet" },
  { "frq_population", _wrap_FreqHolder_frq_population, METH_O, "frq_population(self, arg2) -> FreqSet" },
  { "num_alleles", _wrap_FreqHolder_num_alleles, METH_NOARGS, "num_alleles(self) -> unsigned int" },
  { "allele", _wrap_FreqHolder_allele, METH_O, "allele(self, arg2) -> int" },
  { "get_allele_index", _wrap_FreqHolder_get_allele_index, METH_O, "get_allele_index(self, arg2) -> unsigned int" },
  { "ploidy", _wrap_FreqHolder_ploidy, METH_NOARGS, "ploidy(self) -> unsigned int" },
  { "num_genotypes", _wrap_FreqHolder_num_genotypes, METH_NOARGS, "num_genotypes(self) -> unsigned int" },
  { "genotype_het", _wrap_FreqHolder_genotype_het, METH_O, "genotype_het(self, arg2) -> bool" },
  { "genotype", _wrap_FreqHolder_genotype, METH_O, "genotype(self, arg2) -> int const *" },
  { "genotype_item", _wrap_FreqHolder_genotype_item, METH_VARARGS, "genotype_item(self, arg2, arg3) -> int" },
  { "set_genotype_item", _wrap_FreqHolder_set_genotype_item, METH_VARARGS, "set_genotype_item(self, i, j, a)" },
  { "find_allele", _wrap_FreqHolder_find_allele, METH_O, "find_allele(self, arg2) -> unsigned int" },
  { "cluster_index", _wrap_FreqHolder_cluster_index, METH_O, "cluster_index(self, arg2) -> unsigned int" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__FreqHolder_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.FreqHolder",                  /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_FreqHolder_destructor_closure,                   /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__FreqHolder_type.as_number,            /* tp_as_number */
    &SwigPyBuiltin__egglib__FreqHolder_type.as_sequence,          /* tp_as_sequence */
    &SwigPyBuiltin__egglib__FreqHolder_type.as_mapping,           /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__FreqHolder_type.as_buffer,            /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::FreqHolder",                     /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__FreqHolder_richcompare,                /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__FreqHolder_methods,/* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__FreqHolder_getset, /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_FreqHolder,                     /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__FreqHolder_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__FreqHolder_type};

static SwigPyGetSet AlleleStatus___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__AlleleStatus_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"AlleleStatus", &AlleleStatus___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__AlleleStatus_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__AlleleStatus_methods[] = {
  { "reset", _wrap_AlleleStatus_reset, METH_NOARGS, "reset(self)" },
  { "process", _wrap_AlleleStatus_process, METH_O, "process(self, freqs)" },
  { "total", _wrap_AlleleStatus_total, METH_NOARGS, "total(self)" },
  { "nsites", _wrap_AlleleStatus_nsites, METH_NOARGS, "nsites(self) -> unsigned int" },
  { "nsites_o", _wrap_AlleleStatus_nsites_o, METH_NOARGS, "nsites_o(self) -> unsigned int" },
  { "Sp", _wrap_AlleleStatus_Sp, METH_NOARGS, "Sp(self) -> unsigned int" },
  { "Sp_T1", _wrap_AlleleStatus_Sp_T1, METH_NOARGS, "Sp_T1(self) -> unsigned int" },
  { "Spd", _wrap_AlleleStatus_Spd, METH_NOARGS, "Spd(self) -> unsigned int" },
  { "Spd_T1", _wrap_AlleleStatus_Spd_T1, METH_NOARGS, "Spd_T1(self) -> unsigned int" },
  { "ShP", _wrap_AlleleStatus_ShP, METH_NOARGS, "ShP(self) -> unsigned int" },
  { "ShP_T1", _wrap_AlleleStatus_ShP_T1, METH_NOARGS, "ShP_T1(self) -> unsigned int" },
  { "ShA", _wrap_AlleleStatus_ShA, METH_NOARGS, "ShA(self) -> unsigned int" },
  { "ShA_T1", _wrap_AlleleStatus_ShA_T1, METH_NOARGS, "ShA_T1(self) -> unsigned int" },
  { "FxA", _wrap_AlleleStatus_FxA, METH_NOARGS, "FxA(self) -> unsigned int" },
  { "FxA_T1", _wrap_AlleleStatus_FxA_T1, METH_NOARGS, "FxA_T1(self) -> unsigned int" },
  { "FxD", _wrap_AlleleStatus_FxD, METH_NOARGS, "FxD(self) -> unsigned int" },
  { "FxD_T1", _wrap_AlleleStatus_FxD_T1, METH_NOARGS, "FxD_T1(self) -> unsigned int" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__AlleleStatus_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.AlleleStatus",                /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_AlleleStatus_destructor_closure,                 /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__AlleleStatus_type.as_number,          /* tp_as_number */
    &SwigPyBuiltin__egglib__AlleleStatus_type.as_sequence,        /* tp_as_sequence */
    &SwigPyBuiltin__egglib__AlleleStatus_type.as_mapping,         /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__AlleleStatus_type.as_buffer,          /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::AlleleStatus",                   /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__AlleleStatus_richcompare,              /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__AlleleStatus_methods,                  /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__AlleleStatus_getset,                   /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_AlleleStatus,                   /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__AlleleStatus_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__AlleleStatus_type};

static SwigPyGetSet GeneticCode___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__GeneticCode_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"GeneticCode", &GeneticCode___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__GeneticCode_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__GeneticCode_methods[] = {
  { "get_code", _wrap_GeneticCode_get_code, METH_NOARGS, "get_code(self) -> unsigned int" },
  { "set_code", _wrap_GeneticCode_set_code, METH_O, "set_code(self, index)" },
  { "start", _wrap_GeneticCode_start, METH_O, "start(self, codon) -> bool" },
  { "is_stop_unsmart", _wrap_GeneticCode_is_stop_unsmart, METH_O, "is_stop_unsmart(self, codon) -> bool" },
  { "NSsites", _wrap_GeneticCode_NSsites, METH_VARARGS, "\n"
		"NSsites(self, codon, ignorestop=True) -> double\n"
		"NSsites(self, site, num_samples, ignorestop=True) -> double\n"
		"" },
  { "Ssites", _wrap_GeneticCode_Ssites, METH_VARARGS, "\n"
		"Ssites(self, codon, ignorestop=True) -> double\n"
		"Ssites(self, site, num_samples, ignorestop=True) -> double\n"
		"" },
  { "translate", _wrap_GeneticCode_translate, METH_O, "translate(self, arg2) -> int" },
  { "name", _wrap_GeneticCode_name, METH_NOARGS, "name(self) -> char const *" },
  { "ndiff", (PyCFunction)(void(*)(void))_wrap_GeneticCode_ndiff, METH_STATIC|METH_VARARGS, "ndiff(codon1, codon2) -> unsigned int" },
  { "diff1", (PyCFunction)(void(*)(void))_wrap_GeneticCode_diff1, METH_STATIC|METH_VARARGS, "diff1(codon1, codon2) -> bool" },
  { "diff2", (PyCFunction)(void(*)(void))_wrap_GeneticCode_diff2, METH_STATIC|METH_VARARGS, "diff2(codon1, codon2) -> bool" },
  { "diff3", (PyCFunction)(void(*)(void))_wrap_GeneticCode_diff3, METH_STATIC|METH_VARARGS, "diff3(codon1, codon2) -> bool" },
  { "num_codes", (PyCFunction)(void(*)(void))_wrap_GeneticCode_num_codes, METH_STATIC|METH_NOARGS, "num_codes() -> unsigned int" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__GeneticCode_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.GeneticCode",                 /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_GeneticCode_destructor_closure,                  /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__GeneticCode_type.as_number,           /* tp_as_number */
    &SwigPyBuiltin__egglib__GeneticCode_type.as_sequence,         /* tp_as_sequence */
    &SwigPyBuiltin__egglib__GeneticCode_type.as_mapping,          /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__GeneticCode_type.as_buffer,           /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::GeneticCode",                    /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__GeneticCode_richcompare,               /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__GeneticCode_methods,                   /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__GeneticCode_getset,/* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_GeneticCode,                    /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__GeneticCode_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__GeneticCode_type};

static SwigPyGetSet SiteDiversity___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__SiteDiversity_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"SiteDiversity", &SiteDiversity___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__SiteDiversity_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__SiteDiversity_methods[] = {
  { "toggle_off", _wrap_SiteDiversity_toggle_off, METH_NOARGS, "toggle_off(self)" },
  { "toggle_fstats_diplo", _wrap_SiteDiversity_toggle_fstats_diplo, METH_NOARGS, "toggle_fstats_diplo(self)" },
  { "toggle_fstats_haplo", _wrap_SiteDiversity_toggle_fstats_haplo, METH_NOARGS, "toggle_fstats_haplo(self)" },
  { "toggle_fstats_hier", _wrap_SiteDiversity_toggle_fstats_hier, METH_NOARGS, "toggle_fstats_hier(self)" },
  { "toggle_hstats", _wrap_SiteDiversity_toggle_hstats, METH_NOARGS, "toggle_hstats(self)" },
  { "set_maf", _wrap_SiteDiversity_set_maf, METH_O, "set_maf(self, arg2)" },
  { "process", _wrap_SiteDiversity_process, METH_O, "process(self, arg2) -> unsigned int" },
  { "average", _wrap_SiteDiversity_average, METH_NOARGS, "average(self) -> unsigned int" },
  { "ns", _wrap_SiteDiversity_ns, METH_NOARGS, "ns(self) -> double" },
  { "nso", _wrap_SiteDiversity_nso, METH_NOARGS, "nso(self) -> double" },
  { "k", _wrap_SiteDiversity_k, METH_NOARGS, "k(self) -> unsigned int" },
  { "npop_eff1", _wrap_SiteDiversity_npop_eff1, METH_NOARGS, "npop_eff1(self) -> unsigned int" },
  { "npop_eff2", _wrap_SiteDiversity_npop_eff2, METH_NOARGS, "npop_eff2(self) -> unsigned int" },
  { "npop_eff3", _wrap_SiteDiversity_npop_eff3, METH_NOARGS, "npop_eff3(self) -> unsigned int" },
  { "nclu_eff", _wrap_SiteDiversity_nclu_eff, METH_NOARGS, "nclu_eff(self) -> unsigned int" },
  { "Aglob", _wrap_SiteDiversity_Aglob, METH_NOARGS, "Aglob(self) -> double" },
  { "Aing", _wrap_SiteDiversity_Aing, METH_NOARGS, "Aing(self) -> double" },
  { "Aout", _wrap_SiteDiversity_Aout, METH_NOARGS, "Aout(self) -> double" },
  { "global_allele", _wrap_SiteDiversity_global_allele, METH_O, "global_allele(self, arg2) -> int" },
  { "S", _wrap_SiteDiversity_S, METH_NOARGS, "S(self) -> double" },
  { "Sd", _wrap_SiteDiversity_Sd, METH_NOARGS, "Sd(self) -> double" },
  { "R", _wrap_SiteDiversity_R, METH_NOARGS, "R(self) -> double" },
  { "He", _wrap_SiteDiversity_He, METH_NOARGS, "He(self) -> double" },
  { "Ho", _wrap_SiteDiversity_Ho, METH_NOARGS, "Ho(self) -> double" },
  { "He_pop", _wrap_SiteDiversity_He_pop, METH_O, "He_pop(self, pop) -> double" },
  { "pairdiff_inter", _wrap_SiteDiversity_pairdiff_inter, METH_VARARGS, "pairdiff_inter(self, pop1, pop2) -> double" },
  { "orientable", _wrap_SiteDiversity_orientable, METH_NOARGS, "orientable(self) -> bool" },
  { "derived", _wrap_SiteDiversity_derived, METH_O, "derived(self, arg2) -> double" },
  { "a", _wrap_SiteDiversity_a, METH_NOARGS, "a(self) -> double" },
  { "b", _wrap_SiteDiversity_b, METH_NOARGS, "b(self) -> double" },
  { "c", _wrap_SiteDiversity_c, METH_NOARGS, "c(self) -> double" },
  { "a0", _wrap_SiteDiversity_a0, METH_NOARGS, "a0(self) -> double" },
  { "b1", _wrap_SiteDiversity_b1, METH_NOARGS, "b1(self) -> double" },
  { "b2", _wrap_SiteDiversity_b2, METH_NOARGS, "b2(self) -> double" },
  { "c0", _wrap_SiteDiversity_c0, METH_NOARGS, "c0(self) -> double" },
  { "n", _wrap_SiteDiversity_n, METH_NOARGS, "n(self) -> double" },
  { "d", _wrap_SiteDiversity_d, METH_NOARGS, "d(self) -> double" },
  { "Hst", _wrap_SiteDiversity_Hst, METH_NOARGS, "Hst(self) -> double" },
  { "Gst", _wrap_SiteDiversity_Gst, METH_NOARGS, "Gst(self) -> double" },
  { "Gste", _wrap_SiteDiversity_Gste, METH_NOARGS, "Gste(self) -> double" },
  { "D", _wrap_SiteDiversity_D, METH_NOARGS, "D(self) -> double" },
  { "thetaIAM", _wrap_SiteDiversity_thetaIAM, METH_NOARGS, "thetaIAM(self) -> double" },
  { "thetaSMM", _wrap_SiteDiversity_thetaSMM, METH_NOARGS, "thetaSMM(self) -> double" },
  { "maf", _wrap_SiteDiversity_maf, METH_NOARGS, "maf(self) -> double" },
  { "maf_pop", _wrap_SiteDiversity_maf_pop, METH_O, "maf_pop(self, arg2) -> double" },
  { "num_pop", _wrap_SiteDiversity_num_pop, METH_NOARGS, "num_pop(self) -> unsigned int" },
  { "reset", _wrap_SiteDiversity_reset, METH_NOARGS, "reset(self)" },
  { "nsites1", _wrap_SiteDiversity_nsites1, METH_NOARGS, "nsites1(self) -> unsigned int" },
  { "nsites2", _wrap_SiteDiversity_nsites2, METH_NOARGS, "nsites2(self) -> unsigned int" },
  { "nsites3", _wrap_SiteDiversity_nsites3, METH_NOARGS, "nsites3(self) -> unsigned int" },
  { "nsites4", _wrap_SiteDiversity_nsites4, METH_NOARGS, "nsites4(self) -> unsigned int" },
  { "nsites5", _wrap_SiteDiversity_nsites5, METH_NOARGS, "nsites5(self) -> unsigned int" },
  { "nsites6", _wrap_SiteDiversity_nsites6, METH_NOARGS, "nsites6(self) -> unsigned int" },
  { "nsites7", _wrap_SiteDiversity_nsites7, METH_NOARGS, "nsites7(self) -> unsigned int" },
  { "nsites8", _wrap_SiteDiversity_nsites8, METH_NOARGS, "nsites8(self) -> unsigned int" },
  { "nsites9", _wrap_SiteDiversity_nsites9, METH_NOARGS, "nsites9(self) -> unsigned int" },
  { "nsites10", _wrap_SiteDiversity_nsites10, METH_NOARGS, "nsites10(self) -> unsigned int" },
  { "nsites11", _wrap_SiteDiversity_nsites11, METH_NOARGS, "nsites11(self) -> unsigned int" },
  { "nsites12", _wrap_SiteDiversity_nsites12, METH_NOARGS, "nsites12(self) -> unsigned int" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__SiteDiversity_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.SiteDiversity",               /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_SiteDiversity_destructor_closure,                /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__SiteDiversity_type.as_number,         /* tp_as_number */
    &SwigPyBuiltin__egglib__SiteDiversity_type.as_sequence,       /* tp_as_sequence */
    &SwigPyBuiltin__egglib__SiteDiversity_type.as_mapping,        /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__SiteDiversity_type.as_buffer,         /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::SiteDiversity",                  /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__SiteDiversity_richcompare,             /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__SiteDiversity_methods,                 /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__SiteDiversity_getset,                  /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_SiteDiversity,                  /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__SiteDiversity_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__SiteDiversity_type};

static SwigPyGetSet SiteHolder___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__SiteHolder_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"SiteHolder", &SiteHolder___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__SiteHolder_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__SiteHolder_methods[] = {
  { "reset", _wrap_SiteHolder_reset, METH_NOARGS, "reset(self)" },
  { "get_position", _wrap_SiteHolder_get_position, METH_NOARGS, "get_position(self) -> double" },
  { "set_position", _wrap_SiteHolder_set_position, METH_O, "set_position(self, arg2)" },
  { "add", _wrap_SiteHolder_add, METH_O, "add(self, num)" },
  { "set_sample", _wrap_SiteHolder_set_sample, METH_VARARGS, "set_sample(self, sam, allele)" },
  { "get_ns", _wrap_SiteHolder_get_ns, METH_NOARGS, "get_ns(self) -> unsigned int" },
  { "get_sample", _wrap_SiteHolder_get_sample, METH_O, "get_sample(self, sam) -> int" },
  { "del_sample", _wrap_SiteHolder_del_sample, METH_O, "del_sample(self, arg2)" },
  { "get_missing", _wrap_SiteHolder_get_missing, METH_NOARGS, "get_missing(self) -> unsigned int" },
  { "process_align", _wrap_SiteHolder_process_align, METH_VARARGS, "process_align(self, data, idx, struc=None) -> unsigned int" },
  { "process_vcf", _wrap_SiteHolder_process_vcf, METH_VARARGS, "process_vcf(self, data, start, stop) -> unsigned int" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__SiteHolder_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.SiteHolder",                  /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_SiteHolder_destructor_closure,                   /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__SiteHolder_type.as_number,            /* tp_as_number */
    &SwigPyBuiltin__egglib__SiteHolder_type.as_sequence,          /* tp_as_sequence */
    &SwigPyBuiltin__egglib__SiteHolder_type.as_mapping,           /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__SiteHolder_type.as_buffer,            /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::SiteHolder",                     /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__SiteHolder_richcompare,                /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__SiteHolder_methods,/* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__SiteHolder_getset, /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_SiteHolder,                     /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__SiteHolder_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__SiteHolder_type};

static SwigPyGetSet Genotypes___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__Genotypes_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"Genotypes", &Genotypes___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__Genotypes_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__Genotypes_methods[] = {
  { "process", _wrap_Genotypes_process, METH_VARARGS, "process(self, arg2, arg3)" },
  { "ploidy", _wrap_Genotypes_ploidy, METH_NOARGS, "ploidy(self) -> unsigned int" },
  { "site", _wrap_Genotypes_site, METH_NOARGS, "site(self) -> SiteHolder" },
  { "heter", _wrap_Genotypes_heter, METH_O, "heter(self, arg2) -> bool" },
  { "genot", _wrap_Genotypes_genot, METH_O, "genot(self, arg2) -> int const *const" },
  { "num_genotypes", _wrap_Genotypes_num_genotypes, METH_NOARGS, "num_genotypes(self) -> unsigned int" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__Genotypes_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.Genotypes",                   /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_Genotypes_destructor_closure,/* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__Genotypes_type.as_number,             /* tp_as_number */
    &SwigPyBuiltin__egglib__Genotypes_type.as_sequence,           /* tp_as_sequence */
    &SwigPyBuiltin__egglib__Genotypes_type.as_mapping,            /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__Genotypes_type.as_buffer,             /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::Genotypes",                      /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__Genotypes_richcompare,                 /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__Genotypes_methods, /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__Genotypes_getset,  /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_Genotypes,                      /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__Genotypes_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__Genotypes_type};

static SwigPyGetSet Diversity1___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__Diversity1_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"Diversity1", &Diversity1___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__Diversity1_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__Diversity1_methods[] = {
  { "load", _wrap_Diversity1_load, METH_VARARGS, "load(self, freqs, div, position)" },
  { "compute", _wrap_Diversity1_compute, METH_NOARGS, "compute(self) -> unsigned int" },
  { "reset_stats", _wrap_Diversity1_reset_stats, METH_NOARGS, "reset_stats(self)" },
  { "toggle_off", _wrap_Diversity1_toggle_off, METH_NOARGS, "toggle_off(self)" },
  { "toggle_ori_site", _wrap_Diversity1_toggle_ori_site, METH_NOARGS, "toggle_ori_site(self)" },
  { "toggle_ori_div", _wrap_Diversity1_toggle_ori_div, METH_NOARGS, "toggle_ori_div(self)" },
  { "toggle_basic", _wrap_Diversity1_toggle_basic, METH_NOARGS, "toggle_basic(self)" },
  { "toggle_site_lists", _wrap_Diversity1_toggle_site_lists, METH_NOARGS, "toggle_site_lists(self)" },
  { "set_option_multiple", _wrap_Diversity1_set_option_multiple, METH_O, "set_option_multiple(self, arg2)" },
  { "set_option_ns_set", _wrap_Diversity1_set_option_ns_set, METH_O, "set_option_ns_set(self, arg2)" },
  { "lt", _wrap_Diversity1_lt, METH_NOARGS, "lt(self) -> unsigned int" },
  { "ls", _wrap_Diversity1_ls, METH_NOARGS, "ls(self) -> unsigned int" },
  { "lso", _wrap_Diversity1_lso, METH_NOARGS, "lso(self) -> unsigned int" },
  { "S", _wrap_Diversity1_S, METH_NOARGS, "S(self) -> unsigned int" },
  { "Ss", _wrap_Diversity1_Ss, METH_NOARGS, "Ss(self) -> unsigned int" },
  { "So", _wrap_Diversity1_So, METH_NOARGS, "So(self) -> unsigned int" },
  { "Sso", _wrap_Diversity1_Sso, METH_NOARGS, "Sso(self) -> unsigned int" },
  { "nsingld", _wrap_Diversity1_nsingld, METH_NOARGS, "nsingld(self) -> unsigned int" },
  { "site", _wrap_Diversity1_site, METH_O, "site(self, arg2) -> unsigned int" },
  { "site_o", _wrap_Diversity1_site_o, METH_O, "site_o(self, arg2) -> unsigned int" },
  { "singl", _wrap_Diversity1_singl, METH_O, "singl(self, arg2) -> unsigned int" },
  { "singl_o", _wrap_Diversity1_singl_o, METH_O, "singl_o(self, arg2) -> unsigned int" },
  { "nall", _wrap_Diversity1_nall, METH_O, "nall(self, arg2) -> unsigned int" },
  { "frq", _wrap_Diversity1_frq, METH_VARARGS, "frq(self, arg2, arg3) -> unsigned int" },
  { "frqp", _wrap_Diversity1_frqp, METH_VARARGS, "frqp(self, arg2, arg3, arg4) -> unsigned int" },
  { "eta", _wrap_Diversity1_eta, METH_NOARGS, "eta(self) -> unsigned int" },
  { "etao", _wrap_Diversity1_etao, METH_NOARGS, "etao(self) -> unsigned int" },
  { "D", _wrap_Diversity1_D, METH_NOARGS, "D(self) -> double" },
  { "Deta", _wrap_Diversity1_Deta, METH_NOARGS, "Deta(self) -> double" },
  { "Dstar", _wrap_Diversity1_Dstar, METH_NOARGS, "Dstar(self) -> double" },
  { "Fstar", _wrap_Diversity1_Fstar, METH_NOARGS, "Fstar(self) -> double" },
  { "thetaW", _wrap_Diversity1_thetaW, METH_NOARGS, "thetaW(self) -> double" },
  { "Pi", _wrap_Diversity1_Pi, METH_NOARGS, "Pi(self) -> double" },
  { "Dxy", _wrap_Diversity1_Dxy, METH_NOARGS, "Dxy(self) -> double" },
  { "Da", _wrap_Diversity1_Da, METH_NOARGS, "Da(self) -> double" },
  { "nseff", _wrap_Diversity1_nseff, METH_NOARGS, "nseff(self) -> double" },
  { "nsmax", _wrap_Diversity1_nsmax, METH_NOARGS, "nsmax(self) -> unsigned int" },
  { "nseffo", _wrap_Diversity1_nseffo, METH_NOARGS, "nseffo(self) -> double" },
  { "nsmaxo", _wrap_Diversity1_nsmaxo, METH_NOARGS, "nsmaxo(self) -> unsigned int" },
  { "thetaH", _wrap_Diversity1_thetaH, METH_NOARGS, "thetaH(self) -> double" },
  { "thetaPi", _wrap_Diversity1_thetaPi, METH_NOARGS, "thetaPi(self) -> double" },
  { "thetaL", _wrap_Diversity1_thetaL, METH_NOARGS, "thetaL(self) -> double" },
  { "Hns", _wrap_Diversity1_Hns, METH_NOARGS, "Hns(self) -> double" },
  { "Hsd", _wrap_Diversity1_Hsd, METH_NOARGS, "Hsd(self) -> double" },
  { "E", _wrap_Diversity1_E, METH_NOARGS, "E(self) -> double" },
  { "Dfl", _wrap_Diversity1_Dfl, METH_NOARGS, "Dfl(self) -> double" },
  { "F", _wrap_Diversity1_F, METH_NOARGS, "F(self) -> double" },
  { "pM", _wrap_Diversity1_pM, METH_NOARGS, "pM(self) -> double" },
  { "nM", _wrap_Diversity1_nM, METH_NOARGS, "nM(self) -> unsigned int" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__Diversity1_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.Diversity1",                  /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_Diversity1_destructor_closure,                   /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__Diversity1_type.as_number,            /* tp_as_number */
    &SwigPyBuiltin__egglib__Diversity1_type.as_sequence,          /* tp_as_sequence */
    &SwigPyBuiltin__egglib__Diversity1_type.as_mapping,           /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__Diversity1_type.as_buffer,            /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::Diversity1",                     /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__Diversity1_richcompare,                /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__Diversity1_methods,/* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__Diversity1_getset, /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_Diversity1,                     /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__Diversity1_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__Diversity1_type};

static SwigPyGetSet Diversity2___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__Diversity2_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"Diversity2", &Diversity2___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__Diversity2_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__Diversity2_methods[] = {
  { "reset", _wrap_Diversity2_reset, METH_NOARGS, "reset(self)" },
  { "reset_stats", _wrap_Diversity2_reset_stats, METH_NOARGS, "reset_stats(self)" },
  { "set_structure", _wrap_Diversity2_set_structure, METH_O, "set_structure(self, arg2)" },
  { "load", _wrap_Diversity2_load, METH_VARARGS, "load(self, arg2, arg3, arg4)" },
  { "num_sites", _wrap_Diversity2_num_sites, METH_NOARGS, "num_sites(self) -> unsigned int" },
  { "num_orientable", _wrap_Diversity2_num_orientable, METH_NOARGS, "num_orientable(self) -> unsigned int" },
  { "num_clear", _wrap_Diversity2_num_clear, METH_NOARGS, "num_clear(self) -> unsigned int" },
  { "num_samples", _wrap_Diversity2_num_samples, METH_NOARGS, "num_samples(self) -> unsigned int" },
  { "toggle_off", _wrap_Diversity2_toggle_off, METH_NOARGS, "toggle_off(self)" },
  { "toggle_singletons", _wrap_Diversity2_toggle_singletons, METH_NOARGS, "toggle_singletons(self)" },
  { "toggle_partitions", _wrap_Diversity2_toggle_partitions, METH_NOARGS, "toggle_partitions(self)" },
  { "set_option_multiple", _wrap_Diversity2_set_option_multiple, METH_O, "set_option_multiple(self, arg2)" },
  { "compute", _wrap_Diversity2_compute, METH_NOARGS, "compute(self) -> unsigned int" },
  { "k", _wrap_Diversity2_k, METH_NOARGS, "k(self) -> double" },
  { "ko", _wrap_Diversity2_ko, METH_NOARGS, "ko(self) -> double" },
  { "R2", _wrap_Diversity2_R2, METH_NOARGS, "R2(self) -> double" },
  { "R3", _wrap_Diversity2_R3, METH_NOARGS, "R3(self) -> double" },
  { "R4", _wrap_Diversity2_R4, METH_NOARGS, "R4(self) -> double" },
  { "Ch", _wrap_Diversity2_Ch, METH_NOARGS, "Ch(self) -> double" },
  { "R2E", _wrap_Diversity2_R2E, METH_NOARGS, "R2E(self) -> double" },
  { "R3E", _wrap_Diversity2_R3E, METH_NOARGS, "R3E(self) -> double" },
  { "R4E", _wrap_Diversity2_R4E, METH_NOARGS, "R4E(self) -> double" },
  { "ChE", _wrap_Diversity2_ChE, METH_NOARGS, "ChE(self) -> double" },
  { "B", _wrap_Diversity2_B, METH_NOARGS, "B(self) -> double" },
  { "Q", _wrap_Diversity2_Q, METH_NOARGS, "Q(self) -> double" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__Diversity2_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.Diversity2",                  /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_Diversity2_destructor_closure,                   /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__Diversity2_type.as_number,            /* tp_as_number */
    &SwigPyBuiltin__egglib__Diversity2_type.as_sequence,          /* tp_as_sequence */
    &SwigPyBuiltin__egglib__Diversity2_type.as_mapping,           /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__Diversity2_type.as_buffer,            /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::Diversity2",                     /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__Diversity2_richcompare,                /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__Diversity2_methods,/* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__Diversity2_getset, /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_Diversity2,                     /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__Diversity2_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__Diversity2_type};

static SwigPyGetSet Haplotypes___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__Haplotypes_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"Haplotypes", &Haplotypes___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__Haplotypes_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__Haplotypes_methods[] = {
  { "setup", _wrap_Haplotypes_setup, METH_O, "setup(self, arg2)" },
  { "set_structure", _wrap_Haplotypes_set_structure, METH_O, "set_structure(self, arg2)" },
  { "reset_stats", _wrap_Haplotypes_reset_stats, METH_NOARGS, "reset_stats(self)" },
  { "load", _wrap_Haplotypes_load, METH_O, "load(self, arg2)" },
  { "prepare_impute", _wrap_Haplotypes_prepare_impute, METH_O, "prepare_impute(self, arg2)" },
  { "resolve", _wrap_Haplotypes_resolve, METH_O, "resolve(self, arg2)" },
  { "cp_haplotypes", _wrap_Haplotypes_cp_haplotypes, METH_NOARGS, "cp_haplotypes(self)" },
  { "impute", _wrap_Haplotypes_impute, METH_NOARGS, "impute(self)" },
  { "cp_dist", _wrap_Haplotypes_cp_dist, METH_NOARGS, "cp_dist(self)" },
  { "cp_stats", _wrap_Haplotypes_cp_stats, METH_NOARGS, "cp_stats(self) -> unsigned int" },
  { "n_ing", _wrap_Haplotypes_n_ing, METH_NOARGS, "n_ing(self) -> unsigned int" },
  { "n_otg", _wrap_Haplotypes_n_otg, METH_NOARGS, "n_otg(self) -> unsigned int" },
  { "n_sam", _wrap_Haplotypes_n_sam, METH_NOARGS, "n_sam(self) -> unsigned int" },
  { "nt_hapl", _wrap_Haplotypes_nt_hapl, METH_NOARGS, "nt_hapl(self) -> unsigned int" },
  { "hapl", _wrap_Haplotypes_hapl, METH_VARARGS, "hapl(self, arg2, arg3) -> unsigned int" },
  { "map_sample", _wrap_Haplotypes_map_sample, METH_O, "map_sample(self, arg2) -> unsigned int" },
  { "n_sites", _wrap_Haplotypes_n_sites, METH_NOARGS, "n_sites(self) -> unsigned int" },
  { "ne_ing", _wrap_Haplotypes_ne_ing, METH_NOARGS, "ne_ing(self) -> unsigned int" },
  { "ne_otg", _wrap_Haplotypes_ne_otg, METH_NOARGS, "ne_otg(self) -> unsigned int" },
  { "n_missing", _wrap_Haplotypes_n_missing, METH_O, "n_missing(self, arg2) -> unsigned int" },
  { "n_mis", _wrap_Haplotypes_n_mis, METH_NOARGS, "n_mis(self) -> unsigned int" },
  { "mis_idx", _wrap_Haplotypes_mis_idx, METH_O, "mis_idx(self, arg2) -> unsigned int" },
  { "ni_hapl", _wrap_Haplotypes_ni_hapl, METH_NOARGS, "ni_hapl(self) -> unsigned int" },
  { "ng_hapl", _wrap_Haplotypes_ng_hapl, METH_NOARGS, "ng_hapl(self) -> unsigned int" },
  { "freq_i", _wrap_Haplotypes_freq_i, METH_O, "freq_i(self, arg2) -> unsigned int" },
  { "freq_o", _wrap_Haplotypes_freq_o, METH_O, "freq_o(self, arg2) -> unsigned int" },
  { "n_potential", _wrap_Haplotypes_n_potential, METH_O, "n_potential(self, arg2) -> unsigned int" },
  { "potential", _wrap_Haplotypes_potential, METH_VARARGS, "potential(self, arg2, arg3) -> unsigned int" },
  { "dist", _wrap_Haplotypes_dist, METH_VARARGS, "dist(self, arg2, arg3) -> unsigned int" },
  { "Fst", _wrap_Haplotypes_Fst, METH_NOARGS, "Fst(self) -> double" },
  { "Kst", _wrap_Haplotypes_Kst, METH_NOARGS, "Kst(self) -> double" },
  { "Snn", _wrap_Haplotypes_Snn, METH_NOARGS, "Snn(self) -> double" },
  { "n_pop", _wrap_Haplotypes_n_pop, METH_NOARGS, "n_pop(self) -> unsigned int" },
  { "ne_pop", _wrap_Haplotypes_ne_pop, METH_NOARGS, "ne_pop(self) -> unsigned int" },
  { "get_site", _wrap_Haplotypes_get_site, METH_O, "get_site(self, arg2)" },
  { "invalid", _wrap_Haplotypes_invalid, METH_NOARGS, "invalid(self) -> bool" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__Haplotypes_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.Haplotypes",                  /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_Haplotypes_destructor_closure,                   /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__Haplotypes_type.as_number,            /* tp_as_number */
    &SwigPyBuiltin__egglib__Haplotypes_type.as_sequence,          /* tp_as_sequence */
    &SwigPyBuiltin__egglib__Haplotypes_type.as_mapping,           /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__Haplotypes_type.as_buffer,            /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::Haplotypes",                     /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__Haplotypes_richcompare,                /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__Haplotypes_methods,/* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__Haplotypes_getset, /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_Haplotypes,                     /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__Haplotypes_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__Haplotypes_type};

static SwigPyGetSet PairwiseLD___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__PairwiseLD_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"PairwiseLD", &PairwiseLD___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__PairwiseLD_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__PairwiseLD_methods[] = {
  { "process", _wrap_PairwiseLD_process, METH_VARARGS, "process(self, site1, site2, arg4, arg5, arg6, min_n=2, max_maj=1.0) -> bool" },
  { "num_alleles1", _wrap_PairwiseLD_num_alleles1, METH_NOARGS, "num_alleles1(self) -> unsigned int" },
  { "num_alleles2", _wrap_PairwiseLD_num_alleles2, METH_NOARGS, "num_alleles2(self) -> unsigned int" },
  { "index1", _wrap_PairwiseLD_index1, METH_O, "index1(self, allele) -> unsigned int" },
  { "index2", _wrap_PairwiseLD_index2, METH_O, "index2(self, allele) -> unsigned int" },
  { "freq1", _wrap_PairwiseLD_freq1, METH_O, "freq1(self, allele) -> unsigned int" },
  { "freq2", _wrap_PairwiseLD_freq2, METH_O, "freq2(self, allele) -> unsigned int" },
  { "freq", _wrap_PairwiseLD_freq, METH_VARARGS, "freq(self, allele1, allele2) -> unsigned int" },
  { "nsam", _wrap_PairwiseLD_nsam, METH_NOARGS, "nsam(self) -> unsigned int" },
  { "compute", _wrap_PairwiseLD_compute, METH_VARARGS, "compute(self, allele1, allele2)" },
  { "D", _wrap_PairwiseLD_D, METH_NOARGS, "D(self) -> double" },
  { "Dp", _wrap_PairwiseLD_Dp, METH_NOARGS, "Dp(self) -> double" },
  { "r", _wrap_PairwiseLD_r, METH_NOARGS, "r(self) -> double" },
  { "rsq", _wrap_PairwiseLD_rsq, METH_NOARGS, "rsq(self) -> double" },
  { "reset", _wrap_PairwiseLD_reset, METH_NOARGS, "reset(self)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__PairwiseLD_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.PairwiseLD",                  /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_PairwiseLD_destructor_closure,                   /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__PairwiseLD_type.as_number,            /* tp_as_number */
    &SwigPyBuiltin__egglib__PairwiseLD_type.as_sequence,          /* tp_as_sequence */
    &SwigPyBuiltin__egglib__PairwiseLD_type.as_mapping,           /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__PairwiseLD_type.as_buffer,            /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::PairwiseLD",                     /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__PairwiseLD_richcompare,                /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__PairwiseLD_methods,/* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__PairwiseLD_getset, /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_PairwiseLD,                     /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__PairwiseLD_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__PairwiseLD_type};

static SwigPyGetSet MatrixLD___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__MatrixLD_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"MatrixLD", &MatrixLD___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__MatrixLD_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__MatrixLD_methods[] = {
  { "reset", _wrap_MatrixLD_reset, METH_NOARGS, "reset(self)" },
  { "set_structure", _wrap_MatrixLD_set_structure, METH_O, "set_structure(self, arg2)" },
  { "load", _wrap_MatrixLD_load, METH_VARARGS, "load(self, genosite, position)" },
  { "computeLD", _wrap_MatrixLD_computeLD, METH_VARARGS, "computeLD(self, min_n=2, max_maj=1.0)" },
  { "num_tot", _wrap_MatrixLD_num_tot, METH_NOARGS, "num_tot(self) -> unsigned int" },
  { "num_pairs", _wrap_MatrixLD_num_pairs, METH_NOARGS, "num_pairs(self) -> unsigned int" },
  { "num_alleles", _wrap_MatrixLD_num_alleles, METH_NOARGS, "num_alleles(self) -> unsigned int" },
  { "pairLD", _wrap_MatrixLD_pairLD, METH_O, "pairLD(self, index) -> PairwiseLD" },
  { "index1", _wrap_MatrixLD_index1, METH_O, "index1(self, allele) -> unsigned int" },
  { "index2", _wrap_MatrixLD_index2, METH_O, "index2(self, allele) -> unsigned int" },
  { "distance", _wrap_MatrixLD_distance, METH_O, "distance(self, index) -> unsigned int" },
  { "computeStats", _wrap_MatrixLD_computeStats, METH_VARARGS, "computeStats(self, multiallelic=ignore, min_freq=0)" },
  { "num_allele_pairs", _wrap_MatrixLD_num_allele_pairs, METH_NOARGS, "num_allele_pairs(self) -> unsigned int" },
  { "num_allele_pairs_adj", _wrap_MatrixLD_num_allele_pairs_adj, METH_NOARGS, "num_allele_pairs_adj(self) -> unsigned int" },
  { "ZnS", _wrap_MatrixLD_ZnS, METH_NOARGS, "ZnS(self) -> double" },
  { "ZnS_star1", _wrap_MatrixLD_ZnS_star1, METH_NOARGS, "ZnS_star1(self) -> double" },
  { "ZnS_star2", _wrap_MatrixLD_ZnS_star2, METH_NOARGS, "ZnS_star2(self) -> double" },
  { "Za", _wrap_MatrixLD_Za, METH_NOARGS, "Za(self) -> double" },
  { "ZZ", _wrap_MatrixLD_ZZ, METH_NOARGS, "ZZ(self) -> double" },
  { "computeRmin", _wrap_MatrixLD_computeRmin, METH_VARARGS, "computeRmin(self, oriented=False)" },
  { "Rmin", _wrap_MatrixLD_Rmin, METH_NOARGS, "Rmin(self) -> unsigned int" },
  { "Rmin_num_sites", _wrap_MatrixLD_Rmin_num_sites, METH_NOARGS, "Rmin_num_sites(self) -> unsigned int" },
  { "Rmin_left", _wrap_MatrixLD_Rmin_left, METH_O, "Rmin_left(self, i) -> unsigned int" },
  { "Rmin_right", _wrap_MatrixLD_Rmin_right, METH_O, "Rmin_right(self, i) -> unsigned int" },
  { "toggle_off", _wrap_MatrixLD_toggle_off, METH_NOARGS, "toggle_off(self)" },
  { "toggle_Rmin", _wrap_MatrixLD_toggle_Rmin, METH_NOARGS, "toggle_Rmin(self)" },
  { "toggle_stats", _wrap_MatrixLD_toggle_stats, METH_NOARGS, "toggle_stats(self)" },
  { "process", _wrap_MatrixLD_process, METH_VARARGS, "process(self, min_n, max_maj, multiallelic, min_freq, oriented) -> unsigned int" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__MatrixLD_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.MatrixLD",                    /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_MatrixLD_destructor_closure, /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__MatrixLD_type.as_number,              /* tp_as_number */
    &SwigPyBuiltin__egglib__MatrixLD_type.as_sequence,            /* tp_as_sequence */
    &SwigPyBuiltin__egglib__MatrixLD_type.as_mapping,             /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__MatrixLD_type.as_buffer,              /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::MatrixLD",                       /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__MatrixLD_richcompare,                  /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__MatrixLD_methods,  /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__MatrixLD_getset,   /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_MatrixLD,                       /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__MatrixLD_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__MatrixLD_type};

static SwigPyGetSet ParalogPi___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__ParalogPi_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"ParalogPi", &ParalogPi___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__ParalogPi_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__ParalogPi_methods[] = {
  { "reset", _wrap_ParalogPi_reset, METH_VARARGS, "reset(self, str_prl, str_idv, max_missing)" },
  { "load", _wrap_ParalogPi_load, METH_O, "load(self, arg2)" },
  { "num_sites_tot", _wrap_ParalogPi_num_sites_tot, METH_NOARGS, "num_sites_tot(self) -> unsigned int" },
  { "num_sites_paralog", _wrap_ParalogPi_num_sites_paralog, METH_O, "num_sites_paralog(self, arg2) -> unsigned int" },
  { "num_sites_pair", _wrap_ParalogPi_num_sites_pair, METH_VARARGS, "num_sites_pair(self, arg2, arg3) -> unsigned int" },
  { "num_samples", _wrap_ParalogPi_num_samples, METH_NOARGS, "num_samples(self) -> unsigned int" },
  { "num_paralogs", _wrap_ParalogPi_num_paralogs, METH_NOARGS, "num_paralogs(self) -> unsigned int" },
  { "Piw", _wrap_ParalogPi_Piw, METH_O, "Piw(self, arg2) -> double" },
  { "Pib", _wrap_ParalogPi_Pib, METH_VARARGS, "Pib(self, arg2, arg3) -> double" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__ParalogPi_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.ParalogPi",                   /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_ParalogPi_destructor_closure,/* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__ParalogPi_type.as_number,             /* tp_as_number */
    &SwigPyBuiltin__egglib__ParalogPi_type.as_sequence,           /* tp_as_sequence */
    &SwigPyBuiltin__egglib__ParalogPi_type.as_mapping,            /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__ParalogPi_type.as_buffer,             /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::ParalogPi",                      /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__ParalogPi_richcompare,                 /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__ParalogPi_methods, /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__ParalogPi_getset,  /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_ParalogPi,                      /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__ParalogPi_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__ParalogPi_type};

static SwigPyGetSet Rd___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__Rd_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"Rd", &Rd___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__Rd_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__Rd_methods[] = {
  { "load", _wrap_Rd_load, METH_O, "load(self, arg2)" },
  { "configure", _wrap_Rd_configure, METH_O, "configure(self, arg2)" },
  { "compute", _wrap_Rd_compute, METH_NOARGS, "compute(self) -> double" },
  { "num_loci", _wrap_Rd_num_loci, METH_NOARGS, "num_loci(self) -> unsigned int" },
  { "ploidy", _wrap_Rd_ploidy, METH_NOARGS, "ploidy(self) -> unsigned int" },
  { "num_indiv", _wrap_Rd_num_indiv, METH_NOARGS, "num_indiv(self) -> unsigned int" },
  { "reset", _wrap_Rd_reset, METH_NOARGS, "reset(self)" },
  { "reset_stats", _wrap_Rd_reset_stats, METH_NOARGS, "reset_stats(self)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__Rd_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.Rd",                          /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_Rd_destructor_closure,       /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__Rd_type.as_number,/* tp_as_number */
    &SwigPyBuiltin__egglib__Rd_type.as_sequence,                  /* tp_as_sequence */
    &SwigPyBuiltin__egglib__Rd_type.as_mapping,                   /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__Rd_type.as_buffer,/* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::Rd",                             /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__Rd_richcompare,    /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__Rd_methods,        /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__Rd_getset,         /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_Rd,                             /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__Rd_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__Rd_type};

static SwigPyGetSet Migration___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__Migration_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"Migration", &Migration___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__Migration_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__Migration_methods[] = {
  { "n", _wrap_Migration_n, METH_NOARGS, "n(self) -> unsigned int" },
  { "set_all", _wrap_Migration_set_all, METH_O, "set_all(self, M)" },
  { "set_all_R", _wrap_Migration_set_all_R, METH_O, "set_all_R(self, M)" },
  { "set_row", _wrap_Migration_set_row, METH_VARARGS, "set_row(self, i, M)" },
  { "set_row_R", _wrap_Migration_set_row_R, METH_VARARGS, "set_row_R(self, i, M)" },
  { "set_pair", _wrap_Migration_set_pair, METH_VARARGS, "set_pair(self, i, j, m)" },
  { "set_pair_R", _wrap_Migration_set_pair_R, METH_VARARGS, "set_pair_R(self, i, j, m)" },
  { "get_row", _wrap_Migration_get_row, METH_O, "get_row(self, i) -> double" },
  { "get_pair", _wrap_Migration_get_pair, METH_VARARGS, "get_pair(self, i, j) -> double" },
  { "restore", _wrap_Migration_restore, METH_NOARGS, "restore(self)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__Migration_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.Migration",                   /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_Migration_destructor_closure,/* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__Migration_type.as_number,             /* tp_as_number */
    &SwigPyBuiltin__egglib__Migration_type.as_sequence,           /* tp_as_sequence */
    &SwigPyBuiltin__egglib__Migration_type.as_mapping,            /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__Migration_type.as_buffer,             /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::Migration",                      /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__Migration_richcompare,                 /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__Migration_methods, /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__Migration_getset,  /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_Migration,                      /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__Migration_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__Migration_type};

static SwigPyGetSet Event___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__Event_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"Event", &Event___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__Event_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__Event_methods[] = {
  { "copy", _wrap_Event_copy, METH_O, "copy(self, src)" },
  { "event_type", _wrap_Event_event_type, METH_NOARGS, "event_type(self) -> egglib::Event::Type" },
  { "date", _wrap_Event_date, METH_NOARGS, "date(self) -> double" },
  { "get_param", _wrap_Event_get_param, METH_NOARGS, "get_param(self) -> double" },
  { "get_index", _wrap_Event_get_index, METH_NOARGS, "get_index(self) -> unsigned int" },
  { "get_dest", _wrap_Event_get_dest, METH_NOARGS, "get_dest(self) -> unsigned int" },
  { "get_number1", _wrap_Event_get_number1, METH_NOARGS, "get_number1(self) -> unsigned int" },
  { "get_number2", _wrap_Event_get_number2, METH_NOARGS, "get_number2(self) -> unsigned int" },
  { "perform", _wrap_Event_perform, METH_VARARGS, "perform(self, param, coal)" },
  { "insert", _wrap_Event_insert, METH_O, "insert(self, event)" },
  { "disconnect", _wrap_Event_disconnect, METH_NOARGS, "disconnect(self)" },
  { "next", _wrap_Event_next, METH_NOARGS, "next(self) -> Event" },
  { "prev", _wrap_Event_prev, METH_NOARGS, "prev(self) -> Event" },
  { "move", _wrap_Event_move, METH_O, "move(self, date)" },
  { "set_index", _wrap_Event_set_index, METH_O, "set_index(self, i)" },
  { "set_param", _wrap_Event_set_param, METH_O, "set_param(self, p)" },
  { "set_dest", _wrap_Event_set_dest, METH_O, "set_dest(self, d)" },
  { "set_number1", _wrap_Event_set_number1, METH_O, "set_number1(self, n)" },
  { "set_number2", _wrap_Event_set_number2, METH_O, "set_number2(self, n)" },
  { "get_label", _wrap_Event_get_label, METH_NOARGS, "get_label(self) -> char const *" },
  { "set_label", _wrap_Event_set_label, METH_O, "set_label(self, arg2)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__Event_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.Event",                       /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_Event_destructor_closure,    /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__Event_type.as_number,                 /* tp_as_number */
    &SwigPyBuiltin__egglib__Event_type.as_sequence,               /* tp_as_sequence */
    &SwigPyBuiltin__egglib__Event_type.as_mapping,                /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__Event_type.as_buffer,                 /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::Event",                          /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__Event_richcompare, /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__Event_methods,     /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__Event_getset,      /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_Event,                          /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__Event_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__Event_type};

static SwigPyGetSet Params___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__Params_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"\n"
		"Params\n"
		"Params\n"
		"", &Params___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__Params_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__Params_methods[] = {
  { "summary", _wrap_Params_summary, METH_NOARGS, "summary(self) -> std::string" },
  { "k", _wrap_Params_k, METH_NOARGS, "k(self) -> unsigned int" },
  { "get_K", _wrap_Params_get_K, METH_NOARGS, "get_K(self) -> unsigned int" },
  { "set_K", _wrap_Params_set_K, METH_O, "set_K(self, value)" },
  { "get_max_iter", _wrap_Params_get_max_iter, METH_NOARGS, "get_max_iter(self) -> unsigned long" },
  { "set_max_iter", _wrap_Params_set_max_iter, METH_O, "set_max_iter(self, arg2)" },
  { "get_random_start_allele", _wrap_Params_get_random_start_allele, METH_NOARGS, "get_random_start_allele(self) -> bool" },
  { "set_random_start_allele", _wrap_Params_set_random_start_allele, METH_O, "set_random_start_allele(self, value)" },
  { "set_R", _wrap_Params_set_R, METH_O, "set_R(self, value)" },
  { "R_R", _wrap_Params_R_R, METH_O, "R_R(self, value)" },
  { "get_R", _wrap_Params_get_R, METH_NOARGS, "get_R(self) -> double" },
  { "set_theta", _wrap_Params_set_theta, METH_O, "set_theta(self, value)" },
  { "get_theta", _wrap_Params_get_theta, METH_NOARGS, "get_theta(self) -> double" },
  { "set_fixed", _wrap_Params_set_fixed, METH_O, "set_fixed(self, value)" },
  { "get_fixed", _wrap_Params_get_fixed, METH_NOARGS, "get_fixed(self) -> unsigned int" },
  { "set_mutmodel", _wrap_Params_set_mutmodel, METH_O, "set_mutmodel(self, value)" },
  { "get_mutmodel", _wrap_Params_get_mutmodel, METH_NOARGS, "get_mutmodel(self) -> egglib::Params::MutationModel" },
  { "set_TPMproba", _wrap_Params_set_TPMproba, METH_O, "set_TPMproba(self, value)" },
  { "get_TPMproba", _wrap_Params_get_TPMproba, METH_NOARGS, "get_TPMproba(self) -> double" },
  { "set_TPMparam", _wrap_Params_set_TPMparam, METH_O, "set_TPMparam(self, value)" },
  { "get_TPMparam", _wrap_Params_get_TPMparam, METH_NOARGS, "get_TPMparam(self) -> double" },
  { "set_n1", _wrap_Params_set_n1, METH_VARARGS, "set_n1(self, pop, value)" },
  { "get_n1", _wrap_Params_get_n1, METH_O, "get_n1(self, pop) -> unsigned int" },
  { "set_n2", _wrap_Params_set_n2, METH_VARARGS, "set_n2(self, pop, value)" },
  { "get_n2", _wrap_Params_get_n2, METH_O, "get_n2(self, pop) -> unsigned int" },
  { "set_N", _wrap_Params_set_N, METH_VARARGS, "set_N(self, pop, value)" },
  { "N_R", _wrap_Params_N_R, METH_VARARGS, "N_R(self, pop, value, t)" },
  { "get_N", _wrap_Params_get_N, METH_O, "get_N(self, pop) -> double" },
  { "set_G", _wrap_Params_set_G, METH_VARARGS, "set_G(self, pop, value)" },
  { "G_R", _wrap_Params_G_R, METH_VARARGS, "G_R(self, pop, value, t)" },
  { "get_G", _wrap_Params_get_G, METH_O, "get_G(self, pop) -> double" },
  { "set_s", _wrap_Params_set_s, METH_VARARGS, "set_s(self, pop, value)" },
  { "s_R", _wrap_Params_s_R, METH_VARARGS, "s_R(self, pop, value, t)" },
  { "get_s", _wrap_Params_get_s, METH_O, "get_s(self, pop) -> double" },
  { "get_L", _wrap_Params_get_L, METH_NOARGS, "get_L(self) -> unsigned int" },
  { "set_L", _wrap_Params_set_L, METH_O, "set_L(self, value)" },
  { "autoSitePos", _wrap_Params_autoSitePos, METH_NOARGS, "autoSitePos(self)" },
  { "set_sitePos", _wrap_Params_set_sitePos, METH_VARARGS, "set_sitePos(self, site, value)" },
  { "get_sitePos", _wrap_Params_get_sitePos, METH_O, "get_sitePos(self, site) -> double" },
  { "set_siteW", _wrap_Params_set_siteW, METH_VARARGS, "set_siteW(self, site, value)" },
  { "get_siteW", _wrap_Params_get_siteW, METH_O, "get_siteW(self, site) -> double" },
  { "totalSiteW", _wrap_Params_totalSiteW, METH_NOARGS, "totalSiteW(self) -> double" },
  { "M", _wrap_Params_M, METH_NOARGS, "M(self) -> Migration" },
  { "set_transW_matrix", _wrap_Params_set_transW_matrix, METH_O, "set_transW_matrix(self, flag)" },
  { "get_transW_matrix", _wrap_Params_get_transW_matrix, METH_NOARGS, "get_transW_matrix(self) -> bool" },
  { "set_transW_pair", _wrap_Params_set_transW_pair, METH_VARARGS, "set_transW_pair(self, i, j, value)" },
  { "get_transW_pair", _wrap_Params_get_transW_pair, METH_VARARGS, "get_transW_pair(self, i, j) -> double" },
  { "get_transW_row", _wrap_Params_get_transW_row, METH_O, "get_transW_row(self, i) -> double" },
  { "addChange", _wrap_Params_addChange, METH_O, "addChange(self, e)" },
  { "clearChanges", _wrap_Params_clearChanges, METH_NOARGS, "clearChanges(self)" },
  { "numChanges", _wrap_Params_numChanges, METH_NOARGS, "numChanges(self) -> unsigned int" },
  { "firstChange", _wrap_Params_firstChange, METH_NOARGS, "firstChange(self) -> Event" },
  { "nextChangeDate", _wrap_Params_nextChangeDate, METH_NOARGS, "nextChangeDate(self) -> double" },
  { "nextChangeDo", _wrap_Params_nextChangeDo, METH_O, "nextChangeDo(self, coal)" },
  { "nDSChanges", _wrap_Params_nDSChanges, METH_NOARGS, "nDSChanges(self) -> unsigned int" },
  { "lastChange", _wrap_Params_lastChange, METH_O, "lastChange(self, pop) -> double" },
  { "restore", _wrap_Params_restore, METH_NOARGS, "restore(self)" },
  { "validate", _wrap_Params_validate, METH_NOARGS, "validate(self)" },
  { "get_nsam", _wrap_Params_get_nsam, METH_NOARGS, "get_nsam(self) -> unsigned int" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__Params_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.Params",                      /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_Params_destructor_closure,   /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__Params_type.as_number,                /* tp_as_number */
    &SwigPyBuiltin__egglib__Params_type.as_sequence,              /* tp_as_sequence */
    &SwigPyBuiltin__egglib__Params_type.as_mapping,               /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__Params_type.as_buffer,                /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::Params",                         /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__Params_richcompare,/* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__Params_methods,    /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__Params_getset,     /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_Params,                         /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__Params_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__Params_type};

static SwigPyGetSet Node___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__Node_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"\n"
		"Node\n"
		"Node\n"
		"", &Node___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__Node_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__Node_methods[] = {
  { "set_internal", _wrap_Node_set_internal, METH_VARARGS, "set_internal(self, t, son1, son2)" },
  { "set_terminal", _wrap_Node_set_terminal, METH_VARARGS, "set_terminal(self, t, label)" },
  { "is_terminal", _wrap_Node_is_terminal, METH_NOARGS, "is_terminal(self) -> bool" },
  { "label", _wrap_Node_label, METH_NOARGS, "label(self) -> unsigned int" },
  { "son1", _wrap_Node_son1, METH_NOARGS, "son1(self) -> unsigned int" },
  { "son2", _wrap_Node_son2, METH_NOARGS, "son2(self) -> unsigned int" },
  { "date", _wrap_Node_date, METH_NOARGS, "date(self) -> double" },
  { "get_L", _wrap_Node_get_L, METH_NOARGS, "get_L(self) -> double" },
  { "set_L", _wrap_Node_set_L, METH_O, "set_L(self, value)" },
  { "nmut", _wrap_Node_nmut, METH_NOARGS, "nmut(self) -> unsigned int" },
  { "mutationSite", _wrap_Node_mutationSite, METH_O, "mutationSite(self, mut) -> unsigned int" },
  { "mutationPos", _wrap_Node_mutationPos, METH_O, "mutationPos(self, mut) -> double" },
  { "addMutation", _wrap_Node_addMutation, METH_VARARGS, "addMutation(self, site, pos)" },
  { "reset", _wrap_Node_reset, METH_NOARGS, "reset(self)" },
  { "clear_mutations", _wrap_Node_clear_mutations, METH_NOARGS, "clear_mutations(self)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__Node_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.Node",                        /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_Node_destructor_closure,     /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__Node_type.as_number,                  /* tp_as_number */
    &SwigPyBuiltin__egglib__Node_type.as_sequence,                /* tp_as_sequence */
    &SwigPyBuiltin__egglib__Node_type.as_mapping,                 /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__Node_type.as_buffer,                  /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::Node",                           /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__Node_richcompare,  /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__Node_methods,      /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__Node_getset,       /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_Node,                           /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__Node_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__Node_type};

static SwigPyGetSet Tree___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__Tree_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"\n"
		"Tree\n"
		"Tree\n"
		"", &Tree___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__Tree_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__Tree_methods[] = {
  { "recomb", _wrap_Tree_recomb, METH_VARARGS, "recomb(self, point, new_tree)" },
  { "coal", _wrap_Tree_coal, METH_VARARGS, "coal(self, nodeIndex1, nodeIndex2, time) -> unsigned int" },
  { "mutate", _wrap_Tree_mutate, METH_VARARGS, "mutate(self, site, data, params) -> unsigned int" },
  { "start", _wrap_Tree_start, METH_NOARGS, "start(self) -> double" },
  { "stop", _wrap_Tree_stop, METH_NOARGS, "stop(self) -> double" },
  { "cov", _wrap_Tree_cov, METH_NOARGS, "cov(self) -> double" },
  { "L", _wrap_Tree_L, METH_NOARGS, "L(self) -> double" },
  { "root", _wrap_Tree_root, METH_NOARGS, "root(self) -> Node" },
  { "nnodes", _wrap_Tree_nnodes, METH_NOARGS, "nnodes(self) -> unsigned int" },
  { "node", _wrap_Tree_node, METH_O, "node(self, i) -> Node" },
  { "addNode", _wrap_Tree_addNode, METH_VARARGS, "addNode(self, t, label) -> unsigned int" },
  { "reset", _wrap_Tree_reset, METH_VARARGS, "reset(self, numberOfLeaves, start, stop)" },
  { "clear_mutations", _wrap_Tree_clear_mutations, METH_NOARGS, "clear_mutations(self)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__Tree_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.Tree",                        /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_Tree_destructor_closure,     /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__Tree_type.as_number,                  /* tp_as_number */
    &SwigPyBuiltin__egglib__Tree_type.as_sequence,                /* tp_as_sequence */
    &SwigPyBuiltin__egglib__Tree_type.as_mapping,                 /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__Tree_type.as_buffer,                  /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::Tree",                           /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__Tree_richcompare,  /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__Tree_methods,      /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__Tree_getset,       /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_Tree,                           /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__Tree_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__Tree_type};

static SwigPyGetSet Coalesce___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__Coalesce_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"Coalesce", &Coalesce___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__Coalesce_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__Coalesce_methods[] = {
  { "simul", _wrap_Coalesce_simul, METH_VARARGS, "simul(self, params, mutate=True)" },
  { "mutate", _wrap_Coalesce_mutate, METH_NOARGS, "mutate(self)" },
  { "coalescence", _wrap_Coalesce_coalescence, METH_VARARGS, "coalescence(self, pop, i, j)" },
  { "migrate", _wrap_Coalesce_migrate, METH_VARARGS, "migrate(self, source, i, dest)" },
  { "admixt", _wrap_Coalesce_admixt, METH_VARARGS, "admixt(self, source, dest, proba)" },
  { "bottleneck", _wrap_Coalesce_bottleneck, METH_VARARGS, "bottleneck(self, pop, duration)" },
  { "number_of_trees", _wrap_Coalesce_number_of_trees, METH_NOARGS, "number_of_trees(self) -> unsigned int" },
  { "delayedSample", _wrap_Coalesce_delayedSample, METH_VARARGS, "delayedSample(self, date, pop, n1, n2)" },
  { "tree", _wrap_Coalesce_tree, METH_O, "tree(self, i) -> Tree" },
  { "data", _wrap_Coalesce_data, METH_NOARGS, "data(self) -> DataHolder" },
  { "tree_start", _wrap_Coalesce_tree_start, METH_O, "tree_start(self, i) -> double" },
  { "tree_stop", _wrap_Coalesce_tree_stop, METH_O, "tree_stop(self, i) -> double" },
  { "site_position", _wrap_Coalesce_site_position, METH_O, "site_position(self, mut) -> double" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__Coalesce_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.Coalesce",                    /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_Coalesce_destructor_closure, /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__Coalesce_type.as_number,              /* tp_as_number */
    &SwigPyBuiltin__egglib__Coalesce_type.as_sequence,            /* tp_as_sequence */
    &SwigPyBuiltin__egglib__Coalesce_type.as_mapping,             /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__Coalesce_type.as_buffer,              /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::Coalesce",                       /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__Coalesce_richcompare,                  /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__Coalesce_methods,  /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__Coalesce_getset,   /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_Coalesce,                       /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__Coalesce_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__Coalesce_type};

static SwigPyGetSet EHH___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__EHH_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"EHH", &EHH___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__EHH_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__EHH_methods[] = {
  { "set_core", _wrap_EHH_set_core, METH_VARARGS, "set_core(self, site, struct_indiv, EHH_thr, EHHc_thr, EHHS_thr, EHHG_thr, min_freq, min_sam, crop)" },
  { "load_distant", _wrap_EHH_load_distant, METH_O, "load_distant(self, site)" },
  { "K_core", _wrap_EHH_K_core, METH_NOARGS, "K_core(self) -> unsigned int" },
  { "K_cur", _wrap_EHH_K_cur, METH_NOARGS, "K_cur(self) -> unsigned int" },
  { "num_avail_tot", _wrap_EHH_num_avail_tot, METH_NOARGS, "num_avail_tot(self) -> unsigned int" },
  { "num_avail_core", _wrap_EHH_num_avail_core, METH_O, "num_avail_core(self, arg2) -> unsigned int" },
  { "num_avail_cur", _wrap_EHH_num_avail_cur, METH_O, "num_avail_cur(self, arg2) -> unsigned int" },
  { "EHHS", _wrap_EHH_EHHS, METH_NOARGS, "EHHS(self) -> double" },
  { "EHHG", _wrap_EHH_EHHG, METH_NOARGS, "EHHG(self) -> double" },
  { "EHHi", _wrap_EHH_EHHi, METH_O, "EHHi(self, haplotype) -> double" },
  { "EHHc", _wrap_EHH_EHHc, METH_O, "EHHc(self, haplotype) -> double" },
  { "rEHH", _wrap_EHH_rEHH, METH_O, "rEHH(self, haplotype) -> double" },
  { "iES", _wrap_EHH_iES, METH_NOARGS, "iES(self) -> double" },
  { "iEG", _wrap_EHH_iEG, METH_NOARGS, "iEG(self) -> double" },
  { "iHH", _wrap_EHH_iHH, METH_O, "iHH(self, haplotype) -> double" },
  { "iHHc", _wrap_EHH_iHHc, METH_O, "iHHc(self, haplotype) -> double" },
  { "iHS", _wrap_EHH_iHS, METH_O, "iHS(self, haplotype) -> double" },
  { "num_EHH_done", _wrap_EHH_num_EHH_done, METH_NOARGS, "num_EHH_done(self) -> unsigned int" },
  { "num_EHHc_done", _wrap_EHH_num_EHHc_done, METH_NOARGS, "num_EHHc_done(self) -> unsigned int" },
  { "flag_EHHS_done", _wrap_EHH_flag_EHHS_done, METH_NOARGS, "flag_EHHS_done(self) -> bool" },
  { "flag_EHHG_done", _wrap_EHH_flag_EHHG_done, METH_NOARGS, "flag_EHHG_done(self) -> bool" },
  { "dEHH", _wrap_EHH_dEHH, METH_O, "dEHH(self, haplotype) -> double" },
  { "dEHHc", _wrap_EHH_dEHHc, METH_O, "dEHHc(self, haplotype) -> double" },
  { "dEHH_max", _wrap_EHH_dEHH_max, METH_NOARGS, "dEHH_max(self) -> double" },
  { "dEHH_mean", _wrap_EHH_dEHH_mean, METH_NOARGS, "dEHH_mean(self) -> double" },
  { "dEHHS", _wrap_EHH_dEHHS, METH_NOARGS, "dEHHS(self) -> double" },
  { "dEHHG", _wrap_EHH_dEHHG, METH_NOARGS, "dEHHG(self) -> double" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__EHH_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.EHH",                         /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_EHH_destructor_closure,      /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__EHH_type.as_number,                   /* tp_as_number */
    &SwigPyBuiltin__egglib__EHH_type.as_sequence,                 /* tp_as_sequence */
    &SwigPyBuiltin__egglib__EHH_type.as_mapping,                  /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__EHH_type.as_buffer,                   /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::EHH",                            /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__EHH_richcompare,   /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__EHH_methods,       /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__EHH_getset,        /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_EHH,                            /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__EHH_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__EHH_type};

static SwigPyGetSet BaseFormatter___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__BaseFormatter_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"BaseFormatter", &BaseFormatter___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__BaseFormatter_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__BaseFormatter_methods[] = {
  { "open_file", _wrap_BaseFormatter_open_file, METH_O, "open_file(self, fname) -> bool" },
  { "to_str", _wrap_BaseFormatter_to_str, METH_NOARGS, "to_str(self)" },
  { "get_str", _wrap_BaseFormatter_get_str, METH_NOARGS, "get_str(self) -> std::string" },
  { "to_cout", _wrap_BaseFormatter_to_cout, METH_NOARGS, "to_cout(self)" },
  { "write", _wrap_BaseFormatter_write, METH_VARARGS, "write(self, bit, eol)" },
  { "flush", _wrap_BaseFormatter_flush, METH_NOARGS, "flush(self)" },
  { "close", _wrap_BaseFormatter_close, METH_NOARGS, "close(self)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__BaseFormatter_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.BaseFormatter",               /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_BaseFormatter_destructor_closure,                /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__BaseFormatter_type.as_number,         /* tp_as_number */
    &SwigPyBuiltin__egglib__BaseFormatter_type.as_sequence,       /* tp_as_sequence */
    &SwigPyBuiltin__egglib__BaseFormatter_type.as_mapping,        /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__BaseFormatter_type.as_buffer,         /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::BaseFormatter",                  /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__BaseFormatter_richcompare,             /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__BaseFormatter_methods,                 /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__BaseFormatter_getset,                  /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_BaseFormatter,                  /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__BaseFormatter_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__BaseFormatter_type};

static SwigPyGetSet Export___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__Export_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"Export", &Export___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__Export_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__Export_methods[] = {
  { "newick", _wrap_Export_newick, METH_VARARGS, "newick(self, tree, blen, eol)" },
  { "ms_num_positions", _wrap_Export_ms_num_positions, METH_O, "ms_num_positions(self, n)" },
  { "ms_position", _wrap_Export_ms_position, METH_VARARGS, "ms_position(self, site, position)" },
  { "ms_auto_positions", _wrap_Export_ms_auto_positions, METH_O, "ms_auto_positions(self, n)" },
  { "ms", _wrap_Export_ms, METH_VARARGS, "ms(self, data, spacer)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__Export_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.Export",                      /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_Export_destructor_closure,   /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__Export_type.as_number,                /* tp_as_number */
    &SwigPyBuiltin__egglib__Export_type.as_sequence,              /* tp_as_sequence */
    &SwigPyBuiltin__egglib__Export_type.as_mapping,               /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__Export_type.as_buffer,                /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::Export",                         /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__Export_richcompare,/* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__Export_methods,    /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__Export_getset,     /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_Export,                         /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__Export_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__Export_type};

static SwigPyGetSet FastaFormatter___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__FastaFormatter_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"FastaFormatter", &FastaFormatter___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__FastaFormatter_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__FastaFormatter_methods[] = {
  { "defaults", _wrap_FastaFormatter_defaults, METH_NOARGS, "defaults(self)" },
  { "set_first", _wrap_FastaFormatter_set_first, METH_VARARGS, "set_first(self, first=0)" },
  { "set_last", _wrap_FastaFormatter_set_last, METH_VARARGS, "set_last(self, last=MAX)" },
  { "set_labels", _wrap_FastaFormatter_set_labels, METH_VARARGS, "set_labels(self, labels=True)" },
  { "set_linelength", _wrap_FastaFormatter_set_linelength, METH_VARARGS, "set_linelength(self, linelength=50)" },
  { "write", _wrap_FastaFormatter_write, METH_VARARGS, "write(self, src, alph)" },
  { "write_string", _wrap_FastaFormatter_write_string, METH_VARARGS, "write_string(self, src, alph) -> std::string" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__FastaFormatter_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.FastaFormatter",              /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_FastaFormatter_destructor_closure,               /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__FastaFormatter_type.as_number,        /* tp_as_number */
    &SwigPyBuiltin__egglib__FastaFormatter_type.as_sequence,      /* tp_as_sequence */
    &SwigPyBuiltin__egglib__FastaFormatter_type.as_mapping,       /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__FastaFormatter_type.as_buffer,        /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::FastaFormatter",                 /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__FastaFormatter_richcompare,            /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__FastaFormatter_methods,                /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__FastaFormatter_getset,                 /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_FastaFormatter,                 /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__FastaFormatter_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__FastaFormatter_type};

static SwigPyGetSet FastaParser___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__FastaParser_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"FastaParser", &FastaParser___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__FastaParser_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__FastaParser_methods[] = {
  { "reserve", _wrap_FastaParser_reserve, METH_VARARGS, "reserve(self, ln, ls, ng, lf)" },
  { "open_file", _wrap_FastaParser_open_file, METH_VARARGS, "open_file(self, fname, arg3, offset=0)" },
  { "set_string", _wrap_FastaParser_set_string, METH_VARARGS, "set_string(self, str, arg3)" },
  { "read_sequence", _wrap_FastaParser_read_sequence, METH_VARARGS, "read_sequence(self, groups, dest=None, label_marker='@', label_separator=' ')" },
  { "read_all", _wrap_FastaParser_read_all, METH_VARARGS, "read_all(self, groups, dest, label_marker='@', label_separator=' ')" },
  { "close", _wrap_FastaParser_close, METH_NOARGS, "close(self)" },
  { "good", _wrap_FastaParser_good, METH_NOARGS, "good(self) -> bool" },
  { "name", _wrap_FastaParser_name, METH_NOARGS, "name(self) -> char const *" },
  { "ls", _wrap_FastaParser_ls, METH_NOARGS, "ls(self) -> unsigned int" },
  { "ch", _wrap_FastaParser_ch, METH_O, "ch(self, index) -> char" },
  { "nlabels", _wrap_FastaParser_nlabels, METH_NOARGS, "nlabels(self) -> unsigned int" },
  { "label", _wrap_FastaParser_label, METH_O, "label(self, index) -> char const *" },
  { "clear", _wrap_FastaParser_clear, METH_NOARGS, "clear(self)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__FastaParser_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.FastaParser",                 /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_FastaParser_destructor_closure,                  /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__FastaParser_type.as_number,           /* tp_as_number */
    &SwigPyBuiltin__egglib__FastaParser_type.as_sequence,         /* tp_as_sequence */
    &SwigPyBuiltin__egglib__FastaParser_type.as_mapping,          /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__FastaParser_type.as_buffer,           /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::FastaParser",                    /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__FastaParser_richcompare,               /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__FastaParser_methods,                   /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__FastaParser_getset,/* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_FastaParser,                    /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__FastaParser_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__FastaParser_type};

static SwigPyGetSet Feature___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__Feature_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"Feature", &Feature___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__Feature_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__Feature_methods[] = {
  { "reset", _wrap_Feature_reset, METH_NOARGS, "reset(self)" },
  { "clear", _wrap_Feature_clear, METH_NOARGS, "clear(self)" },
  { "get_seqid", _wrap_Feature_get_seqid, METH_NOARGS, "get_seqid(self) -> char const *" },
  { "set_seqid", _wrap_Feature_set_seqid, METH_O, "set_seqid(self, str)" },
  { "get_source", _wrap_Feature_get_source, METH_NOARGS, "get_source(self) -> char const *" },
  { "set_source", _wrap_Feature_set_source, METH_O, "set_source(self, str)" },
  { "get_type", _wrap_Feature_get_type, METH_NOARGS, "get_type(self) -> char const *" },
  { "set_type", _wrap_Feature_set_type, METH_O, "set_type(self, str)" },
  { "get_num_fragments", _wrap_Feature_get_num_fragments, METH_NOARGS, "get_num_fragments(self) -> unsigned int" },
  { "set_num_fragments", _wrap_Feature_set_num_fragments, METH_O, "set_num_fragments(self, num)" },
  { "get_start", _wrap_Feature_get_start, METH_O, "get_start(self, i) -> unsigned int" },
  { "set_start", _wrap_Feature_set_start, METH_VARARGS, "set_start(self, i, val)" },
  { "get_end", _wrap_Feature_get_end, METH_O, "get_end(self, i) -> unsigned int" },
  { "set_end", _wrap_Feature_set_end, METH_VARARGS, "set_end(self, i, val)" },
  { "get_score", _wrap_Feature_get_score, METH_NOARGS, "get_score(self) -> double" },
  { "set_score", _wrap_Feature_set_score, METH_O, "set_score(self, d)" },
  { "get_strand", _wrap_Feature_get_strand, METH_NOARGS, "get_strand(self) -> egglib::Feature::STRAND" },
  { "set_strand", _wrap_Feature_set_strand, METH_O, "set_strand(self, s)" },
  { "get_phase", _wrap_Feature_get_phase, METH_NOARGS, "get_phase(self) -> egglib::Feature::PHASE" },
  { "set_phase", _wrap_Feature_set_phase, METH_O, "set_phase(self, p)" },
  { "get_ID", _wrap_Feature_get_ID, METH_NOARGS, "get_ID(self) -> char const *" },
  { "set_ID", _wrap_Feature_set_ID, METH_O, "set_ID(self, str)" },
  { "get_Name", _wrap_Feature_get_Name, METH_NOARGS, "get_Name(self) -> char const *" },
  { "set_Name", _wrap_Feature_set_Name, METH_O, "set_Name(self, str)" },
  { "get_num_Alias", _wrap_Feature_get_num_Alias, METH_NOARGS, "get_num_Alias(self) -> unsigned int" },
  { "set_num_Alias", _wrap_Feature_set_num_Alias, METH_O, "set_num_Alias(self, num)" },
  { "get_Alias", _wrap_Feature_get_Alias, METH_O, "get_Alias(self, i) -> char const *" },
  { "set_Alias", _wrap_Feature_set_Alias, METH_VARARGS, "set_Alias(self, i, s)" },
  { "get_num_Parent", _wrap_Feature_get_num_Parent, METH_NOARGS, "get_num_Parent(self) -> unsigned int" },
  { "set_num_Parent", _wrap_Feature_set_num_Parent, METH_O, "set_num_Parent(self, num)" },
  { "get_Parent", _wrap_Feature_get_Parent, METH_O, "get_Parent(self, i) -> char const *" },
  { "set_Parent", _wrap_Feature_set_Parent, METH_VARARGS, "set_Parent(self, i, s)" },
  { "get_Target", _wrap_Feature_get_Target, METH_NOARGS, "get_Target(self) -> char const *" },
  { "set_Target", _wrap_Feature_set_Target, METH_O, "set_Target(self, str)" },
  { "get_Gap", _wrap_Feature_get_Gap, METH_NOARGS, "get_Gap(self) -> char const *" },
  { "set_Gap", _wrap_Feature_set_Gap, METH_O, "set_Gap(self, str)" },
  { "get_Derives_from", _wrap_Feature_get_Derives_from, METH_NOARGS, "get_Derives_from(self) -> char const *" },
  { "set_Derives_from", _wrap_Feature_set_Derives_from, METH_O, "set_Derives_from(self, str)" },
  { "get_num_Note", _wrap_Feature_get_num_Note, METH_NOARGS, "get_num_Note(self) -> unsigned int" },
  { "set_num_Note", _wrap_Feature_set_num_Note, METH_O, "set_num_Note(self, num)" },
  { "get_Note", _wrap_Feature_get_Note, METH_O, "get_Note(self, i) -> char const *" },
  { "set_Note", _wrap_Feature_set_Note, METH_VARARGS, "set_Note(self, i, s)" },
  { "get_num_Dbxref", _wrap_Feature_get_num_Dbxref, METH_NOARGS, "get_num_Dbxref(self) -> unsigned int" },
  { "set_num_Dbxref", _wrap_Feature_set_num_Dbxref, METH_O, "set_num_Dbxref(self, num)" },
  { "get_Dbxref", _wrap_Feature_get_Dbxref, METH_O, "get_Dbxref(self, i) -> char const *" },
  { "set_Dbxref", _wrap_Feature_set_Dbxref, METH_VARARGS, "set_Dbxref(self, i, s)" },
  { "get_num_Ontology_term", _wrap_Feature_get_num_Ontology_term, METH_NOARGS, "get_num_Ontology_term(self) -> unsigned int" },
  { "set_num_Ontology_term", _wrap_Feature_set_num_Ontology_term, METH_O, "set_num_Ontology_term(self, num)" },
  { "get_Ontology_term", _wrap_Feature_get_Ontology_term, METH_O, "get_Ontology_term(self, i) -> char const *" },
  { "set_Ontology_term", _wrap_Feature_set_Ontology_term, METH_VARARGS, "set_Ontology_term(self, i, s)" },
  { "get_Is_circular", _wrap_Feature_get_Is_circular, METH_NOARGS, "get_Is_circular(self) -> bool" },
  { "set_Is_circular", _wrap_Feature_set_Is_circular, METH_O, "set_Is_circular(self, b)" },
  { "get_num_attributes", _wrap_Feature_get_num_attributes, METH_NOARGS, "get_num_attributes(self) -> unsigned int" },
  { "set_num_attributes", _wrap_Feature_set_num_attributes, METH_O, "set_num_attributes(self, num)" },
  { "get_num_items_attribute", _wrap_Feature_get_num_items_attribute, METH_O, "get_num_items_attribute(self, i) -> unsigned int" },
  { "set_num_items_attribute", _wrap_Feature_set_num_items_attribute, METH_VARARGS, "set_num_items_attribute(self, i, num)" },
  { "get_attribute_key", _wrap_Feature_get_attribute_key, METH_O, "get_attribute_key(self, i) -> char const *" },
  { "set_attribute_key", _wrap_Feature_set_attribute_key, METH_VARARGS, "set_attribute_key(self, i, str)" },
  { "get_attribute_value", _wrap_Feature_get_attribute_value, METH_VARARGS, "get_attribute_value(self, attr, item) -> char const *" },
  { "set_attribute_value", _wrap_Feature_set_attribute_value, METH_VARARGS, "set_attribute_value(self, attr, item, str)" },
  { "get_num_parents", _wrap_Feature_get_num_parents, METH_NOARGS, "get_num_parents(self) -> unsigned int" },
  { "set_num_parents", _wrap_Feature_set_num_parents, METH_O, "set_num_parents(self, num)" },
  { "get_parent", _wrap_Feature_get_parent, METH_O, "get_parent(self, i) -> Feature" },
  { "set_parent", _wrap_Feature_set_parent, METH_VARARGS, "set_parent(self, i, feat)" },
  { "get_num_parts", _wrap_Feature_get_num_parts, METH_NOARGS, "get_num_parts(self) -> unsigned int" },
  { "set_num_parts", _wrap_Feature_set_num_parts, METH_O, "set_num_parts(self, num)" },
  { "get_part", _wrap_Feature_get_part, METH_O, "get_part(self, i) -> Feature" },
  { "set_part", _wrap_Feature_set_part, METH_VARARGS, "set_part(self, i, feat)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__Feature_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.Feature",                     /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_Feature_destructor_closure,  /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__Feature_type.as_number,               /* tp_as_number */
    &SwigPyBuiltin__egglib__Feature_type.as_sequence,             /* tp_as_sequence */
    &SwigPyBuiltin__egglib__Feature_type.as_mapping,              /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__Feature_type.as_buffer,               /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::Feature",                        /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__Feature_richcompare,                   /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__Feature_methods,   /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__Feature_getset,    /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_Feature,                        /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__Feature_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__Feature_type};

static SwigPyGetSet GFF3___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__GFF3_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"GFF3", &GFF3___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__GFF3_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__GFF3_methods[] = {
  { "liberal", _wrap_GFF3_liberal, METH_O, "liberal(self, flag)" },
  { "parse", _wrap_GFF3_parse, METH_VARARGS, "\n"
		"parse(self, fname)\n"
		"parse(self, stream)\n"
		"" },
  { "parse_string", _wrap_GFF3_parse_string, METH_O, "parse_string(self, string)" },
  { "reset", _wrap_GFF3_reset, METH_NOARGS, "reset(self)" },
  { "clear", _wrap_GFF3_clear, METH_NOARGS, "clear(self)" },
  { "num_metadata", _wrap_GFF3_num_metadata, METH_NOARGS, "num_metadata(self) -> unsigned int" },
  { "metadata_key", _wrap_GFF3_metadata_key, METH_O, "metadata_key(self, i) -> char const *" },
  { "metadata_value", _wrap_GFF3_metadata_value, METH_O, "metadata_value(self, i) -> char const *" },
  { "num_genes", _wrap_GFF3_num_genes, METH_NOARGS, "num_genes(self) -> unsigned int" },
  { "gene", _wrap_GFF3_gene, METH_O, "gene(self, i) -> Feature" },
  { "num_features", _wrap_GFF3_num_features, METH_NOARGS, "num_features(self) -> unsigned int" },
  { "feature", _wrap_GFF3_feature, METH_O, "feature(self, i) -> Feature" },
  { "sequences", _wrap_GFF3_sequences, METH_NOARGS, "sequences(self) -> DataHolder" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__GFF3_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.GFF3",                        /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_GFF3_destructor_closure,     /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__GFF3_type.as_number,                  /* tp_as_number */
    &SwigPyBuiltin__egglib__GFF3_type.as_sequence,                /* tp_as_sequence */
    &SwigPyBuiltin__egglib__GFF3_type.as_mapping,                 /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__GFF3_type.as_buffer,                  /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::GFF3",                           /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__GFF3_richcompare,  /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__GFF3_methods,      /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__GFF3_getset,       /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_GFF3,                           /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__GFF3_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__GFF3_type};

static SwigPyGetSet VcfFilter___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__vcf__Filter_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"\n"
		"Filter\n"
		"Filter\n"
		"Filter\n"
		"", &VcfFilter___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__vcf__Filter_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__vcf__Filter_methods[] = {
  { "update", _wrap_VcfFilter_update, METH_VARARGS, "update(self, id, descr)" },
  { "get_ID", _wrap_VcfFilter_get_ID, METH_NOARGS, "get_ID(self) -> char const *" },
  { "set_ID", _wrap_VcfFilter_set_ID, METH_O, "set_ID(self, id)" },
  { "get_description", _wrap_VcfFilter_get_description, METH_NOARGS, "get_description(self) -> char const *" },
  { "set_description", _wrap_VcfFilter_set_description, METH_O, "set_description(self, descr)" },
  { "clear", _wrap_VcfFilter_clear, METH_NOARGS, "clear(self)" },
  { "set_extra", _wrap_VcfFilter_set_extra, METH_VARARGS, "set_extra(self, key, value)" },
  { "get_extra_key", _wrap_VcfFilter_get_extra_key, METH_O, "get_extra_key(self, idx) -> char const *" },
  { "get_extra_value", _wrap_VcfFilter_get_extra_value, METH_O, "get_extra_value(self, idx) -> char const *" },
  { "get_num_extra", _wrap_VcfFilter_get_num_extra, METH_NOARGS, "get_num_extra(self) -> unsigned int" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__vcf__Filter_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.VcfFilter",                   /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_VcfFilter_destructor_closure,/* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__vcf__Filter_type.as_number,           /* tp_as_number */
    &SwigPyBuiltin__egglib__vcf__Filter_type.as_sequence,         /* tp_as_sequence */
    &SwigPyBuiltin__egglib__vcf__Filter_type.as_mapping,          /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__vcf__Filter_type.as_buffer,           /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::vcf::Filter",                    /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__vcf__Filter_richcompare,               /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__vcf__Filter_methods,                   /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__vcf__Filter_getset,/* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_VcfFilter,                      /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__vcf__Filter_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__Filter_type};

static SwigPyGetSet Info___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__vcf__Info_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"\n"
		"Info\n"
		"Info\n"
		"Info\n"
		"", &Info___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__vcf__Info_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__vcf__Info_methods[] = {
  { "update", _wrap_Info_update, METH_VARARGS, "update(self, id, num, t, descr)" },
  { "get_number", _wrap_Info_get_number, METH_NOARGS, "get_number(self) -> unsigned int" },
  { "set_number", _wrap_Info_set_number, METH_O, "set_number(self, num)" },
  { "get_type", _wrap_Info_get_type, METH_NOARGS, "get_type(self) -> egglib::vcf::Info::Type" },
  { "set_type", _wrap_Info_set_type, METH_O, "set_type(self, t)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__vcf__Info_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.Info",                        /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_Info_destructor_closure,     /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__vcf__Info_type.as_number,             /* tp_as_number */
    &SwigPyBuiltin__egglib__vcf__Info_type.as_sequence,           /* tp_as_sequence */
    &SwigPyBuiltin__egglib__vcf__Info_type.as_mapping,            /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__vcf__Info_type.as_buffer,             /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::vcf::Info",                      /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__vcf__Info_richcompare,                 /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__vcf__Info_methods, /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__vcf__Info_getset,  /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_Info,                           /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__vcf__Info_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__Info_type};

static SwigPyGetSet Format___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__vcf__Format_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"\n"
		"Format\n"
		"Format\n"
		"Format\n"
		"Format\n"
		"", &Format___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__vcf__Format_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__vcf__Format_methods[] = {
  { "update", _wrap_Format_update, METH_VARARGS, "update(self, id, num, t, descr)" },
  { "get_type", _wrap_Format_get_type, METH_NOARGS, "get_type(self) -> egglib::vcf::Info::Type" },
  { "set_type", _wrap_Format_set_type, METH_O, "set_type(self, t)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__vcf__Format_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.Format",                      /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_Format_destructor_closure,   /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__vcf__Format_type.as_number,           /* tp_as_number */
    &SwigPyBuiltin__egglib__vcf__Format_type.as_sequence,         /* tp_as_sequence */
    &SwigPyBuiltin__egglib__vcf__Format_type.as_mapping,          /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__vcf__Format_type.as_buffer,           /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::vcf::Format",                    /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__vcf__Format_richcompare,               /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__vcf__Format_methods,                   /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__vcf__Format_getset,/* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_Format,                         /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__vcf__Format_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__Format_type};

static SwigPyGetSet Meta___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__vcf__Meta_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"\n"
		"Meta\n"
		"Meta\n"
		"Meta\n"
		"", &Meta___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__vcf__Meta_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__vcf__Meta_methods[] = {
  { "update", _wrap_Meta_update, METH_VARARGS, "update(self, k, v)" },
  { "set_key", _wrap_Meta_set_key, METH_O, "set_key(self, k)" },
  { "set_value", _wrap_Meta_set_value, METH_O, "set_value(self, v)" },
  { "get_key", _wrap_Meta_get_key, METH_NOARGS, "get_key(self) -> char const *" },
  { "get_value", _wrap_Meta_get_value, METH_NOARGS, "get_value(self) -> char const *" },
  { "clear", _wrap_Meta_clear, METH_NOARGS, "clear(self)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__vcf__Meta_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.Meta",                        /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_Meta_destructor_closure,     /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__vcf__Meta_type.as_number,             /* tp_as_number */
    &SwigPyBuiltin__egglib__vcf__Meta_type.as_sequence,           /* tp_as_sequence */
    &SwigPyBuiltin__egglib__vcf__Meta_type.as_mapping,            /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__vcf__Meta_type.as_buffer,             /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::vcf::Meta",                      /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__vcf__Meta_richcompare,                 /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__vcf__Meta_methods, /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__vcf__Meta_getset,  /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_Meta,                           /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__vcf__Meta_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__Meta_type};

static SwigPyGetSet Alt___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__vcf__Alt_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"\n"
		"Alt\n"
		"Alt\n"
		"Alt\n"
		"Alt\n"
		"", &Alt___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__vcf__Alt_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__vcf__Alt_methods[] = {
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__vcf__Alt_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.Alt",                         /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_Alt_destructor_closure,      /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__vcf__Alt_type.as_number,              /* tp_as_number */
    &SwigPyBuiltin__egglib__vcf__Alt_type.as_sequence,            /* tp_as_sequence */
    &SwigPyBuiltin__egglib__vcf__Alt_type.as_mapping,             /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__vcf__Alt_type.as_buffer,              /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::vcf::Alt",                       /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__vcf__Alt_richcompare,                  /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__vcf__Alt_methods,  /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__vcf__Alt_getset,   /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_Alt,                            /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__vcf__Alt_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__Alt_type};

static SwigPyGetSet FlagInfo___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__vcf__FlagInfo_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"\n"
		"FlagInfo\n"
		"FlagInfo\n"
		"", &FlagInfo___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__vcf__FlagInfo_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__vcf__FlagInfo_methods[] = {
  { "set_ID", _wrap_FlagInfo_set_ID, METH_O, "set_ID(self, id)" },
  { "get_ID", _wrap_FlagInfo_get_ID, METH_NOARGS, "get_ID(self) -> char const *" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__vcf__FlagInfo_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.FlagInfo",                    /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_FlagInfo_destructor_closure, /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__vcf__FlagInfo_type.as_number,         /* tp_as_number */
    &SwigPyBuiltin__egglib__vcf__FlagInfo_type.as_sequence,       /* tp_as_sequence */
    &SwigPyBuiltin__egglib__vcf__FlagInfo_type.as_mapping,        /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__vcf__FlagInfo_type.as_buffer,         /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::vcf::FlagInfo",                  /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__vcf__FlagInfo_richcompare,             /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__vcf__FlagInfo_methods,                 /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__vcf__FlagInfo_getset,                  /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_FlagInfo,                       /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__vcf__FlagInfo_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__FlagInfo_type};

static SwigPyGetSet TypeInfoString___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__vcf__TypeInfoT_char_p_t_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"\n"
		"TypeInfo<(p.char)>\n"
		"TypeInfoString\n"
		"", &TypeInfoString___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__vcf__TypeInfoT_char_p_t_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__vcf__TypeInfoT_char_p_t_methods[] = {
  { "num_items", _wrap_TypeInfoString_num_items, METH_NOARGS, "num_items(self) -> unsigned int" },
  { "item", _wrap_TypeInfoString_item, METH_O, "item(self, i) -> char *const &" },
  { "reset", _wrap_TypeInfoString_reset, METH_NOARGS, "reset(self)" },
  { "get_expected_number", _wrap_TypeInfoString_get_expected_number, METH_NOARGS, "get_expected_number(self) -> unsigned int" },
  { "set_expected_number", _wrap_TypeInfoString_set_expected_number, METH_O, "set_expected_number(self, arg2)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__vcf__TypeInfoT_char_p_t_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.TypeInfoString",              /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_TypeInfoString_destructor_closure,               /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__vcf__TypeInfoT_char_p_t_type.as_number,/* tp_as_number */
    &SwigPyBuiltin__egglib__vcf__TypeInfoT_char_p_t_type.as_sequence,/* tp_as_sequence */
    &SwigPyBuiltin__egglib__vcf__TypeInfoT_char_p_t_type.as_mapping,/* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__vcf__TypeInfoT_char_p_t_type.as_buffer,/* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::vcf::TypeInfo< char * >",        /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__vcf__TypeInfoT_char_p_t_richcompare,   /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__vcf__TypeInfoT_char_p_t_methods,       /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__vcf__TypeInfoT_char_p_t_getset,        /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_TypeInfoString,                 /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__vcf__TypeInfoT_char_p_t_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__TypeInfoT_char_p_t_type};

static SwigPyGetSet TypeInfoInt___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__vcf__TypeInfoT_int_t_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"\n"
		"TypeInfo<(int)>\n"
		"TypeInfoInt\n"
		"", &TypeInfoInt___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__vcf__TypeInfoT_int_t_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__vcf__TypeInfoT_int_t_methods[] = {
  { "num_items", _wrap_TypeInfoInt_num_items, METH_NOARGS, "num_items(self) -> unsigned int" },
  { "item", _wrap_TypeInfoInt_item, METH_O, "item(self, i) -> int const &" },
  { "reset", _wrap_TypeInfoInt_reset, METH_NOARGS, "reset(self)" },
  { "get_expected_number", _wrap_TypeInfoInt_get_expected_number, METH_NOARGS, "get_expected_number(self) -> unsigned int" },
  { "set_expected_number", _wrap_TypeInfoInt_set_expected_number, METH_O, "set_expected_number(self, arg2)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__vcf__TypeInfoT_int_t_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.TypeInfoInt",                 /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_TypeInfoInt_destructor_closure,                  /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__vcf__TypeInfoT_int_t_type.as_number,  /* tp_as_number */
    &SwigPyBuiltin__egglib__vcf__TypeInfoT_int_t_type.as_sequence,/* tp_as_sequence */
    &SwigPyBuiltin__egglib__vcf__TypeInfoT_int_t_type.as_mapping, /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__vcf__TypeInfoT_int_t_type.as_buffer,  /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::vcf::TypeInfo< int >",           /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__vcf__TypeInfoT_int_t_richcompare,      /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__vcf__TypeInfoT_int_t_methods,          /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__vcf__TypeInfoT_int_t_getset,           /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_TypeInfoInt,                    /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__vcf__TypeInfoT_int_t_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__TypeInfoT_int_t_type};

static SwigPyGetSet TypeInfoChar___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__vcf__TypeInfoT_char_t_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"\n"
		"TypeInfo<(char)>\n"
		"TypeInfoChar\n"
		"", &TypeInfoChar___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__vcf__TypeInfoT_char_t_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__vcf__TypeInfoT_char_t_methods[] = {
  { "num_items", _wrap_TypeInfoChar_num_items, METH_NOARGS, "num_items(self) -> unsigned int" },
  { "item", _wrap_TypeInfoChar_item, METH_O, "item(self, i) -> char const &" },
  { "reset", _wrap_TypeInfoChar_reset, METH_NOARGS, "reset(self)" },
  { "get_expected_number", _wrap_TypeInfoChar_get_expected_number, METH_NOARGS, "get_expected_number(self) -> unsigned int" },
  { "set_expected_number", _wrap_TypeInfoChar_set_expected_number, METH_O, "set_expected_number(self, arg2)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__vcf__TypeInfoT_char_t_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.TypeInfoChar",                /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_TypeInfoChar_destructor_closure,                 /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__vcf__TypeInfoT_char_t_type.as_number, /* tp_as_number */
    &SwigPyBuiltin__egglib__vcf__TypeInfoT_char_t_type.as_sequence,/* tp_as_sequence */
    &SwigPyBuiltin__egglib__vcf__TypeInfoT_char_t_type.as_mapping,/* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__vcf__TypeInfoT_char_t_type.as_buffer, /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::vcf::TypeInfo< char >",          /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__vcf__TypeInfoT_char_t_richcompare,     /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__vcf__TypeInfoT_char_t_methods,         /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__vcf__TypeInfoT_char_t_getset,          /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_TypeInfoChar,                   /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__vcf__TypeInfoT_char_t_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__TypeInfoT_char_t_type};

static SwigPyGetSet TypeInfoDouble___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__vcf__TypeInfoT_double_t_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"\n"
		"TypeInfo<(double)>\n"
		"TypeInfoDouble\n"
		"", &TypeInfoDouble___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__vcf__TypeInfoT_double_t_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__vcf__TypeInfoT_double_t_methods[] = {
  { "num_items", _wrap_TypeInfoDouble_num_items, METH_NOARGS, "num_items(self) -> unsigned int" },
  { "item", _wrap_TypeInfoDouble_item, METH_O, "item(self, i) -> double const &" },
  { "reset", _wrap_TypeInfoDouble_reset, METH_NOARGS, "reset(self)" },
  { "get_expected_number", _wrap_TypeInfoDouble_get_expected_number, METH_NOARGS, "get_expected_number(self) -> unsigned int" },
  { "set_expected_number", _wrap_TypeInfoDouble_set_expected_number, METH_O, "set_expected_number(self, arg2)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__vcf__TypeInfoT_double_t_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.TypeInfoDouble",              /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_TypeInfoDouble_destructor_closure,               /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__vcf__TypeInfoT_double_t_type.as_number,/* tp_as_number */
    &SwigPyBuiltin__egglib__vcf__TypeInfoT_double_t_type.as_sequence,/* tp_as_sequence */
    &SwigPyBuiltin__egglib__vcf__TypeInfoT_double_t_type.as_mapping,/* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__vcf__TypeInfoT_double_t_type.as_buffer,/* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::vcf::TypeInfo< double >",        /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__vcf__TypeInfoT_double_t_richcompare,   /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__vcf__TypeInfoT_double_t_methods,       /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__vcf__TypeInfoT_double_t_getset,        /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_TypeInfoDouble,                 /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__vcf__TypeInfoT_double_t_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__TypeInfoT_double_t_type};

static SwigPyGetSet StringInfo___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__vcf__StringInfo_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"\n"
		"StringInfo\n"
		"StringInfo\n"
		"", &StringInfo___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__vcf__StringInfo_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__vcf__StringInfo_methods[] = {
  { "item", _wrap_StringInfo_item, METH_O, "item(self, item) -> char const *" },
  { "change", _wrap_StringInfo_change, METH_VARARGS, "change(self, item, position, value)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__vcf__StringInfo_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.StringInfo",                  /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_StringInfo_destructor_closure,                   /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__vcf__StringInfo_type.as_number,       /* tp_as_number */
    &SwigPyBuiltin__egglib__vcf__StringInfo_type.as_sequence,     /* tp_as_sequence */
    &SwigPyBuiltin__egglib__vcf__StringInfo_type.as_mapping,      /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__vcf__StringInfo_type.as_buffer,       /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::vcf::StringInfo",                /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__vcf__StringInfo_richcompare,           /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__vcf__StringInfo_methods,               /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__vcf__StringInfo_getset,                /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_StringInfo,                     /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__vcf__StringInfo_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__StringInfo_type};

static SwigPyGetSet SampleInfo___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__vcf__SampleInfo_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"\n"
		"SampleInfo\n"
		"SampleInfo\n"
		"", &SampleInfo___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__vcf__SampleInfo_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__vcf__SampleInfo_methods[] = {
  { "reset", _wrap_SampleInfo_reset, METH_NOARGS, "reset(self)" },
  { "clear", _wrap_SampleInfo_clear, METH_NOARGS, "clear(self)" },
  { "num_IntegerEntries", _wrap_SampleInfo_num_IntegerEntries, METH_NOARGS, "num_IntegerEntries(self) -> unsigned int" },
  { "num_IntegerItems", _wrap_SampleInfo_num_IntegerItems, METH_O, "num_IntegerItems(self, i) -> unsigned int" },
  { "IntegerItem", _wrap_SampleInfo_IntegerItem, METH_VARARGS, "IntegerItem(self, i, j) -> int" },
  { "num_FloatEntries", _wrap_SampleInfo_num_FloatEntries, METH_NOARGS, "num_FloatEntries(self) -> unsigned int" },
  { "num_FloatItems", _wrap_SampleInfo_num_FloatItems, METH_O, "num_FloatItems(self, i) -> unsigned int" },
  { "FloatItem", _wrap_SampleInfo_FloatItem, METH_VARARGS, "FloatItem(self, i, j) -> double" },
  { "num_CharacterEntries", _wrap_SampleInfo_num_CharacterEntries, METH_NOARGS, "num_CharacterEntries(self) -> unsigned int" },
  { "num_CharacterItems", _wrap_SampleInfo_num_CharacterItems, METH_O, "num_CharacterItems(self, i) -> unsigned int" },
  { "CharacterItem", _wrap_SampleInfo_CharacterItem, METH_VARARGS, "CharacterItem(self, i, j) -> char" },
  { "num_StringEntries", _wrap_SampleInfo_num_StringEntries, METH_NOARGS, "num_StringEntries(self) -> unsigned int" },
  { "num_StringItems", _wrap_SampleInfo_num_StringItems, METH_O, "num_StringItems(self, i) -> unsigned int" },
  { "StringItem", _wrap_SampleInfo_StringItem, METH_VARARGS, "StringItem(self, i, j) -> char const *" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__vcf__SampleInfo_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.SampleInfo",                  /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_SampleInfo_destructor_closure,                   /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__vcf__SampleInfo_type.as_number,       /* tp_as_number */
    &SwigPyBuiltin__egglib__vcf__SampleInfo_type.as_sequence,     /* tp_as_sequence */
    &SwigPyBuiltin__egglib__vcf__SampleInfo_type.as_mapping,      /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__vcf__SampleInfo_type.as_buffer,       /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::vcf::SampleInfo",                /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__vcf__SampleInfo_richcompare,           /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__vcf__SampleInfo_methods,               /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__vcf__SampleInfo_getset,                /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_SampleInfo,                     /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__vcf__SampleInfo_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__SampleInfo_type};

static SwigPyGetSet VcfParser___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__VcfParser_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"VcfParser", &VcfParser___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__VcfParser_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__VcfParser_methods[] = {
  { "open_file", _wrap_VcfParser_open_file, METH_O, "open_file(self, fname)" },
  { "set_stream", _wrap_VcfParser_set_stream, METH_O, "set_stream(self, stream)" },
  { "read_header", _wrap_VcfParser_read_header, METH_O, "read_header(self, arg2)" },
  { "set_threshold_PL", _wrap_VcfParser_set_threshold_PL, METH_O, "set_threshold_PL(self, arg2)" },
  { "set_threshold_GL", _wrap_VcfParser_set_threshold_GL, METH_O, "set_threshold_GL(self, arg2)" },
  { "get_threshold_PL", _wrap_VcfParser_get_threshold_PL, METH_NOARGS, "get_threshold_PL(self) -> unsigned int" },
  { "get_threshold_GL", _wrap_VcfParser_get_threshold_GL, METH_NOARGS, "get_threshold_GL(self) -> unsigned int" },
  { "allow_X", _wrap_VcfParser_allow_X, METH_O, "allow_X(self, arg2)" },
  { "allow_gap", _wrap_VcfParser_allow_gap, METH_O, "allow_gap(self, arg2)" },
  { "read", _wrap_VcfParser_read, METH_NOARGS, "read(self)" },
  { "readline", _wrap_VcfParser_readline, METH_O, "readline(self, arg2)" },
  { "get_currline", _wrap_VcfParser_get_currline, METH_NOARGS, "get_currline(self) -> unsigned long" },
  { "good", _wrap_VcfParser_good, METH_NOARGS, "good(self) -> bool" },
  { "has_data", _wrap_VcfParser_has_data, METH_NOARGS, "has_data(self) -> bool" },
  { "reset", _wrap_VcfParser_reset, METH_NOARGS, "reset(self)" },
  { "reset_variant", _wrap_VcfParser_reset_variant, METH_NOARGS, "reset_variant(self)" },
  { "clear", _wrap_VcfParser_clear, METH_NOARGS, "clear(self)" },
  { "file_format", _wrap_VcfParser_file_format, METH_NOARGS, "file_format(self) -> char const *" },
  { "num_filter", _wrap_VcfParser_num_filter, METH_NOARGS, "num_filter(self) -> unsigned int" },
  { "get_filter", _wrap_VcfParser_get_filter, METH_O, "get_filter(self, i) -> VcfFilter" },
  { "num_info", _wrap_VcfParser_num_info, METH_NOARGS, "num_info(self) -> unsigned int" },
  { "get_info", _wrap_VcfParser_get_info, METH_O, "get_info(self, i) -> Info" },
  { "num_format", _wrap_VcfParser_num_format, METH_NOARGS, "num_format(self) -> unsigned int" },
  { "get_format", _wrap_VcfParser_get_format, METH_O, "get_format(self, i) -> Format" },
  { "num_meta", _wrap_VcfParser_num_meta, METH_NOARGS, "num_meta(self) -> unsigned int" },
  { "get_meta", _wrap_VcfParser_get_meta, METH_O, "get_meta(self, i) -> Meta" },
  { "num_alt", _wrap_VcfParser_num_alt, METH_NOARGS, "num_alt(self) -> unsigned int" },
  { "get_alt", _wrap_VcfParser_get_alt, METH_O, "get_alt(self, i) -> Alt" },
  { "add_filter", _wrap_VcfParser_add_filter, METH_VARARGS, "add_filter(self, id, descr)" },
  { "add_alt", _wrap_VcfParser_add_alt, METH_VARARGS, "add_alt(self, id, descr)" },
  { "add_info", _wrap_VcfParser_add_info, METH_VARARGS, "add_info(self, id, num, type, descr)" },
  { "add_format", _wrap_VcfParser_add_format, METH_VARARGS, "add_format(self, id, num, type, descr)" },
  { "add_meta", _wrap_VcfParser_add_meta, METH_VARARGS, "add_meta(self, key, val)" },
  { "num_samples", _wrap_VcfParser_num_samples, METH_NOARGS, "num_samples(self) -> unsigned int" },
  { "get_sample", _wrap_VcfParser_get_sample, METH_O, "get_sample(self, i) -> char const *" },
  { "find_filter", _wrap_VcfParser_find_filter, METH_O, "find_filter(self, id) -> VcfFilter" },
  { "find_format", _wrap_VcfParser_find_format, METH_O, "find_format(self, id) -> Format" },
  { "find_info", _wrap_VcfParser_find_info, METH_O, "find_info(self, id) -> Info" },
  { "find_meta", _wrap_VcfParser_find_meta, METH_O, "find_meta(self, key) -> Meta" },
  { "find_alt", _wrap_VcfParser_find_alt, METH_O, "find_alt(self, id) -> Alt" },
  { "chromosome", _wrap_VcfParser_chromosome, METH_NOARGS, "chromosome(self) -> char const *" },
  { "position", _wrap_VcfParser_position, METH_NOARGS, "position(self) -> unsigned long" },
  { "num_ID", _wrap_VcfParser_num_ID, METH_NOARGS, "num_ID(self) -> unsigned int" },
  { "ID", _wrap_VcfParser_ID, METH_O, "ID(self, i) -> char const *" },
  { "len_reference", _wrap_VcfParser_len_reference, METH_NOARGS, "len_reference(self) -> unsigned int" },
  { "reference", _wrap_VcfParser_reference, METH_NOARGS, "reference(self) -> char const *" },
  { "num_alternate", _wrap_VcfParser_num_alternate, METH_NOARGS, "num_alternate(self) -> unsigned int" },
  { "alternate_type", _wrap_VcfParser_alternate_type, METH_O, "alternate_type(self, i) -> egglib::vcf::AltType" },
  { "alternate", _wrap_VcfParser_alternate, METH_O, "alternate(self, i) -> char const *" },
  { "quality", _wrap_VcfParser_quality, METH_NOARGS, "quality(self) -> double" },
  { "num_failed_tests", _wrap_VcfParser_num_failed_tests, METH_NOARGS, "num_failed_tests(self) -> unsigned int" },
  { "failed_test", _wrap_VcfParser_failed_test, METH_O, "failed_test(self, i) -> char const *" },
  { "num_FlagInfo", _wrap_VcfParser_num_FlagInfo, METH_NOARGS, "num_FlagInfo(self) -> unsigned int" },
  { "FlagInfo", _wrap_VcfParser_FlagInfo, METH_O, "FlagInfo(self, arg2) -> FlagInfo" },
  { "num_IntegerInfo", _wrap_VcfParser_num_IntegerInfo, METH_NOARGS, "num_IntegerInfo(self) -> unsigned int" },
  { "IntegerInfo", _wrap_VcfParser_IntegerInfo, METH_O, "IntegerInfo(self, arg2) -> TypeInfoInt" },
  { "num_FloatInfo", _wrap_VcfParser_num_FloatInfo, METH_NOARGS, "num_FloatInfo(self) -> unsigned int" },
  { "FloatInfo", _wrap_VcfParser_FloatInfo, METH_O, "FloatInfo(self, arg2) -> TypeInfoDouble" },
  { "num_CharacterInfo", _wrap_VcfParser_num_CharacterInfo, METH_NOARGS, "num_CharacterInfo(self) -> unsigned int" },
  { "CharacterInfo", _wrap_VcfParser_CharacterInfo, METH_O, "CharacterInfo(self, arg2) -> TypeInfoChar" },
  { "num_StringInfo", _wrap_VcfParser_num_StringInfo, METH_NOARGS, "num_StringInfo(self) -> unsigned int" },
  { "StringInfo", _wrap_VcfParser_StringInfo, METH_O, "StringInfo(self, arg2) -> StringInfo" },
  { "num_fields", _wrap_VcfParser_num_fields, METH_NOARGS, "num_fields(self) -> unsigned int" },
  { "field", _wrap_VcfParser_field, METH_O, "field(self, i) -> Format" },
  { "field_index", _wrap_VcfParser_field_index, METH_O, "field_index(self, ID) -> unsigned int" },
  { "field_rank", _wrap_VcfParser_field_rank, METH_VARARGS, "\n"
		"field_rank(self, i) -> unsigned int\n"
		"field_rank(self, ID) -> unsigned int\n"
		"" },
  { "sample_info", _wrap_VcfParser_sample_info, METH_O, "sample_info(self, i) -> SampleInfo" },
  { "has_AN", _wrap_VcfParser_has_AN, METH_NOARGS, "has_AN(self) -> bool" },
  { "AN", _wrap_VcfParser_AN, METH_NOARGS, "AN(self) -> unsigned int" },
  { "has_AA", _wrap_VcfParser_has_AA, METH_NOARGS, "has_AA(self) -> bool" },
  { "AA_string", _wrap_VcfParser_AA_string, METH_NOARGS, "AA_string(self) -> char const *" },
  { "AA_index", _wrap_VcfParser_AA_index, METH_NOARGS, "AA_index(self) -> unsigned int" },
  { "has_AC", _wrap_VcfParser_has_AC, METH_NOARGS, "has_AC(self) -> bool" },
  { "num_AC", _wrap_VcfParser_num_AC, METH_NOARGS, "num_AC(self) -> unsigned int" },
  { "AC", _wrap_VcfParser_AC, METH_O, "AC(self, i) -> unsigned int" },
  { "has_AF", _wrap_VcfParser_has_AF, METH_NOARGS, "has_AF(self) -> bool" },
  { "num_AF", _wrap_VcfParser_num_AF, METH_NOARGS, "num_AF(self) -> unsigned int" },
  { "AF", _wrap_VcfParser_AF, METH_O, "AF(self, i) -> double" },
  { "has_GT", _wrap_VcfParser_has_GT, METH_NOARGS, "has_GT(self) -> bool" },
  { "ploidy", _wrap_VcfParser_ploidy, METH_NOARGS, "ploidy(self) -> unsigned int" },
  { "num_genotypes", _wrap_VcfParser_num_genotypes, METH_NOARGS, "num_genotypes(self) -> unsigned int" },
  { "GT_phased", _wrap_VcfParser_GT_phased, METH_VARARGS, "\n"
		"GT_phased(self, i) -> bool\n"
		"GT_phased(self) -> bool\n"
		"" },
  { "GT", _wrap_VcfParser_GT, METH_VARARGS, "GT(self, sample, allele) -> unsigned int" },
  { "has_PL", _wrap_VcfParser_has_PL, METH_NOARGS, "has_PL(self) -> bool" },
  { "has_GL", _wrap_VcfParser_has_GL, METH_NOARGS, "has_GL(self) -> bool" },
  { "PL", _wrap_VcfParser_PL, METH_VARARGS, "PL(self, sample, genotype) -> unsigned int" },
  { "GL", _wrap_VcfParser_GL, METH_VARARGS, "GL(self, sample, genotype) -> double" },
  { "unread", _wrap_VcfParser_unread, METH_NOARGS, "unread(self)" },
  { "get_index", _wrap_VcfParser_get_index, METH_NOARGS, "get_index(self) -> VcfIndex" },
  { "rewind", _wrap_VcfParser_rewind, METH_NOARGS, "rewind(self)" },
  { "type_alleles", _wrap_VcfParser_type_alleles, METH_NOARGS, "type_alleles(self) -> unsigned int" },
  { "set_alleles", _wrap_VcfParser_set_alleles, METH_O, "set_alleles(self, arg2)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__VcfParser_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.VcfParser",                   /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_VcfParser_destructor_closure,/* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__VcfParser_type.as_number,             /* tp_as_number */
    &SwigPyBuiltin__egglib__VcfParser_type.as_sequence,           /* tp_as_sequence */
    &SwigPyBuiltin__egglib__VcfParser_type.as_mapping,            /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__VcfParser_type.as_buffer,             /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::VcfParser",                      /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__VcfParser_richcompare,                 /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__VcfParser_methods, /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__VcfParser_getset,  /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_VcfParser,                      /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__VcfParser_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__VcfParser_type};

static SwigPyGetSet VcfIndex___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__VcfIndex_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"go_to", &VcfIndex___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__VcfIndex_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__VcfIndex_methods[] = {
  { "go_to", _wrap_VcfIndex_go_to, METH_VARARGS, "go_to(self, arg2, arg3)" },
  { "num", _wrap_VcfIndex_num, METH_NOARGS, "num(self) -> unsigned long" },
  { "has_data", _wrap_VcfIndex_has_data, METH_NOARGS, "has_data(self) -> bool" },
  { "load_data", _wrap_VcfIndex_load_data, METH_VARARGS, "load_data(self, arg2, fname)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__VcfIndex_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.VcfIndex",                    /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_VcfIndex_destructor_closure, /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__VcfIndex_type.as_number,              /* tp_as_number */
    &SwigPyBuiltin__egglib__VcfIndex_type.as_sequence,            /* tp_as_sequence */
    &SwigPyBuiltin__egglib__VcfIndex_type.as_mapping,             /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__VcfIndex_type.as_buffer,              /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::VcfIndex",                       /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__VcfIndex_richcompare,                  /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__VcfIndex_methods,  /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__VcfIndex_getset,   /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_VcfIndex,                       /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__VcfIndex_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__VcfIndex_type};

static SwigPyGetSet WSite___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__WSite_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"site", &WSite___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__WSite_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__WSite_methods[] = {
  { "site", _wrap_WSite_site, METH_NOARGS, "site(self) -> SiteHolder" },
  { "alphabet", _wrap_WSite_alphabet, METH_NOARGS, "alphabet(self) -> StringAlphabet" },
  { "get_pos", _wrap_WSite_get_pos, METH_NOARGS, "get_pos(self) -> unsigned int" },
  { "next", _wrap_WSite_next, METH_NOARGS, "next(self) -> WSite" },
  { "prev", _wrap_WSite_prev, METH_NOARGS, "prev(self) -> WSite" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__WSite_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.WSite",                       /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_WSite_destructor_closure,    /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__WSite_type.as_number,                 /* tp_as_number */
    &SwigPyBuiltin__egglib__WSite_type.as_sequence,               /* tp_as_sequence */
    &SwigPyBuiltin__egglib__WSite_type.as_mapping,                /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__WSite_type.as_buffer,                 /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::WSite",                          /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__WSite_richcompare, /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__WSite_methods,     /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__WSite_getset,      /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    SwigPyBuiltin_BadInit,                    /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__WSite_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__WSite_type};

static SwigPyGetSet VcfWindowBase___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__VcfWindowBase_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"~VcfWindowBase", &VcfWindowBase___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__VcfWindowBase_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__VcfWindowBase_methods[] = {
  { "setup_base", _wrap_VcfWindowBase_setup_base, METH_VARARGS, "setup_base(self, arg2, max_missing, mask)" },
  { "chromosome", _wrap_VcfWindowBase_chromosome, METH_NOARGS, "chromosome(self) -> char const *" },
  { "num_samples", _wrap_VcfWindowBase_num_samples, METH_NOARGS, "num_samples(self) -> unsigned int" },
  { "num_sites", _wrap_VcfWindowBase_num_sites, METH_NOARGS, "num_sites(self) -> unsigned int" },
  { "win_start", _wrap_VcfWindowBase_win_start, METH_NOARGS, "win_start(self) -> unsigned long" },
  { "win_stop", _wrap_VcfWindowBase_win_stop, METH_NOARGS, "win_stop(self) -> unsigned long" },
  { "first_site", _wrap_VcfWindowBase_first_site, METH_NOARGS, "first_site(self) -> WSite" },
  { "last_site", _wrap_VcfWindowBase_last_site, METH_NOARGS, "last_site(self) -> WSite" },
  { "good", _wrap_VcfWindowBase_good, METH_NOARGS, "good(self) -> bool" },
  { "get_site", _wrap_VcfWindowBase_get_site, METH_O, "get_site(self, arg2) -> WSite" },
  { "next_window", _wrap_VcfWindowBase_next_window, METH_NOARGS, "next_window(self)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__VcfWindowBase_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.VcfWindowBase",               /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_VcfWindowBase_destructor_closure,                /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__VcfWindowBase_type.as_number,         /* tp_as_number */
    &SwigPyBuiltin__egglib__VcfWindowBase_type.as_sequence,       /* tp_as_sequence */
    &SwigPyBuiltin__egglib__VcfWindowBase_type.as_mapping,        /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__VcfWindowBase_type.as_buffer,         /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::VcfWindowBase",                  /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__VcfWindowBase_richcompare,             /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__VcfWindowBase_methods,                 /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__VcfWindowBase_getset,                  /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    SwigPyBuiltin_BadInit,                    /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__VcfWindowBase_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__VcfWindowBase_type};

static SwigPyGetSet VcfWindowSliderBase___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__VcfWindowSliderBase_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"~VcfWindowSliderBase", &VcfWindowSliderBase___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__VcfWindowSliderBase_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__VcfWindowSliderBase_methods[] = {
  { "setup", _wrap_VcfWindowSliderBase_setup, METH_VARARGS, "setup(self, parser, wsize, wstep, start_pos, stop_pos, max_missing, mask)" },
  { "next_window", _wrap_VcfWindowSliderBase_next_window, METH_NOARGS, "next_window(self)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__VcfWindowSliderBase_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.VcfWindowSliderBase",         /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_VcfWindowSliderBase_destructor_closure,          /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__VcfWindowSliderBase_type.as_number,   /* tp_as_number */
    &SwigPyBuiltin__egglib__VcfWindowSliderBase_type.as_sequence, /* tp_as_sequence */
    &SwigPyBuiltin__egglib__VcfWindowSliderBase_type.as_mapping,  /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__VcfWindowSliderBase_type.as_buffer,   /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::VcfWindowSliderBase",            /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__VcfWindowSliderBase_richcompare,       /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__VcfWindowSliderBase_methods,           /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__VcfWindowSliderBase_getset,            /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    SwigPyBuiltin_BadInit,                    /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__VcfWindowSliderBase_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__VcfWindowSliderBase_type};

static SwigPyGetSet VcfWindowSlider___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__VcfWindowSlider_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"VcfWindowSlider", &VcfWindowSlider___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__VcfWindowSlider_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__VcfWindowSlider_methods[] = {
  { "setup", _wrap_VcfWindowSlider_setup, METH_VARARGS, "setup(self, parser, wsize, wstep, start_pos, stop_pos, max_missing, mask)" },
  { "next_window", _wrap_VcfWindowSlider_next_window, METH_NOARGS, "next_window(self)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__VcfWindowSlider_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.VcfWindowSlider",             /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_VcfWindowSlider_destructor_closure,              /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__VcfWindowSlider_type.as_number,       /* tp_as_number */
    &SwigPyBuiltin__egglib__VcfWindowSlider_type.as_sequence,     /* tp_as_sequence */
    &SwigPyBuiltin__egglib__VcfWindowSlider_type.as_mapping,      /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__VcfWindowSlider_type.as_buffer,       /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::VcfWindowSlider",                /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__VcfWindowSlider_richcompare,           /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__VcfWindowSlider_methods,               /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__VcfWindowSlider_getset,                /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_VcfWindowSlider,                /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__VcfWindowSlider_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__VcfWindowSlider_type};

static SwigPyGetSet VcfWindowSliderPerSite___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__VcfWindowSliderPerSite_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"VcfWindowSliderPerSite", &VcfWindowSliderPerSite___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__VcfWindowSliderPerSite_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__VcfWindowSliderPerSite_methods[] = {
  { "next_window", _wrap_VcfWindowSliderPerSite_next_window, METH_NOARGS, "next_window(self)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__VcfWindowSliderPerSite_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.VcfWindowSliderPerSite",      /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_VcfWindowSliderPerSite_destructor_closure,       /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__VcfWindowSliderPerSite_type.as_number,/* tp_as_number */
    &SwigPyBuiltin__egglib__VcfWindowSliderPerSite_type.as_sequence,/* tp_as_sequence */
    &SwigPyBuiltin__egglib__VcfWindowSliderPerSite_type.as_mapping,/* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__VcfWindowSliderPerSite_type.as_buffer,/* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::VcfWindowSliderPerSite",         /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__VcfWindowSliderPerSite_richcompare,    /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__VcfWindowSliderPerSite_methods,        /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__VcfWindowSliderPerSite_getset,         /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_VcfWindowSliderPerSite,         /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__VcfWindowSliderPerSite_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__VcfWindowSliderPerSite_type};

static SwigPyGetSet VcfWindowBED___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__VcfWindowBED_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"VcfWindowBED", &VcfWindowBED___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__VcfWindowBED_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__VcfWindowBED_methods[] = {
  { "setup", _wrap_VcfWindowBED_setup, METH_VARARGS, "setup(self, parser, arg3, max_missing, mask)" },
  { "next_window", _wrap_VcfWindowBED_next_window, METH_NOARGS, "next_window(self)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__VcfWindowBED_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.VcfWindowBED",                /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_VcfWindowBED_destructor_closure,                 /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__VcfWindowBED_type.as_number,          /* tp_as_number */
    &SwigPyBuiltin__egglib__VcfWindowBED_type.as_sequence,        /* tp_as_sequence */
    &SwigPyBuiltin__egglib__VcfWindowBED_type.as_mapping,         /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__VcfWindowBED_type.as_buffer,          /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::VcfWindowBED",                   /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__VcfWindowBED_richcompare,              /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__VcfWindowBED_methods,                  /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__VcfWindowBED_getset,                   /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_VcfWindowBED,                   /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__VcfWindowBED_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__VcfWindowBED_type};

static SwigPyGetSet BedParser___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__BedParser_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"BedParser", &BedParser___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__BedParser_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__BedParser_methods[] = {
  { "get_bed_file", _wrap_BedParser_get_bed_file, METH_O, "get_bed_file(self, filename)" },
  { "n_bed_data", _wrap_BedParser_n_bed_data, METH_NOARGS, "n_bed_data(self) -> unsigned int" },
  { "get_chrom", _wrap_BedParser_get_chrom, METH_O, "get_chrom(self, arg2) -> char const *" },
  { "get_start", _wrap_BedParser_get_start, METH_O, "get_start(self, arg2) -> unsigned long" },
  { "get_end", _wrap_BedParser_get_end, METH_O, "get_end(self, arg2) -> unsigned long" },
  { "append", _wrap_BedParser_append, METH_VARARGS, "append(self, arg2, arg3, arg4)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__BedParser_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.BedParser",                   /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_BedParser_destructor_closure,/* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__BedParser_type.as_number,             /* tp_as_number */
    &SwigPyBuiltin__egglib__BedParser_type.as_sequence,           /* tp_as_sequence */
    &SwigPyBuiltin__egglib__BedParser_type.as_mapping,            /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__BedParser_type.as_buffer,             /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::BedParser",                      /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__BedParser_richcompare,                 /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__BedParser_methods, /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__BedParser_getset,  /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_BedParser,                      /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__BedParser_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__BedParser_type};

static SwigPyGetSet ComputeV___dict___getset = { SwigPyObject_get___dict__, 0 };
SWIGINTERN PyGetSetDef SwigPyBuiltin__egglib__ComputeV_getset[] = {
    { (char *)"__dict__", SwigPyBuiltin_FunpackGetterClosure, 0, (char *)"ComputeV", &ComputeV___dict___getset },
    { NULL, NULL, NULL, NULL, NULL } /* Sentinel */
};

SWIGINTERN PyObject *
SwigPyBuiltin__egglib__ComputeV_richcompare(PyObject *self, PyObject *other, int op) {
  PyObject *result = NULL;
  if (!result) {
    if (SwigPyObject_Check(self) && SwigPyObject_Check(other)) {
      result = SwigPyObject_richcompare((SwigPyObject *)self, (SwigPyObject *)other, op);
    } else {
      result = Py_NotImplemented;
      Py_INCREF(result);
    }
  }
  return result;
}

SWIGINTERN PyMethodDef SwigPyBuiltin__egglib__ComputeV_methods[] = {
  { "reset", _wrap_ComputeV_reset, METH_NOARGS, "reset(self)" },
  { "compute", _wrap_ComputeV_compute, METH_VARARGS, "compute(self, arg2, arg3) -> bool" },
  { "average_V", _wrap_ComputeV_average_V, METH_NOARGS, "average_V(self) -> double" },
  { "average_Ar", _wrap_ComputeV_average_Ar, METH_NOARGS, "average_Ar(self) -> double" },
  { "average_M", _wrap_ComputeV_average_M, METH_NOARGS, "average_M(self) -> double" },
  { "average_Rst", _wrap_ComputeV_average_Rst, METH_NOARGS, "average_Rst(self) -> double" },
  { "curr_V", _wrap_ComputeV_curr_V, METH_NOARGS, "curr_V(self) -> double" },
  { "curr_Ar", _wrap_ComputeV_curr_Ar, METH_NOARGS, "curr_Ar(self) -> double" },
  { "curr_M", _wrap_ComputeV_curr_M, METH_NOARGS, "curr_M(self) -> double" },
  { "curr_Rst", _wrap_ComputeV_curr_Rst, METH_NOARGS, "curr_Rst(self) -> double" },
  { "num_sites", _wrap_ComputeV_num_sites, METH_NOARGS, "num_sites(self) -> unsigned int" },
  { "num_sites_m", _wrap_ComputeV_num_sites_m, METH_NOARGS, "num_sites_m(self) -> unsigned int" },
  { "num_sites_rst", _wrap_ComputeV_num_sites_rst, METH_NOARGS, "num_sites_rst(self) -> unsigned int" },
  { "set_maf", _wrap_ComputeV_set_maf, METH_O, "set_maf(self, arg2)" },
  { NULL, NULL, 0, NULL } /* Sentinel */
};

static PyHeapTypeObject SwigPyBuiltin__egglib__ComputeV_type = {
  {
#if PY_VERSION_HEX >= 0x03000000
    PyVarObject_HEAD_INIT(NULL, 0)
#else
    PyObject_HEAD_INIT(NULL)
    0,                                        /* ob_size */
#endif
    "eggwrapper.ComputeV",                    /* tp_name */
    sizeof(SwigPyObject),                     /* tp_basicsize */
    0,                                        /* tp_itemsize */
    _wrap_delete_ComputeV_destructor_closure, /* tp_dealloc */
    (printfunc) 0,                            /* tp_print */
    (getattrfunc) 0,                          /* tp_getattr */
    (setattrfunc) 0,                          /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
    0,                                        /* tp_compare */
#else
    (cmpfunc) 0,                              /* tp_compare */
#endif
    (reprfunc) 0,                             /* tp_repr */
    &SwigPyBuiltin__egglib__ComputeV_type.as_number,              /* tp_as_number */
    &SwigPyBuiltin__egglib__ComputeV_type.as_sequence,            /* tp_as_sequence */
    &SwigPyBuiltin__egglib__ComputeV_type.as_mapping,             /* tp_as_mapping */
    SwigPyObject_hash,                        /* tp_hash */
    (ternaryfunc) 0,                          /* tp_call */
    (reprfunc) 0,                             /* tp_str */
    (getattrofunc) 0,                         /* tp_getattro */
    (setattrofunc) 0,                         /* tp_setattro */
    &SwigPyBuiltin__egglib__ComputeV_type.as_buffer,              /* tp_as_buffer */
#if PY_VERSION_HEX >= 0x03000000
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,   /* tp_flags */
#else
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_CHECKTYPES, /* tp_flags */
#endif
    "egglib::ComputeV",                       /* tp_doc */
    (traverseproc) 0,                         /* tp_traverse */
    (inquiry) 0,                              /* tp_clear */
    SwigPyBuiltin__egglib__ComputeV_richcompare,                  /* tp_richcompare */
    0,                                        /* tp_weaklistoffset */
    (getiterfunc) 0,                          /* tp_iter */
    (iternextfunc) 0,                         /* tp_iternext */
    SwigPyBuiltin__egglib__ComputeV_methods,  /* tp_methods */
    0,                                        /* tp_members */
    SwigPyBuiltin__egglib__ComputeV_getset,   /* tp_getset */
    0,                                        /* tp_base */
    0,                                        /* tp_dict */
    (descrgetfunc) 0,                         /* tp_descr_get */
    (descrsetfunc) 0,                         /* tp_descr_set */
    offsetof(SwigPyObject, dict),             /* tp_dictoffset */
    _wrap_new_ComputeV,                       /* tp_init */
    (allocfunc) 0,                            /* tp_alloc */
    (newfunc) 0,                              /* tp_new */
    (freefunc) 0,                             /* tp_free */
    (inquiry) 0,                              /* tp_is_gc */
    (PyObject *) 0,                           /* tp_bases */
    (PyObject *) 0,                           /* tp_mro */
    (PyObject *) 0,                           /* tp_cache */
    (PyObject *) 0,                           /* tp_subclasses */
    (PyObject *) 0,                           /* tp_weaklist */
    (destructor) 0,                           /* tp_del */
    (int) 0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
    (destructor) 0,                           /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
    (vectorcallfunc) 0,                       /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
    0,                                        /* tp_print */
#endif
#ifdef COUNT_ALLOCS
    (Py_ssize_t) 0,                           /* tp_allocs */
    (Py_ssize_t) 0,                           /* tp_frees */
    (Py_ssize_t) 0,                           /* tp_maxalloc */
    0,                                        /* tp_prev */
    0,                                        /* tp_next */
#endif
  },
#if PY_VERSION_HEX >= 0x03050000
  {
    (unaryfunc) 0,                            /* am_await */
    (unaryfunc) 0,                            /* am_aiter */
    (unaryfunc) 0,                            /* am_anext */
  },
#endif
  {
    (binaryfunc) 0,                           /* nb_add */
    (binaryfunc) 0,                           /* nb_subtract */
    (binaryfunc) 0,                           /* nb_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_divide */
#endif
    (binaryfunc) 0,                           /* nb_remainder */
    (binaryfunc) 0,                           /* nb_divmod */
    (ternaryfunc) 0,                          /* nb_power */
    (unaryfunc) 0,                            /* nb_negative */
    (unaryfunc) 0,                            /* nb_positive */
    (unaryfunc) 0,                            /* nb_absolute */
    (inquiry) 0,                              /* nb_nonzero */
    (unaryfunc) 0,                            /* nb_invert */
    (binaryfunc) 0,                           /* nb_lshift */
    (binaryfunc) 0,                           /* nb_rshift */
    (binaryfunc) 0,                           /* nb_and */
    (binaryfunc) 0,                           /* nb_xor */
    (binaryfunc) 0,                           /* nb_or */
#if PY_VERSION_HEX < 0x03000000
    (coercion) 0,                             /* nb_coerce */
#endif
    (unaryfunc) 0,                            /* nb_int */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* nb_reserved */
#else
    (unaryfunc) 0,                            /* nb_long */
#endif
    (unaryfunc) 0,                            /* nb_float */
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc) 0,                            /* nb_oct */
    (unaryfunc) 0,                            /* nb_hex */
#endif
    (binaryfunc) 0,                           /* nb_inplace_add */
    (binaryfunc) 0,                           /* nb_inplace_subtract */
    (binaryfunc) 0,                           /* nb_inplace_multiply */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc) 0,                           /* nb_inplace_divide */
#endif
    (binaryfunc) 0,                           /* nb_inplace_remainder */
    (ternaryfunc) 0,                          /* nb_inplace_power */
    (binaryfunc) 0,                           /* nb_inplace_lshift */
    (binaryfunc) 0,                           /* nb_inplace_rshift */
    (binaryfunc) 0,                           /* nb_inplace_and */
    (binaryfunc) 0,                           /* nb_inplace_xor */
    (binaryfunc) 0,                           /* nb_inplace_or */
    (binaryfunc) 0,                           /* nb_floor_divide */
    (binaryfunc) 0,                           /* nb_true_divide */
    (binaryfunc) 0,                           /* nb_inplace_floor_divide */
    (binaryfunc) 0,                           /* nb_inplace_true_divide */
    (unaryfunc) 0,                            /* nb_index */
#if PY_VERSION_HEX >= 0x03050000
    (binaryfunc) 0,                           /* nb_matrix_multiply */
    (binaryfunc) 0,                           /* nb_inplace_matrix_multiply */
#endif
  },
  {
    (lenfunc) 0,                              /* mp_length */
    (binaryfunc) 0,                           /* mp_subscript */
    (objobjargproc) 0,                        /* mp_ass_subscript */
  },
  {
    (lenfunc) 0,                              /* sq_length */
    (binaryfunc) 0,                           /* sq_concat */
    (ssizeargfunc) 0,                         /* sq_repeat */
    (ssizeargfunc) 0,                         /* sq_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_slice */
#else
    (ssizessizeargfunc) 0,                    /* sq_slice */
#endif
    (ssizeobjargproc) 0,                      /* sq_ass_item */
#if PY_VERSION_HEX >= 0x03000000
    (void *) 0,                               /* was_sq_ass_slice */
#else
    (ssizessizeobjargproc) 0,                 /* sq_ass_slice */
#endif
    (objobjproc) 0,                           /* sq_contains */
    (binaryfunc) 0,                           /* sq_inplace_concat */
    (ssizeargfunc) 0,                         /* sq_inplace_repeat */
  },
  {
#if PY_VERSION_HEX < 0x03000000
    (readbufferproc) 0,                       /* bf_getreadbuffer */
    (writebufferproc) 0,                      /* bf_getwritebuffer */
    (segcountproc) 0,                         /* bf_getsegcount */
    (charbufferproc) 0,                       /* bf_getcharbuffer */
#endif
    (getbufferproc) 0,                        /* bf_getbuffer */
    (releasebufferproc) 0,                    /* bf_releasebuffer */
  },
    (PyObject *) 0,                           /* ht_name */
    (PyObject *) 0,                           /* ht_slots */
#if PY_VERSION_HEX >= 0x03030000
    (PyObject *) 0,                           /* ht_qualname */
    0,                                        /* ht_cached_keys */
#endif
};

SWIGINTERN SwigPyClientData SwigPyBuiltin__egglib__ComputeV_clientdata = {0, 0, 0, 0, 0, 0, (PyTypeObject *)&SwigPyBuiltin__egglib__ComputeV_type};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (BEGIN) -------- */

static void *_p_egglib__FastaFormatterTo_p_egglib__BaseFormatter(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::BaseFormatter *)  ((egglib::FastaFormatter *) x));
}
static void *_p_egglib__ExportTo_p_egglib__BaseFormatter(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::BaseFormatter *)  ((egglib::Export *) x));
}
static void *_p_egglib__RangeAlphabetTo_p_egglib__AbstractTemplateAlphabetT_int_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractTemplateAlphabet< int > *)  ((egglib::RangeAlphabet *) x));
}
static void *_p_egglib__FiniteAlphabetT_int_tTo_p_egglib__AbstractTemplateAlphabetT_int_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractTemplateAlphabet< int > *)  ((egglib::FiniteAlphabet< int > *) x));
}
static void *_p_egglib__vcf__AltTo_p_egglib__vcf__Filter(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::vcf::Filter *)  ((egglib::vcf::Alt *) x));
}
static void *_p_egglib__vcf__FormatTo_p_egglib__vcf__Filter(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::vcf::Filter *) (egglib::vcf::Info *) ((egglib::vcf::Format *) x));
}
static void *_p_egglib__vcf__InfoTo_p_egglib__vcf__Filter(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::vcf::Filter *)  ((egglib::vcf::Info *) x));
}
static void *_p_egglib__CodonAlphabetTo_p_egglib__StringAlphabet(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::StringAlphabet *)  ((egglib::CodonAlphabet *) x));
}
static void *_p_egglib__CaseInsensitiveStringAlphabetTo_p_egglib__StringAlphabet(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::StringAlphabet *)  ((egglib::CaseInsensitiveStringAlphabet *) x));
}
static void *_p_egglib__CustomStringAlphabetTo_p_egglib__StringAlphabet(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::StringAlphabet *)  ((egglib::CustomStringAlphabet *) x));
}
static void *_p_egglib__DNAAlphabetTo_p_egglib__CaseInsensitiveCharAlphabet(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::CaseInsensitiveCharAlphabet *)  ((egglib::DNAAlphabet *) x));
}
static void *_p_egglib__AbstractTemplateAlphabetT_char_tTo_p_egglib__AbstractBaseAlphabet(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractBaseAlphabet *)  ((egglib::AbstractTemplateAlphabet< char > *) x));
}
static void *_p_egglib__AbstractTemplateAlphabetT_char_p_tTo_p_egglib__AbstractBaseAlphabet(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractBaseAlphabet *)  ((egglib::AbstractTemplateAlphabet< char * > *) x));
}
static void *_p_egglib__FiniteAlphabetT_char_tTo_p_egglib__AbstractBaseAlphabet(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractBaseAlphabet *) (egglib::AbstractTemplateAlphabet< char > *) ((egglib::FiniteAlphabet< char > *) x));
}
static void *_p_egglib__FiniteAlphabetT_char_p_tTo_p_egglib__AbstractBaseAlphabet(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractBaseAlphabet *) (egglib::AbstractTemplateAlphabet< char * > *) ((egglib::FiniteAlphabet< char * > *) x));
}
static void *_p_egglib__CaseInsensitiveCharAlphabetTo_p_egglib__AbstractBaseAlphabet(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractBaseAlphabet *) (egglib::AbstractTemplateAlphabet< char > *)(egglib::FiniteAlphabet< char > *) ((egglib::CaseInsensitiveCharAlphabet *) x));
}
static void *_p_egglib__DNAAlphabetTo_p_egglib__AbstractBaseAlphabet(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractBaseAlphabet *) (egglib::AbstractTemplateAlphabet< char > *)(egglib::FiniteAlphabet< char > *)(egglib::CaseInsensitiveCharAlphabet *) ((egglib::DNAAlphabet *) x));
}
static void *_p_egglib__StringAlphabetTo_p_egglib__AbstractBaseAlphabet(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractBaseAlphabet *) (egglib::AbstractTemplateAlphabet< char * > *)(egglib::FiniteAlphabet< char * > *) ((egglib::StringAlphabet *) x));
}
static void *_p_egglib__CodonAlphabetTo_p_egglib__AbstractBaseAlphabet(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractBaseAlphabet *) (egglib::AbstractTemplateAlphabet< char * > *)(egglib::FiniteAlphabet< char * > *)(egglib::StringAlphabet *) ((egglib::CodonAlphabet *) x));
}
static void *_p_egglib__CaseInsensitiveStringAlphabetTo_p_egglib__AbstractBaseAlphabet(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractBaseAlphabet *) (egglib::AbstractTemplateAlphabet< char * > *)(egglib::FiniteAlphabet< char * > *)(egglib::StringAlphabet *) ((egglib::CaseInsensitiveStringAlphabet *) x));
}
static void *_p_egglib__CustomStringAlphabetTo_p_egglib__AbstractBaseAlphabet(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractBaseAlphabet *) (egglib::AbstractTemplateAlphabet< char * > *)(egglib::FiniteAlphabet< char * > *)(egglib::StringAlphabet *) ((egglib::CustomStringAlphabet *) x));
}
static void *_p_egglib__RangeAlphabetTo_p_egglib__AbstractBaseAlphabet(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractBaseAlphabet *) (egglib::AbstractTemplateAlphabet< int > *) ((egglib::RangeAlphabet *) x));
}
static void *_p_egglib__AbstractTemplateAlphabetT_int_tTo_p_egglib__AbstractBaseAlphabet(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractBaseAlphabet *)  ((egglib::AbstractTemplateAlphabet< int > *) x));
}
static void *_p_egglib__FiniteAlphabetT_int_tTo_p_egglib__AbstractBaseAlphabet(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractBaseAlphabet *) (egglib::AbstractTemplateAlphabet< int > *) ((egglib::FiniteAlphabet< int > *) x));
}
static void *_p_egglib__vcf__StringInfoTo_p_egglib__vcf__TypeInfoT_char_p_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::vcf::TypeInfo< char * > *)  ((egglib::vcf::StringInfo *) x));
}
static void *_p_egglib__StringAlphabetTo_p_egglib__FiniteAlphabetT_char_p_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::FiniteAlphabet< char * > *)  ((egglib::StringAlphabet *) x));
}
static void *_p_egglib__CodonAlphabetTo_p_egglib__FiniteAlphabetT_char_p_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::FiniteAlphabet< char * > *) (egglib::StringAlphabet *) ((egglib::CodonAlphabet *) x));
}
static void *_p_egglib__CaseInsensitiveStringAlphabetTo_p_egglib__FiniteAlphabetT_char_p_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::FiniteAlphabet< char * > *) (egglib::StringAlphabet *) ((egglib::CaseInsensitiveStringAlphabet *) x));
}
static void *_p_egglib__CustomStringAlphabetTo_p_egglib__FiniteAlphabetT_char_p_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::FiniteAlphabet< char * > *) (egglib::StringAlphabet *) ((egglib::CustomStringAlphabet *) x));
}
static void *_p_egglib__CaseInsensitiveCharAlphabetTo_p_egglib__FiniteAlphabetT_char_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::FiniteAlphabet< char > *)  ((egglib::CaseInsensitiveCharAlphabet *) x));
}
static void *_p_egglib__DNAAlphabetTo_p_egglib__FiniteAlphabetT_char_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::FiniteAlphabet< char > *) (egglib::CaseInsensitiveCharAlphabet *) ((egglib::DNAAlphabet *) x));
}
static void *_p_egglib__FiniteAlphabetT_char_p_tTo_p_egglib__AbstractTemplateAlphabetT_char_p_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractTemplateAlphabet< char * > *)  ((egglib::FiniteAlphabet< char * > *) x));
}
static void *_p_egglib__StringAlphabetTo_p_egglib__AbstractTemplateAlphabetT_char_p_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractTemplateAlphabet< char * > *) (egglib::FiniteAlphabet< char * > *) ((egglib::StringAlphabet *) x));
}
static void *_p_egglib__CodonAlphabetTo_p_egglib__AbstractTemplateAlphabetT_char_p_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractTemplateAlphabet< char * > *) (egglib::FiniteAlphabet< char * > *)(egglib::StringAlphabet *) ((egglib::CodonAlphabet *) x));
}
static void *_p_egglib__CaseInsensitiveStringAlphabetTo_p_egglib__AbstractTemplateAlphabetT_char_p_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractTemplateAlphabet< char * > *) (egglib::FiniteAlphabet< char * > *)(egglib::StringAlphabet *) ((egglib::CaseInsensitiveStringAlphabet *) x));
}
static void *_p_egglib__CustomStringAlphabetTo_p_egglib__AbstractTemplateAlphabetT_char_p_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractTemplateAlphabet< char * > *) (egglib::FiniteAlphabet< char * > *)(egglib::StringAlphabet *) ((egglib::CustomStringAlphabet *) x));
}
static void *_p_egglib__FiniteAlphabetT_char_tTo_p_egglib__AbstractTemplateAlphabetT_char_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractTemplateAlphabet< char > *)  ((egglib::FiniteAlphabet< char > *) x));
}
static void *_p_egglib__CaseInsensitiveCharAlphabetTo_p_egglib__AbstractTemplateAlphabetT_char_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractTemplateAlphabet< char > *) (egglib::FiniteAlphabet< char > *) ((egglib::CaseInsensitiveCharAlphabet *) x));
}
static void *_p_egglib__DNAAlphabetTo_p_egglib__AbstractTemplateAlphabetT_char_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::AbstractTemplateAlphabet< char > *) (egglib::FiniteAlphabet< char > *)(egglib::CaseInsensitiveCharAlphabet *) ((egglib::DNAAlphabet *) x));
}
static void *_p_egglib__VcfWindowSliderPerSiteTo_p_egglib__VcfWindowSliderBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::VcfWindowSliderBase *)  ((egglib::VcfWindowSliderPerSite *) x));
}
static void *_p_egglib__VcfWindowSliderTo_p_egglib__VcfWindowSliderBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::VcfWindowSliderBase *)  ((egglib::VcfWindowSlider *) x));
}
static void *_p_egglib__vcf__TypeInfoT_char_p_tTo_p_egglib__vcf__FlagInfo(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::vcf::FlagInfo *)  ((egglib::vcf::TypeInfo< char * > *) x));
}
static void *_p_egglib__vcf__TypeInfoT_char_tTo_p_egglib__vcf__FlagInfo(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::vcf::FlagInfo *)  ((egglib::vcf::TypeInfo< char > *) x));
}
static void *_p_egglib__vcf__TypeInfoT_int_tTo_p_egglib__vcf__FlagInfo(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::vcf::FlagInfo *)  ((egglib::vcf::TypeInfo< int > *) x));
}
static void *_p_egglib__vcf__TypeInfoT_double_tTo_p_egglib__vcf__FlagInfo(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::vcf::FlagInfo *)  ((egglib::vcf::TypeInfo< double > *) x));
}
static void *_p_egglib__vcf__StringInfoTo_p_egglib__vcf__FlagInfo(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::vcf::FlagInfo *) (egglib::vcf::TypeInfo< char * > *) ((egglib::vcf::StringInfo *) x));
}
static void *_p_egglib__VcfWindowSliderBaseTo_p_egglib__VcfWindowBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::VcfWindowBase *)  ((egglib::VcfWindowSliderBase *) x));
}
static void *_p_egglib__VcfWindowSliderPerSiteTo_p_egglib__VcfWindowBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::VcfWindowBase *) (egglib::VcfWindowSliderBase *) ((egglib::VcfWindowSliderPerSite *) x));
}
static void *_p_egglib__VcfWindowBEDTo_p_egglib__VcfWindowBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::VcfWindowBase *)  ((egglib::VcfWindowBED *) x));
}
static void *_p_egglib__VcfWindowSliderTo_p_egglib__VcfWindowBase(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::VcfWindowBase *) (egglib::VcfWindowSliderBase *) ((egglib::VcfWindowSlider *) x));
}
static void *_p_egglib__vcf__FormatTo_p_egglib__vcf__Info(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((egglib::vcf::Info *)  ((egglib::vcf::Format *) x));
}
static swig_type_info _swigt__p_SwigPyObject = {"_p_SwigPyObject", "SwigPyObject *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_char = {"_p_char", "char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_egglib__AbstractBaseAlphabet = {"_p_egglib__AbstractBaseAlphabet", "egglib::AbstractBaseAlphabet *", 0, 0, (void*)&SwigPyBuiltin__egglib__AbstractBaseAlphabet_clientdata, 0};
static swig_type_info _swigt__p_egglib__AbstractTemplateAlphabetT_char_p_t = {"_p_egglib__AbstractTemplateAlphabetT_char_p_t", "egglib::AbstractTemplateAlphabet< char * > *", 0, 0, (void*)&SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_p_t_clientdata, 0};
static swig_type_info _swigt__p_egglib__AbstractTemplateAlphabetT_char_t = {"_p_egglib__AbstractTemplateAlphabetT_char_t", "egglib::AbstractTemplateAlphabet< char > *", 0, 0, (void*)&SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_t_clientdata, 0};
static swig_type_info _swigt__p_egglib__AbstractTemplateAlphabetT_int_t = {"_p_egglib__AbstractTemplateAlphabetT_int_t", "egglib::AbstractTemplateAlphabet< int > *", 0, 0, (void*)&SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_int_t_clientdata, 0};
static swig_type_info _swigt__p_egglib__AlleleStatus = {"_p_egglib__AlleleStatus", "egglib::AlleleStatus *", 0, 0, (void*)&SwigPyBuiltin__egglib__AlleleStatus_clientdata, 0};
static swig_type_info _swigt__p_egglib__BaseFormatter = {"_p_egglib__BaseFormatter", "egglib::BaseFormatter *", 0, 0, (void*)&SwigPyBuiltin__egglib__BaseFormatter_clientdata, 0};
static swig_type_info _swigt__p_egglib__BedParser = {"_p_egglib__BedParser", "egglib::BedParser *", 0, 0, (void*)&SwigPyBuiltin__egglib__BedParser_clientdata, 0};
static swig_type_info _swigt__p_egglib__CaseInsensitiveCharAlphabet = {"_p_egglib__CaseInsensitiveCharAlphabet", "egglib::CaseInsensitiveCharAlphabet *", 0, 0, (void*)&SwigPyBuiltin__egglib__CaseInsensitiveCharAlphabet_clientdata, 0};
static swig_type_info _swigt__p_egglib__CaseInsensitiveStringAlphabet = {"_p_egglib__CaseInsensitiveStringAlphabet", "egglib::CaseInsensitiveStringAlphabet *", 0, 0, (void*)&SwigPyBuiltin__egglib__CaseInsensitiveStringAlphabet_clientdata, 0};
static swig_type_info _swigt__p_egglib__Coalesce = {"_p_egglib__Coalesce", "egglib::Coalesce *", 0, 0, (void*)&SwigPyBuiltin__egglib__Coalesce_clientdata, 0};
static swig_type_info _swigt__p_egglib__CodonAlphabet = {"_p_egglib__CodonAlphabet", "egglib::CodonAlphabet *", 0, 0, (void*)&SwigPyBuiltin__egglib__CodonAlphabet_clientdata, 0};
static swig_type_info _swigt__p_egglib__ComputeV = {"_p_egglib__ComputeV", "egglib::ComputeV *", 0, 0, (void*)&SwigPyBuiltin__egglib__ComputeV_clientdata, 0};
static swig_type_info _swigt__p_egglib__CustomStringAlphabet = {"_p_egglib__CustomStringAlphabet", "egglib::CustomStringAlphabet *", 0, 0, (void*)&SwigPyBuiltin__egglib__CustomStringAlphabet_clientdata, 0};
static swig_type_info _swigt__p_egglib__DNAAlphabet = {"_p_egglib__DNAAlphabet", "egglib::DNAAlphabet *", 0, 0, (void*)&SwigPyBuiltin__egglib__DNAAlphabet_clientdata, 0};
static swig_type_info _swigt__p_egglib__DataHolder = {"_p_egglib__DataHolder", "egglib::DataHolder *", 0, 0, (void*)&SwigPyBuiltin__egglib__DataHolder_clientdata, 0};
static swig_type_info _swigt__p_egglib__Diversity1 = {"_p_egglib__Diversity1", "egglib::Diversity1 *", 0, 0, (void*)&SwigPyBuiltin__egglib__Diversity1_clientdata, 0};
static swig_type_info _swigt__p_egglib__Diversity2 = {"_p_egglib__Diversity2", "egglib::Diversity2 *", 0, 0, (void*)&SwigPyBuiltin__egglib__Diversity2_clientdata, 0};
static swig_type_info _swigt__p_egglib__EHH = {"_p_egglib__EHH", "egglib::EHH *", 0, 0, (void*)&SwigPyBuiltin__egglib__EHH_clientdata, 0};
static swig_type_info _swigt__p_egglib__Event = {"_p_egglib__Event", "egglib::Event *", 0, 0, (void*)&SwigPyBuiltin__egglib__Event_clientdata, 0};
static swig_type_info _swigt__p_egglib__Export = {"_p_egglib__Export", "egglib::Export *", 0, 0, (void*)&SwigPyBuiltin__egglib__Export_clientdata, 0};
static swig_type_info _swigt__p_egglib__FastaFormatter = {"_p_egglib__FastaFormatter", "egglib::FastaFormatter *", 0, 0, (void*)&SwigPyBuiltin__egglib__FastaFormatter_clientdata, 0};
static swig_type_info _swigt__p_egglib__FastaParser = {"_p_egglib__FastaParser", "egglib::FastaParser *", 0, 0, (void*)&SwigPyBuiltin__egglib__FastaParser_clientdata, 0};
static swig_type_info _swigt__p_egglib__Feature = {"_p_egglib__Feature", "egglib::Feature *", 0, 0, (void*)&SwigPyBuiltin__egglib__Feature_clientdata, 0};
static swig_type_info _swigt__p_egglib__FiniteAlphabetT_char_p_t = {"_p_egglib__FiniteAlphabetT_char_p_t", "egglib::FiniteAlphabet< char * > *", 0, 0, (void*)&SwigPyBuiltin__egglib__FiniteAlphabetT_char_p_t_clientdata, 0};
static swig_type_info _swigt__p_egglib__FiniteAlphabetT_char_t = {"_p_egglib__FiniteAlphabetT_char_t", "egglib::FiniteAlphabet< char > *", 0, 0, (void*)&SwigPyBuiltin__egglib__FiniteAlphabetT_char_t_clientdata, 0};
static swig_type_info _swigt__p_egglib__FiniteAlphabetT_int_t = {"_p_egglib__FiniteAlphabetT_int_t", "egglib::FiniteAlphabet< int > *", 0, 0, (void*)&SwigPyBuiltin__egglib__FiniteAlphabetT_int_t_clientdata, 0};
static swig_type_info _swigt__p_egglib__FreqHolder = {"_p_egglib__FreqHolder", "egglib::FreqHolder *", 0, 0, (void*)&SwigPyBuiltin__egglib__FreqHolder_clientdata, 0};
static swig_type_info _swigt__p_egglib__FreqSet = {"_p_egglib__FreqSet", "egglib::FreqSet *", 0, 0, (void*)&SwigPyBuiltin__egglib__FreqSet_clientdata, 0};
static swig_type_info _swigt__p_egglib__GFF3 = {"_p_egglib__GFF3", "egglib::GFF3 *", 0, 0, (void*)&SwigPyBuiltin__egglib__GFF3_clientdata, 0};
static swig_type_info _swigt__p_egglib__GeneticCode = {"_p_egglib__GeneticCode", "egglib::GeneticCode *", 0, 0, (void*)&SwigPyBuiltin__egglib__GeneticCode_clientdata, 0};
static swig_type_info _swigt__p_egglib__Genotypes = {"_p_egglib__Genotypes", "egglib::Genotypes *", 0, 0, (void*)&SwigPyBuiltin__egglib__Genotypes_clientdata, 0};
static swig_type_info _swigt__p_egglib__Haplotypes = {"_p_egglib__Haplotypes", "egglib::Haplotypes *", 0, 0, (void*)&SwigPyBuiltin__egglib__Haplotypes_clientdata, 0};
static swig_type_info _swigt__p_egglib__IntersperseAlign = {"_p_egglib__IntersperseAlign", "egglib::IntersperseAlign *", 0, 0, (void*)&SwigPyBuiltin__egglib__IntersperseAlign_clientdata, 0};
static swig_type_info _swigt__p_egglib__MatrixLD = {"_p_egglib__MatrixLD", "egglib::MatrixLD *", 0, 0, (void*)&SwigPyBuiltin__egglib__MatrixLD_clientdata, 0};
static swig_type_info _swigt__p_egglib__Migration = {"_p_egglib__Migration", "egglib::Migration *", 0, 0, (void*)&SwigPyBuiltin__egglib__Migration_clientdata, 0};
static swig_type_info _swigt__p_egglib__Node = {"_p_egglib__Node", "egglib::Node *", 0, 0, (void*)&SwigPyBuiltin__egglib__Node_clientdata, 0};
static swig_type_info _swigt__p_egglib__PairwiseLD = {"_p_egglib__PairwiseLD", "egglib::PairwiseLD *", 0, 0, (void*)&SwigPyBuiltin__egglib__PairwiseLD_clientdata, 0};
static swig_type_info _swigt__p_egglib__ParalogPi = {"_p_egglib__ParalogPi", "egglib::ParalogPi *", 0, 0, (void*)&SwigPyBuiltin__egglib__ParalogPi_clientdata, 0};
static swig_type_info _swigt__p_egglib__Params = {"_p_egglib__Params", "egglib::Params *", 0, 0, (void*)&SwigPyBuiltin__egglib__Params_clientdata, 0};
static swig_type_info _swigt__p_egglib__RangeAlphabet = {"_p_egglib__RangeAlphabet", "egglib::RangeAlphabet *", 0, 0, (void*)&SwigPyBuiltin__egglib__RangeAlphabet_clientdata, 0};
static swig_type_info _swigt__p_egglib__Rd = {"_p_egglib__Rd", "egglib::Rd *", 0, 0, (void*)&SwigPyBuiltin__egglib__Rd_clientdata, 0};
static swig_type_info _swigt__p_egglib__SiteDiversity = {"_p_egglib__SiteDiversity", "egglib::SiteDiversity *", 0, 0, (void*)&SwigPyBuiltin__egglib__SiteDiversity_clientdata, 0};
static swig_type_info _swigt__p_egglib__SiteHolder = {"_p_egglib__SiteHolder", "egglib::SiteHolder *", 0, 0, (void*)&SwigPyBuiltin__egglib__SiteHolder_clientdata, 0};
static swig_type_info _swigt__p_egglib__StringAlphabet = {"_p_egglib__StringAlphabet", "egglib::StringAlphabet *", 0, 0, (void*)&SwigPyBuiltin__egglib__StringAlphabet_clientdata, 0};
static swig_type_info _swigt__p_egglib__StructureCluster = {"_p_egglib__StructureCluster", "egglib::StructureCluster *", 0, 0, (void*)&SwigPyBuiltin__egglib__StructureCluster_clientdata, 0};
static swig_type_info _swigt__p_egglib__StructureHolder = {"_p_egglib__StructureHolder", "egglib::StructureHolder *", 0, 0, (void*)&SwigPyBuiltin__egglib__StructureHolder_clientdata, 0};
static swig_type_info _swigt__p_egglib__StructureIndiv = {"_p_egglib__StructureIndiv", "egglib::StructureIndiv *", 0, 0, (void*)&SwigPyBuiltin__egglib__StructureIndiv_clientdata, 0};
static swig_type_info _swigt__p_egglib__StructurePopulation = {"_p_egglib__StructurePopulation", "egglib::StructurePopulation *", 0, 0, (void*)&SwigPyBuiltin__egglib__StructurePopulation_clientdata, 0};
static swig_type_info _swigt__p_egglib__Tree = {"_p_egglib__Tree", "egglib::Tree *", 0, 0, (void*)&SwigPyBuiltin__egglib__Tree_clientdata, 0};
static swig_type_info _swigt__p_egglib__VcfIndex = {"_p_egglib__VcfIndex", "egglib::VcfIndex *", 0, 0, (void*)&SwigPyBuiltin__egglib__VcfIndex_clientdata, 0};
static swig_type_info _swigt__p_egglib__VcfParser = {"_p_egglib__VcfParser", "egglib::VcfParser *", 0, 0, (void*)&SwigPyBuiltin__egglib__VcfParser_clientdata, 0};
static swig_type_info _swigt__p_egglib__VcfWindowBED = {"_p_egglib__VcfWindowBED", "egglib::VcfWindowBED *", 0, 0, (void*)&SwigPyBuiltin__egglib__VcfWindowBED_clientdata, 0};
static swig_type_info _swigt__p_egglib__VcfWindowBase = {"_p_egglib__VcfWindowBase", "egglib::VcfWindowBase *", 0, 0, (void*)&SwigPyBuiltin__egglib__VcfWindowBase_clientdata, 0};
static swig_type_info _swigt__p_egglib__VcfWindowSlider = {"_p_egglib__VcfWindowSlider", "egglib::VcfWindowSlider *", 0, 0, (void*)&SwigPyBuiltin__egglib__VcfWindowSlider_clientdata, 0};
static swig_type_info _swigt__p_egglib__VcfWindowSliderBase = {"_p_egglib__VcfWindowSliderBase", "egglib::VcfWindowSliderBase *", 0, 0, (void*)&SwigPyBuiltin__egglib__VcfWindowSliderBase_clientdata, 0};
static swig_type_info _swigt__p_egglib__VcfWindowSliderPerSite = {"_p_egglib__VcfWindowSliderPerSite", "egglib::VcfWindowSliderPerSite *", 0, 0, (void*)&SwigPyBuiltin__egglib__VcfWindowSliderPerSite_clientdata, 0};
static swig_type_info _swigt__p_egglib__VectorInt = {"_p_egglib__VectorInt", "egglib::VectorInt *", 0, 0, (void*)&SwigPyBuiltin__egglib__VectorInt_clientdata, 0};
static swig_type_info _swigt__p_egglib__WSite = {"_p_egglib__WSite", "egglib::WSite *", 0, 0, (void*)&SwigPyBuiltin__egglib__WSite_clientdata, 0};
static swig_type_info _swigt__p_egglib__vcf__Alt = {"_p_egglib__vcf__Alt", "egglib::vcf::Alt *", 0, 0, (void*)&SwigPyBuiltin__egglib__vcf__Alt_clientdata, 0};
static swig_type_info _swigt__p_egglib__vcf__Filter = {"_p_egglib__vcf__Filter", "egglib::vcf::Filter *", 0, 0, (void*)&SwigPyBuiltin__egglib__vcf__Filter_clientdata, 0};
static swig_type_info _swigt__p_egglib__vcf__FlagInfo = {"_p_egglib__vcf__FlagInfo", "egglib::vcf::FlagInfo *", 0, 0, (void*)&SwigPyBuiltin__egglib__vcf__FlagInfo_clientdata, 0};
static swig_type_info _swigt__p_egglib__vcf__Format = {"_p_egglib__vcf__Format", "egglib::vcf::Format *", 0, 0, (void*)&SwigPyBuiltin__egglib__vcf__Format_clientdata, 0};
static swig_type_info _swigt__p_egglib__vcf__Info = {"_p_egglib__vcf__Info", "egglib::vcf::Info *", 0, 0, (void*)&SwigPyBuiltin__egglib__vcf__Info_clientdata, 0};
static swig_type_info _swigt__p_egglib__vcf__Meta = {"_p_egglib__vcf__Meta", "egglib::vcf::Meta *", 0, 0, (void*)&SwigPyBuiltin__egglib__vcf__Meta_clientdata, 0};
static swig_type_info _swigt__p_egglib__vcf__SampleInfo = {"_p_egglib__vcf__SampleInfo", "egglib::vcf::SampleInfo *", 0, 0, (void*)&SwigPyBuiltin__egglib__vcf__SampleInfo_clientdata, 0};
static swig_type_info _swigt__p_egglib__vcf__StringInfo = {"_p_egglib__vcf__StringInfo", "egglib::vcf::StringInfo *", 0, 0, (void*)&SwigPyBuiltin__egglib__vcf__StringInfo_clientdata, 0};
static swig_type_info _swigt__p_egglib__vcf__TypeInfoT_char_p_t = {"_p_egglib__vcf__TypeInfoT_char_p_t", "egglib::vcf::TypeInfo< char * > *", 0, 0, (void*)&SwigPyBuiltin__egglib__vcf__TypeInfoT_char_p_t_clientdata, 0};
static swig_type_info _swigt__p_egglib__vcf__TypeInfoT_char_t = {"_p_egglib__vcf__TypeInfoT_char_t", "egglib::vcf::TypeInfo< char > *", 0, 0, (void*)&SwigPyBuiltin__egglib__vcf__TypeInfoT_char_t_clientdata, 0};
static swig_type_info _swigt__p_egglib__vcf__TypeInfoT_double_t = {"_p_egglib__vcf__TypeInfoT_double_t", "egglib::vcf::TypeInfo< double > *", 0, 0, (void*)&SwigPyBuiltin__egglib__vcf__TypeInfoT_double_t_clientdata, 0};
static swig_type_info _swigt__p_egglib__vcf__TypeInfoT_int_t = {"_p_egglib__vcf__TypeInfoT_int_t", "egglib::vcf::TypeInfo< int > *", 0, 0, (void*)&SwigPyBuiltin__egglib__vcf__TypeInfoT_int_t_clientdata, 0};
static swig_type_info _swigt__p_int = {"_p_int", "int *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__istream = {"_p_std__istream", "std::istream *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__string = {"_p_std__string", "std::string *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_int = {"_p_unsigned_int", "unsigned int *", 0, 0, (void*)0, 0};

static swig_type_info *swig_type_initial[] = {
  &_swigt__p_SwigPyObject,
  &_swigt__p_char,
  &_swigt__p_egglib__AbstractBaseAlphabet,
  &_swigt__p_egglib__AbstractTemplateAlphabetT_char_p_t,
  &_swigt__p_egglib__AbstractTemplateAlphabetT_char_t,
  &_swigt__p_egglib__AbstractTemplateAlphabetT_int_t,
  &_swigt__p_egglib__AlleleStatus,
  &_swigt__p_egglib__BaseFormatter,
  &_swigt__p_egglib__BedParser,
  &_swigt__p_egglib__CaseInsensitiveCharAlphabet,
  &_swigt__p_egglib__CaseInsensitiveStringAlphabet,
  &_swigt__p_egglib__Coalesce,
  &_swigt__p_egglib__CodonAlphabet,
  &_swigt__p_egglib__ComputeV,
  &_swigt__p_egglib__CustomStringAlphabet,
  &_swigt__p_egglib__DNAAlphabet,
  &_swigt__p_egglib__DataHolder,
  &_swigt__p_egglib__Diversity1,
  &_swigt__p_egglib__Diversity2,
  &_swigt__p_egglib__EHH,
  &_swigt__p_egglib__Event,
  &_swigt__p_egglib__Export,
  &_swigt__p_egglib__FastaFormatter,
  &_swigt__p_egglib__FastaParser,
  &_swigt__p_egglib__Feature,
  &_swigt__p_egglib__FiniteAlphabetT_char_p_t,
  &_swigt__p_egglib__FiniteAlphabetT_char_t,
  &_swigt__p_egglib__FiniteAlphabetT_int_t,
  &_swigt__p_egglib__FreqHolder,
  &_swigt__p_egglib__FreqSet,
  &_swigt__p_egglib__GFF3,
  &_swigt__p_egglib__GeneticCode,
  &_swigt__p_egglib__Genotypes,
  &_swigt__p_egglib__Haplotypes,
  &_swigt__p_egglib__IntersperseAlign,
  &_swigt__p_egglib__MatrixLD,
  &_swigt__p_egglib__Migration,
  &_swigt__p_egglib__Node,
  &_swigt__p_egglib__PairwiseLD,
  &_swigt__p_egglib__ParalogPi,
  &_swigt__p_egglib__Params,
  &_swigt__p_egglib__RangeAlphabet,
  &_swigt__p_egglib__Rd,
  &_swigt__p_egglib__SiteDiversity,
  &_swigt__p_egglib__SiteHolder,
  &_swigt__p_egglib__StringAlphabet,
  &_swigt__p_egglib__StructureCluster,
  &_swigt__p_egglib__StructureHolder,
  &_swigt__p_egglib__StructureIndiv,
  &_swigt__p_egglib__StructurePopulation,
  &_swigt__p_egglib__Tree,
  &_swigt__p_egglib__VcfIndex,
  &_swigt__p_egglib__VcfParser,
  &_swigt__p_egglib__VcfWindowBED,
  &_swigt__p_egglib__VcfWindowBase,
  &_swigt__p_egglib__VcfWindowSlider,
  &_swigt__p_egglib__VcfWindowSliderBase,
  &_swigt__p_egglib__VcfWindowSliderPerSite,
  &_swigt__p_egglib__VectorInt,
  &_swigt__p_egglib__WSite,
  &_swigt__p_egglib__vcf__Alt,
  &_swigt__p_egglib__vcf__Filter,
  &_swigt__p_egglib__vcf__FlagInfo,
  &_swigt__p_egglib__vcf__Format,
  &_swigt__p_egglib__vcf__Info,
  &_swigt__p_egglib__vcf__Meta,
  &_swigt__p_egglib__vcf__SampleInfo,
  &_swigt__p_egglib__vcf__StringInfo,
  &_swigt__p_egglib__vcf__TypeInfoT_char_p_t,
  &_swigt__p_egglib__vcf__TypeInfoT_char_t,
  &_swigt__p_egglib__vcf__TypeInfoT_double_t,
  &_swigt__p_egglib__vcf__TypeInfoT_int_t,
  &_swigt__p_int,
  &_swigt__p_std__istream,
  &_swigt__p_std__string,
  &_swigt__p_unsigned_int,
};

static swig_cast_info _swigc__p_SwigPyObject[] = {  {&_swigt__p_SwigPyObject, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_char[] = {  {&_swigt__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__AbstractBaseAlphabet[] = {  {&_swigt__p_egglib__AbstractTemplateAlphabetT_char_t, _p_egglib__AbstractTemplateAlphabetT_char_tTo_p_egglib__AbstractBaseAlphabet, 0, 0},  {&_swigt__p_egglib__FiniteAlphabetT_char_t, _p_egglib__FiniteAlphabetT_char_tTo_p_egglib__AbstractBaseAlphabet, 0, 0},  {&_swigt__p_egglib__AbstractTemplateAlphabetT_char_p_t, _p_egglib__AbstractTemplateAlphabetT_char_p_tTo_p_egglib__AbstractBaseAlphabet, 0, 0},  {&_swigt__p_egglib__FiniteAlphabetT_char_p_t, _p_egglib__FiniteAlphabetT_char_p_tTo_p_egglib__AbstractBaseAlphabet, 0, 0},  {&_swigt__p_egglib__AbstractTemplateAlphabetT_int_t, _p_egglib__AbstractTemplateAlphabetT_int_tTo_p_egglib__AbstractBaseAlphabet, 0, 0},  {&_swigt__p_egglib__FiniteAlphabetT_int_t, _p_egglib__FiniteAlphabetT_int_tTo_p_egglib__AbstractBaseAlphabet, 0, 0},  {&_swigt__p_egglib__RangeAlphabet, _p_egglib__RangeAlphabetTo_p_egglib__AbstractBaseAlphabet, 0, 0},  {&_swigt__p_egglib__CustomStringAlphabet, _p_egglib__CustomStringAlphabetTo_p_egglib__AbstractBaseAlphabet, 0, 0},  {&_swigt__p_egglib__CaseInsensitiveStringAlphabet, _p_egglib__CaseInsensitiveStringAlphabetTo_p_egglib__AbstractBaseAlphabet, 0, 0},  {&_swigt__p_egglib__CodonAlphabet, _p_egglib__CodonAlphabetTo_p_egglib__AbstractBaseAlphabet, 0, 0},  {&_swigt__p_egglib__StringAlphabet, _p_egglib__StringAlphabetTo_p_egglib__AbstractBaseAlphabet, 0, 0},  {&_swigt__p_egglib__DNAAlphabet, _p_egglib__DNAAlphabetTo_p_egglib__AbstractBaseAlphabet, 0, 0},  {&_swigt__p_egglib__CaseInsensitiveCharAlphabet, _p_egglib__CaseInsensitiveCharAlphabetTo_p_egglib__AbstractBaseAlphabet, 0, 0},  {&_swigt__p_egglib__AbstractBaseAlphabet, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__AbstractTemplateAlphabetT_char_p_t[] = {  {&_swigt__p_egglib__AbstractTemplateAlphabetT_char_p_t, 0, 0, 0},  {&_swigt__p_egglib__FiniteAlphabetT_char_p_t, _p_egglib__FiniteAlphabetT_char_p_tTo_p_egglib__AbstractTemplateAlphabetT_char_p_t, 0, 0},  {&_swigt__p_egglib__StringAlphabet, _p_egglib__StringAlphabetTo_p_egglib__AbstractTemplateAlphabetT_char_p_t, 0, 0},  {&_swigt__p_egglib__CodonAlphabet, _p_egglib__CodonAlphabetTo_p_egglib__AbstractTemplateAlphabetT_char_p_t, 0, 0},  {&_swigt__p_egglib__CaseInsensitiveStringAlphabet, _p_egglib__CaseInsensitiveStringAlphabetTo_p_egglib__AbstractTemplateAlphabetT_char_p_t, 0, 0},  {&_swigt__p_egglib__CustomStringAlphabet, _p_egglib__CustomStringAlphabetTo_p_egglib__AbstractTemplateAlphabetT_char_p_t, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__AbstractTemplateAlphabetT_char_t[] = {  {&_swigt__p_egglib__CaseInsensitiveCharAlphabet, _p_egglib__CaseInsensitiveCharAlphabetTo_p_egglib__AbstractTemplateAlphabetT_char_t, 0, 0},  {&_swigt__p_egglib__DNAAlphabet, _p_egglib__DNAAlphabetTo_p_egglib__AbstractTemplateAlphabetT_char_t, 0, 0},  {&_swigt__p_egglib__AbstractTemplateAlphabetT_char_t, 0, 0, 0},  {&_swigt__p_egglib__FiniteAlphabetT_char_t, _p_egglib__FiniteAlphabetT_char_tTo_p_egglib__AbstractTemplateAlphabetT_char_t, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__AbstractTemplateAlphabetT_int_t[] = {  {&_swigt__p_egglib__RangeAlphabet, _p_egglib__RangeAlphabetTo_p_egglib__AbstractTemplateAlphabetT_int_t, 0, 0},  {&_swigt__p_egglib__AbstractTemplateAlphabetT_int_t, 0, 0, 0},  {&_swigt__p_egglib__FiniteAlphabetT_int_t, _p_egglib__FiniteAlphabetT_int_tTo_p_egglib__AbstractTemplateAlphabetT_int_t, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__AlleleStatus[] = {  {&_swigt__p_egglib__AlleleStatus, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__BaseFormatter[] = {  {&_swigt__p_egglib__BaseFormatter, 0, 0, 0},  {&_swigt__p_egglib__FastaFormatter, _p_egglib__FastaFormatterTo_p_egglib__BaseFormatter, 0, 0},  {&_swigt__p_egglib__Export, _p_egglib__ExportTo_p_egglib__BaseFormatter, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__BedParser[] = {  {&_swigt__p_egglib__BedParser, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__CaseInsensitiveCharAlphabet[] = {  {&_swigt__p_egglib__CaseInsensitiveCharAlphabet, 0, 0, 0},  {&_swigt__p_egglib__DNAAlphabet, _p_egglib__DNAAlphabetTo_p_egglib__CaseInsensitiveCharAlphabet, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__CaseInsensitiveStringAlphabet[] = {  {&_swigt__p_egglib__CaseInsensitiveStringAlphabet, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__Coalesce[] = {  {&_swigt__p_egglib__Coalesce, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__CodonAlphabet[] = {  {&_swigt__p_egglib__CodonAlphabet, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__ComputeV[] = {  {&_swigt__p_egglib__ComputeV, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__CustomStringAlphabet[] = {  {&_swigt__p_egglib__CustomStringAlphabet, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__DNAAlphabet[] = {  {&_swigt__p_egglib__DNAAlphabet, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__DataHolder[] = {  {&_swigt__p_egglib__DataHolder, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__Diversity1[] = {  {&_swigt__p_egglib__Diversity1, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__Diversity2[] = {  {&_swigt__p_egglib__Diversity2, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__EHH[] = {  {&_swigt__p_egglib__EHH, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__Event[] = {  {&_swigt__p_egglib__Event, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__Export[] = {  {&_swigt__p_egglib__Export, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__FastaFormatter[] = {  {&_swigt__p_egglib__FastaFormatter, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__FastaParser[] = {  {&_swigt__p_egglib__FastaParser, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__Feature[] = {  {&_swigt__p_egglib__Feature, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__FiniteAlphabetT_char_p_t[] = {  {&_swigt__p_egglib__FiniteAlphabetT_char_p_t, 0, 0, 0},  {&_swigt__p_egglib__StringAlphabet, _p_egglib__StringAlphabetTo_p_egglib__FiniteAlphabetT_char_p_t, 0, 0},  {&_swigt__p_egglib__CodonAlphabet, _p_egglib__CodonAlphabetTo_p_egglib__FiniteAlphabetT_char_p_t, 0, 0},  {&_swigt__p_egglib__CaseInsensitiveStringAlphabet, _p_egglib__CaseInsensitiveStringAlphabetTo_p_egglib__FiniteAlphabetT_char_p_t, 0, 0},  {&_swigt__p_egglib__CustomStringAlphabet, _p_egglib__CustomStringAlphabetTo_p_egglib__FiniteAlphabetT_char_p_t, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__FiniteAlphabetT_char_t[] = {  {&_swigt__p_egglib__CaseInsensitiveCharAlphabet, _p_egglib__CaseInsensitiveCharAlphabetTo_p_egglib__FiniteAlphabetT_char_t, 0, 0},  {&_swigt__p_egglib__DNAAlphabet, _p_egglib__DNAAlphabetTo_p_egglib__FiniteAlphabetT_char_t, 0, 0},  {&_swigt__p_egglib__FiniteAlphabetT_char_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__FiniteAlphabetT_int_t[] = {  {&_swigt__p_egglib__FiniteAlphabetT_int_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__FreqHolder[] = {  {&_swigt__p_egglib__FreqHolder, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__FreqSet[] = {  {&_swigt__p_egglib__FreqSet, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__GFF3[] = {  {&_swigt__p_egglib__GFF3, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__GeneticCode[] = {  {&_swigt__p_egglib__GeneticCode, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__Genotypes[] = {  {&_swigt__p_egglib__Genotypes, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__Haplotypes[] = {  {&_swigt__p_egglib__Haplotypes, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__IntersperseAlign[] = {  {&_swigt__p_egglib__IntersperseAlign, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__MatrixLD[] = {  {&_swigt__p_egglib__MatrixLD, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__Migration[] = {  {&_swigt__p_egglib__Migration, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__Node[] = {  {&_swigt__p_egglib__Node, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__PairwiseLD[] = {  {&_swigt__p_egglib__PairwiseLD, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__ParalogPi[] = {  {&_swigt__p_egglib__ParalogPi, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__Params[] = {  {&_swigt__p_egglib__Params, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__RangeAlphabet[] = {  {&_swigt__p_egglib__RangeAlphabet, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__Rd[] = {  {&_swigt__p_egglib__Rd, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__SiteDiversity[] = {  {&_swigt__p_egglib__SiteDiversity, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__SiteHolder[] = {  {&_swigt__p_egglib__SiteHolder, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__StringAlphabet[] = {  {&_swigt__p_egglib__StringAlphabet, 0, 0, 0},  {&_swigt__p_egglib__CodonAlphabet, _p_egglib__CodonAlphabetTo_p_egglib__StringAlphabet, 0, 0},  {&_swigt__p_egglib__CaseInsensitiveStringAlphabet, _p_egglib__CaseInsensitiveStringAlphabetTo_p_egglib__StringAlphabet, 0, 0},  {&_swigt__p_egglib__CustomStringAlphabet, _p_egglib__CustomStringAlphabetTo_p_egglib__StringAlphabet, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__StructureCluster[] = {  {&_swigt__p_egglib__StructureCluster, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__StructureHolder[] = {  {&_swigt__p_egglib__StructureHolder, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__StructureIndiv[] = {  {&_swigt__p_egglib__StructureIndiv, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__StructurePopulation[] = {  {&_swigt__p_egglib__StructurePopulation, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__Tree[] = {  {&_swigt__p_egglib__Tree, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__VcfIndex[] = {  {&_swigt__p_egglib__VcfIndex, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__VcfParser[] = {  {&_swigt__p_egglib__VcfParser, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__VcfWindowBED[] = {  {&_swigt__p_egglib__VcfWindowBED, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__VcfWindowBase[] = {  {&_swigt__p_egglib__VcfWindowSliderBase, _p_egglib__VcfWindowSliderBaseTo_p_egglib__VcfWindowBase, 0, 0},  {&_swigt__p_egglib__VcfWindowSliderPerSite, _p_egglib__VcfWindowSliderPerSiteTo_p_egglib__VcfWindowBase, 0, 0},  {&_swigt__p_egglib__VcfWindowBase, 0, 0, 0},  {&_swigt__p_egglib__VcfWindowBED, _p_egglib__VcfWindowBEDTo_p_egglib__VcfWindowBase, 0, 0},  {&_swigt__p_egglib__VcfWindowSlider, _p_egglib__VcfWindowSliderTo_p_egglib__VcfWindowBase, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__VcfWindowSlider[] = {  {&_swigt__p_egglib__VcfWindowSlider, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__VcfWindowSliderBase[] = {  {&_swigt__p_egglib__VcfWindowSliderBase, 0, 0, 0},  {&_swigt__p_egglib__VcfWindowSliderPerSite, _p_egglib__VcfWindowSliderPerSiteTo_p_egglib__VcfWindowSliderBase, 0, 0},  {&_swigt__p_egglib__VcfWindowSlider, _p_egglib__VcfWindowSliderTo_p_egglib__VcfWindowSliderBase, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__VcfWindowSliderPerSite[] = {  {&_swigt__p_egglib__VcfWindowSliderPerSite, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__VectorInt[] = {  {&_swigt__p_egglib__VectorInt, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__WSite[] = {  {&_swigt__p_egglib__WSite, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__vcf__Alt[] = {  {&_swigt__p_egglib__vcf__Alt, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__vcf__Filter[] = {  {&_swigt__p_egglib__vcf__Info, _p_egglib__vcf__InfoTo_p_egglib__vcf__Filter, 0, 0},  {&_swigt__p_egglib__vcf__Filter, 0, 0, 0},  {&_swigt__p_egglib__vcf__Alt, _p_egglib__vcf__AltTo_p_egglib__vcf__Filter, 0, 0},  {&_swigt__p_egglib__vcf__Format, _p_egglib__vcf__FormatTo_p_egglib__vcf__Filter, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__vcf__FlagInfo[] = {  {&_swigt__p_egglib__vcf__TypeInfoT_char_p_t, _p_egglib__vcf__TypeInfoT_char_p_tTo_p_egglib__vcf__FlagInfo, 0, 0},  {&_swigt__p_egglib__vcf__TypeInfoT_double_t, _p_egglib__vcf__TypeInfoT_double_tTo_p_egglib__vcf__FlagInfo, 0, 0},  {&_swigt__p_egglib__vcf__FlagInfo, 0, 0, 0},  {&_swigt__p_egglib__vcf__TypeInfoT_char_t, _p_egglib__vcf__TypeInfoT_char_tTo_p_egglib__vcf__FlagInfo, 0, 0},  {&_swigt__p_egglib__vcf__TypeInfoT_int_t, _p_egglib__vcf__TypeInfoT_int_tTo_p_egglib__vcf__FlagInfo, 0, 0},  {&_swigt__p_egglib__vcf__StringInfo, _p_egglib__vcf__StringInfoTo_p_egglib__vcf__FlagInfo, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__vcf__Format[] = {  {&_swigt__p_egglib__vcf__Format, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__vcf__Info[] = {  {&_swigt__p_egglib__vcf__Info, 0, 0, 0},  {&_swigt__p_egglib__vcf__Format, _p_egglib__vcf__FormatTo_p_egglib__vcf__Info, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__vcf__Meta[] = {  {&_swigt__p_egglib__vcf__Meta, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__vcf__SampleInfo[] = {  {&_swigt__p_egglib__vcf__SampleInfo, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__vcf__StringInfo[] = {  {&_swigt__p_egglib__vcf__StringInfo, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__vcf__TypeInfoT_char_p_t[] = {  {&_swigt__p_egglib__vcf__TypeInfoT_char_p_t, 0, 0, 0},  {&_swigt__p_egglib__vcf__StringInfo, _p_egglib__vcf__StringInfoTo_p_egglib__vcf__TypeInfoT_char_p_t, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__vcf__TypeInfoT_char_t[] = {  {&_swigt__p_egglib__vcf__TypeInfoT_char_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__vcf__TypeInfoT_double_t[] = {  {&_swigt__p_egglib__vcf__TypeInfoT_double_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_egglib__vcf__TypeInfoT_int_t[] = {  {&_swigt__p_egglib__vcf__TypeInfoT_int_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_int[] = {  {&_swigt__p_int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__istream[] = {  {&_swigt__p_std__istream, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__string[] = {  {&_swigt__p_std__string, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_int[] = {  {&_swigt__p_unsigned_int, 0, 0, 0},{0, 0, 0, 0}};

static swig_cast_info *swig_cast_initial[] = {
  _swigc__p_SwigPyObject,
  _swigc__p_char,
  _swigc__p_egglib__AbstractBaseAlphabet,
  _swigc__p_egglib__AbstractTemplateAlphabetT_char_p_t,
  _swigc__p_egglib__AbstractTemplateAlphabetT_char_t,
  _swigc__p_egglib__AbstractTemplateAlphabetT_int_t,
  _swigc__p_egglib__AlleleStatus,
  _swigc__p_egglib__BaseFormatter,
  _swigc__p_egglib__BedParser,
  _swigc__p_egglib__CaseInsensitiveCharAlphabet,
  _swigc__p_egglib__CaseInsensitiveStringAlphabet,
  _swigc__p_egglib__Coalesce,
  _swigc__p_egglib__CodonAlphabet,
  _swigc__p_egglib__ComputeV,
  _swigc__p_egglib__CustomStringAlphabet,
  _swigc__p_egglib__DNAAlphabet,
  _swigc__p_egglib__DataHolder,
  _swigc__p_egglib__Diversity1,
  _swigc__p_egglib__Diversity2,
  _swigc__p_egglib__EHH,
  _swigc__p_egglib__Event,
  _swigc__p_egglib__Export,
  _swigc__p_egglib__FastaFormatter,
  _swigc__p_egglib__FastaParser,
  _swigc__p_egglib__Feature,
  _swigc__p_egglib__FiniteAlphabetT_char_p_t,
  _swigc__p_egglib__FiniteAlphabetT_char_t,
  _swigc__p_egglib__FiniteAlphabetT_int_t,
  _swigc__p_egglib__FreqHolder,
  _swigc__p_egglib__FreqSet,
  _swigc__p_egglib__GFF3,
  _swigc__p_egglib__GeneticCode,
  _swigc__p_egglib__Genotypes,
  _swigc__p_egglib__Haplotypes,
  _swigc__p_egglib__IntersperseAlign,
  _swigc__p_egglib__MatrixLD,
  _swigc__p_egglib__Migration,
  _swigc__p_egglib__Node,
  _swigc__p_egglib__PairwiseLD,
  _swigc__p_egglib__ParalogPi,
  _swigc__p_egglib__Params,
  _swigc__p_egglib__RangeAlphabet,
  _swigc__p_egglib__Rd,
  _swigc__p_egglib__SiteDiversity,
  _swigc__p_egglib__SiteHolder,
  _swigc__p_egglib__StringAlphabet,
  _swigc__p_egglib__StructureCluster,
  _swigc__p_egglib__StructureHolder,
  _swigc__p_egglib__StructureIndiv,
  _swigc__p_egglib__StructurePopulation,
  _swigc__p_egglib__Tree,
  _swigc__p_egglib__VcfIndex,
  _swigc__p_egglib__VcfParser,
  _swigc__p_egglib__VcfWindowBED,
  _swigc__p_egglib__VcfWindowBase,
  _swigc__p_egglib__VcfWindowSlider,
  _swigc__p_egglib__VcfWindowSliderBase,
  _swigc__p_egglib__VcfWindowSliderPerSite,
  _swigc__p_egglib__VectorInt,
  _swigc__p_egglib__WSite,
  _swigc__p_egglib__vcf__Alt,
  _swigc__p_egglib__vcf__Filter,
  _swigc__p_egglib__vcf__FlagInfo,
  _swigc__p_egglib__vcf__Format,
  _swigc__p_egglib__vcf__Info,
  _swigc__p_egglib__vcf__Meta,
  _swigc__p_egglib__vcf__SampleInfo,
  _swigc__p_egglib__vcf__StringInfo,
  _swigc__p_egglib__vcf__TypeInfoT_char_p_t,
  _swigc__p_egglib__vcf__TypeInfoT_char_t,
  _swigc__p_egglib__vcf__TypeInfoT_double_t,
  _swigc__p_egglib__vcf__TypeInfoT_int_t,
  _swigc__p_int,
  _swigc__p_std__istream,
  _swigc__p_std__string,
  _swigc__p_unsigned_int,
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (END) -------- */

static swig_const_info swig_const_table[] = {
{0, 0, 0, 0.0, 0, 0}};

#ifdef __cplusplus
}
#endif
static PyTypeObject *builtin_bases[3];

/* -----------------------------------------------------------------------------
 * Type initialization:
 * This problem is tough by the requirement that no dynamic
 * memory is used. Also, since swig_type_info structures store pointers to
 * swig_cast_info structures and swig_cast_info structures store pointers back
 * to swig_type_info structures, we need some lookup code at initialization.
 * The idea is that swig generates all the structures that are needed.
 * The runtime then collects these partially filled structures.
 * The SWIG_InitializeModule function takes these initial arrays out of
 * swig_module, and does all the lookup, filling in the swig_module.types
 * array with the correct data and linking the correct swig_cast_info
 * structures together.
 *
 * The generated swig_type_info structures are assigned statically to an initial
 * array. We just loop through that array, and handle each type individually.
 * First we lookup if this type has been already loaded, and if so, use the
 * loaded structure instead of the generated one. Then we have to fill in the
 * cast linked list. The cast data is initially stored in something like a
 * two-dimensional array. Each row corresponds to a type (there are the same
 * number of rows as there are in the swig_type_initial array). Each entry in
 * a column is one of the swig_cast_info structures for that type.
 * The cast_initial array is actually an array of arrays, because each row has
 * a variable number of columns. So to actually build the cast linked list,
 * we find the array of casts associated with the type, and loop through it
 * adding the casts to the list. The one last trick we need to do is making
 * sure the type pointer in the swig_cast_info struct is correct.
 *
 * First off, we lookup the cast->type name to see if it is already loaded.
 * There are three cases to handle:
 *  1) If the cast->type has already been loaded AND the type we are adding
 *     casting info to has not been loaded (it is in this module), THEN we
 *     replace the cast->type pointer with the type pointer that has already
 *     been loaded.
 *  2) If BOTH types (the one we are adding casting info to, and the
 *     cast->type) are loaded, THEN the cast info has already been loaded by
 *     the previous module so we just ignore it.
 *  3) Finally, if cast->type has not already been loaded, then we add that
 *     swig_cast_info to the linked list (because the cast->type) pointer will
 *     be correct.
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#if 0
} /* c-mode */
#endif
#endif

#if 0
#define SWIGRUNTIME_DEBUG
#endif


SWIGRUNTIME void
SWIG_InitializeModule(void *clientdata) {
  size_t i;
  swig_module_info *module_head, *iter;
  int init;
  
  /* check to see if the circular list has been setup, if not, set it up */
  if (swig_module.next==0) {
    /* Initialize the swig_module */
    swig_module.type_initial = swig_type_initial;
    swig_module.cast_initial = swig_cast_initial;
    swig_module.next = &swig_module;
    init = 1;
  } else {
    init = 0;
  }
  
  /* Try and load any already created modules */
  module_head = SWIG_GetModule(clientdata);
  if (!module_head) {
    /* This is the first module loaded for this interpreter */
    /* so set the swig module into the interpreter */
    SWIG_SetModule(clientdata, &swig_module);
  } else {
    /* the interpreter has loaded a SWIG module, but has it loaded this one? */
    iter=module_head;
    do {
      if (iter==&swig_module) {
        /* Our module is already in the list, so there's nothing more to do. */
        return;
      }
      iter=iter->next;
    } while (iter!= module_head);
    
    /* otherwise we must add our module into the list */
    swig_module.next = module_head->next;
    module_head->next = &swig_module;
  }
  
  /* When multiple interpreters are used, a module could have already been initialized in
       a different interpreter, but not yet have a pointer in this interpreter.
       In this case, we do not want to continue adding types... everything should be
       set up already */
  if (init == 0) return;
  
  /* Now work on filling in swig_module.types */
#ifdef SWIGRUNTIME_DEBUG
  printf("SWIG_InitializeModule: size %lu\n", (unsigned long)swig_module.size);
#endif
  for (i = 0; i < swig_module.size; ++i) {
    swig_type_info *type = 0;
    swig_type_info *ret;
    swig_cast_info *cast;
    
#ifdef SWIGRUNTIME_DEBUG
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
#endif
    
    /* if there is another module already loaded */
    if (swig_module.next != &swig_module) {
      type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, swig_module.type_initial[i]->name);
    }
    if (type) {
      /* Overwrite clientdata field */
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found type %s\n", type->name);
#endif
      if (swig_module.type_initial[i]->clientdata) {
        type->clientdata = swig_module.type_initial[i]->clientdata;
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: found and overwrite type %s \n", type->name);
#endif
      }
    } else {
      type = swig_module.type_initial[i];
    }
    
    /* Insert casting types */
    cast = swig_module.cast_initial[i];
    while (cast->type) {
      /* Don't need to add information already in the list */
      ret = 0;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: look cast %s\n", cast->type->name);
#endif
      if (swig_module.next != &swig_module) {
        ret = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, cast->type->name);
#ifdef SWIGRUNTIME_DEBUG
        if (ret) printf("SWIG_InitializeModule: found cast %s\n", ret->name);
#endif
      }
      if (ret) {
        if (type == swig_module.type_initial[i]) {
#ifdef SWIGRUNTIME_DEBUG
          printf("SWIG_InitializeModule: skip old type %s\n", ret->name);
#endif
          cast->type = ret;
          ret = 0;
        } else {
          /* Check for casting already in the list */
          swig_cast_info *ocast = SWIG_TypeCheck(ret->name, type);
#ifdef SWIGRUNTIME_DEBUG
          if (ocast) printf("SWIG_InitializeModule: skip old cast %s\n", ret->name);
#endif
          if (!ocast) ret = 0;
        }
      }
      
      if (!ret) {
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: adding cast %s\n", cast->type->name);
#endif
        if (type->cast) {
          type->cast->prev = cast;
          cast->next = type->cast;
        }
        type->cast = cast;
      }
      cast++;
    }
    /* Set entry in modules->types array equal to the type */
    swig_module.types[i] = type;
  }
  swig_module.types[i] = 0;
  
#ifdef SWIGRUNTIME_DEBUG
  printf("**** SWIG_InitializeModule: Cast List ******\n");
  for (i = 0; i < swig_module.size; ++i) {
    int j = 0;
    swig_cast_info *cast = swig_module.cast_initial[i];
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
    while (cast->type) {
      printf("SWIG_InitializeModule: cast type %s\n", cast->type->name);
      cast++;
      ++j;
    }
    printf("---- Total casts: %d\n",j);
  }
  printf("**** SWIG_InitializeModule: Cast List ******\n");
#endif
}

/* This function will propagate the clientdata field of type to
* any new swig_type_info structures that have been added into the list
* of equivalent types.  It is like calling
* SWIG_TypeClientData(type, clientdata) a second time.
*/
SWIGRUNTIME void
SWIG_PropagateClientData(void) {
  size_t i;
  swig_cast_info *equiv;
  static int init_run = 0;
  
  if (init_run) return;
  init_run = 1;
  
  for (i = 0; i < swig_module.size; i++) {
    if (swig_module.types[i]->clientdata) {
      equiv = swig_module.types[i]->cast;
      while (equiv) {
        if (!equiv->converter) {
          if (equiv->type && !equiv->type->clientdata)
          SWIG_TypeClientData(equiv->type, swig_module.types[i]->clientdata);
        }
        equiv = equiv->next;
      }
    }
  }
}

#ifdef __cplusplus
#if 0
{
  /* c-mode */
#endif
}
#endif



#ifdef __cplusplus
extern "C" {
#endif
  
  /* Python-specific SWIG API */
#define SWIG_newvarlink()                             SWIG_Python_newvarlink()
#define SWIG_addvarlink(p, name, get_attr, set_attr)  SWIG_Python_addvarlink(p, name, get_attr, set_attr)
#define SWIG_InstallConstants(d, constants)           SWIG_Python_InstallConstants(d, constants)
  
  /* -----------------------------------------------------------------------------
   * global variable support code.
   * ----------------------------------------------------------------------------- */
  
  typedef struct swig_globalvar {
    char       *name;                  /* Name of global variable */
    PyObject *(*get_attr)(void);       /* Return the current value */
    int       (*set_attr)(PyObject *); /* Set the value */
    struct swig_globalvar *next;
  } swig_globalvar;
  
  typedef struct swig_varlinkobject {
    PyObject_HEAD
    swig_globalvar *vars;
  } swig_varlinkobject;
  
  SWIGINTERN PyObject *
  swig_varlink_repr(swig_varlinkobject *SWIGUNUSEDPARM(v)) {
#if PY_VERSION_HEX >= 0x03000000
    return PyUnicode_InternFromString("<Swig global variables>");
#else
    return PyString_FromString("<Swig global variables>");
#endif
  }
  
  SWIGINTERN PyObject *
  swig_varlink_str(swig_varlinkobject *v) {
#if PY_VERSION_HEX >= 0x03000000
    PyObject *str = PyUnicode_InternFromString("(");
    PyObject *tail;
    PyObject *joined;
    swig_globalvar *var;
    for (var = v->vars; var; var=var->next) {
      tail = PyUnicode_FromString(var->name);
      joined = PyUnicode_Concat(str, tail);
      Py_DecRef(str);
      Py_DecRef(tail);
      str = joined;
      if (var->next) {
        tail = PyUnicode_InternFromString(", ");
        joined = PyUnicode_Concat(str, tail);
        Py_DecRef(str);
        Py_DecRef(tail);
        str = joined;
      }
    }
    tail = PyUnicode_InternFromString(")");
    joined = PyUnicode_Concat(str, tail);
    Py_DecRef(str);
    Py_DecRef(tail);
    str = joined;
#else
    PyObject *str = PyString_FromString("(");
    swig_globalvar *var;
    for (var = v->vars; var; var=var->next) {
      PyString_ConcatAndDel(&str,PyString_FromString(var->name));
      if (var->next) PyString_ConcatAndDel(&str,PyString_FromString(", "));
    }
    PyString_ConcatAndDel(&str,PyString_FromString(")"));
#endif
    return str;
  }
  
  SWIGINTERN void
  swig_varlink_dealloc(swig_varlinkobject *v) {
    swig_globalvar *var = v->vars;
    while (var) {
      swig_globalvar *n = var->next;
      free(var->name);
      free(var);
      var = n;
    }
  }
  
  SWIGINTERN PyObject *
  swig_varlink_getattr(swig_varlinkobject *v, char *n) {
    PyObject *res = NULL;
    swig_globalvar *var = v->vars;
    while (var) {
      if (strcmp(var->name,n) == 0) {
        res = (*var->get_attr)();
        break;
      }
      var = var->next;
    }
    if (res == NULL && !PyErr_Occurred()) {
      PyErr_Format(PyExc_AttributeError, "Unknown C global variable '%s'", n);
    }
    return res;
  }
  
  SWIGINTERN int
  swig_varlink_setattr(swig_varlinkobject *v, char *n, PyObject *p) {
    int res = 1;
    swig_globalvar *var = v->vars;
    while (var) {
      if (strcmp(var->name,n) == 0) {
        res = (*var->set_attr)(p);
        break;
      }
      var = var->next;
    }
    if (res == 1 && !PyErr_Occurred()) {
      PyErr_Format(PyExc_AttributeError, "Unknown C global variable '%s'", n);
    }
    return res;
  }
  
  SWIGINTERN PyTypeObject*
  swig_varlink_type(void) {
    static char varlink__doc__[] = "Swig var link object";
    static PyTypeObject varlink_type;
    static int type_init = 0;
    if (!type_init) {
      const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
        PyVarObject_HEAD_INIT(NULL, 0)
#else
        PyObject_HEAD_INIT(NULL)
        0,                                  /* ob_size */
#endif
        "swigvarlink",                      /* tp_name */
        sizeof(swig_varlinkobject),         /* tp_basicsize */
        0,                                  /* tp_itemsize */
        (destructor) swig_varlink_dealloc,  /* tp_dealloc */
        0,                                  /* tp_print */
        (getattrfunc) swig_varlink_getattr, /* tp_getattr */
        (setattrfunc) swig_varlink_setattr, /* tp_setattr */
        0,                                  /* tp_compare */
        (reprfunc) swig_varlink_repr,       /* tp_repr */
        0,                                  /* tp_as_number */
        0,                                  /* tp_as_sequence */
        0,                                  /* tp_as_mapping */
        0,                                  /* tp_hash */
        0,                                  /* tp_call */
        (reprfunc) swig_varlink_str,        /* tp_str */
        0,                                  /* tp_getattro */
        0,                                  /* tp_setattro */
        0,                                  /* tp_as_buffer */
        0,                                  /* tp_flags */
        varlink__doc__,                     /* tp_doc */
        0,                                  /* tp_traverse */
        0,                                  /* tp_clear */
        0,                                  /* tp_richcompare */
        0,                                  /* tp_weaklistoffset */
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* tp_iter -> tp_weaklist */
        0,                                  /* tp_del */
        0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
        0,                                  /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
        0,                                  /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
        0,                                  /* tp_print */
#endif
#ifdef COUNT_ALLOCS
        0,                                  /* tp_allocs */
        0,                                  /* tp_frees */
        0,                                  /* tp_maxalloc */
        0,                                  /* tp_prev */
        0                                   /* tp_next */
#endif
      };
      varlink_type = tmp;
      type_init = 1;
      if (PyType_Ready(&varlink_type) < 0)
      return NULL;
    }
    return &varlink_type;
  }
  
  /* Create a variable linking object for use later */
  SWIGINTERN PyObject *
  SWIG_Python_newvarlink(void) {
    swig_varlinkobject *result = PyObject_NEW(swig_varlinkobject, swig_varlink_type());
    if (result) {
      result->vars = 0;
    }
    return ((PyObject*) result);
  }
  
  SWIGINTERN void 
  SWIG_Python_addvarlink(PyObject *p, const char *name, PyObject *(*get_attr)(void), int (*set_attr)(PyObject *p)) {
    swig_varlinkobject *v = (swig_varlinkobject *) p;
    swig_globalvar *gv = (swig_globalvar *) malloc(sizeof(swig_globalvar));
    if (gv) {
      size_t size = strlen(name)+1;
      gv->name = (char *)malloc(size);
      if (gv->name) {
        memcpy(gv->name, name, size);
        gv->get_attr = get_attr;
        gv->set_attr = set_attr;
        gv->next = v->vars;
      }
    }
    v->vars = gv;
  }
  
  SWIGINTERN PyObject *
  SWIG_globals(void) {
    static PyObject *globals = 0;
    if (!globals) {
      globals = SWIG_newvarlink();
    }
    return globals;
  }
  
  /* -----------------------------------------------------------------------------
   * constants/methods manipulation
   * ----------------------------------------------------------------------------- */
  
  /* Install Constants */
  SWIGINTERN void
  SWIG_Python_InstallConstants(PyObject *d, swig_const_info constants[]) {
    PyObject *obj = 0;
    size_t i;
    for (i = 0; constants[i].type; ++i) {
      switch(constants[i].type) {
      case SWIG_PY_POINTER:
        obj = SWIG_InternalNewPointerObj(constants[i].pvalue, *(constants[i]).ptype,0);
        break;
      case SWIG_PY_BINARY:
        obj = SWIG_NewPackedObj(constants[i].pvalue, constants[i].lvalue, *(constants[i].ptype));
        break;
      default:
        obj = 0;
        break;
      }
      if (obj) {
        PyDict_SetItemString(d, constants[i].name, obj);
        Py_DECREF(obj);
      }
    }
  }
  
  /* -----------------------------------------------------------------------------*/
  /* Fix SwigMethods to carry the callback ptrs when needed */
  /* -----------------------------------------------------------------------------*/
  
  SWIGINTERN void
  SWIG_Python_FixMethods(PyMethodDef *methods,
    swig_const_info *const_table,
    swig_type_info **types,
    swig_type_info **types_initial) {
    size_t i;
    for (i = 0; methods[i].ml_name; ++i) {
      const char *c = methods[i].ml_doc;
      if (!c) continue;
      c = strstr(c, "swig_ptr: ");
      if (c) {
        int j;
        swig_const_info *ci = 0;
        const char *name = c + 10;
        for (j = 0; const_table[j].type; ++j) {
          if (strncmp(const_table[j].name, name, 
              strlen(const_table[j].name)) == 0) {
            ci = &(const_table[j]);
            break;
          }
        }
        if (ci) {
          void *ptr = (ci->type == SWIG_PY_POINTER) ? ci->pvalue : 0;
          if (ptr) {
            size_t shift = (ci->ptype) - types;
            swig_type_info *ty = types_initial[shift];
            size_t ldoc = (c - methods[i].ml_doc);
            size_t lptr = strlen(ty->name)+2*sizeof(void*)+2;
            char *ndoc = (char*)malloc(ldoc + lptr + 10);
            if (ndoc) {
              char *buff = ndoc;
              memcpy(buff, methods[i].ml_doc, ldoc);
              buff += ldoc;
              memcpy(buff, "swig_ptr: ", 10);
              buff += 10;
              SWIG_PackVoidPtr(buff, ptr, ty->name, lptr);
              methods[i].ml_doc = ndoc;
            }
          }
        }
      }
    }
  } 
  
  /* -----------------------------------------------------------------------------
   * Method creation and docstring support functions
   * ----------------------------------------------------------------------------- */
  
  /* -----------------------------------------------------------------------------
   * Function to find the method definition with the correct docstring for the
   * proxy module as opposed to the low-level API
   * ----------------------------------------------------------------------------- */
  
  SWIGINTERN PyMethodDef *SWIG_PythonGetProxyDoc(const char *name) {
    /* Find the function in the modified method table */
    size_t offset = 0;
    int found = 0;
    while (SwigMethods_proxydocs[offset].ml_meth != NULL) {
      if (strcmp(SwigMethods_proxydocs[offset].ml_name, name) == 0) {
        found = 1;
        break;
      }
      offset++;
    }
    /* Use the copy with the modified docstring if available */
    return found ? &SwigMethods_proxydocs[offset] : NULL;
  }
  
  /* -----------------------------------------------------------------------------
   * Wrapper of PyInstanceMethod_New() used in Python 3
   * It is exported to the generated module, used for -fastproxy
   * ----------------------------------------------------------------------------- */
  
  SWIGINTERN PyObject *SWIG_PyInstanceMethod_New(PyObject *SWIGUNUSEDPARM(self), PyObject *func) {
    if (PyCFunction_Check(func)) {
      PyCFunctionObject *funcobj = (PyCFunctionObject *)func;
      PyMethodDef *ml = SWIG_PythonGetProxyDoc(funcobj->m_ml->ml_name);
      if (ml)
      func = PyCFunction_NewEx(ml, funcobj->m_self, funcobj->m_module);
    }
#if PY_VERSION_HEX >= 0x03000000
    return PyInstanceMethod_New(func);
#else
    return PyMethod_New(func, NULL, NULL);
#endif
  }
  
  /* -----------------------------------------------------------------------------
   * Wrapper of PyStaticMethod_New()
   * It is exported to the generated module, used for -fastproxy
   * ----------------------------------------------------------------------------- */
  
  SWIGINTERN PyObject *SWIG_PyStaticMethod_New(PyObject *SWIGUNUSEDPARM(self), PyObject *func) {
    if (PyCFunction_Check(func)) {
      PyCFunctionObject *funcobj = (PyCFunctionObject *)func;
      PyMethodDef *ml = SWIG_PythonGetProxyDoc(funcobj->m_ml->ml_name);
      if (ml)
      func = PyCFunction_NewEx(ml, funcobj->m_self, funcobj->m_module);
    }
    return PyStaticMethod_New(func);
  }
  
#ifdef __cplusplus
}
#endif

/* -----------------------------------------------------------------------------*
 *  Partial Init method
 * -----------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C"
#endif

SWIGEXPORT 
#if PY_VERSION_HEX >= 0x03000000
PyObject*
#else
void
#endif
SWIG_init(void) {
  PyObject *m, *d, *md, *globals;
  
#if PY_VERSION_HEX >= 0x03000000
  static struct PyModuleDef SWIG_module = {
    PyModuleDef_HEAD_INIT,
    SWIG_name,
    NULL,
    -1,
    SwigMethods,
    NULL,
    NULL,
    NULL,
    NULL
  };
#endif
  
#if defined(SWIGPYTHON_BUILTIN)
  static SwigPyClientData SwigPyObject_clientdata = {
    0, 0, 0, 0, 0, 0, 0
  };
  static PyGetSetDef this_getset_def = {
    (char *)"this", &SwigPyBuiltin_ThisClosure, NULL, NULL, NULL
  };
  static SwigPyGetSet thisown_getset_closure = {
    SwigPyObject_own,
    SwigPyObject_own
  };
  static PyGetSetDef thisown_getset_def = {
    (char *)"thisown", SwigPyBuiltin_GetterClosure, SwigPyBuiltin_SetterClosure, NULL, &thisown_getset_closure
  };
  PyTypeObject *builtin_pytype;
  int builtin_base_count;
  swig_type_info *builtin_basetype;
  PyObject *tuple;
  PyGetSetDescrObject *static_getset;
  PyTypeObject *metatype;
  PyTypeObject *swigpyobject;
  SwigPyClientData *cd;
  PyObject *public_interface, *public_symbol;
  PyObject *this_descr;
  PyObject *thisown_descr;
  PyObject *self = 0;
  int i;
  
  (void)builtin_pytype;
  (void)builtin_base_count;
  (void)builtin_basetype;
  (void)tuple;
  (void)static_getset;
  (void)self;
  
  /* Metaclass is used to implement static member variables */
  metatype = SwigPyObjectType();
  assert(metatype);
#endif
  
  (void)globals;
  
  /* Create singletons now to avoid potential deadlocks with multi-threaded usage after module initialization */
  SWIG_This();
  SWIG_Python_TypeCache();
  SwigPyPacked_type();
#ifndef SWIGPYTHON_BUILTIN
  SwigPyObject_type();
#endif
  
  /* Fix SwigMethods to carry the callback ptrs when needed */
  SWIG_Python_FixMethods(SwigMethods, swig_const_table, swig_types, swig_type_initial);
  
#if PY_VERSION_HEX >= 0x03000000
  m = PyModule_Create(&SWIG_module);
#else
  m = Py_InitModule(SWIG_name, SwigMethods);
#endif
  
  md = d = PyModule_GetDict(m);
  (void)md;
  
  SWIG_InitializeModule(0);
  
#ifdef SWIGPYTHON_BUILTIN
  swigpyobject = SwigPyObject_TypeOnce();
  
  SwigPyObject_stype = SWIG_MangledTypeQuery("_p_SwigPyObject");
  assert(SwigPyObject_stype);
  cd = (SwigPyClientData*) SwigPyObject_stype->clientdata;
  if (!cd) {
    SwigPyObject_stype->clientdata = &SwigPyObject_clientdata;
    SwigPyObject_clientdata.pytype = swigpyobject;
  } else if (swigpyobject->tp_basicsize != cd->pytype->tp_basicsize) {
    PyErr_SetString(PyExc_RuntimeError, "Import error: attempted to load two incompatible swig-generated modules.");
# if PY_VERSION_HEX >= 0x03000000
    return NULL;
# else
    return;
# endif
  }
  
  /* All objects have a 'this' attribute */
  this_descr = PyDescr_NewGetSet(SwigPyObject_type(), &this_getset_def);
  (void)this_descr;
  
  /* All objects have a 'thisown' attribute */
  thisown_descr = PyDescr_NewGetSet(SwigPyObject_type(), &thisown_getset_def);
  (void)thisown_descr;
  
  public_interface = PyList_New(0);
  public_symbol = 0;
  (void)public_symbol;
  
  PyDict_SetItemString(md, "__all__", public_interface);
  Py_DECREF(public_interface);
  for (i = 0; SwigMethods[i].ml_name != NULL; ++i)
  SwigPyBuiltin_AddPublicSymbol(public_interface, SwigMethods[i].ml_name);
  for (i = 0; swig_const_table[i].name != 0; ++i)
  SwigPyBuiltin_AddPublicSymbol(public_interface, swig_const_table[i].name);
#endif
  
  SWIG_InstallConstants(d,swig_const_table);
  
  
  /* type 'egglib::VectorInt' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__VectorInt_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'VectorInt'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "VectorInt", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "VectorInt");
  d = md;
  
  /* type 'egglib::AbstractBaseAlphabet' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__AbstractBaseAlphabet_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'AbstractBaseAlphabet'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "AbstractBaseAlphabet", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "AbstractBaseAlphabet");
  d = md;
  
  /* type 'egglib::AbstractTemplateAlphabet< int >' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_int_t_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__AbstractBaseAlphabet");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'IntBaseAlphabet' as base 'egglib::AbstractBaseAlphabet' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'IntBaseAlphabet'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "IntBaseAlphabet", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "IntBaseAlphabet");
  d = md;
  
  /* type 'egglib::AbstractTemplateAlphabet< char >' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_t_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__AbstractBaseAlphabet");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'CharBaseAlphabet' as base 'egglib::AbstractBaseAlphabet' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'CharBaseAlphabet'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "CharBaseAlphabet", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "CharBaseAlphabet");
  d = md;
  
  /* type 'egglib::AbstractTemplateAlphabet< char * >' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__AbstractTemplateAlphabetT_char_p_t_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__AbstractBaseAlphabet");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'CStringBaseAlphabet' as base 'egglib::AbstractBaseAlphabet' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'CStringBaseAlphabet'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "CStringBaseAlphabet", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "CStringBaseAlphabet");
  d = md;
  
  /* type 'egglib::FiniteAlphabet< int >' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__FiniteAlphabetT_int_t_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__AbstractTemplateAlphabetT_int_t");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'IntAlphabet' as base 'egglib::AbstractTemplateAlphabet<(int)>' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'IntAlphabet'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "IntAlphabet", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "IntAlphabet");
  d = md;
  
  /* type 'egglib::FiniteAlphabet< char >' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__FiniteAlphabetT_char_t_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__AbstractTemplateAlphabetT_char_t");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'CharAlphabet' as base 'egglib::AbstractTemplateAlphabet<(char)>' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'CharAlphabet'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "CharAlphabet", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "CharAlphabet");
  d = md;
  
  /* type 'egglib::FiniteAlphabet< char * >' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__FiniteAlphabetT_char_p_t_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__AbstractTemplateAlphabetT_char_p_t");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'CStringAlphabet' as base 'egglib::AbstractTemplateAlphabet<(p.char)>' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'CStringAlphabet'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "CStringAlphabet", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "CStringAlphabet");
  d = md;
  
  /* type 'egglib::CaseInsensitiveCharAlphabet' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__CaseInsensitiveCharAlphabet_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__FiniteAlphabetT_char_t");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'CaseInsensitiveCharAlphabet' as base 'egglib::FiniteAlphabet<(char)>' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'CaseInsensitiveCharAlphabet'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "CaseInsensitiveCharAlphabet", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "CaseInsensitiveCharAlphabet");
  d = md;
  
  /* type 'egglib::DNAAlphabet' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__DNAAlphabet_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__CaseInsensitiveCharAlphabet");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'DNAAlphabet' as base 'egglib::CaseInsensitiveCharAlphabet' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'DNAAlphabet'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "DNAAlphabet", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "DNAAlphabet");
  d = md;
  
  /* type 'egglib::StringAlphabet' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__StringAlphabet_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__FiniteAlphabetT_char_p_t");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'StringAlphabet' as base 'egglib::FiniteAlphabet<(p.char)>' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'StringAlphabet'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "StringAlphabet", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "StringAlphabet");
  d = md;
  
  /* type 'egglib::CodonAlphabet' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__CodonAlphabet_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__StringAlphabet");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'CodonAlphabet' as base 'egglib::StringAlphabet' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'CodonAlphabet'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "CodonAlphabet", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "CodonAlphabet");
  d = md;
  
  /* type 'egglib::CaseInsensitiveStringAlphabet' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__CaseInsensitiveStringAlphabet_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__StringAlphabet");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'CaseInsensitiveStringAlphabet' as base 'egglib::StringAlphabet' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'CaseInsensitiveStringAlphabet'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "CaseInsensitiveStringAlphabet", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "CaseInsensitiveStringAlphabet");
  d = md;
  
  /* type 'egglib::CustomStringAlphabet' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__CustomStringAlphabet_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__StringAlphabet");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'CustomStringAlphabet' as base 'egglib::StringAlphabet' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'CustomStringAlphabet'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "CustomStringAlphabet", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "CustomStringAlphabet");
  d = md;
  
  /* type 'egglib::RangeAlphabet' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__RangeAlphabet_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__AbstractTemplateAlphabetT_int_t");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'RangeAlphabet' as base 'egglib::AbstractTemplateAlphabet<(int)>' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'RangeAlphabet'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "RangeAlphabet", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "RangeAlphabet");
  d = md;
  
  /* type 'egglib::DataHolder' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__DataHolder_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'DataHolder'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "DataHolder", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "DataHolder");
  d = md;
  
  /* type 'egglib::IntersperseAlign' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__IntersperseAlign_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'IntersperseAlign'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "IntersperseAlign", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "IntersperseAlign");
  d = md;
  
  /* type 'egglib::StructureHolder' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__StructureHolder_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'StructureHolder'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "StructureHolder", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "StructureHolder");
  d = md;
  
  /* type 'egglib::StructureCluster' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__StructureCluster_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'StructureCluster'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "StructureCluster", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "StructureCluster");
  d = md;
  
  /* type 'egglib::StructurePopulation' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__StructurePopulation_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'StructurePopulation'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "StructurePopulation", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "StructurePopulation");
  d = md;
  
  /* type 'egglib::StructureIndiv' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__StructureIndiv_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'StructureIndiv'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "StructureIndiv", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "StructureIndiv");
  d = md;
  
  /* type 'egglib::FreqSet' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__FreqSet_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'FreqSet'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "FreqSet", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "FreqSet");
  d = md;
  
  /* type 'egglib::FreqHolder' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__FreqHolder_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'FreqHolder'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "FreqHolder", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "FreqHolder");
  d = md;
  
  /* type 'egglib::AlleleStatus' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__AlleleStatus_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'AlleleStatus'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "AlleleStatus", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "AlleleStatus");
  d = md;
  
  /* type 'egglib::GeneticCode' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__GeneticCode_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'GeneticCode'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "GeneticCode", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "GeneticCode");
  d = md;
  
  /* type 'egglib::SiteDiversity' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__SiteDiversity_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'SiteDiversity'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "SiteDiversity", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "SiteDiversity");
  d = md;
  
  /* type 'egglib::SiteHolder' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__SiteHolder_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'SiteHolder'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "SiteHolder", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "SiteHolder");
  d = md;
  
  /* type 'egglib::Genotypes' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__Genotypes_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'Genotypes'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "Genotypes", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "Genotypes");
  d = md;
  
  /* type 'egglib::Diversity1' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__Diversity1_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'Diversity1'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "Diversity1", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "Diversity1");
  d = md;
  
  /* type 'egglib::Diversity2' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__Diversity2_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'Diversity2'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "Diversity2", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "Diversity2");
  d = md;
  
  /* type 'egglib::Haplotypes' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__Haplotypes_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'Haplotypes'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "Haplotypes", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "Haplotypes");
  d = md;
  
  /* type 'egglib::PairwiseLD' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__PairwiseLD_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'PairwiseLD'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "PairwiseLD", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "PairwiseLD");
  d = md;
  
  /* type 'egglib::MatrixLD' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__MatrixLD_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "ignore",SWIG_From_int(static_cast< int >(egglib::MatrixLD::ignore)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "use_main",SWIG_From_int(static_cast< int >(egglib::MatrixLD::use_main)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "use_all",SWIG_From_int(static_cast< int >(egglib::MatrixLD::use_all)));
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'MatrixLD'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "MatrixLD", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "MatrixLD");
  d = md;
  
  /* type 'egglib::ParalogPi' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__ParalogPi_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'ParalogPi'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "ParalogPi", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "ParalogPi");
  d = md;
  
  /* type 'egglib::Rd' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__Rd_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'Rd'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "Rd", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "Rd");
  d = md;
  
  /* type 'egglib::Migration' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__Migration_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'Migration'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "Migration", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "Migration");
  d = md;
  
  /* type 'egglib::Event' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__Event_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "none",SWIG_From_int(static_cast< int >(egglib::Event::none)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "change_N",SWIG_From_int(static_cast< int >(egglib::Event::change_N)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "change_M",SWIG_From_int(static_cast< int >(egglib::Event::change_M)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "change_Mp",SWIG_From_int(static_cast< int >(egglib::Event::change_Mp)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "change_G",SWIG_From_int(static_cast< int >(egglib::Event::change_G)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "change_s",SWIG_From_int(static_cast< int >(egglib::Event::change_s)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "change_R",SWIG_From_int(static_cast< int >(egglib::Event::change_R)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "bottleneck",SWIG_From_int(static_cast< int >(egglib::Event::bottleneck)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "admixture",SWIG_From_int(static_cast< int >(egglib::Event::admixture)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "delayed",SWIG_From_int(static_cast< int >(egglib::Event::delayed)));
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'Event'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "Event", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "Event");
  d = md;
  
  /* type 'egglib::Params' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__Params_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "KAM",SWIG_From_int(static_cast< int >(egglib::Params::KAM)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "IAM",SWIG_From_int(static_cast< int >(egglib::Params::IAM)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "SMM",SWIG_From_int(static_cast< int >(egglib::Params::SMM)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "TPM",SWIG_From_int(static_cast< int >(egglib::Params::TPM)));
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'Params'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "Params", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "Params");
  d = md;
  
  /* type 'egglib::Node' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__Node_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'Node'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "Node", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "Node");
  d = md;
  
  /* type 'egglib::Tree' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__Tree_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'Tree'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "Tree", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "Tree");
  d = md;
  
  /* type 'egglib::Coalesce' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__Coalesce_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'Coalesce'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "Coalesce", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "Coalesce");
  d = md;
  
  /* type 'egglib::EHH' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__EHH_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'EHH'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "EHH", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "EHH");
  d = md;
  
  /* type 'egglib::BaseFormatter' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__BaseFormatter_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'BaseFormatter'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "BaseFormatter", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "BaseFormatter");
  d = md;
  
  /* type 'egglib::Export' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__Export_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__BaseFormatter");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'Export' as base 'egglib::BaseFormatter' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'Export'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "Export", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "Export");
  d = md;
  
  /* type 'egglib::FastaFormatter' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__FastaFormatter_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__BaseFormatter");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'FastaFormatter' as base 'egglib::BaseFormatter' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'FastaFormatter'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "FastaFormatter", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "FastaFormatter");
  d = md;
  
  /* type 'egglib::FastaParser' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__FastaParser_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'FastaParser'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "FastaParser", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "FastaParser");
  d = md;
  
  /* type 'egglib::Feature' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__Feature_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "zero",SWIG_From_int(static_cast< int >(egglib::Feature::zero)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "one",SWIG_From_int(static_cast< int >(egglib::Feature::one)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "two",SWIG_From_int(static_cast< int >(egglib::Feature::two)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "no_phase",SWIG_From_int(static_cast< int >(egglib::Feature::no_phase)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "plus",SWIG_From_int(static_cast< int >(egglib::Feature::plus)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "minus",SWIG_From_int(static_cast< int >(egglib::Feature::minus)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "no_strand",SWIG_From_int(static_cast< int >(egglib::Feature::no_strand)));
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'Feature'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "Feature", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "Feature");
  d = md;
  
  /* type 'egglib::GFF3' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__GFF3_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'GFF3'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "GFF3", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "GFF3");
  d = md;
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "Default",SWIG_From_int(static_cast< int >(egglib::vcf::Default)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "Referred",SWIG_From_int(static_cast< int >(egglib::vcf::Referred)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "Breakend",SWIG_From_int(static_cast< int >(egglib::vcf::Breakend)));
  globals = SWIG_globals();
  if (!globals) {
    PyErr_SetString(PyExc_TypeError, "Failure to create SWIG globals.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  PyDict_SetItemString(md, "cvar", globals);
  Py_DECREF(globals);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "cvar");
  SWIG_addvarlink(globals, "NUM_ALTERNATE", Swig_var_NUM_ALTERNATE_get, Swig_var_NUM_ALTERNATE_set);
  PyDict_SetItemString(md, "NUM_ALTERNATE", PyObject_GetAttrString(globals, "NUM_ALTERNATE"));
  SwigPyBuiltin_AddPublicSymbol(public_interface, "NUM_ALTERNATE");
  SWIG_addvarlink(globals, "NUM_GENOTYPES", Swig_var_NUM_GENOTYPES_get, Swig_var_NUM_GENOTYPES_set);
  PyDict_SetItemString(md, "NUM_GENOTYPES", PyObject_GetAttrString(globals, "NUM_GENOTYPES"));
  SwigPyBuiltin_AddPublicSymbol(public_interface, "NUM_GENOTYPES");
  SWIG_addvarlink(globals, "NUM_POSSIBLE_ALLELES", Swig_var_NUM_POSSIBLE_ALLELES_get, Swig_var_NUM_POSSIBLE_ALLELES_set);
  PyDict_SetItemString(md, "NUM_POSSIBLE_ALLELES", PyObject_GetAttrString(globals, "NUM_POSSIBLE_ALLELES"));
  SwigPyBuiltin_AddPublicSymbol(public_interface, "NUM_POSSIBLE_ALLELES");
  
  /* type 'egglib::vcf::Filter' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__Filter_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'VcfFilter'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "VcfFilter", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "VcfFilter");
  d = md;
  
  /* type 'egglib::vcf::Info' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__Info_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "Integer",SWIG_From_int(static_cast< int >(egglib::vcf::Info::Integer)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "Float",SWIG_From_int(static_cast< int >(egglib::vcf::Info::Float)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "Flag",SWIG_From_int(static_cast< int >(egglib::vcf::Info::Flag)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "Character",SWIG_From_int(static_cast< int >(egglib::vcf::Info::Character)));
  SWIG_Python_SetConstant(d, d == md ? public_interface : NULL, "String",SWIG_From_int(static_cast< int >(egglib::vcf::Info::String)));
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__vcf__Filter");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'Info' as base 'egglib::vcf::Filter' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'Info'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "Info", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "Info");
  d = md;
  
  /* type 'egglib::vcf::Format' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__Format_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__vcf__Info");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'Format' as base 'egglib::vcf::Info' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'Format'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "Format", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "Format");
  d = md;
  
  /* type 'egglib::vcf::Meta' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__Meta_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'Meta'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "Meta", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "Meta");
  d = md;
  
  /* type 'egglib::vcf::Alt' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__Alt_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__vcf__Filter");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'Alt' as base 'egglib::vcf::Filter' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'Alt'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "Alt", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "Alt");
  d = md;
  
  /* type 'egglib::vcf::FlagInfo' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__FlagInfo_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'FlagInfo'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "FlagInfo", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "FlagInfo");
  d = md;
  
  /* type 'egglib::vcf::TypeInfo< char * >' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__TypeInfoT_char_p_t_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__vcf__FlagInfo");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'TypeInfoString' as base 'egglib::vcf::FlagInfo' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'TypeInfoString'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "TypeInfoString", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "TypeInfoString");
  d = md;
  
  /* type 'egglib::vcf::TypeInfo< int >' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__TypeInfoT_int_t_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__vcf__FlagInfo");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'TypeInfoInt' as base 'egglib::vcf::FlagInfo' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'TypeInfoInt'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "TypeInfoInt", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "TypeInfoInt");
  d = md;
  
  /* type 'egglib::vcf::TypeInfo< char >' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__TypeInfoT_char_t_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__vcf__FlagInfo");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'TypeInfoChar' as base 'egglib::vcf::FlagInfo' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'TypeInfoChar'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "TypeInfoChar", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "TypeInfoChar");
  d = md;
  
  /* type 'egglib::vcf::TypeInfo< double >' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__TypeInfoT_double_t_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__vcf__FlagInfo");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'TypeInfoDouble' as base 'egglib::vcf::FlagInfo' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'TypeInfoDouble'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "TypeInfoDouble", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "TypeInfoDouble");
  d = md;
  
  /* type 'egglib::vcf::StringInfo' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__StringInfo_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__vcf__TypeInfoT_char_p_t");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'StringInfo' as base 'egglib::vcf::TypeInfo<(p.char)>' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'StringInfo'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "StringInfo", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "StringInfo");
  d = md;
  
  /* type 'egglib::vcf::SampleInfo' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__vcf__SampleInfo_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'SampleInfo'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "SampleInfo", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "SampleInfo");
  d = md;
  
  /* type 'egglib::VcfParser' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__VcfParser_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'VcfParser'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "VcfParser", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "VcfParser");
  d = md;
  
  /* type 'egglib::VcfIndex' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__VcfIndex_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'VcfIndex'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "VcfIndex", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "VcfIndex");
  d = md;
  
  /* type 'egglib::WSite' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__WSite_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'WSite'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "WSite", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "WSite");
  d = md;
  
  /* type 'egglib::VcfWindowBase' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__VcfWindowBase_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'VcfWindowBase'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "VcfWindowBase", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "VcfWindowBase");
  d = md;
  
  /* type 'egglib::VcfWindowSliderBase' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__VcfWindowSliderBase_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__VcfWindowBase");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'VcfWindowSliderBase' as base 'egglib::VcfWindowBase' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'VcfWindowSliderBase'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "VcfWindowSliderBase", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "VcfWindowSliderBase");
  d = md;
  
  /* type 'egglib::VcfWindowSlider' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__VcfWindowSlider_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__VcfWindowSliderBase");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'VcfWindowSlider' as base 'egglib::VcfWindowSliderBase' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'VcfWindowSlider'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "VcfWindowSlider", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "VcfWindowSlider");
  d = md;
  
  /* type 'egglib::VcfWindowSliderPerSite' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__VcfWindowSliderPerSite_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__VcfWindowSliderBase");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'VcfWindowSliderPerSite' as base 'egglib::VcfWindowSliderBase' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'VcfWindowSliderPerSite'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "VcfWindowSliderPerSite", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "VcfWindowSliderPerSite");
  d = md;
  
  /* type 'egglib::VcfWindowBED' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__VcfWindowBED_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_basetype = SWIG_MangledTypeQuery("_p_egglib__VcfWindowBase");
  if (builtin_basetype && builtin_basetype->clientdata && ((SwigPyClientData *) builtin_basetype->clientdata)->pytype) {
    builtin_bases[builtin_base_count++] = ((SwigPyClientData *) builtin_basetype->clientdata)->pytype;
  } else {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'VcfWindowBED' as base 'egglib::VcfWindowBase' has not been initialized.\n");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'VcfWindowBED'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "VcfWindowBED", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "VcfWindowBED");
  d = md;
  
  /* type 'egglib::BedParser' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__BedParser_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'BedParser'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "BedParser", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "BedParser");
  d = md;
  
  /* type 'egglib::ComputeV' */
  builtin_pytype = (PyTypeObject *)&SwigPyBuiltin__egglib__ComputeV_type;
  builtin_pytype->tp_dict = d = PyDict_New();
  SwigPyBuiltin_SetMetaType(builtin_pytype, metatype);
  builtin_pytype->tp_new = PyType_GenericNew;
  builtin_base_count = 0;
  builtin_bases[builtin_base_count] = NULL;
  SwigPyBuiltin_InitBases(builtin_pytype, builtin_bases);
  PyDict_SetItemString(d, "this", this_descr);
  PyDict_SetItemString(d, "thisown", thisown_descr);
  if (PyType_Ready(builtin_pytype) < 0) {
    PyErr_SetString(PyExc_TypeError, "Could not create type 'ComputeV'.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  Py_INCREF(builtin_pytype);
  PyModule_AddObject(m, "ComputeV", (PyObject *)builtin_pytype);
  SwigPyBuiltin_AddPublicSymbol(public_interface, "ComputeV");
  d = md;
  SWIG_addvarlink(globals, "MISSINGDATA", Swig_var_MISSINGDATA_get, Swig_var_MISSINGDATA_set);
  PyDict_SetItemString(md, "MISSINGDATA", PyObject_GetAttrString(globals, "MISSINGDATA"));
  SwigPyBuiltin_AddPublicSymbol(public_interface, "MISSINGDATA");
  SWIG_addvarlink(globals, "MAX_ALLELE_RANGE", Swig_var_MAX_ALLELE_RANGE_get, Swig_var_MAX_ALLELE_RANGE_set);
  PyDict_SetItemString(md, "MAX_ALLELE_RANGE", PyObject_GetAttrString(globals, "MAX_ALLELE_RANGE"));
  SwigPyBuiltin_AddPublicSymbol(public_interface, "MAX_ALLELE_RANGE");
  SWIG_addvarlink(globals, "UNDEF", Swig_var_UNDEF_get, Swig_var_UNDEF_set);
  PyDict_SetItemString(md, "UNDEF", PyObject_GetAttrString(globals, "UNDEF"));
  SwigPyBuiltin_AddPublicSymbol(public_interface, "UNDEF");
  SWIG_addvarlink(globals, "UNKNOWN", Swig_var_UNKNOWN_get, Swig_var_UNKNOWN_set);
  PyDict_SetItemString(md, "UNKNOWN", PyObject_GetAttrString(globals, "UNKNOWN"));
  SwigPyBuiltin_AddPublicSymbol(public_interface, "UNKNOWN");
  SWIG_addvarlink(globals, "MAX", Swig_var_MAX_get, Swig_var_MAX_set);
  PyDict_SetItemString(md, "MAX", PyObject_GetAttrString(globals, "MAX"));
  SwigPyBuiltin_AddPublicSymbol(public_interface, "MAX");
  SWIG_addvarlink(globals, "MAXCHAR", Swig_var_MAXCHAR_get, Swig_var_MAXCHAR_set);
  PyDict_SetItemString(md, "MAXCHAR", PyObject_GetAttrString(globals, "MAXCHAR"));
  SwigPyBuiltin_AddPublicSymbol(public_interface, "MAXCHAR");
  SWIG_addvarlink(globals, "OUTGROUP", Swig_var_OUTGROUP_get, Swig_var_OUTGROUP_set);
  PyDict_SetItemString(md, "OUTGROUP", PyObject_GetAttrString(globals, "OUTGROUP"));
  SwigPyBuiltin_AddPublicSymbol(public_interface, "OUTGROUP");
  SWIG_addvarlink(globals, "MISSING", Swig_var_MISSING_get, Swig_var_MISSING_set);
  PyDict_SetItemString(md, "MISSING", PyObject_GetAttrString(globals, "MISSING"));
  SwigPyBuiltin_AddPublicSymbol(public_interface, "MISSING");
  SWIG_addvarlink(globals, "BEFORE", Swig_var_BEFORE_get, Swig_var_BEFORE_set);
  PyDict_SetItemString(md, "BEFORE", PyObject_GetAttrString(globals, "BEFORE"));
  SwigPyBuiltin_AddPublicSymbol(public_interface, "BEFORE");
  SWIG_addvarlink(globals, "FIRST", Swig_var_FIRST_get, Swig_var_FIRST_set);
  PyDict_SetItemString(md, "FIRST", PyObject_GetAttrString(globals, "FIRST"));
  SwigPyBuiltin_AddPublicSymbol(public_interface, "FIRST");
  SWIG_addvarlink(globals, "LAST", Swig_var_LAST_get, Swig_var_LAST_set);
  PyDict_SetItemString(md, "LAST", PyObject_GetAttrString(globals, "LAST"));
  SwigPyBuiltin_AddPublicSymbol(public_interface, "LAST");
  SWIG_addvarlink(globals, "DEBUG_FLAG", Swig_var_DEBUG_FLAG_get, Swig_var_DEBUG_FLAG_set);
  PyDict_SetItemString(md, "DEBUG_FLAG", PyObject_GetAttrString(globals, "DEBUG_FLAG"));
  SwigPyBuiltin_AddPublicSymbol(public_interface, "DEBUG_FLAG");
#if PY_VERSION_HEX >= 0x03000000
  return m;
#else
  return;
#endif
}

